/*
BUFFERS.C -- Display buffer chain, and count BUFFERS=
See also COUNTF.C to determine value of FILES=
Andrew Schulman, revised July 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

#pragma pack(1)

typedef struct dskbuf3 {
    struct dskbuf3 far *next;
    BYTE drive, flags;
    } DSKBUF3;
    
typedef struct {
    WORD next, prev;
    BYTE drive, flags;
    DWORD sector;
    } DSKBUF4;
    
void fail(const char *s) { puts(s); exit(1); }

char *buff_status(BYTE flags)
{
    static char buf[128];
    buf[0] = '\0';
    if (flags & 2) strcpy(buf, "FAT ");
    if (flags & 4) strcat(buf, "DIR ");
    if (flags & 8) strcat(buf, "DATA ");
    if (flags & 16) strcat(buf, "REF ");    // referenced
    if (flags & 32) strcat(buf, "DIRTY ");  // modified, not yet written
    if (flags & 64) strcat(buf, "REMOTE ");
    return buf;
}
    
/* if XMS present, store entry point in xms_entrypoint */
static void (far *xms_entrypoint)(void) = (void (far*)()) 0;

int xms_is_present(void)
	{
	int	available = 0;

	if (xms_entrypoint)	return 1;

	_asm {
		mov 	ax, 4300h
		int		2fh
		cmp		al, 80h
		je		present
		jmp		SHORT done
		}
	
present:	
	available = 1;
	_asm {
		mov		ax, 4310h;
		int		2fh;
		push	ds;
		mov		ax, seg xms_entrypoint;
		mov		ds, ax;
		mov		word ptr xms_entrypoint, bx;
		mov		word ptr xms_entrypoint+2, es;
		pop		ds;
		}
done:
	return available;
	}
	
int xms_local_a20(int func)
{
	if (! xms_entrypoint)
		if (! xms_is_present())
			return 0;	// no XMS
				
	_asm mov ah, byte ptr func
	(*xms_entrypoint)();
	// return value in AX; success=1
}

int xms_local_enable_a20(void)	{ return xms_local_a20(5); }

int xms_local_disable_a20(void) { return xms_local_a20(6); }

unsigned buffers(void)
{
    BYTE far *doslist;
    WORD buffers;
    
    _asm mov ah, 52h
    _asm int 21h
    _asm mov word ptr doslist+2, es
    _asm mov word ptr doslist, bx
    
    /* pointer to first disk buffer in List of Lists */
    if (_osmajor < 3)
        fail("This program requires DOS 3.0 or higher");
    else if (_osmajor == 3)
    {
        DSKBUF3 far *dskbuf3;
        if (_osminor == 0)
            dskbuf3 = *((DSKBUF3 far * far *) (doslist + 0x13));
        else 
            dskbuf3 = *((DSKBUF3 far * far *) (doslist + 0x12));
        for (buffers=1; ; buffers++)
        {
            printf("buffer @ %Fp\n", dskbuf3);
            if ((dskbuf3 = dskbuf3->next) == (DSKBUF3 far *) -1L)
                break;
        }
    }
    else    // DOS 4, 5, 6
    {
        DSKBUF4 far *dskbuf4 = *((DSKBUF4 far * far *) (doslist + 0x12));
        WORD seg, first, in_use;
        
        if (_osmajor >= 5)  // DOS 5+ disk buffer info struct
            dskbuf4 = *((DSKBUF4 far * far *) dskbuf4);
            
        seg = FP_SEG(dskbuf4);
        first = FP_OFF(dskbuf4);
		in_use = 0;
		
#if 1
{
	DWORD lin = (DWORD) seg << 4L;
	lin += first;
	if (lin > 0x100000L)
	{
		DWORD far *low = (DWORD far *) 0x00000080L;
		DWORD far *high = (DWORD far *) 0xFFFF0090L;
		puts("BUFFERS are HIGH");
#if 1		
		xms_local_disable_a20();	// testing!!
#endif			
		if ((low[0] == high[0]) && (low[1] == high[1]))
		{
			puts("A20 is off; turning A20 on");
#if 0
			if (! xms_local_enable_a20())
				fail("Can't enable A20");
#else
			// as we know from chapter 6, this will work too!!
			_asm mov ax, 3000h
			_asm int 21h
#endif
			if ((low[0] == high[0]) && (low[1] == high[1]))	// still!
				fail("A20 still disabled!");
		}
	}
}
#endif
        
        for (buffers=1; ; buffers++)
        {
            if (dskbuf4->drive == 0xFF) 
                printf("%Fp -- unused\n", dskbuf4);
            else
			{
				in_use++;
                printf("%Fp -- %c: #%lu -- %s\n", 
                    dskbuf4,                        // pointer
                    'A' + dskbuf4->drive,           // drive
                    dskbuf4->sector,                // sector number
                    buff_status(dskbuf4->flags));   // status
			}
            if (dskbuf4->next == first) // LRU list wraps around
                break;
            else
                dskbuf4 = (DSKBUF4 far *) MK_FP(seg, dskbuf4->next);
        }
        if (buffers != doslist[0x3F])	// number of buffers
            fail("Something wrong!");
		printf("%u buffers in use\n", in_use);
        return buffers;
    }
}

main(int argc, char *argv[])
{
	WORD num_buf;
	if (argc > 1)	// to experiment in DEBUG
	{
		static char space[1024];
		printf("Space at %Fp\n", (char far *) space);
		system(getenv("COMSPEC"));
	}
    num_buf = buffers();
    printf("BUFFERS=%d\n", num_buf);
    return num_buf;     // can check with ERRORLEVEL
}
