// TODO: MAKE CMDLINE SWITCHES TO TURN OFF INDIVIDUAL HOOKS


/* 
    CALDERA.C -- combination of DOSVER.C from UndocDOS 2e (to change
    DOS version number to 7.10) and code to fake Get_SysDat_Path, etc.
        
    Changed from version printed in chapter 2 of UndocDOS:
    Added slimy -3306 switch to also fake out the 21/3306 Get Version call
        
    Andrew Schulman, December 1997, to run Windows 4.x on OpenDOS 7.x
        
    bcc -P- -Ox -B caldera.c
        
    Run "caldera win" instead of plain "win". Can also 
    run "caldera winice" instead of plain "winice".
        
(1) Change DOS version to 7.0 or 7.10       
        
(2) Fake support for INT 2Fh AX=1613h (Get_SYSDAT_Path call,
    mentioned but not documented in DDK INT2FAPI.H)
        
    Entry: 
        AX = 1613h
        ES:DI -> "SYSTEM.DAT" and buffer
        CX = length of buffer at ES:DI
        INT 2Fh
    Exit:
        AX = 0
        ES:DI replaced with fullpathname, up to CX length
            
(3) Temporarily, use device=c:\windows\ifshlp.sys for IFS$HLP$ support
    (later, use Ralf Brown Interrupt list documentation 21/3F, 21/4402)
        
(4) Support for CONFIG$ (Ralf Brown Interrupt list documentation
    21/4402, 21/4403). Right now, I'm just hacking in this with 
    Soft-ICE!
        
(5) Add support for INT 2Fh AX=1612h, called by VTD real-mode
    init? (INT 2Fh AX=1612h sets AX=0, return ES:BX -> 6 bytes
    of data, with ES:[BX+2] -> some IO.SYS buffer????
        
(6) Add stub support for INT 2Fh AX=1690h (see 2f1690.lst). Also
    added one entry for IFSHLP.SYS, to see if helps with message
    from SYSDM.

(7) Add stub support for INT 2Fh AX=1607h BX=0Dh (VKD callout; see
    vxd.lst). Must be inside INT 16h handler.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <dos.h>

#pragma pack(1)

static int do_3306 = 0;     // fake out 21/3306 Get Version call too
static void (interrupt far *old21)();
static void (interrupt far *old2f)();
static void (interrupt far *old16)();
static unsigned dosver, old_bx, old_cx, old_revflags;

typedef struct {
#ifdef __TURBOC__
    unsigned short bp,di,si,ds,es,dx,cx,bx,ax;
#else
    unsigned short es,ds,di,si,bp,sp,bx,dx,cx,ax;   /* PUSHA order */
#endif
    unsigned short ip,cs,flags;
    } REG_PARAMS;
    
void interrupt far dos(REG_PARAMS r)
{
    if ((r.ax >> 8) == 0x30)
    {
        r.ax = dosver;
        r.bx = old_bx;
        r.cx = old_cx;
    }
    else if (r.ax == 0x3306 && do_3306)
    {
        r.bx = dosver;
        r.dx = old_revflags;
    }
#if 1
    // FOLLOWING NOT NECESSARY FOR WIN4 TO RUN ON OPENDOS ANYWAY??

    // right now, just hacking it in with Soft-ICE
    // when BPINT 21 AH=3D hit with DS:DX -> 'CONFIG$'
    // just change DS:DX on the fly to 'IFS$HLP$',0 so succeeds
    // then after IOCTL (21 AX=4402) call, 
    // R AX = 5
    // A (assemble) to change JB 379 to JNB 379
    // E 0E 0,0,FE,C8,03
    // this will get past CONFIGMG VxD check for CONFIG$ IOCTL
#else
    else if ((r.ax >> 8) == 0x3D)
    {
        char far *dsdx = MK_FP(r.ds, r.dx);
        if (fmemcpy(dsdx, "CONFIG$", 7) == 0)
        {
        }
        else
            _chain_intr(old21);
    }
#endif  
    else
        _chain_intr(old21);
}

void interrupt far int16(REG_PARAMS r)
{
    _asm addr_for_2f16070d: 
    _asm pushf
    _asm dec word ptr cs:[idle_word]
    _asm jnz skip_nop
    _asm nop
skip_nop:       
    _asm popf
        
    _chain_intr(old16);

    _asm idle_word:
    _asm dw 0
}

char *reg_path = "C:\\WINDOWS\\SYSTEM.DAT"; // hard-wired; make cmdline param?

char fake_data1[1024] = {0} ;
int fake_data2[6]; // need to set so fake_data2[1] -> fake_data1
char far *fp_fake_data = (char far *) fake_data2;

// one fake struct, with blank 8.3 name, 0 type byte
char fake_1690[] = {
    7,
    0,0,0,0,
    2,                          // number of structures
        
    '1','2','3','4','5','6','7','8','1','2','3', // fake name
    0,0,0,
    1,                          // number of 5-byte groups
    0,0,0,0,0,                   // one 5-byte group, first byte not 2 or 3
        
    'I','F','S','H','L','P',' ',' ','S','Y','S', // IFSHLP.SYS
    0,0,0,
    1,
    8,0,0,0,0,                  // just what MS-DOS does
    } ;

unsigned long fake_1690_2[2] = {0, 0};

#if 0
// NO!!! THIS WILL ACTUALLY NEED TO BE CODE PART OF AN INT 16 HANDLER
unsigned short fake_1607_0d[] = {
    0x2e9c, 0x0eff,             // signature 1 = 0x0eff2e9c
    0x050b,                     // offset for V86 bp???
    0x0175, 0x9d90,             // signature 2 = 0x9d900175
    } ; 
#endif
        
void interrupt far int2f(REG_PARAMS r)
{
    if (r.ax == 0x1613)
    {
        r.ax = 0;
        _fmemcpy(MK_FP(r.es, r.di), reg_path, 
            min(r.cx, strlen(reg_path) + 1));
    }
    else if (r.ax == 0x1612)
    {
        // should watch this in Soft-ICE to see if ever used
            
        *((char far **) &fake_data2[1]) = (char far *) fake_data1;
        
        r.ax = 0;
        // for now
        r.es = FP_SEG(fp_fake_data);
        r.bx = FP_OFF(fp_fake_data);
    }
    else if (r.ax == 0x1690)
    {
        void far *fp = (void far *) fake_1690_2;
        r.es = FP_SEG(fp);
        r.bx = FP_OFF(fp);
        fake_1690_2[0] = (unsigned long) (void far *) fake_1690;
    }
    else if ((r.ax == 0x1607) && (r.bx == 0x0D)) // VKD
    {
#if 1
        unsigned short segm, offs;
        _asm push cs
        _asm pop ax
        _asm mov segm, ax
        _asm mov ax, offset addr_for_2f16070d
        _asm mov offs, ax
        r.es = segm;
        r.di = offs;
#else           
        void far *fp = (void far *) fake_1607_0d;
        r.es = FP_SEG(fp);
        r.di = FP_OFF(fp);
#endif      
    }
    else
        _chain_intr(old2f);
}

void fail(char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    int new_major, new_minor;
    unsigned char old_major, old_minor;
    
    if (argc < 2)
        fail("usage: caldera [command line]");
    
#define MAJOR 7
#define MINOR 10
// #define MINOR 0
    
    new_major = MAJOR;
    new_minor = MINOR;
    if (new_minor < 10) new_minor *= 10;
    dosver = (new_minor << 8) + new_major;
    
    _asm mov ax, 3000h
    _asm int 21h
    _asm mov old_major, al
    _asm mov old_minor, ah
    _asm mov old_cx, cx                     /* OEM, serial# */
    _asm mov old_bx, bx

    do_3306++;
    _asm mov ax, 3306h
    _asm int 21h
    _asm mov old_major, bl
    _asm mov old_minor, bh
    _asm mov old_revflags, dx

    old21 = _dos_getvect(0x21);                       /* save INT 21h */
    old2f = _dos_getvect(0x2f);
    old16 = _dos_getvect(0x16);
    _dos_setvect(0x21, dos);                        /* hook INT 21h */
    _dos_setvect(0x2f, int2f);
    _dos_setvect(0x16, int16);
    if (spawnvp(P_WAIT, argv[1], &argv[1]) == -1)   /* run command */
        perror("Can't run program!");   // changed from puts() in book
    _dos_setvect(0x16, old16);
    _dos_setvect(0x2f, old2f);
    _dos_setvect(0x21, old21);                        /* unhook INT 21h */
    return 0;
}
