/*
CLIPTEST.C
Andrew Schulman
Phar Lap Software
(617) 661-1510 x238 (w)
(617) 876-2102 (h)

(c) Andrew Schulman 1990

This is a non-Windows DOS program that puts a string into the
Windows. It also times the overhead of this operation. Must be used
from inside Windows 3.0 enhanced mode (or Windows/386 2.0). Won't work
in standard or real modes. To use:

C:\>cliptest [string] [iterations]

For example:

C:\>cliptest "This is a test" 1000

The iterations test is here just to see whether this mechanism is quick
enough to used as a generalized way of communicating between DOS programs
and Windows. 

This is an old hacked-up version that (yuk!) doesn't include subroutines. All
the INT 2Fh calls are inline. Sorry!
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <malloc.h>
#include <dos.h>

/* old application clipboard functions */
#define WINOLDAP_VERS	0x1700
#define OPEN_CLIP		0x1701
#define EMPTY_CLIP		0x1702
#define SET_CLIP_DATA	0x1703
#define GET_CLIP_SIZE	0x1704
#define GET_CLIP_DATA	0x1705
#define CLOSE_CLIP		0x1708
#define CLIP_COMPACT	0x1709
#define GET_DEV_CAPS	0x170A

/* predefined clipboard formats */
#define CF_TEXT			1
#define CF_BITMAP		2
#define CF_METAFILEPICT 3
#define CF_SYLK			4
#define CF_DIF			5
#define CF_TIFF			6
#define CF_OEMTEXT		7

#define MAKELONG(a, b)	((long)(((unsigned)(a)) | ((unsigned long)((unsigned)(b))) << 16))

int main(int argc, char *argv[])
{
	union REGS regs;
	struct SREGS sregs;
	unsigned long size;
	char far *buf = argv[1];
	int count = atoi(argv[2]);
	
	regs.x.ax = WINOLDAP_VERS;
	int86(0x2F, &regs, &regs);
	if (regs.x.ax == WINOLDAP_VERS)
	{
		/* either Windows not installed, or old version */
		printf("WINOLDAP.MOD clipboard functions not available\n");
		return 1;
	}
	printf("winoldap.mod version %d.%d\n",
		regs.h.al, regs.h.ah);

	regs.x.ax = OPEN_CLIP;
	int86(0x2F, &regs, &regs);

{
	time_t t1, t2;
	int i;
	time(&t1);
	for (i=count; i--; )
	{
		regs.x.ax = SET_CLIP_DATA;
		regs.x.dx = CF_TEXT;
		segread(&sregs);
		sregs.es = FP_SEG(buf);
		regs.x.bx = FP_OFF(buf);
		regs.x.si = 0;
		regs.x.cx = strlen(buf);
		int86x(0x2F, &regs, &regs, &sregs);
		if (regs.x.ax == 0)
			puts("Data not copied into clipboard");
	}
	time(&t2);
	printf("%d iter: %lu seconds\n", count, t2 - t1);
}
	
	regs.x.ax = GET_CLIP_SIZE;
	regs.x.dx = CF_TEXT;
	int86(0x2F, &regs, &regs);
	size = MAKELONG(regs.x.ax, regs.x.dx);
	if (! size)
		puts("No text in clipboard");
	else
	{
		printf("Text in clipboard: %lu bytes\n", size);
		if (buf = _fmalloc(size))
		{
			regs.x.ax = GET_CLIP_DATA;
			regs.x.dx = CF_TEXT;
			sregs.es = FP_SEG(buf);
			regs.x.bx = FP_OFF(buf);
			int86x(0x2F, &regs, &regs, &sregs);
			if (! regs.x.ax)
				printf("Get clipboard data failed\n");
			else
				printf("Clipboard contents:\n%Fs\n", buf);	/* Far string */
			_ffree(buf);
		}
	}
	
	regs.x.ax = CLOSE_CLIP;
	int86(0x2F, &regs, &regs);

	return 0;
}

