/* 
CLUSTNAM.C -- Convert cluster # to full pathname
See also NAMCLUST.C (Convert pathname to cluster #)
Andrew Schulman, July 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include "diskstuf.h"

void search_for_cluster();
void do_cluster(WORD cluster);
void do_dir(WORD cluster, DIR_ENTRY *buffer);
void read_cluster(int drive, BYTE far *buffer, WORD cluster);

void fail(const char *s) { puts(s); exit(1); }

static STRING glob_name[64] = {0};
static DPB far *glob_dpb;
static WORD want_cluster = 0xFFFF;
static WORD bytes_per_clust = 0, sect_per_clust = 0, first_sect = 0;
static int level = 0, num_dir = 0, glob_drive;

main(int argc, char *argv[])
{
    int i;
    
    if (argc < 3)
        fail("usage: clusters [drive] [cluster]");
    
    // program uses DOS 3.31 style of INT 25h
    if ((_osmajor < 3) || (_osmajor == 3 && _osminor < 31))
        fail("This program requires DOS 3.31 or higher");
    
    for (i=0; i<64; i++)
        glob_name[i] = (STRING) malloc(16);
    
    glob_drive = toupper(argv[1][0]) - 'A';
    want_cluster = atoi(argv[2]);
    if (glob_dpb = get_dpb(glob_drive+1))
	{
		if (want_cluster <= glob_dpb->highest_cluster)
        	search_for_cluster();
		else
			fail("cluster number too high");
	}
    else
        fail("can't get DPB");
	fail("couldn't find cluster");	// error if get here
    return 0;
}

void search_for_cluster(void)
{
    glob_drive = glob_dpb->drive;
    sect_per_clust = glob_dpb->sectors_per_cluster + 1;
    bytes_per_clust = glob_dpb->bytes_per_sect * sect_per_clust;
    first_sect = glob_dpb->first_data_sector;
    num_dir = bytes_per_clust / sizeof(DIR_ENTRY);  
    do_cluster(0);  // phony cluster 0 to kick-start
}

void do_cluster(WORD cluster)
{
    BYTE *buffer;
    if (! (buffer = (BYTE far *) calloc(bytes_per_clust, 1)))
        fail("insufficient memory");
    read_cluster(glob_drive, buffer, cluster);
    do_dir(cluster, (DIR_ENTRY *) buffer);
    free(buffer);
}

void do_dir(WORD cluster, DIR_ENTRY *buffer)
{
    DIR_ENTRY *dir, *p;
    char name[9], ext[4];
    WORD next;
    int entry, i;
    
    dir = (DIR_ENTRY *) buffer;
    for (entry=0, p=dir; entry<num_dir; p++, entry++)
    {
        if (p->name[0] == '\0')	return;		// end of dir
        if (p->name[0] == 0xE5)	continue;	// deleted entry
        if (p->name[0] == '.')	continue;	// don't bother with . and ..
        
        memcpy(name, p->name, 8); name[8] = 0;
        if (name[0] == 0x05) name[0] = 0xE5;    // 0xE5 is valid first char
        memcpy(ext, p->ext, 3);   ext[3] = 0;

        strcpy(glob_name[level], name);
        if (ext[0] != ' ' != 0)
        {
            strcat(glob_name[level], ".");
            strcat(glob_name[level], ext);
        }
        
        if (p->cluster == want_cluster) // found it!
        {
            char *s;
            int i;
            putchar(glob_drive + 'A'); putchar(':'); putchar('\\');
            for (i=0; i<level; i++)
            {
                s = glob_name[i];       // trim spaces out of name
                while (*s) { if (*s != ' ') putchar(*s); s++; }
                putchar('\\');
            }
            s = glob_name[level];       // trim spaces out of name
            while (*s) { if (*s != ' ') putchar(*s); s++; }
            putchar('\n');
            exit(0);                    // done: success!
        }
        
        if (p->attrib & DIRECTORY)      // already know not . or ..
        {
            level++;
            do_cluster(p->cluster);     // if directory, recurse
            level--;
        }
    }
    
    // locate next directory cluster
    next = get_fat_entry(glob_drive, glob_dpb, cluster);
    if (next < 0xFFF0)                  // end of dir
        do_cluster(next);
}

// Hummel, Assembly Language Lab Notes, p. 88
#define CLUSTER_TO_LSN(clust)   \
    ((((DWORD) (clust) - 2) * sect_per_clust) + first_sect)

void read_cluster(int drive, BYTE far *buffer, WORD cluster)
{
    DWORD sector;
    if (cluster)            sector = CLUSTER_TO_LSN(cluster);
    else if (_osmajor >= 4) sector = glob_dpb->vers.dos4.first_dir_sector;
    else                    sector = glob_dpb->vers.dos3.first_dir_sector;
    if (! read_sectors(drive, buffer, sect_per_clust, sector))
        fail("can't read cluster");
}
