comment~
CR3.ASM -- Virtual CR3 Device
Andrew Schulman
from Microsoft Systems Journal, October 1992
See also Microsoft Systems Journal, December 1992

Provides applications running under Windows Enhanced mode (Windows
apps, real-mode DOS apps, and protected-mode DOS apps) with access
to the CR3 register.  CR3 contains the _physical_ base address of
the page directory, which in turn contains physical addresses of
page tables.  A program calls CR3.386 using its V86 or PM API; the
API entry point is retrieved with INT 2Fh AX=1684h BX=28C2h (CR3.386's
device ID).  The entry point is called with AX=1 to get CR3 into DX:AX:

   unsigned long (far *cr3_vxd)(void);
   unsigned long cr3;
   cr3_vxd = GetDeviceAPI(0x28C2);    // call 2F/1684
   if (cr3_vxd) {
      _asm mov ax, 1
      cr3 = (*cr3_vxd)();
      }
   else fail("Requires device=cr3.386 in SYSTEM.INI [386Enh]");

The returned address is _physical_, not linear, and must be mapped in
using a service such as DPMI INT 31h AX=0800h.     

 To build:
   set include=\ddk\include
   masm5 -p -w2 cr3;
   link386 cr3,cr3.386,,,cr3.def
   addhdr cr3.386
 Add device=cr3.386 to SYSTEM.INI [386Enh]
~

.386p

INCLUDE VMM.INC

;*****************************************************************************
;                   E Q U A T E S
;*****************************************************************************

; get your own ID from vxdid@microsoft.com

RUNNING_DOG_OEM     EQU 326
CR3_DEVICE_ID       EQU 2
CR3_VxD_ID          EQU (RUNNING_DOG_OEM SHL 5) OR CR3_DEVICE_ID

CFlag               EQU 1

;*****************************************************************************
;        V I R T U A L   D E V I C E   D E C L A R A T I O N
;*****************************************************************************

Declare_Virtual_Device CR3, 1, 0, CR3_VxD_Control, \
    CR3_VxD_ID, Undefined_Init_Order, \
    CR3_API_Handler, CR3_API_Handler,

;*****************************************************************************
;           R E A L   M O D E   I N I T I A L I Z A T I O N
;*****************************************************************************

VXD_REAL_INIT_SEG

real_init proc near
IF 0
    mov ah, 9
    mov dx, offset copyright
    int 21h
    xor ax, ax
    xor bx, bx
    xor si, si
    xor edx, edx
ENDIF
    ret
real_init endp

copyright   db  'CR3.386 -- CR3 Access VxD v. 1.0',0dh,0ah
            db  'Copyright (c) 1992 Andrew Schulman. All Rights Reserved.'
            db  0dh,0ah,'$'

VXD_REAL_INIT_ENDS

;*****************************************************************************
;           L O C K E D   C O D E   S E G M E N T
;*****************************************************************************

VxD_LOCKED_CODE_SEG

BeginProc CR3_VxD_Control
    clc
    ret
EndProc CR3_VxD_Control

VxD_LOCKED_CODE_ENDS

;*****************************************************************************
;               O N L Y   D A T A   S E G M E N T
;*****************************************************************************

VxD_DATA_SEG

CR3_API_Call    label dword
                dd  offset32 Get_Version
                dd  offset32 Get_CR3

CR3_API_MaxCall EQU ( $ - CR3_API_Call ) / 4

VxD_DATA_ENDS

;*****************************************************************************
;          M O V E A B L E   C O D E   S E G M E N T
;*****************************************************************************

VxD_CODE_SEG

BeginProc   CR3_API_Handler
    movzx   eax, [ebp.Client_AX]
    cmp     eax, CR3_API_MaxCall
    jae     short fail
    and     [ebp.Client_EFlags], NOT CFlag
    call    CR3_API_Call[eax * 4]
    ret
fail:
    or      [ebp.Client_EFlags], CFlag
    ret
EndProc     CR3_API_Handler

BeginProc   Get_Version
    mov     [ebp.Client_AX], 0001h
    ret
EndProc     Get_Version

BeginProc   Get_CR3
    mov     eax, cr3
    mov     edx, eax
    shr     edx, 16
    mov     [ebp.Client_DX], dx
    mov     [ebp.Client_AX], ax
    ret
EndProc     Get_CR3

VxD_CODE_ENDS

    END
