/****************************************************************
 *    DEVLOD.C - Jim Kyle - 08/20/90                            *
 *      Copyright 1990,1992 by Jim Kyle - All Rights Reserved   *
 *    (minor revisions by Andrew Schulman - 9/12/90)            *
 *    (major rewrite by Jim Kyle - July-Aug 1992)               *
 *    Dynamic loader for device drivers                         *
 *           Requires Turbo C or BC++; see DEVLOD.MAK also.     *
 ****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>

#include "undocdos.h"   /* defines DOS internal structures      */

#define GETFLAGS __emit__(0x9F)
#define FIXDS    __emit__(0x16,0x1F)    /* PUSH SS, POP DS      */
#define PUSH_BP  __emit__(0x55)
#define POP_BP   __emit__(0x5D)
#define GO_DOS   geninterrupt( 0x21 )

unsigned _stklen = 0x200;
unsigned _heaplen = 0;

LPPSP PSPptr;           /* used to access fields of the PSP     */
char  FileName[65];     /* filename global buffer               */
char  * dvrarg;         /* points to char after name in cmdline */
WORD  movsize;          /* number of bytes to be moved up       */
void  (far * driver)(); /* used as pointer to call driver code  */
LPDDVR drvptr;          /* holds pointer to device driver       */
LPDDVR NULptr;          /* pointer to NUL device (chain start)  */
FP    nuldrvr;          /* additional driver pointers           */
FP    nxtdrvr;
BYTE  far * nblkdrs;    /* points to block device count in List */
WORD  lastdrive;        /* value of LASTDRIVE in List of Lists  */
BYTE  far * CDSbase;    /* base of Current Dir Structure        */
int   CDSsize;          /* size of CDS element                  */
WORD  nulseg;           /* hold parts of ListOfLists pointer    */
WORD  nulofs;
WORD  LoLofs;
LPLOL LOLptr;
DDCMDPKT CmdPkt;
int SFT_size;           /* used by nex FixSFT, set by GetList   */

extern unsigned _psp;           /* established by startup code  */
extern unsigned _heaptop;       /* established by startup code  */
extern BYTE _osmajor;           /* established by startup code  */
extern BYTE _osminor;           /* established by startup code  */
void _exit( int );              /* established by startup code  */
void abort( void );             /* established by startup code  */

void movup( LPPSP, FP, int );   /* in MOVUP.ASM                 */

int  TestName( LPDDVR, LPSFT ); /* in TESTNAME.ASM for FixSFT   */
void ChgSFT( LPSFT, LPDDVR );   /* in TESTNAME.ASM for FixSFT   */

void copyptr( FPP src, FPP dst )        /* copy far pointer     */
{ *dst = *src; }

void exit(int c)        /* called by startup code's sequence    */
{ _exit(c);}

void Put_Msg ( char *msg )  /* replaces printf(), uses DOS only */
{ _AH = 2;              /* doesn't need to be inside loop       */
  while (*msg)
  { _DL = *msg++;
    GO_DOS;
  }
}

BOOL Get_Driver_Name ( void )
{ char *nameptr;
  int i, j, cmdlinesz;

  nameptr = (char *)&(PSPptr->Tailc);   /* set up to parse      */
  cmdlinesz = *nameptr++;
  if (cmdlinesz < 1)        /* if nothing there, return FALSE   */
    return FALSE;
  for (i=0; i<cmdlinesz && nameptr[i]<'!'; i++) /* skip blanks  */
    ;
  dvrarg = (char *)&nameptr[i]; /* save to put in SI            */
  for ( j=0; i<cmdlinesz && nameptr[i]>' '; i++)  /* copy name  */
    FileName[j++] = nameptr[i];
  FileName[j] = '\0';   /* name copied, but good time to        */
  for (; i<cmdlinesz && nameptr[i]>=' '; i++)   /*  make all UC */
    if( nameptr[i] >= 'a' && nameptr[i] <= 'z' )
      nameptr[i] &= 0x5F;               /* take out case bit    */
  return TRUE;          /* and return TRUE to keep going        */
}

void Err_Halt ( char *msg )     /* print message and abort      */
{ Put_Msg ( msg );
  Put_Msg ( "\r\n" );           /* send CR,LF                   */
  abort();
}

void Move_Loader ( void )       /* vacate lower part of RAM     */
{ WORD movsize, destseg;
  movsize = _heaptop - _psp;    /* size of loader in paragraphs */
  destseg = PSPptr->Nxtgraf;    /* end of memory                */
  movup ( PSPptr, MK_FP( destseg - movsize, 0 ),
          movsize << 4);        /* move and fix segregs         */
}

void Load_Drvr ( void )         /* load driver file into RAM    */
{ WORD handle;
  struct {
    WORD LoadSeg;
    WORD RelocSeg;
  } ExecBlock;

  ExecBlock.LoadSeg = _psp + 0x10;
  ExecBlock.RelocSeg = _psp + 0x10;
  _DX = (WORD)&FileName[0];     /* ds:dx point to filename      */
  _BX = (WORD)&ExecBlock;       /* es:bx point to ExecBlock     */
  _ES = _SS;                    /* which, being local, is SS    */
  _AX = 0x4B03;                 /* load overlay (COM, SYS, EXE) */
  GO_DOS;                       /* DS is okay on this call      */
  GETFLAGS;                     /* check what happened          */
  if ( _AH & 1 )                /* if carry flag set...         */
    Err_Halt ( "Unable to load driver file." );
}

void Get_List ( void )          /* set up pointers via List     */
{ _AH = 0x52;                   /* find DOS List of Lists       */
  GO_DOS;
  nulseg = _ES;                 /* DOS data segment             */
  LoLofs = _BX;                 /* current drive table offset   */
  LOLptr = (LPLOL)MK_FP( nulseg, LoLofs );

  switch( _osmajor )            /* NUL adr varies with version  */
    {
    case  0:
      Err_Halt ( "Drivers not used in DOS V1." );
    case  2:
      nblkdrs = (FP)0L;
      lastdrive = (LOLptr->ver.v2.lastdrv);
      NULptr = (LPDDVR)&(LOLptr->ver.v2.nul);
      nulofs = (WORD)NULptr;    /* just the offset part         */
      SFT_size = 0x28;
      break;
    case  3:
      if (_osminor == 0)
      {
          nblkdrs = (BYTE far *)&(LOLptr->ver.v30.blk_dev);
          lastdrive = (LOLptr->ver.v30.lastdrv);
          NULptr = (LPDDVR)&(LOLptr->ver.v30.nul);
          nulofs = (WORD)NULptr;        /* just the offset part */
          SFT_size = 0x38;
      }
      else
      {
          nblkdrs = (BYTE far *)&(LOLptr->ver.v31up.blk_dev);
          lastdrive = (LOLptr->ver.v31up.lastdrv);
          NULptr = (LPDDVR)&(LOLptr->ver.v31up.nul);
          nulofs = (WORD)NULptr;        /* just the offset part */
          SFT_size = 0x35;
      }
      CDSbase = (BYTE far *)(LOLptr->ver.v31up.cds);
      CDSsize = sizeof( CDS );  /* defined for DOS3.1 struct    */
      break;
    case  4:
    case  5:
    case  6:
      nblkdrs = (BYTE far *)&(LOLptr->ver.v31up.blk_dev);
      lastdrive = (LOLptr->ver.v31up.lastdrv);
      NULptr = (LPDDVR)&(LOLptr->ver.v31up.nul);
      nulofs = (WORD)NULptr;            /* just the offset part */
      CDSbase = (BYTE far *)(LOLptr->ver.v31up.cds);
      CDSsize = sizeof( CDS ) + 7;      /* V4,5 7 bytes bigger  */
      SFT_size = 0x3B;
      break;
    case 10:
    case 20:
      Err_Halt ( "OS2 DOS Box not supported." );
    case 30:
      Err_Halt ( "Windows NT not supported." );
    default:
      Err_Halt ( "Unknown version of DOS!");
    }
}

void Fix_DOS_Chain ( void )     /* patches driver into DOS chn  */
{ WORD i;

  nuldrvr = MK_FP( nulseg, nulofs+0x0A );       /* verify drvr  */
  drvptr = "NUL     ";
  for ( i=0; i<8; ++i )
    if ( *((BYTE far *)nuldrvr+i) != *((BYTE far *)drvptr+i) )
      Err_Halt ( "Failed to find NUL driver." );

  nuldrvr = MK_FP( nulseg, nulofs );    /* point to NUL driver  */
  drvptr  = MK_FP( _psp+0x10, 0 );      /* new driver's address */

  copyptr( (FP)nuldrvr, (FP)&nxtdrvr ); /* hold old head now    */
  copyptr( (FP)&drvptr, (FP)nuldrvr );  /* put new after NUL    */
  copyptr( (FP)&nxtdrvr, (FP)drvptr );  /* and old after new    */
}

// returns number of next free drive, -1 if none available
int Next_Drive ( void )
{
#ifdef USE_BLKDEV
  return (nblkdrs && (*nblkdrs < lastdrive)) ? *nblkdrs : -1;
#else
  /* This approach takes account of SUBSTed and network-redirector
   * drives by finding the first unused entry in the CDS structure.
   */
  LPCDS cds;
  int i=0;
  for ( cds=(LPCDS)CDSbase;     /* start at front of struct     */
        i<lastdrive;            /* go all way through it        */
        i++, ((BYTE far *)cds)+=CDSsize)        /* count up     */
    if (! cds->flags)           /* found a free drive           */
        break;
  return (i == lastdrive) ? -1 : i;     /* return number, or -1 */
#endif
}

/*  This routine initializes the device driver and returns TRUE
 *  if all went well.  If initialization fails for any reason,
 *  this function returns FALSE and the driver will not be linked
 *  into the chain maintained by DOS.
 */
BOOL Init_Drvr ( void )
{ WORD tmp;
  drvptr  = MK_FP( _psp+0x10, 0 );      /* new driver's address */
  CmdPkt.command = DD_INIT;     /* defined in UNDOCDOS.H file   */
  CmdPkt.hdrlen = sizeof( DDCMDPKT );
  CmdPkt.unit = 0;
  CmdPkt.status = 0;            /* clear status just in case    */
  CmdPkt.inpofs = (WORD)dvrarg; /* points into command line     */
  CmdPkt.inpseg = _psp;
  CmdPkt.NextDrv = (BYTE)Next_Drive();  /* for block dev init   */
  if( CmdPkt.NextDrv == 0xFF &&
      ((drvptr->attr & CHAR_DEV) == 0))
    { Put_Msg( "Current Directory Structure is full, cannot install." );
      return FALSE;
    }

#if 1
  if (_osmajor >= 5)
  {
	  // In DOS 5+, DOS passes the device driver irEndAddress 
	  // (see DOS Programmer's Reference, p. 400)
	  CmdPkt.brkofs = 0;
	  CmdPkt.brkseg = 0xA000;		// is this right????
  }
#endif  

  tmp = drvptr->stratofs;       /* STRATEGY pointer in driver   */
  driver = MK_FP( FP_SEG( drvptr ), tmp );
  _ES = FP_SEG( (void far *)&CmdPkt );
  _BX = FP_OFF( (void far *)&CmdPkt );
  (*driver)();                  /* set up the packet address    */

  tmp = drvptr->introfs;        /* INTERRUPT pointer in driver  */
  driver = MK_FP( FP_SEG( drvptr ), tmp );
  (*driver)();                  /* do the initialization        */

  /* In the first edition version of DEVLOD, this function checked
   * the status code in the command packet to determine whether the
   * installation had failed. Actually, the status code is NOT
   * checked by DOS itself! Thanks to Geoff Chappell for pointing
   * out that SYSINIT does not check the status returned by drivers
   * after initialization -- a block device driver is not retained if
   * its unit count is found to be zero, while a character device
   * should set its break address to offset 0 in its load segment.
   */
  return( drvptr->attr & CHAR_DEV ?
          MK_FP(CmdPkt.brkseg,CmdPkt.brkofs) != MK_FP(FP_SEG(drvptr), 0) :
          CmdPkt.nunits != 0 );
}

/*  This routine looks far more complicated than it actually is.
 *  It's used only when block-device drivers are being installed,
 *  and does the housekeeping of DPBs and CDS entries that such
 *  devices require.  Major changes were made here in the second
 *  edition of the program, to accomodate changes in DPB size that
 *  happened at Version 4.0 but went unnoticed until DOS5 appeared.
 *  Special thanks are due Dan B. Wright, Nathaniel Polish, and
 *  Geoff Chappell for spotting problem areas here.
 *
 *  This routine returns FALSE if all goes well, or TRUE if any
 *  error condition is detected.
 */
BOOL Put_Blk_Dev ( void )
{ int newdrv;
  int i;
  int retval = TRUE;        /* pre-set for failure              */
  int BufferSize;
  int unit = 0;
  LPDPB oldDPB, newDPB, endmark = (LPDPB)0xFFFFFFFFL;
  LPCDS cds;
  LPDDVR newdriver = (LPDDVR)MK_FP( _psp+0x10, 0 );

  if ((Next_Drive() == -1) || CmdPkt.nunits == 0)
    return retval;              /* cannot install block driver  */
  if (CmdPkt.brkofs != 0)       /* align to next paragraph      */
  {
    CmdPkt.brkseg += (CmdPkt.brkofs >> 4) + 1;
    CmdPkt.brkofs = 0;
  }
  while( CmdPkt.nunits-- )  /* repeat this loop for each unit   */
  {
    if ((newdrv = Next_Drive()) == -1)
      return TRUE;      /* no room for another drive, quit      */
    if( nblkdrs )               /* if not a null pointer,       */
      (*nblkdrs)++;             /* ...tally into drive counter  */

#ifdef ORIGINAL
    /*  Tell DOS to get the DPB of the last drive in CDS.  This
     *  technique, used in the first version of DEVLOD, creates a
     *  problem if the final drive is a JOINed or SUBSTed entry in
     *  the list.  The alternate method of finding the last DPB is
     *  not subject to this problem, but may be a bit slower.  The
     *  address of the last-drive DPB is saved in "oldDPB".
     */
    _AH = 0x32;         /* get DPB of last drive in CDS         */
    _DL = newdrv;
    GO_DOS;
    _AX = _DS;          /* save segment to make the pointer     */
    FIXDS;
    oldDPB = MK_FP(_AX, _BX);   /* this is base address of DPB  */
#else
    /*  Trace out entire chain each time around the loop.  While
     *  this is possibly slower than the original method, it will
     *  not be deceived by CDS entries.  The address of the last
     *  DPB is saved in "oldDPB".  Note that only the offset words
     *  of the link pointers are compared; DOS itself does not put
     *  "endmark" in the segment word, although this routine does.
     */
    oldDPB = LOLptr->dpb;       /* always start at first DPB    */
    if (_osmajor < 4)           /* trace through to the end     */
      while( (WORD)oldDPB->ver.v3.next != (WORD)endmark )
        oldDPB = oldDPB->ver.v3.next;
    else
      while( (WORD)oldDPB->ver.v45.next != (WORD)endmark )
        oldDPB = oldDPB->ver.v45.next;
#endif

    /*  Tell DOS to create the DPB, passing it BPB info from a
     *  list of near pointers passed back by the driver.  Note
     *  that DS must be set after all memory references are done,
     *  because it's used to access the globals.  Similarly, AX
     *  must be set after all segment registers, because it is
     *  used to load the segreg.
     */
    newDPB = (LPDPB)MK_FP( CmdPkt.brkseg, 0 );
    _SI = *(WORD far *)MK_FP(CmdPkt.inpseg, CmdPkt.inpofs);
    _ES = CmdPkt.brkseg;        /* ES:BP is DPB address to use  */
    _DS = CmdPkt.inpseg;        /* DS:SI is adr of BPB to read  */
    PUSH_BP;                    /* save stack-frame pointer     */
    _BP = 0;                    /* DPB offset value             */
    _AH = 0x53;                 /* build the DPB for this unit  */
    GO_DOS;
    POP_BP;                     /* restore stack-frame pointer  */
    FIXDS;

    /*  Check to be sure that block sector size is acceptable.
     *  If bigger than BUFFERS were built for, refuse to install
     *  the driver...  No such check was made in the original
     *  version of this program. Geoff Chappell spotted the omission.
     */
    switch( _osmajor )          /* get BUFFERS size from LOL    */
    {                           /* location in LOL will vary... */
    case  2:
      BufferSize = LOLptr->ver.v2.secsiz;
      break;
    case  3:
      if (_osminor == 0)
        BufferSize = LOLptr->ver.v30.secsiz;
      else
        BufferSize = LOLptr->ver.v31up.secsiz;
      break;
    case  4:
    case  5:
	case  6:
      BufferSize = LOLptr->ver.v31up.secsiz;
      break;
    default:
      Err_Halt ( "Unknown version of DOS!");
    }
    if( newDPB->bytes_per_sect > BufferSize )
      return TRUE;                      /* get out if too big   */

    /*  Now set the DPB address into the old last-DPB's link
     *  address field ("next" pointer).
     */
    if (_osmajor < 4)           /* link new DPB to chain        */
      oldDPB->ver.v3.next = newDPB;
    else
      oldDPB->ver.v45.next = newDPB;

    /*  Set up the Current Directory Structure for this drive and
     *  tag it as a physical drive.  Clear IFS area if DOS V4 or
     *  V5.
     */
    if (_osmajor > 2)           /* Version 2 did not use CDS    */
      { cds = (LPCDS)(CDSbase + (newdrv * CDSsize));
        cds->flags = CDS_PHYS;  /* defined in UNDOCDOS.H file   */
        cds->dpb = newDPB;      /* set DPB adr into CDS         */
        cds->in.loc.start_cluster = 0xFFFF; /* not accessed yet */
        cds->in.loc.ffff = -1L;
        cds->slash_offset = 2;  /* start in root directory      */
        if (_osmajor > 3)       /* zero out IFS stuff           */
          { *((WORD far *)(&(cds->slash_offset))+1) = 0;
            *((WORD far *)(&(cds->slash_offset))+2) = 0;
            *((WORD far *)(&(cds->slash_offset))+3) = 0;
            *((BYTE far *)(&(cds->slash_offset))+8) = 0;
          }
      }

    /*  Finally, set up pointers for the DPB and the driver so
     *  that they can find each other, and adjust space
     *  reservations so that the DPB won't be wiped out upon
     *  return to DOS.  Step the BPB list pointer in case the
     *  driver has multiple units.
     */
    newDPB->drive = newdrv;     /* set in drive number          */
    newDPB->unit = unit++;      /* and also the unit number     */
    if (_osmajor < 4)           /* Versions 2 and 3 are alike   */
      { newDPB->ver.v3.driver = newdriver;
               newDPB->ver.v3.next = endmark;
        CmdPkt.brkseg += 2;     /* was 32 bytes, exact fit      */
      }
    else                        /* Versions 4 and 5 are alike   */
      { newDPB->ver.v45.driver = newdriver;
        newDPB->ver.v45.next = endmark;
        CmdPkt.brkseg += 3;     /* 33 bytes each now            */
      }
    CmdPkt.inpofs += 2;         /* point to next BPB pointer    */
  }                             /* end of nunits loop           */
  return FALSE;                 /* all went okay                */
}

/*  This function is called for a character driver only. It searches
 *  every entry in the SFT, looking for any reference to the named
 *  driver, and if such a reference is found, modifies 6 bytes in the
 *  SFT entry to point to the just-added new driver. This is necessary
 *  for proper operation of the critical error handler. It is based on
 *  code by Dan Winter, published in the July, 1992, issue of Dr. Dobb's
 *  Journal, but Dan's original code was not compatible with DEVLOD.
 *
 *  SFT_size is established by the GetList() function earlier.
 */
void FixSFT ( void )
{ LPSFTB blk, next_blk;         /* use typedefs for simplicity  */
  LPSFT sft;
  WORD num_items;

  for( next_blk = LOLptr->sft;          /* first SFT block      */
       (WORD)next_blk != 0xFFFF; )      /* check all blocks     */
    { blk = next_blk;                   /* start this block     */
      next_blk = blk->next;             /* link for next one    */
      num_items = blk->here;            /* get size of block    */
      sft = &(blk->first);              /* first SFT in it now  */
      while( num_items-- > 0 )          /* test all items here  */
        { if( sft->ver.v2.nusers )      /* is this SFT in use?  */
            if( TestName( drvptr, sft ))        /* yes, match?  */
              ChgSFT( sft, drvptr );    /* yes, fix the pointer */
          (char far *)sft += SFT_size;  /* to next item in list */
        }
    }
}

/*  This function is called only when the driver has been fully
 *  installed with no detected errors.  If the driver is a block
 *  device, Put_Blk_Dev() is called to create its drive letter
 *  and CDS and DPB structures; otherwise the CON and CLOCK$
 *  pointers in the List of Lists are updated as applicable.  If
 *  the block device installation fails, DEVLOD quits without
 *  linking the driver into the DOS chain.  Otherwise this driver
 *  is put at the head of the chain right after NUL, and the program
 *  returns to DOS, leaving the driver resident.  In either case,
 *  this function will never return to the main() procedure.
 */
void Get_Out ( void )
{ WORD temp;

  temp = drvptr->attr;          /* attribute word               */
  if( (temp & CHAR_DEV) == 0 )  /* if block device, set up tbls */
    { if( Put_Blk_Dev() )       /* fails if cannot do so        */
             Err_Halt( "Could not install block device" );
    }
  else                          /* not block, check for updates */
    { if( (temp & IS_STDIN) )
      LOLptr->con = drvptr;     /* this is for CTRL-C checking  */
      else if( (temp & IS_CLOCK) )
      LOLptr->clock = drvptr;   /* this is for fast time access */
      FixSFT();                 /* Dan Winter's fix for SFT     */
    }

  Fix_DOS_Chain();              /* all okay so patch into DOS   */

  _ES = PSPptr->EnvSeg;         /* release environment space    */
  _AH = 0x49;
  GO_DOS;
  PSPptr->EnvSeg = 0;           /* zero out the address in PSP  */

  /* then set up regs for KEEP function, and go resident        */
  temp = (CmdPkt.brkofs + 15);  /* normalize the offset         */
  temp >>= 4;
  temp += CmdPkt.brkseg;        /* add the segment address      */
  temp -= _psp;                 /* convert to paragraph count   */
  _DX = (WORD)temp;             /* paragraphs to retain         */
  _AX = 0x3100;                 /* KEEP function of DOS         */
  GO_DOS;                       /* won't come back from here!   */
}

void main( void )               /* usual argc, argv not used!   */
{ PSPptr = (LPPSP)MK_FP( _psp, 0 );     /* create global ptr    */
  if( !Get_Driver_Name() )
    Err_Halt( "Device driver name required.");
  Move_Loader();                /* move code high and jump      */
  Load_Drvr();                  /* bring driver into freed RAM  */
  Get_List();                   /* get DOS internal variables   */
  if(Init_Drvr())               /* let driver do its thing      */
      Get_Out();                /* check init status, go TSR    */
  else
      Err_Halt( "Driver initialization failed." );
}
