/*
DISKSTUF.C -- Some functions and structures for low-level disk access
Andrew Schulman, July 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <dos.h>
#include "diskstuf.h"

DPB far *get_dpb(int drive)
{
    DPB far *dpb;

    _asm push ds
    _asm mov ah, 32h
    _asm mov dl, byte ptr drive
    _asm int 21h
    _asm mov dx, ds
    _asm pop ds
    _asm cmp al, 0FFh
    _asm je failure
    _asm mov word ptr dpb+2, dx
    _asm mov word ptr dpb, bx
    return dpb;
failure:
    return (DPB far *) 0;   // failure
}

char far *truename(char far *s, char far *d)
{
    /* INT 21h AH=60h doesn't like leading or trailing blanks */
    char far *s2;
    while (isspace(*s)) s++;            // ltrim
    s2 = s; while (*s2) s2++; s2--;     // go to end
    while (isspace(*s2)) *s2-- = 0;     // rtrim

    /* BTW, apparently some versions of DR DOS insist on ES:DI != DS:SI */
    
    _asm push di
    _asm push si
    _asm push ds
    _asm push es
    _asm les di, d
    _asm lds si, s
    _asm mov ah, 60h
    _asm int 21h
    _asm pop es
    _asm pop ds
    _asm pop si
    _asm pop di
    _asm jc error
    return d;
error:
    return (char far *) 0;
}

typedef struct {
    DWORD diStartSector;
    WORD diSectors;
    BYTE far *diBuffer;
    } DISKIO;

int read_sectors(int drive, BYTE far *buf, int sectors, DWORD first_sector)
{
    DISKIO diskio, far *pdiskio = &diskio;
    
    diskio.diStartSector = first_sector;
    diskio.diSectors = sectors;
    diskio.diBuffer = buf;
    
    _asm push ds
    _asm pusha                      /* requires 286 or higher */
    _asm mov al, byte ptr drive
    _asm lds bx, pdiskio
    _asm mov cx, 0FFFFh
    _asm mov dx, 0
    _asm int 25h
    _asm jc fail
    _asm popf
    _asm popa
    _asm pop ds
    return 1;
fail:
    _asm popf
    _asm popa
    _asm pop ds
    return 0;
}

WORD get_fat_entry(int drive, DPB far *dpb, WORD cluster)
{
    static WORD *fat_sect = (WORD *) -1;
    static WORD prev_drive = (WORD) -1;
    static DPB far *prev_dpb = (DPB far *) -1L;
    static WORD prev_sect = (WORD) -1;
    WORD first_fat_sect, entry_per_sect, sect, fat_size;
    
    if (fat_sect == (WORD *) -1)    // one-time initialization
        if (! (fat_sect = (WORD *) malloc(dpb->bytes_per_sect * 2)))
            fail("insufficient memory");
		
    fat_size = (dpb->highest_cluster >> 12 == 0) ? 12 : 16;
    
    if (_osmajor >= 4)
        first_fat_sect = dpb->vers.dos4.first_dir_sector -
            (dpb->copies_fat * dpb->vers.dos4.sectors_per_fat);
    else
        first_fat_sect = dpb->vers.dos3.first_dir_sector -
            (dpb->copies_fat * dpb->vers.dos3.sectors_per_fat);

    if (fat_size == 12)
        sect = first_fat_sect + (((cluster * 3) / 2) / dpb->bytes_per_sect);
    else
    {
        entry_per_sect = dpb->bytes_per_sect / 2;
        sect = first_fat_sect + (cluster / entry_per_sect);
    }

    if (! (drive == prev_drive && dpb == prev_dpb && sect == prev_sect))
    {
        int num_sect = (fat_size == 12) ? 2 : 1;    // for possible slop
        if (! read_sectors(drive, (BYTE far *) fat_sect, num_sect, sect))
            fail("can't read FAT");     // could try FAT #2
    }

    prev_sect = sect; prev_drive = drive; prev_dpb = dpb;

    if (fat_size == 12)
    {
        BYTE *fat = (BYTE *) fat_sect;
        WORD ofs = ((cluster * 3) / 2) % dpb->bytes_per_sect;
        WORD retval = *((WORD *) &fat[ofs]);
        
        if (cluster & 1)        // odd cluster #
            retval >>= 4;       //    take top 12 bits
        else                    // even cluster #
            retval &= 0x0FFF;   //    take bottom 12 bits
				
        // return 12-bit FF0-FFF as 16-bit FFF0-FFFF
        return (retval >= 0x0FF0) ? (retval | 0xF000) : retval;
    }
    else    // gosh, 16-bit FATs are easy
        return fat_sect[cluster % entry_per_sect];
}

