/* 
DISKSTUF.H -- Some functions and structures for low-level disk access
Andrew Schulman, July 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)
*/

#ifndef DISKSTUF_H
#define DISKSTUF_H

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef char *STRING;

extern void fail(const char *s);    // to be defined by app

#pragma pack(1)

typedef struct {
    BYTE name[8], ext[3], attrib;
#ifdef OS2
    BYTE reserved[8];
    WORD ea_handle;     // OS/2 uses reserved bytes for extended attributes
#else       
    BYTE reserved[10];  // Chicago long filenames DON'T seem to use these
#endif
    WORD time, date, cluster;
    DWORD size;
    } DIR_ENTRY;

#define VOLUME_ATTR      0x08
#define DIRECTORY_ATTR   0x10

#ifndef MK_FP
#define MK_FP(seg,ofs) \
    ((void far *)(((DWORD)(seg) << 16) | (ofs)))
#endif

typedef struct dpb {                // Drive Parameter Block
    BYTE drive, unit;
    WORD bytes_per_sect;
    BYTE sectors_per_cluster;       // plus 1
    BYTE shift;                     // for sectors per cluster
    WORD boot_sectors;          
    BYTE copies_fat;
    WORD max_root_dir, first_data_sector, highest_cluster;
    union {
        struct {
            BYTE sectors_per_fat;
            WORD first_dir_sector;  // root dir
            void far *device_driver;
            BYTE media_descriptor, access_flag;
            struct dpb far *next;
            DWORD reserved;
            } dos3;
        struct {
            WORD sectors_per_fat;       // WORD, not BYTE!
            WORD first_dir_sector;
            void far *device_driver;
            BYTE media_descriptor, access_flag;
            struct dpb far *next;
            DWORD reserved;
            } dos4;
        } vers;
    } DPB;
	
#pragma pack()	
    
DPB far *get_dpb(int drive);
int _dos_driveremoveable(int drive);
int _dos_getdrivemap(int drive);
char far *truename(char far *s, char far *d);
WORD get_fat_entry(int drive, DPB far *dpb, WORD cluster);
int read_sectors(int drive, BYTE far *buf, int sectors, DWORD first_sector);

/* critical error stuff */
void watch_crit_err(void);
void unwatch_crit_err(void);
int get_failed(void);
void reset_failed(void);

#endif /* DISKSTUF_H */
