/* DOSVER.C -- set different DOS version numbers */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <dos.h>

#pragma pack(1)

static int do_3306 = 0;		// fake out 21/3306 Get Version call too
static void (interrupt far *old)();
static unsigned dosver, old_bx, old_cx, old_revflags;

typedef struct {
#ifdef __TURBOC__
    unsigned short bp,di,si,ds,es,dx,cx,bx,ax;
#else
    unsigned short es,ds,di,si,bp,sp,bx,dx,cx,ax;   /* PUSHA order */
#endif
    unsigned short ip,cs,flags;
    } REG_PARAMS;
	
void interrupt far dos(REG_PARAMS r)
{
    if ((r.ax >> 8) == 0x30)
    {
        r.ax = dosver;
        r.bx = old_bx;
        r.cx = old_cx;
    }
    else if (r.ax == 0x3306 && do_3306)
	{
		r.bx = dosver;
		r.dx = old_revflags;
	}
	else
        _chain_intr(old);
}

void fail(char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    int new_major, new_minor;
    unsigned char old_major, old_minor;
	
	#define PROGNAME	"DOSVER"
	#define DESC		"Change DOS version for run of specified application"
	#define AUTHOR		"Andrew Schulman"
printf("%s v. %d.%02d -- %s\n", PROGNAME, 2, 1, DESC);
printf("Copyright (c) %s 1993.  All rights reserved.\n", AUTHOR);
printf("From \"Undocumented DOS\" 2nd edition (Addison-Wesley, 1993)\n\n");
    
    if (argc < 4)
        fail("usage: dosver <major> <minor> [-3306] <command...>\n"
             "example: dosver 3 31 exe2bin devlod.exe devlod.com");

    if (! (new_major = atoi(argv[1])))
        fail("bad version number");
    if ((new_minor = atoi(argv[2])) < 10)       /* e.g. 3.1 to 3.10 */
        new_minor *= 10;                        
    dosver = (new_minor << 8) + new_major;
    
    _asm mov ax, 3000h
    _asm int 21h
    _asm mov old_major, al
    _asm mov old_minor, ah
    _asm mov old_cx, cx                     /* OEM, serial# */
    _asm mov old_bx, bx

    if ((new_major == old_major) && (new_minor == old_minor))
        puts("no change");
    printf("Changing DOS version from %u.%02u to %u.%02u\n",
        old_major, old_minor, new_major, new_minor);
	
	if (strcmp(strupr(argv[3]), "-3306") == 0)
	{
		do_3306++;
		_asm mov ax, 3306h
		_asm int 21h
		_asm mov old_major, bl
		_asm mov old_minor, bh
		_asm mov old_revflags, dx
		argv++;
	
		if ((new_major == old_major) && (new_minor == old_minor))
			puts("3306: no change");
		printf("Changing 21/3306 version from %u.%02u to %u.%02u\n",
			old_major, old_minor, new_major, new_minor);
	}
    
    old = _dos_getvect(0x21);               /* save INT 21h */
    _dos_setvect(0x21, dos);                /* hook INT 21h */
    if (spawnvp(P_WAIT, argv[3], &argv[3]) == -1)     /* run command */
		puts("Can't run program!");
    _dos_setvect(0x21, old);                /* unhook INT 21h */
    return 0;
}
