/*
DPBTEST.C -- uses undocumented INT 21h Function 32h (Get DPB)
    to display bytes per drive; but first walks the DPB chain, 
    showing the difference between the two access methods
Andrew Schulman, updated July 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)
*/

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#ifdef __TURBOC__
#include <dir.h>
#endif
#include "diskstuf.h"			// DPB structure, get_dpb()

void fail(const char *s) { puts(s); exit(1); }

void display(DPB far *dpb)
{
    DWORD bytes_per_clust = 
        dpb->bytes_per_sect * (dpb->sectors_per_cluster + 1);

    printf("Drive %c: ", 'A' + dpb->drive);
    printf("%u bytes/sector * ", dpb->bytes_per_sect);
    printf("%u sectors/cluster = \n", 
        dpb->sectors_per_cluster + 1);
    printf("         %lu bytes/cluster * ", bytes_per_clust);
    printf("%u clusters = ", dpb->highest_cluster - 1);
    printf("%lu bytes\n\n",
        bytes_per_clust * (dpb->highest_cluster - 1));
}

main()
{
	BYTE far *sysvars;
    DPB far *dpb;
    int i;
    
    /* floppy = single disk drive logical drive indicator 0=a 1=b */
	BYTE floppy;
	if (_osmajor > 3 || (_osmajor == 3 && _osminor >= 2))
	{
		_asm mov ax, 440Eh		/* Get Logical Drive Map */
		_asm mov bl, 1			/* drive A: */
		_asm int 21h
		_asm jnc okay
		fail("can't get logical drive map");
okay:		
		_asm mov floppy, al		/* 1=A, 2=B, etc. */
		floppy--;				/* make 0=A etc. */
	}
	else
		floppy = *((BYTE far *) 0x504L);	// use low-memory variable
    
#ifdef __TURBOC__
    WORD lastdrive = setdisk(0xFF);
#else
    WORD lastdrive;
    _dos_setdrive(0xFF, &lastdrive);
#endif

    puts("Using DPB linked list");

	_asm mov ah, 52h			/* get SysVars */
	_asm int 21h
	_asm mov word ptr sysvars+2, es
	_asm mov word ptr sysvars, bx
    /* pointer to first DPB at offset 0h in SysVars */
    if (! (dpb = *((DPB far * far *) sysvars)))
        return 1;
    do {
        /* skip either drive A: or drive B: */
        if (((floppy == 1) && (dpb->drive != 0)) ||
            ((floppy == 0) && (dpb->drive != 1)))
                display(dpb);
        if (_osmajor < 4) 
            dpb = dpb->vers.dos3.next;
        else
            dpb = dpb->vers.dos4.next;
    } while (FP_OFF(dpb) != 0xFFFF);

    
    puts("Using INT 21h Function 32h");
	/* do either A: or B: */
	if (dpb = get_dpb(floppy + 1))	
		display(dpb);
	/* do drive C: through lastdrive */
    for (i=3; i<=lastdrive; i++)
		if (dpb = get_dpb(i))
			display(dpb);

    return 0;
}
