/*
DYNLINK32.C
Win32 Console app -- demonstrates Win32 run-time dynamic links to Win16
Via undocumented LoadLibrary16, GetProcAddress16 APIs in KERNEL32

from Schulman, Unauthorized Windows, 1994
*/

#include <stdlib.h>
#include <stdio.h>
#define  WIN32_LEAN_AND_MEAN
#include "windows.h"

HINSTANCE (WINAPI *LoadLibrary16)(LPCSTR lpLibFileName);
FARPROC (WINAPI *GetProcAddress16)(HINSTANCE hModule, LPCSTR lpProcName);

void fail(const char *s) { puts(s); exit(1); }

#define GET_PROC(mod, func) GetProcAddress(GetModuleHandle(mod), (func))

main(int argc, char *argv[])
{
    HINSTANCE mod;
    FARPROC fp;
    DWORD arg, retval;
    
    if (argc < 3) 
        fail("usage: dynlnk32 [module name] [function name]");
    
    #define GET(func, str) \
        if (! (func = GET_PROC("KERNEL32", str))) \
            fail("Cannot link to " str);
        
    GET(LoadLibrary16, "LoadLibrary16");
    GET(GetProcAddress16, "GetProcAddress16");
    
    if (! (mod = LoadLibrary16(argv[1])))
        fail("LoadLibrary16 failed");
    printf("LoadLibrary16(\"%s\") ==> 0x%08X\n", argv[1], mod);
    if (! (fp = GetProcAddress16(mod, argv[2])))
        fail("GetProcAddress16 failed");
    printf("GetProcAddress16(0x%08X, \"%s\") ==> 0x%08X\n", 
        mod, argv[2], fp);
    
    return 0;
}

