/* 
ENUMDRV.C
Version of ENUMDRV for multiple VMs in Windows Enhanced mode
Andrew Schulman, April 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

DPMI version:
bcc -2 -DDPMI_APP -c vxdcalls.c vmwalk.c iswin.c prot.c dpmish.c ctrl_c.asm
tlib dpmi_app-+vxdcalls-+vmwalk-+iswin-+prot-+dpmish-+ctrl_c
bcc -2 -DDPMI_APP enumdrv.c dpmi_app.lib
copy enumdrv.exe denumdrv.exe
        
WVMWALK version (uses EasyWin; no generic VxD):
bcc -W -DWINDOWS -DUSE_WVMWALK -2 enumdrv.c wvmwalk.c prot.c
copy enumdrv.exe wenumdrv.exe

Windows version (uses EasyWin):
bcc -2 -W -DWINDOWS enumdrv.c prot.c vxdcalls.c vmwalk.c

Microsoft version (uses QuickWin; had to hack _osmajor because QuickWin
	uses it for the Windows version number instead of the DOS version!):
cl -G2 -Mq -DWINDOWS enumdrv.c prot.c vxdcalls.c vmwalk.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

#ifdef WINDOWS
#include "windows.h"
#else
#define DPMI_APP
#include "dpmish.h"
#endif

#include "prot.h"
#ifndef USE_WVMWALK
#include "vxdcalls.h"
#endif
#include "vmwalk.h"

#define NETWORK			(1 << 15)
#define PHYSICAL		(1 << 14)
#define JOIN			(1 << 13)
#define SUBST			(1 << 12)
#define REDIR_NOT_NET   (1 << 7)		// not really CDROM bit!

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef BYTE far *DPB;      // provide actual DPB struct if needed

#pragma pack(1)

typedef struct {
    BYTE current_path[67];  // current path
    WORD flags;             // NETWORK, PHYSICAL, JOIN, SUBST
    DPB  far *dpb;          // pointer to Drive Parameter Block
    union {
        struct {
            WORD start_cluster; // root: 0000; never accessed: FFFFh
            DWORD unknown;
            } LOCAL;        // if (! (cds[drive].flags & NETWORK))
        struct {
            DWORD redirifs_record_ptr;
            WORD parameter;
            } NET;          // if (cds[drive].flags & NETWORK)
        } u;
    WORD backslash_offset;  // offset in current_path of '\'
    // DOS4 fields for IFS
    // 7 extra bytes...
    } CDS;                  // Current Directory Structure

#pragma pack()

CDS far *currdir(unsigned drive);
void set_vm(DWORD vm);

#define ENHANCED_MODE   3

// see ISWIN.C
extern int is_win(int *pmaj, int *pmin, int *pmode);

void fail(const char *s, ...)
{ 
    puts(s); 
#ifdef DPMI_APP
    _dos_exit(1); 
#else
    exit(1);
#endif
}

static void far *sysvars_real = 0;
static void far *sda_real = 0;
static int currdir_size = 0;

#ifdef DPMI_APP
// going to do initialization from real mode (real_main())!
#define CALL_DOS(x,y,z)     int86x(0x21, x, y, z)
#else
#define CALL_DOS(x,y,z)     real_int86x(0x21, x, y, z)
#endif

void far *get_sysvars(void)
{
    union REGS r;
    struct SREGS s;
    memset(&r, 0, sizeof(r));
    memset(&s, 0, sizeof(s));
    r.h.ah = 0x52;
    CALL_DOS(&r, &r, &s);
    return MK_FP(s.es, r.x.bx);
}

void far *get_sda(void)
{
    union REGS r;
    struct SREGS s;
    memset(&r, 0, sizeof(r));
    memset(&s, 0, sizeof(s));
    r.x.ax = 0x5d06;
    CALL_DOS(&r, &r, &s);
    return (r.x.cflag) ? (void far *) 0 : MK_FP(s.ds, r.x.si);
}

void do_init(void)
{
#ifndef USE_WVMWALK 
    if (! GenericVxDInstalled())
        fail("This program requires the generic VxD (VXD.386)\n"
             "Put device=vxd.386 in [386Enh] section of SYSTEM.INI");
#endif       

    // assumes SysVars at same address in all VMs
    // but CDS not necessarily at same address!
    if (! (sysvars_real = get_sysvars()))
        fail("21/52 failed: can't get SysVars!\n");
    
    // assumes SDA at same address in all VMs
    if (! (sda_real = get_sda()))
        fail("21/5D06 failed: can't get SDA!\n");

	// problem!  Microsoft C QuickWin defines _osmajor, _osminor
	// with the Windows version, not the DOS version!
	_asm mov ax, 3000h
	_asm int 21h
	_asm mov byte ptr _osmajor, al
    currdir_size = (_osmajor >= 4) ? 0x58 : 0x51;
}
    
#pragma argsused
#ifdef WINDOWS
int main(int argc, char *argv[])
{
	#define PROGNAME	"ENUMDRV"
	#define DESC		"Enumerate drives in all Windows virtual machines"
	#define AUTHOR		"Andrew Schulman"
printf("%s v. %d.%02d -- %s\n", PROGNAME, 2, 0, DESC);
printf("Copyright (c) %s 1993.  All rights reserved.\n", AUTHOR);
printf("From \"Undocumented DOS\" 2nd edition (Addison-Wesley, 1993)\n\n");
	
    if (! (GetWinFlags() & WF_PMODE))
        fail("This program requires Windows 3.x Enhanced mode");
    
    do_init();

#ifdef USE_WVMWALK
    return wvmwalk(wwalkfunc);
#else
    return vmwalk(walkfunc);
#endif
}
#else
int real_main(int argc, char *argv[])
{
    int maj, min, mode;
    
    puts("ENUMDRV -- Enumerate CDS in all Windows VMs");
    puts("From \"Undocumented DOS\", 2nd edition (Addison-Wesley, 1993)\n");
    
    if ((! is_win(&maj, &min, &mode)) || (maj < 3) || 
        (mode != ENHANCED_MODE))
        fail("This program requires Windows 3.x Enhanced mode");
    
    do_init();  // call in real mode
    return 0;   // okay to switch into protected mode
}

#pragma argsused
int pmode_main(int argc, char *argv[])
{
    vmwalk(walkfunc);
    return 0;
}
#endif

#pragma argsused
#ifdef USE_WVMWALK
BOOL wwalkfunc(HWND hwnd, char *title, DWORD vm, VM_CB far *vm_cb)
#else
BOOL walkfunc(DWORD vm, VM_CB far *vm_cb)
#endif
{
    CDS far *cds, far *dir;
    BYTE far *sysvars, far *cds_real, far *sda;
    int lastdrv, currdrv, i;
	
#ifndef USE_WVMWALK
    static DWORD sys_vm = 0;
    static DWORD cur_vm = 0;
	
    if (! sys_vm) sys_vm = GetSysVMHandle();
    if (! cur_vm) cur_vm = GetCurVMHandle();
#endif  
    
#define CDS_OFS         0x16
#define LASTDRV_OFS     0x21

    // Set the base address.  Could also have used map_linear_vm
    // or map_real_vm.  This one call to set_base will affect all
    // subsequent calls to map_linear.
    set_base(vm_cb->CB_High_Linear);
    if (! (sysvars = (BYTE far *) map_real(sysvars_real, LASTDRV_OFS+1)))
		fail("can't map SysVars");
	
    // Note: does _not_ assume that CDS at same address in all VMs!
    // But moving a CDS breaks in Windows anyway, because DOSMGR assumes
    // that CDS never moves after Windows initialization.  *Sigh*
		
    cds_real = *((BYTE far * far *) &sysvars[CDS_OFS]);
    lastdrv = sysvars[LASTDRV_OFS];

    free_mapped_linear(sysvars);
	
    printf("\n%s (VM %lu) -- CDS at %08lX\n",
#ifdef USE_WVMWALK
		title,
#else
        (vm == sys_vm) ? "System VM" :
        (vm == cur_vm) ? "Current VM" :
        /* default */    "DOS Box",
#endif			
        vm_cb->CB_VMID,
        vm_cb->CB_High_Linear + MK_LIN(cds_real));
    
#define CURRDRV_OFS     0x16

    // get current drive for this VM from SDA:0016
    // for System VM, will depend on current task
    // (each Windows app has its own current drive/directory)
    // already did set_base(vm_cb->CB_High_Linear);
    if (! (sda = (BYTE far *) map_real(sda_real, CURRDRV_OFS+1)))
		fail("can't map SDA");
    currdrv = sda[CURRDRV_OFS];
    free_mapped_linear(sda);
	
    // map in cds_real for this VM, print out, free it
    if (! (cds = (CDS far *) map_real(cds_real, lastdrv * currdir_size)))
		fail("can't map CDS");
    for (i=0; i<lastdrv; i++)
    {
        dir = (CDS far *) (((BYTE far *) cds) + (i * currdir_size));
        if (dir->flags)
        {
            // highlight current drive for this VM
            printf( (i == currdrv) ? "==> " : "    " ); 

            printf("%c\t%-50Fs", 'A' + i, dir->current_path);
            if (dir->flags & REDIR_NOT_NET) printf("REDIR_NOT_NET ");
            else if (dir->flags & NETWORK) printf("NETWORK ");
            if (dir->flags & JOIN) printf("JOIN ");
            if (dir->flags & SUBST) printf("SUBST");
            printf("\n");
        }
    }
    free_mapped_linear(cds);
	
    return TRUE;
}
