/* 
ENUMDRV2.C
Version of ENUMDRV for multiple VMs in Windows Enhanced mode
Andrew Schulman, February 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

bcc -2 -DDPMI_APP enumdrv2.c dpmish.obj prot.obj vxdcalls.obj \
	iswin.obj tasksw.obj ctrl_c.obj 

To do:
Make general-purpose VM walker for use with ENUMDRV, FILES, DEV?
Test with app (FILES, LASTDRV from QEMM) that moves CDS!
Possibly write own version to show how works!
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <dos.h>
#define DPMI_APP
#include "dpmish.h"
#include "prot.h"
#include "vxdcalls.h"
#include "currdir.h"

#define ENHANCED_MODE	3

// see ISWIN.C
extern int is_win(int *pmaj, int *pmin, int *pmode);

// from ddk vmm.inc
typedef struct {
    DWORD CB_VM_Status;
    DWORD CB_High_Linear;
    DWORD CB_Client_Pointer;
    DWORD CB_VMID;
    } VM_CB;    // Virtual Machine Control Block
        
typedef BOOL (*WALKFUNC)(DWORD vm, VM_CB far *);        

int vmwalk(WALKFUNC walkfunc);
BOOL walkfunc(DWORD vm, VM_CB far *vm_cb);

DWORD GetSysVMHandle(void);
DWORD GetCurVMHandle(void);
DWORD GetNextVMHandle(DWORD vm);

void fail(const char *s)
{ 
    puts(s); 
    _dos_exit(1); 
}

static DWORD sysvars_lin = 0;
static DWORD sda_lin = 0;
static int currdir_size = 0;

void far *get_sysvars(void)
{
	_asm xor bx, bx
	_asm mov es, bx
	_asm mov ah, 52h
	_asm int 21h
	_asm mov dx, es
	_asm mov ax, bx
	// return value in DX:AX
}

void far *get_sda(void)
{
	_asm push ds
	_asm push si
	_asm mov ax, 5d06h
	_asm int 21h
	_asm jc error
	_asm mov dx, ds
	_asm mov ax, si
	_asm jmp short done
error:
	_asm xor ax, ax
	_asm mov dx, ax
done:
	_asm pop si
	_asm pop ds
	// return value in DX:AX
}

int real_main(int argc, char *argv[])
{
	void far *sda;
	void far *sysvars;
	int maj, min, mode;
	
	// need banner -- ENUMDRV2
	
	if ((! is_win(&maj, &min, &mode)) || (maj < 3) || 
		(mode != ENHANCED_MODE))
        fail("This program requires Windows 3.x Enhanced mode");
    
    if (! GenericVxDInstalled())
        fail("This program requires the generic VxD (VXD.386)\n"
             "Put device=vxd.386 in [386Enh] section of SYSTEM.INI");

	// assumes SysVars at same address in all VMs
	// but CDS not necessarily at same address!
	if (! (sysvars = get_sysvars()))
		fail("21/52 failed: can't get SysVars!\n");
	else
		sysvars_lin = ((DWORD) FP_SEG(sysvars) << 4) + FP_OFF(sysvars);
	
	// assumes SDA at same address in all VMs
	if (! (sda = get_sda()))
		fail("21/5D06 failed: can't get SDA!\n");	// 5D0B?
	else
		sda_lin = ((DWORD) FP_SEG(sda) << 4) + FP_OFF(sda);
	
    currdir_size = (_osmajor >= 4) ? 0x58 : 0x51;
    
    return 0;   // okay to switch into protected mode
}

int pmode_main(int argc, char *argv[])
{
    vmwalk(walkfunc);
    return 0;
}

int vmwalk(WALKFUNC walkfunc)
{
    DWORD sys_vm = GetSysVMHandle();
    void far *vm_cb;
    VM_CB my_vm_cb;
    DWORD vm;
    int num_vm;
    for (vm=sys_vm, num_vm=0;; num_vm++)
    {
        vm_cb = map_linear(vm, sizeof(VM_CB));
        _fmemcpy(&my_vm_cb, vm_cb, sizeof(VM_CB));
        free_mapped_linear(vm_cb);
        if (! (*walkfunc)(vm, &my_vm_cb))
            return num_vm;
        if ((vm = GetNextVMHandle(vm)) == sys_vm)
            break;  // circular list
    }
    return num_vm;
}

BOOL walkfunc(DWORD vm, VM_CB far *vm_cb)
{
	CDS far *cds, far *dir;
	BYTE far *sysvars_vm;
	BYTE far *cds_real;
	BYTE far *sda_vm;
	DWORD sda_lin_vm;
	DWORD sysvars_lin_vm;
	DWORD cds_lin, cds_lin_vm;
    DWORD total, notmapped;
	int lastdrv, currdrv;
	int i;
    
    static DWORD sys_vm = 0;
    static DWORD cur_vm = 0;
    if (! sys_vm) sys_vm = GetSysVMHandle();
    if (! cur_vm) cur_vm = GetCurVMHandle();
    
	sysvars_lin_vm = vm_cb->CB_High_Linear + sysvars_lin;
	sysvars_vm = (BYTE far *) map_linear(sysvars_lin_vm, 1024);
	cds_real = *((BYTE far * far *) &sysvars_vm[0x16]);
	/* printf("CDS at %Fp\n", cds_real); */
	lastdrv = sysvars_vm[0x21];
	/* printf("LASTDRIVE = %c\n", 'A' - 1 + lastdrv); */
	free_mapped_linear(sysvars_vm);
	
	cds_lin = ((DWORD) FP_SEG(cds_real) << 4) + FP_OFF(cds_real);
	cds_lin_vm = vm_cb->CB_High_Linear + cds_lin;
	
	printf("\n%s (VM %2lu) -- CDS at %08lX\n",
		(vm == sys_vm) ? "System VM" :
		(vm == cur_vm) ? "Current VM" :
		/* default */    "DOS Box",
        vm_cb->CB_VMID,
		cds_lin_vm);

	// get current drive for this VM from SDA:0016
	// for System VM, will depend on current task
	// (each Windows app has its own current drive/directory)
	sda_lin_vm = vm_cb->CB_High_Linear + sda_lin;
	sda_vm = (BYTE far *) map_linear(sda_lin_vm, 1024);
	currdrv = sda_vm[0x16];
	free_mapped_linear(sda_vm);

	// map in cds_lin_vm, print out, free it
	cds = (CDS far *) map_linear(cds_lin_vm, 1024);
	for (i=0; i<lastdrv; i++)
	{
		char buf[128];
		dir = (CDS far *) (((BYTE far *) cds) + (i * currdir_size));
		if (dir->flags)			// any way to check for garbage??
		{
			if (i == currdrv)	// for this VM
				printf("==> ");
			else
				printf("    ");
			sprintf(buf, "%c\t%-50Fs", 'A' + i, dir->current_path);
			if (dir->flags & NETWORK) strcat(buf, "NETWORK ");
			if (dir->flags & JOIN) strcat(buf, "JOIN ");
			if (dir->flags & SUBST) strcat(buf, "SUBST");
			printf("%s\n", buf);
		}
	}
	free_mapped_linear(cds);
	
    return TRUE;
}
        
DWORD GetSysVMHandle(void)
{
    VxDParams p;
    p.CallNum = Get_Sys_VM_Handle;
    return (VxDCall(&p)) ? p.OutEBX : 0;
}

DWORD GetCurVMHandle(void)
{
    VxDParams p;
    p.CallNum = Get_Cur_VM_Handle;
    return (VxDCall(&p)) ? p.OutEBX : 0;
}

DWORD GetNextVMHandle(DWORD vm)
{
    VxDParams p;
    p.CallNum = Get_Next_VM_Handle;
    p.InEBX = vm;
    return (VxDCall(&p)) ? p.OutEBX : 0;
}

