/*
ENUMDRV.C -- uses currdir() in CURRDIR.C
Added CD-ROM bit, DoubleSpace test, Stacker test, primitive RAM disk test
Andrew Schulman, revised July 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)

bcc enumdrv.c currdir.c is_stack.c is_dspac.c diskstuf.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

#include "currdir.h"    // includes DoubleSpace, Stacker too
#include "diskstuf.h"   // get_dpb

void fail(const char *s) { puts(s); exit(1); }

int maybe_ram_disk(int drive);

main()
{
    CDS far *dir;
    BYTE host;
    BOOL swapped;
    unsigned lastdrv;
    int i, seq;
    
    _dos_setdrive(0xFF, &lastdrv);
    for (i=0; i<lastdrv; i++)
        if (! (dir = currdir(i)))
            fail("can't get current directory structure");
        else if (dir->flags)    /* is this a valid drive? */
        {
            printf("%c   %-40Fs   ", 'A' + i, dir->current_path);
            
            if (dir->flags & JOIN) printf("JOIN ");
            if (dir->flags & SUBST) printf("SUBST ");
            if (dir->flags & REDIR_NOT_NET) printf("REDIR_NOT_NET ");
            
            if (dir->flags & NETWORK) printf("NETWORK ");
            else switch (dir->u.LOCAL.start_cluster)
            {
                case 0:      printf("[ROOT] "); break;      // root dir
                case 0xFFFF: printf("[NOT_ACC] "); break;   // not accessed
                default:     printf("[#%u] ", dir->u.LOCAL.start_cluster);
            }   // cluster numbers can be checked by running CLUSTNAM
            
            if (double_space_drive(i, &swapped, &host, &seq))
                printf("DBLSPACE %c:\\dblspace.%03u", 'A' + host, seq);
            
            if (stacker_drive(i, &swapped, &host))
            {
                printf("STACKER");
                if (swapped) printf(" (swapped %c:)", 'A' + host);
            }
            
            if (maybe_ram_disk(i)) printf("RAM disk?");
            
            putchar('\n');
        }
    return 0;
}

int maybe_ram_disk(int drive)
{
#ifdef USE_CDS_DPB
    return (currdir(drive)->dpb->copies_fat == 1);
#else
    DPB far *dpb = get_dpb(drive+1);
    return (dpb) ? (dpb->copies_fat == 1) : 0;
#endif  
}
