/*
euler.c
Andrew Schulman
3 May 1989

Assignment:
Write a program to determine if an Euler path exists in a connected graph G
with n nodes, using:
    a)  its adjacency matrix.
    b)  its adjacency list representation.
        
Comments:
    i)  Graphs are input without redundancy.  That is, if input has already
        indicated that nodes 3 and 4 are adjacent, we don't have to specify
        that 4 and 3 are adjacent.  In fact, a node x adjacent to a node
        y should only be input if x > y.

        The redundancy will be put in when the adjacency matrix and
        adjacency list are constructed.  Without this redundancy, it
		is difficult to calculate the degree of a node, which is used to
		determine if an Euler path exists.
            
    ii) There is one ugly wrinkle to input.  To make programming easier,
        graph should be preceded by number of nodes.
            
    iii) See other routines in graph.c (definitions in graph.h).
        
    iv) It does not appear to be possible to solve the Konigsberg bridge
        problem using these representations, since the problem requires
        two separate undirected edges to and from the same vertices.  This
        would appear to be a major limitation, since all graph theory
        starts from the Konigsberg bridge problem!  But if we use a matrix
        m, where m[i][j] represents not the _existence_ of an edge between
        i and j (as in a Boolean matrix), but the _number_ of edges between
        i and j, and if we use lists where duplicates are allowed, then we
        can support multigraphs, and thus can solve the Konigsberg bridge
        problem.  Code to support this is flagged #ifdef MULTIGRAPH.

Input:
    5 nodes
    1 2 3
    2 3     
    3 4 5   
    4 5

    5 nodes
    1 3 4 5
    2 3 4 5

Output:
    0 1 1 0 0 
    1 0 1 0 0 
    1 1 0 1 1 
    0 0 1 0 1 
    0 0 1 1 0 
    Graph has Euler path

    [1] 2 3 
    [2] 1 3 
    [3] 1 2 4 5 
    [4] 3 5 
    [5] 3 4 
    Graph has Euler path

    0 0 1 1 1 
    0 0 1 1 1 
    1 1 0 0 0 
    1 1 0 0 0 
    1 1 0 0 0 
    Graph has Euler path

    [1] 3 4 5 
    [2] 3 4 5 
    [3] 1 2 
    [4] 1 2 
    [5] 1 2 
    Graph has Euler path
        
Test multigraph support with Konigsberg bridge problem:
Input:
    4 nodes
    1 2 2 4
    2 3 3 4
    3 4
Output: 
    0 2 0 1 
    2 0 2 1 
    0 2 0 1 
    1 1 1 0 
    No Euler path

    [1] 2 2 4 
    [2] 1 1 3 3 4 
    [3] 2 2 4 
    [4] 1 2 3 
    No Euler path
*/

#include <stdio.h>
#include "graph.h"

#define ISODD(x)        ((x) % 2)

void show(MATRIX m, LIST l);
BOOL euler_path_mat(MATRIX m);
BOOL euler_path_list(LIST l);

char *has_path = "Graph has Euler path";
char *no_path = "No Euler path";

main(void)
{
    MATRIX graph;

    while (graph = read_graph())
    {
        show(graph, mat_to_list(graph));
        free_matrix(graph);
        putchar('\n');
    }
}

void show(MATRIX m, LIST l)
{
    print_graph(m);
    puts(euler_path_mat(m) ? has_path : no_path);
    putchar('\n');
    print_list(l);
    puts(euler_path_list(l) ? has_path : no_path);
}

BOOL euler_path_mat(MATRIX m)
{
    WORD i,j,c;
    WORD odd = 0;
    for (i=1; i<=ROWS(m); i++)
    {
        for (c=0, j=1; j<=COLS(m); j++)
#ifdef MULTIGRAPH
            c += m[i][j];
#else
            if (m[i][j])
                c++;
#endif          
        if (ISODD(c))
            odd++;
        if (odd > 2)
            return FALSE;
    }
    return ((odd == 0) || (odd == 2));
}

BOOL euler_path_list(LIST l)
{
    NODE *p;
    WORD i,c;
    WORD odd = 0;
    
    for (i=1; i<= SIZE(l); i++)
    {
        p = l+i;
        c = 0;
        do
        {
            if (p->info)
                c++;
        } while (p = p->ptr);
        if (ISODD(c))
            odd++;
        if (odd > 2)
            return FALSE;
    }
    return ((odd == 0) || (odd == 2));
}
