/* 
EXEDUMP.C

Apologies:

A lot of the complexity here (and some of the bugs in earlier versions!)
comes from trying to read in large chunks at one time from disk, and then
seek in memory, rather than doing small reads and seeking on disk.  Watch
the drive light to see the difference!

The other reason some of this code is so convoluted is that the NE
format seems to have been designed to minimize storage space on disk,
so several key structures are variable length.

Basic organization:
main 
|   newexe 
|   |   open 
|   |   read
|   |   do_enttab 
|   |   |   ne_read 
|   |   |   do_nres (description string comes from here)
|   |   |   |   exe_read 
|   |   |   do_resnam (module name comes from here)
|   |   |   |   ne_read
|   |   |   mangled_name 
|   |   do_segtab 
|   |   |   ne_read
|   |   do_modref 
|   |   |   ne_read
|   |   do_impnam 
|   |   |   ne_read
|   |   |   count2null 
|   |   do_cv_sym 
|   |   |   read
|   |   |   do_subsection 
|   |   |   |   read
|   |   do_seg 
|   |   |   exe_read
|   |   |   print_reloc 
|   |   |   |   in_entry_segofs 
|   |   |   |   print_reloc (recursive)
|   |   |   |   print_reloc_2 
|   |   close
*/

#include <stdlib.h>
#ifdef WINDOWS
#include "windows.h"
#include "winio.h"
#else
#include <stdio.h>
#endif
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <dos.h>

#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC		0x454C			/* linear executable */
#define LXMAGIC		0x584C			/* IBM OS/2 2.0 linear executable */
#define PEMAGIC		0x4550			/* NT portable executable */
#define W3MAGIC		0x3357			/* WIN386.EXE signature */

#define DLL_FLAG    0x8000
#define NEI386		0x0040
#define NEI286		0x0020
#define NEI086		0x0010

#define NSTYPE		0x0007
#define NSCODE		0x0000
#define NSDATA		0x0001
#define NSRELOC		0x0100
#define NS32BIT		0x2000

#define	NERESBYTES	9

#pragma pack(1)

typedef struct
{
    unsigned    ne_magic;
    char        ne_ver;
    char        ne_rev;
    unsigned    ne_enttab;
    unsigned    ne_cbenttab;
    long        ne_crc;
    unsigned    ne_flags;
    unsigned    ne_autodata;			/* dgroup */
    unsigned    ne_heap;
    unsigned    ne_stack;
    void far *  ne_csip;
    void far *  ne_sssp;
    unsigned    ne_cseg;
    unsigned    ne_cmod;
    unsigned    ne_cbnrestab;           /* size of nonres name table */
    unsigned    ne_segtab;
    unsigned    ne_rsrctab;             /* offset of resource table */
    unsigned    ne_restab;              /* offset of resident name table */
    unsigned    ne_modref;
    unsigned    ne_impname;
    long        ne_nonrestab;           /* offset of nres table from begin */
	unsigned	ne_cmovent;
	unsigned	ne_align;				/* segment align shift count */
	unsigned	ne_cres;
	unsigned char	ne_exetyp;
	char		ne_res[NERESBYTES];		/* pad struct to 64 bytes */
	/* more stuff, but we don't care */
} NEWEXE;

typedef struct {
	unsigned short ns_sector;
	unsigned short ns_cbseg;
	unsigned short ns_flags;
	unsigned short ns_minalloc;
	} SEG; 
	
typedef struct {
	unsigned char ep_flags;
	unsigned short ep_int_3f;
	unsigned char ep_segnum;
	unsigned short ep_offset;
	} ENTRY_MOVEABLE;

typedef	struct {
	unsigned char ep_flags;
	unsigned short ep_offset;
	} ENTRY_FIXED;
			
typedef struct {
	union {
		ENTRY_MOVEABLE ep_moveable;
		ENTRY_FIXED ep_fixed;
		} ep_union;
	} ENTRY_POINT;
	
typedef struct {
	unsigned char eb_nentry;
	unsigned char eb_segnum;
	} BUNDLE_HEADER;
	
typedef struct {
	BUNDLE_HEADER eb_hdr;
	ENTRY_POINT eb_epoint [1] ;
	} ENTRY_BUNDLE;
	
typedef struct {
	char nr_stype;
	char nr_flags;
	unsigned short nr_soff;
	union {
		struct {
			char nr_segno;
			char nr_res;
			unsigned short nr_entry;
			} nr_intref;
		struct {
			unsigned short nr_mod;
			unsigned short nr_proc;
			} nr_import;
		struct {
			unsigned short nr_ostype;
			unsigned short nr_osres;
			} nr_osfix;
		} nr_union;
	} RELOC_ITEM;
	
typedef struct {
	unsigned short seg, ofs;
	char *name;
	} ENTRY_SEGOFS;

#define NRSTYP		0x0f
#define NRSBYT		0x00
#define NRSSEG		0x02
#define NRSPTR		0x03
#define NRSOFF		0x05
#define NRSPTR48	0x0b
#define NRSOFF32	0x0d

#define NRADD		0x04
#define NRRTYP		0x03
#define NRRINT		0x00
#define NRRORD		0x01
#define NRRNAM		0x02
#define NRROSF		0x03

#define MOVEABLE	0xFF

#define IMPNAME(x)	(&impnam[(x)+1])
#define MODNAME(x)	(strupr(IMPNAME(modref[x-1])))
#define FUNCNAME(x)	(entry_segofs[x].name)
	
void failret(int ret, char *s)
{ 
	printf("exedump fail: %s\n", s);
	exit(ret); 
}

void fail(char *s) { failret(1, s); }

void banner(void)
{
#ifndef WINDOWS
	fputs("Segmented-Executable (NE) Mapper version 1.2\n", stderr);
	fputs("Copyright (c) 1991 Andrew Schulman.  All rights reserved.\n\n", 
		stderr);
#endif	
}
	
void usage()
{
	banner();
	puts("usage: exedump [-exports] [-noreloc] [-cvblanks] [-magic] <exe>");
	exit(1);
}
	
void newexe(char *name);
void print_reloc(unsigned seg, unsigned ofs, unsigned type,
	unsigned p1, unsigned p2, RELOC_ITEM *r, unsigned char *code);
void print_reloc_2(unsigned ofs, RELOC_ITEM *r, unsigned char *code);
void do_seg(int segnum, SEG *seg);
void do_segtab(void);
void do_enttab(void);
void do_modref(void);
void do_impnam(void);
void do_cv_sym(void);
void do_nres(void);
void do_resnam(void);
int ne_read(int f, void *p, unsigned long ne_offset, unsigned bytes);
int exe_read(int f, void *p, unsigned long exe_offset, unsigned bytes);
void *allocate(size_t x, size_t y);
void count2null(char *s, int len);

/* key global structures and variables */
static NEWEXE hdr = {0} ;
static ENTRY_SEGOFS *entry_segofs;	
static SEG *segtab;
static unsigned long start_newexe = 0;
static unsigned *modref;
static unsigned char *impnam;
static char *this_modname;
static unsigned max_ord;
static unsigned long align;
static int f;

/* command-line options */
static int verbose = 0;
static int export_only = 0;
static int no_reloc = 0;
static int do_all = 1;
static int no_banner = 0;
static int do_cv_blanks = 0;
static int do_type_only = 0;

/* arbitrary limits */
#define MAX_IMPNAME		8092
#define MAX_ENTRY		4096

main(int argc, char *argv[])
{
	int i;
	
	if (argc < 2)
		usage();
	
	for (i=1; i<argc; i++)
		if ((argv[i][0] == '-') || (argv[i][0] == '/'))
		{
			if (argv[i][1] == '?')
				usage();
			else if (strcmp(&argv[i][1], "verbose") == 0)
				verbose = 1;
			else if (strcmp(&argv[i][1], "exports") == 0)
			{
				export_only = 1;
				do_all = 0;
			}
			else if (strcmp(&argv[i][1], "noreloc") == 0)
				no_reloc = 1;
			else if (strcmp(&argv[i][1], "nobanner") == 0)
				no_banner = 1;
			else if (strcmp(&argv[i][1], "cvblanks") == 0)
				do_cv_blanks = 1;
			else if (strcmp(&argv[i][1], "magic") == 0)
			{
				no_banner = 1;
				do_type_only = 1;
			}
			else
				usage();
		}
		else
		{
			if (! no_banner)
				banner();
			newexe(argv[i]);
			exit(0);
		}
	return 0;
}

void newexe(char *name)
{
	SEG *seg;
    unsigned w;
	unsigned i;
	
    if ((f = open(name, O_RDONLY | O_BINARY)) == -1)
		fail("can't find file");
    read(f, &w, 2);
    if (w != EMAGIC)
		fail("not an executable");
    lseek(f, ENEWHDR, SEEK_SET);
    read(f, &start_newexe, 4);
    lseek(f, start_newexe, SEEK_SET);
    read(f, &hdr, sizeof(NEWEXE));
	
	if (do_type_only)
	{
		switch (hdr.ne_magic)
		{
			case LEMAGIC: puts("LE -- Linear Executable (Windows VxD)"); 
						  break;
			case LXMAGIC: puts("LX -- OS/2 2.0 Linear Executable");
						  break;
			case W3MAGIC: puts("W3 -- WIN386 (collection of LE files)");
				          break;
			case PEMAGIC: puts("PE -- Win32 NT Portable Executable");
						  break;
			case NEMAGIC: puts("NE -- New (Segmented) Executable");
						  break;
			default:      puts("MZ -- not a new executable");
						  break;
		}
		exit(0);
	}
	
	switch (hdr.ne_magic)
	{
		case LEMAGIC:
			failret(2, "Use LEDUMP to examine this file");
			break;
		case LXMAGIC:
			fail("sorry, this is an OS/2 2.0 linear executable, not "
				 "a segmented executable");
			break;
		case W3MAGIC:
			failret(3, "Use W3MAP to examine this file");
			break;
		case PEMAGIC:
			fail("use COFF -DUMP to examine this NT portable executable "
				 "(PE)");
			break;
		case NEMAGIC:
			// this is the one we want
			break;
		default:
			fail("not a segmented (new) executable");
	}
	
	align = 1 << hdr.ne_align;
	
	if (do_all)
	{
		printf("name %s\n", name);

		printf("%s\n", hdr.ne_flags & DLL_FLAG ? "library" : "program");

		printf("80%d instructions\n",
			(hdr.ne_flags & NEI386) ? 386 :
			(hdr.ne_flags & NEI286) ? 286 :
			(hdr.ne_flags & NEI086) ? 86 :
			/* default */			  286);

		printf("dgroup %u\n", hdr.ne_autodata);
		
		printf("start_csip %u:%04x\n", 
			FP_SEG(hdr.ne_csip),
			FP_OFF(hdr.ne_csip));

		do_enttab();
		do_segtab();
		do_modref();
		do_impnam();
		do_cv_sym();
	}
	else if (export_only)
	{
		printf(";\n; %s\n", name);
		do_enttab();
		exit(0);
	}

	if (! no_reloc)
	{
		printf("begin reloc\n");

		printf("\t%u\t%04x\t[e] start\n",	// program entry point
			FP_SEG(hdr.ne_csip),
			FP_OFF(hdr.ne_csip));

		for (i=0, seg=segtab; i<hdr.ne_cseg; i++, seg++)
			if (seg->ns_flags & NSRELOC)
				do_seg(i+1, seg);

		putchar('\n');
	}
	
	close(f);
}

void do_modref(void)
{
	modref = allocate(hdr.ne_cmod, sizeof(unsigned));
	ne_read(f, modref, hdr.ne_modref, hdr.ne_cmod * sizeof(unsigned));
}

void do_impnam(void)
{
	impnam = allocate(MAX_IMPNAME, 1);	
	ne_read(f, impnam, hdr.ne_impname, MAX_IMPNAME);
	count2null(impnam, MAX_IMPNAME);
}
		
void do_segtab(void)
{
	SEG *seg;
	unsigned i;
	
	segtab = allocate(hdr.ne_cseg, sizeof(SEG));
	ne_read(f, segtab, hdr.ne_segtab, hdr.ne_cseg * sizeof(SEG));
	
	printf("begin segtab\n");
	for (i=0, seg=segtab; i<hdr.ne_cseg; i++, seg++)
		printf("\t%u\t%lx\t%04x\t%s\t%s\n",
			i+1,
			align * seg->ns_sector,
			seg->ns_cbseg,
			((seg->ns_flags & NSTYPE) == NSCODE) ? "CODE" : "DATA",
			(seg->ns_flags & NS32BIT) ? "USE32" : "USE16");
	putchar('\n');
}

/**********************************************************************/
/* from Microsoft C Developer's Toolkit Reference, Chapter 3, 
   "Extended .EXE Format for Debug Information" */

typedef struct {
	unsigned char sign[4];
	unsigned long offset;
	} CV_HDR;
	
typedef struct {
	unsigned short SubSection;
	unsigned short iMod;
	signed long lfo;
	unsigned short cb;
	} DNT;
	
typedef struct {
	unsigned short seg;
	unsigned short ofs;
	unsigned short len;
	unsigned short overlay;
	unsigned short index;
	unsigned char  cseg;
	unsigned char  reserved;
	unsigned char  str[1];
	} MOD;
	
typedef struct {
	unsigned short ofs;
	unsigned short seg;
	unsigned short typ;
	unsigned char str[1];
	} PUB;
	
#define SSTMODULES		0x101
#define SSTPUBLIC		0x102
#define SSTTYPES		0x103
#define SSTSYMBOLS		0x104
#define SSTSRCLINES		0x105
#define SSTLIBRARIES	0x106
#define SSTCOMPACTED	0x107
#define SSTSRCLNSEG		0x108

void do_subsection(DNT *pdnt);

static unsigned long base;

void do_cv_sym(void)		// CodeView symbols
{
	CV_HDR cv_hdr;
	DNT *pdnt, *p;
	unsigned long filesize;
	unsigned long IfoDIR;
	unsigned short cdnt;
	int i;
	
	// go to last eight bytes of file, look for NB0 signature
	lseek(f, filelength(f) - 8, SEEK_SET);	// problems with SEEK_END
	read(f, &cv_hdr, sizeof(CV_HDR));
	if (memcmp(cv_hdr.sign, "NB0", 3) != 0)
		return;	// no CodeView debug information
	
	// go to CV info base address, look for NB0 signature again
	base = filelength(f) - cv_hdr.offset;
	lseek(f, base, SEEK_SET);
	read(f, &cv_hdr, sizeof(CV_HDR));
	if (memcmp(cv_hdr.sign, "NB0", 3) != 0)
		return;	// error finding second CV debug signature
	if (cv_hdr.offset == -1L)
		return;	// not CodeView:  Turbo Debug?
	
	printf("begin CodeView\n");

	// find the SubSection directory
	IfoDIR = base + cv_hdr.offset;
	lseek(f, IfoDIR, SEEK_SET);
	read(f, &cdnt, sizeof(cdnt));	// number of directory entries
		
	// read the dnt
	pdnt = allocate(cdnt, sizeof(DNT));
	read(f, pdnt, cdnt * sizeof(DNT));
	for (i=0, p=pdnt; i<cdnt; i++, p++)
		do_subsection(p);
	printf("\n");
	free(pdnt);
}

int allspace(char *s)
{
	while (*s)
		if (*s == ' ')
			s++;
		else
			return 0;
	/* still here */
	return 1;
}

void do_subsection(DNT *pdnt)
{
	MOD *pmod;
	PUB *ppub, *psav;
	char name[128];
	unsigned short offset;
	
	if (pdnt->SubSection == SSTMODULES)
	{
		pmod = allocate(1, pdnt->cb);
		lseek(f, base + pdnt->lfo, SEEK_SET);
		read(f, pmod, pdnt->cb);
		if (pmod->cseg > 1)
			;	// ok?
		strncpy(name, &pmod->str[1], pmod->str[0]);
		name[pmod->str[0]] = '\0';
		printf("\t%d\t%04x\t%04x\t%s\n", 
			pmod->seg, pmod->ofs, pmod->len, name);
		free(pmod);
	}
	else if (pdnt->SubSection == SSTPUBLIC)
	{
		ppub = psav = allocate(1, pdnt->cb);
		offset = 0;
		lseek(f, base + pdnt->lfo, SEEK_SET);
		read(f, ppub, pdnt->cb);
		while (offset < pdnt->cb)
		{
			strncpy(name, &ppub->str[1], ppub->str[0]);
			name[ppub->str[0]] = '\0';
			if (! allspace(name))	// don't print blank symbols!
				printf("\t%d\t%04x\t\t%s\n",
					ppub->seg, ppub->ofs, name);
			else if (do_cv_blanks)	// show blank symbols
				printf("\t\t\t\t[%d spaces]\n",
					strlen(name));
			offset += (sizeof(PUB) + ppub->str[0]);
			((unsigned char *) ppub) += (sizeof(PUB) + ppub->str[0]);
		}
		free(psav);
	}
}

/**********************************************************************/
	
int mangled_name(char *s)	// look for mangled names like "mmmmmm"
{
	int c = *s;
	char *s2 = s;
	int len = strlen(s);
	if (len < 3)
		return 0;
	while (*s2 == c)
		s2++;
	return (s + len == s2);
}
					
void do_enttab(void)
{
	ENTRY_BUNDLE *enttab = 0;
	ENTRY_BUNDLE *ep;
	ENTRY_SEGOFS *eso;
	unsigned eb;
	int ord = 0;
	int i;
	
	entry_segofs = allocate(MAX_ENTRY, sizeof(ENTRY_SEGOFS));
	eso = entry_segofs;

	enttab = allocate(hdr.ne_cbenttab, 1);	/* number BYTES in entry table */ 
	ne_read(f, enttab, hdr.ne_enttab, hdr.ne_cbenttab);
	for (eb=0, ep=enttab; ; )
	{
		BUNDLE_HEADER bhdr = ep->eb_hdr;
		if (ord >= MAX_ENTRY)
			fail("too many entry points");
		if (bhdr.eb_nentry == 0)		// end of table
			break;
		else if (bhdr.eb_segnum == 0)	// null entry
		{
			((unsigned char *) ep) += sizeof(BUNDLE_HEADER);
			ord += bhdr.eb_nentry;		// for ordinal numbers
			continue;
		}
		eb++;
		((unsigned char *) ep) += sizeof(BUNDLE_HEADER);
		if (bhdr.eb_segnum == MOVEABLE)
		{
			unsigned i;
			ENTRY_MOVEABLE *em;
			for (i=0, em=ep; i<bhdr.eb_nentry; i++, em++)
			{
				ord++;
				eso[ord].seg = em->ep_segnum;
				eso[ord].ofs = em->ep_offset;
			}
			((unsigned char *) ep) += 
				(bhdr.eb_nentry * sizeof(ENTRY_MOVEABLE));
		}
		else							// fixed
		{
			unsigned i;
			ENTRY_FIXED *ef;
			for (i=0, ef=ep; i<bhdr.eb_nentry; i++, ef++)
			{
				ord++;
				eso[ord].seg = bhdr.eb_segnum;
				eso[ord].ofs = ef->ep_offset;
			}
			((unsigned char *) ep) += 
				(bhdr.eb_nentry * sizeof(ENTRY_FIXED));
		}
	}
	max_ord = ord;
	free(enttab);
	
	eso = entry_segofs;

	/* walk non-resident name table */
	do_nres();
	
	/* walk resident names table */
	/* get module name from first entry */
	/* don't know size of resident names table! */
	do_resnam();
	
	if (do_all)
	{
		printf("modname %s\n\n", this_modname);
	
		printf("begin entry\n");
		for (i=1, eso=entry_segofs+1; i<=max_ord; i++, eso++)
			if (eso->seg)
			{
				printf("\t%u\t%04x\t%u", 
					eso->seg, eso->ofs, i);
				if (eso->name)
					printf("\t%-30s", eso->name);
				putchar('\n');
			}
		putchar('\n');
	}
	else if (export_only)
	{
		for (i=1, eso=entry_segofs+1; i<=max_ord; i++, eso++)
			if (eso->seg && eso->name)
			{
#ifdef WINDOWS
				printf("%s . %u\t%s\n", this_modname, i, eso->name);
#else
				if (mangled_name(eso->name))
					fprintf(stderr, "Warning: mangled name? "
									"%s.%u: \"%s\"\n",
								this_modname, i, eso->name);
				else
					printf("%s . %u\t%s\n", this_modname, i, eso->name);
#endif				
			}
	}
}

void do_nres(void)
{
	ENTRY_SEGOFS *eso = entry_segofs;
	unsigned char *nres = allocate(hdr.ne_cbnrestab, 1);
	unsigned char *nr;
	unsigned i;
	unsigned char c;
	unsigned ndx;

	exe_read(f, nres, hdr.ne_nonrestab, hdr.ne_cbnrestab);
	
	for (i=0, nr=nres; i<hdr.ne_cbnrestab; )
	{
		if (! (c = nr[0]))
			break;
		
		ndx = *((unsigned *) &nr[c+1]);		// ordinal number
		nr[c+1] = '\0';		// terminate string, replacing ord#
			
		if ((ndx == 0) && nr[1])
		{
			if (do_all)
				printf("description \"%s\"\n", &nr[1]);
			else if (export_only)
				printf("; \"%s\"\n;\n", &nr[1]);	// show desc as comment
		}
		else if (ndx < MAX_ENTRY)
		{
			if (eso[ndx].name && eso[ndx].name[0])	// already there!
				if (ndx > 1)
					printf(";; conflict on %u - %s\n", ndx, eso[ndx].name);
			if (! isprint(nr[1]))
				fail("internal error: bad nres name!");
			eso[ndx].name = &nr[1];
		}
		else if (&nr[1] && nr[1])	// internal error: ordnum too big
			printf(";; discarding nres %u - %s\n", ndx, &nr[1]);
		
		nr += (c+3);
		i += (c+3);
	}
}

void do_resnam(void)
{
	ENTRY_SEGOFS *eso = entry_segofs;
	unsigned res;
	unsigned char c;
	unsigned char *s;
	unsigned ndx;
	
	for (res=hdr.ne_restab;;)
	{
		ne_read(f, &c, res, 1);
		if (c == 0)
			break;
		s = allocate(c+2, 1);
		ne_read(f, s, ++res, c+2);
		ndx = *((unsigned *) &s[c]);		// WORD, not a BYTE!
		s[c] = '\0';			// now turn ord# into NULL terminate
		if (ndx == 0)			// module name
			this_modname = s;
		else if (ndx < MAX_ENTRY)
		{
			if (eso[ndx].name && eso[ndx].name[0])	// if slot already used
				printf(";; conflict on %u - %s\n", ndx, eso[ndx].name);
			eso[ndx].name = s;
		}
		else if (s && *s)	// ordnum too big
			printf(";; discarding resnam %u - %s\n", ndx, s);
		res += (c+2);
	}
}

/**********************************************************************/

void do_seg(int segnum, SEG *seg)
{
	/* multiply seg->ns_sector by hdr.ne_align to get _file_ offset; 
	   _not_ NE offset */
	unsigned long reloc = (align * seg->ns_sector) + seg->ns_cbseg;
	unsigned char *code;
	RELOC_ITEM *rel;
	unsigned short nr_nreloc;
	unsigned i;

	exe_read(f, &nr_nreloc, reloc, sizeof(unsigned short));
	if (nr_nreloc == 0)
		return;
	if (nr_nreloc > (0xFFFFU / sizeof(RELOC_ITEM)))
	{
		printf("; seg %d -- compressed relocation records?\n", segnum);
		exit(1);
	}
	
	code = allocate(seg->ns_cbseg, 1);
	exe_read(f, code, align * seg->ns_sector, seg->ns_cbseg);
	rel = allocate(nr_nreloc, sizeof(RELOC_ITEM));
	exe_read(f, rel, reloc+2, nr_nreloc * sizeof(RELOC_ITEM));
	for (i=0; i<nr_nreloc; i++)
	{
		RELOC_ITEM *r = &rel[i];
		void far *fp = *((void far **) &code[r->nr_soff]);

		/*
			relocation prefix:
				[i]		module.funcname
				[m] 	module.ordinal	
				[n]		funcname // thismod.ord
				[s] 	segment:offset
		*/

		switch (r->nr_flags & 0x03)
		{
			case NRRINT:
			{
				int ordnum = r->nr_union.nr_intref.nr_entry;
				int segno = r->nr_union.nr_intref.nr_segno;
				if (segno == (char) MOVEABLE)
				{
					if (FUNCNAME(ordnum))
					{
						print_reloc(segnum, r->nr_soff, 
							'n', 
							0,
							ordnum,
							r,
							code);
					}
					else
					{
						print_reloc(segnum, r->nr_soff, 
							'm', 
							-1,		/* this_modname */
							ordnum,
							r,
							code);
					}
				}
				else	// fixed
				{
					print_reloc(segnum, r->nr_soff, 
						's', 
						segno,
						ordnum,
						r,
						code);
				}
			}
				break;
			case NRRORD:
				print_reloc(segnum, r->nr_soff, 
					'm', 
					r->nr_union.nr_import.nr_mod,
					r->nr_union.nr_import.nr_proc,
					r,
					code);
				break;
			case NRRNAM:
				print_reloc(segnum, r->nr_soff, 
					'i', 
					r->nr_union.nr_import.nr_mod,
					r->nr_union.nr_import.nr_proc,
					r,
					code);
				break;
			case NRROSF:
				/* do nothing: ok as is */
				break;
		}

		/* follow possible relocation chain */
		if (! (r->nr_flags & NRADD))
		if (FP_SEG(fp) == 0)
		{
			while (FP_OFF(fp) != 0xFFFF)
			{
				print_reloc(segnum, FP_OFF(fp), -1, 0, 0, 0, code);
				fp = *((void far **) &code[FP_OFF(fp)]);
				if (fp == 0)
					fail("got zero!");
			}
		}
	}

	free(code);
	free(rel);
}

unsigned in_entry_segofs(unsigned seg, unsigned ofs)
{
	unsigned i;
	ENTRY_SEGOFS *eso;
//	printout helps establish that only called when necessary
//	printf("in_entry_so %04x:%04x\n", seg, ofs);
	for (i=1, eso=entry_segofs+1; i<=max_ord; i++, eso++)
		if ((seg == eso->seg) && (ofs == eso->ofs))
			return i;
	/* still here */
	return 0;
}
		
void print_reloc(unsigned seg, unsigned ofs, unsigned type,
	unsigned p1, unsigned p2, RELOC_ITEM *r, unsigned char *code)
{
	static unsigned s_type, s_p1, s_p2;
	static RELOC_ITEM *s_r;
	unsigned ord;
	
	if (type == (unsigned) -1)
	{	/* use same as last time, except for seg and ofs */
		type=s_type; p1=s_p1; p2=s_p2; r=s_r;
	}
	else
	{	/* new relocation chain	*/  
		s_type=type; s_p1=p1; s_p2=p2; s_r=r;
	}
	
	if (type == 's')
	{
		if (ord = in_entry_segofs(p1, p2))
		{
			// turning [s] into [n]
			print_reloc(seg, ofs, 'n', 0, ord, r, code);
			return;		// NB! don't print reloc type again!
		}
		else
			printf("\t%u\t%04x\t[%c] %u %04x", seg, ofs, type, p1, p2);
	}
	else
		printf("\t%u\t%04x\t[%c] ", seg, ofs, type);

	if (type == 'i')
		printf("%s %s", MODNAME(p1), IMPNAME(p2));
	else if (type == 'm')
		printf("%s %u", 
			(p1 == 0xFFFF) ? this_modname : MODNAME(p1), p2);
	else if (type == 'n')
		printf("%s /* %s %u */", FUNCNAME(p2), this_modname, p2);
	
	putchar('\t');
	print_reloc_2(ofs, r, code);
	putchar('\n');
}

void print_reloc_2(unsigned ofs, RELOC_ITEM *r, unsigned char *code)
{
	char *typ;
	
	// display relocation type -- crucial for WINPP!
	switch (r->nr_stype)
	{
		case NRSPTR:	typ = "PTR";	break;
		case NRSBYT:	typ = "BYT";	break;
		case NRSSEG:	typ = "SEG";	break;	// BASE
		case NRSOFF:	typ = "OFF";	break;
		case NRSPTR48:	typ = "P48";	break;
		case NRSOFF32:	typ = "O32";	break;
		default:		typ = "???";	break;
	}
	printf("%s", typ);
	
	if (r->nr_flags & NRADD)
	{
		void *cod = &code[ofs];
		switch (r->nr_stype)
		{
			case NRSPTR:	
			case NRSOFF32:
				printf(" ADD+%08lx", (*(unsigned long *) cod)); break;
			case NRSBYT:	
				printf(" ADD+%02x", (*(unsigned char *) cod)); break;
			case NRSSEG:
			case NRSOFF:
				if (*((unsigned *) cod) != 0)	// don't print if ADD+0000
					printf(" ADD+%04x", (*(unsigned *) cod)); break;
			case NRSPTR48:
				printf(" ADD+%04x:", (*(unsigned *) cod));
				printf("%08lx", (*(unsigned long *) cod+2)); break;
			default:
				printf(" ADD????"); break;
		}
	}
}
/**********************************************************************/

/* read at offset from NE */
int ne_read(int f, void *p, unsigned long ne_offset, unsigned bytes)
{
    lseek(f, start_newexe + ne_offset, SEEK_SET);
	return read(f, p, bytes);
}

/* read at offset from beginning of file */
int exe_read(int f, void *p, unsigned long exe_offset, unsigned bytes)
{
    lseek(f, exe_offset, SEEK_SET);
	return read(f, p, bytes);
}

void *allocate(size_t x, size_t y)
{
	void *p;
	if (! x) x = 1;
	if (! y) y = 1;
	if (((long) x * (long) y) > 0xFFFFUL)
		fail("internal error");
	if ((p = calloc(x, y)) == 0)
		fail("insufficient memory");
	return p;
}

/* in place, turn counted strings into null-terminated strings */
void count2null(char *s, int len)
{
	unsigned char *s2; 
	int i;
	for (i=0, s2=s; i<len; )
	{
		int len = *s2;	// get string length
		*s2 = '\0';		// replace with NULL
		i += (len+1);	// increment ptr
		s2 += (len+1);
	}
	*s2 = '\0';			// get last one -- table must include room for it
}
