/* 
FAKETSR.C -- for FAKEWIN
from Schulman, Unauthorized Windows, 1994


Some information on 2F/160B in Matt Pietrek, "Windows Internals,"
chapter 3.  Figured out rest by disassembling AutoLoadTSRApps()
internal routine in USER.EXE, which calls 2F/160B via DPMI 31/0300
(real-mode call).  Input: ES:DI=0; Output: TSr info struct in ES:DI.

See David Long article on MSDN CD for doc
*/

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef void far *FP;

#pragma pack(1)

typedef struct {
	WORD size;
	BYTE str[1];
	} TSR_ID_Block_Struc;

typedef struct _TSRINFO {
	struct _TSRINFO far *TSR_Next;
	WORD TSR_PSP_Segment;
	WORD TSR_API_Ver_ID;	/* 0x100 */
	WORD TSR_Exec_Flags;
	WORD TSR_Exec_Cmd_Show;
	char far *TSR_Exec_Cmd;
	DWORD TSR_Reserved;
	TSR_ID_Block_Struc far *TSR_ID_Block;
	FP TSR_Data_Block;
} TSR_Info_Struc;

#pragma pack()

// TSR_Exec_Flag equates
#define TSR_WINEXEC			1
#define TSR_LOADLIBRARY		2
#define TSR_OPENDRIVER		4

// TSR_Exec_Cmd_Show equates
#define SW_HIDE				0
#define SW_SHOWNORMAL		1
#define SW_SHOWMINIMIZED	2
#define SW_SHOWMAXIMIZED    3
#define SW_SHOWNOACTIVATE   4
#define SW_SHOW             5
#define SW_MINIMIZE         6
#define SW_SHOWMINNOACTIVE  7
#define SW_SHOWNA           8
#define SW_RESTORE          9

char *sw_str[SW_RESTORE+1] = {
	"SW_HIDE", "SW_SHOWNORMAL", "SW_SHOWMINIMIZED", "SW_SHOWMAXIMIZED",
	"SW_SHOWNOACTIVATE", "SW_SHOW", "SW_MINIMIZE", "SW_SHOWMINNOACTIVE",
	"SW_SHOWNA", "SW_RESTORE"
	} ;

// in FAKEWIN.C
extern char *owner(FP fp);

void print_tsr_info(TSR_Info_Struc far *tsr_info)
{
	if (tsr_info->TSR_API_Ver_ID < 0x100)
		printf("%Fp not a valid TSR structure!\n", tsr_info);
	printf("\nTSR_Info_Struc at %Fp (PSP %04Xh) %s\n", 
		tsr_info, tsr_info->TSR_PSP_Segment, owner(tsr_info));
	if (tsr_info->TSR_Exec_Cmd)
	{
		switch (tsr_info->TSR_Exec_Flags)
		{
			case TSR_WINEXEC:		
				printf("TSR_WINEXEC ");
				// Exec_Cmd_Show only used for TSR_WINEXEC
				if (tsr_info->TSR_Exec_Cmd_Show <= SW_RESTORE)
					printf("%s ", sw_str[tsr_info->TSR_Exec_Cmd_Show]);
				else
					printf("TSR_Exec_Cmd_Show: %04Xh ", 
						tsr_info->TSR_Exec_Cmd_Show);
				break;
			case TSR_LOADLIBRARY:	printf("TSR_LOADLIBRARY "); break;
			case TSR_OPENDRIVER:	printf("TSR_OPENDRIVER "); break;
			default:				printf("TSR_Exec_Flags %04Xh ",
				tsr_info->TSR_Exec_Flags); break;
		}
		printf("\n\"%Fs\"\n", tsr_info->TSR_Exec_Cmd);
	}
	if (tsr_info->TSR_ID_Block)
		printf("TSR_ID_Block: \"%Fs\"\n", tsr_info->TSR_ID_Block->str);
	if (tsr_info->TSR_Data_Block)
		printf("TSR_Data_Block: %Fp\n", tsr_info->TSR_Data_Block);
}

void tsr_identify(void)
{
    TSR_Info_Struc far *tsr_info;
    _asm push di
    _asm xor cx, cx
    _asm mov es, cx
    _asm mov di, cx
    _asm mov ax, 160bh
    _asm int 2Fh
    _asm mov word ptr tsr_info+2, es
    _asm mov word ptr tsr_info, di
    _asm pop di
    if (tsr_info != 0)
	{
		do {
			print_tsr_info(tsr_info);
			tsr_info = tsr_info->TSR_Next;
		} while (tsr_info != 0);
	}
}

