/*
FAKEVXD.C
Fake DOSMGR VxD callout used with FAKEWIN.C
from Schulman, Unauthorized Windows, 1994
*/

#include <stdlib.h>
#include <stdio.h>

typedef unsigned short WORD;

#define VXD_CALLOUT(vxd_id) \
{ \
    _asm { mov bx, vxd_id } ; \
    _asm { mov ax, 1607h; } ; \
    _asm { int 2fh; } ; \
}

#define DOSMGR_CALLOUT(func) \
{ \
    _asm { mov cx, func } ; \
    VXD_CALLOUT(0x15); \
}

#define NUM_PATCH	6

// names from Microsoft document "API to Identify MS-DOS Instance Data"
char *patch_str[] = {
	"SAVEDS", "SAVEBX", "INDOS", "USER_ID", "CRITPATCH", "UMB_HEAD",
	} ;

void fake_dosmgr_callouts(void)
{
    WORD w, w2, w3;
    WORD far *patchtab;
    WORD far *patch;
	int i;
    
    DOSMGR_CALLOUT(0);      // query instance processing
    _asm mov w, cx
    if (w != 0)
    {
        _asm mov w, dx
        _asm mov word ptr patchtab+2, es
        _asm mov word ptr patchtab, bx
			
        printf("\nDOSMGR instance interface ON\n");
        if (w != 0)
            printf("Segment of DOS drivers: %04X\n", w);    // usually 0
				
        printf("Patch table: %Fp\n", patchtab);
		printf("DOS version %u.%02u\n", 
			(patchtab[0] & 0xFF), (patchtab[0] >> 8));
		
		for (i=0, patch = &patchtab[1]; i<NUM_PATCH; i++, patch++)
			printf("   %04X (%s)\n", *patch, patch_str[i]);
        
        // probably not safe to try DOSMGR_CALLOUT(1), except
        // perhaps if immediately followed by DOSMGR_CALLOUT(2)
            
        _asm mov dx, 1
        DOSMGR_CALLOUT(3);      // get size of DOS data structure
        _asm mov w, ax
        _asm mov w2, cx
        _asm mov w3, dx
        if ((w != 0xB97C) && (w3 != 0xa2ab))
            printf("DOSMGR callout 3 failed: signature wrong!\n"); // AX:DX
        else
            printf("Current Directory Structure = %u bytes\n", w2); // CX
                
        DOSMGR_CALLOUT(4);  // determine instanced data structures
        _asm mov w, ax
        _asm mov w2, bx
        _asm mov w3, dx
		if ((w3 == 0) || (w2 == 0))
			printf("No DOS data structures instanced via DOSMGR API\n");
        else if ((w != 0xB97C) && (w3 != 0xa2ab))
            printf("DOSMGR callout 4 failed: signature wrong!\n"); // AX:DX
        else
        {
			printf("DOS data structures instanced: %04X\n", w2);
			if (w2 & 1)     printf("   CDS\n");
			if (w2 & 2)     printf("   SFT\n");
			if (w2 & 4)     printf("   Device chain\n");
			if (w2 & 8)     printf("   SDA\n");
        }
    }
    else
        printf("\nNo DOSMGR instance interface\n");
}
