/*
FAKEWIN.C
from Schulman, Unauthorized Windows, 1994

bcc -2 -P- -c iswin.c dump.c map.c
tlib ..\lib\unauthwr-+iswin-+dump-+map
bcc -2 -DTEST_1684 -P- -L..\lib fakewin.c fakevxd.c faketsr.c fakeemm.c \
test1684.c unauthwr.lib

Using INT 2Fh broadcasts, FAKEWIN pretends to be Windows.  It
simulates Windows initializing and terminating in order to find out
how DOS and various device drivers and TSRs want to affect Windows
startup.  For each Win386_Startup_Info_Struc, FAKEWIN prints out 
the name of any Windows virtual device driver (VxD) that has been
asked to be loaded, plus the addresses and sizes of any instance
data items.

Can run with -WIN30 and/or -STANDARD switch to be fake Windows 3.0
and/or fake Standard mode.  
-DUMP switch does hex dump of instanced data areas.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef void far *FP;

#pragma pack(1)

typedef struct {
    FP IIS_Ptr;
    WORD IIS_Size;
    } Instance_Item_Struc;

// the Task Switcher Identify Instance Data (2F/4B05)
// function uses the same structure
typedef struct _WINFO {
    WORD SIS_Version;       // should be 3
    struct _WINFO far *SIS_Next_Dev_Ptr;
    DWORD SIS_Virt_Dev_File_Ptr;
    DWORD SIS_Reference_Data;
    Instance_Item_Struc far *SIS_Instance_Data_Ptr;
    } Win386_Startup_Info_Struc, far *FPWININFO;

#pragma pack()	

static FPWININFO fpinfo;
static void (far *switch_func)(void);

#define ENHANCED_MODE       0
#define STANDARD_MODE       1

#define WIN30               0x0300
#define WIN31               0x030A

// in FAKEVXD.C
extern void fake_dosmgr_callouts(void);

// in FAKETSR.C
extern void tsr_identify(void);

// in FAKEEMM.C
extern void check_emm_import(void);

// in DUMP.C
extern void dump(unsigned char far *fp, unsigned bytes, 
    char *mask, unsigned long addr, int width);

// in ISWIN.C
extern int is_win(int *pmaj, int *pmin, int *pmode);

// in MAP.C
extern char *find_owner(DWORD lin_addr);

#define MK_LIN(fp)		((((DWORD) FP_SEG(fp)) << 4) + FP_OFF(fp))

char *owner(FP fp)
{
	static char buf[32];
	char *s = find_owner(MK_LIN(fp));
	if (s) sprintf(buf, "(%s)", s);
	else buf[0] = '\0';
	return buf;
}

WORD win_init_notify(WORD vers, WORD mode)
{
    WORD retval;
    WORD handler_ds;
    _asm push si
    _asm push di
    _asm push ds
    _asm xor bx, bx
    _asm mov es, bx
    _asm mov cx, bx
    _asm mov dx, mode
    _asm mov di, vers
    _asm xor si, si
    _asm mov ds, si
    _asm mov ax, 1605h
    _asm int 2fh
    _asm mov retval, cx
    _asm mov handler_ds, ds
    _asm pop ds
    _asm cmp cx, 0
    _asm jne no_init
init:
    _asm mov word ptr fpinfo+2, es
    _asm mov word ptr fpinfo, bx
    _asm mov bx, handler_ds
    _asm mov word ptr switch_func+2, bx
    _asm mov word ptr switch_func, si
done:
    _asm pop di
    _asm pop si
    return retval;
no_init:
    goto done;
}

#define V86_TO_REAL		0
#define REAL_TO_V86		1

// FAKEEMM.C will call to get Global EMM Import data
void call_v86_switch(int flag)
{
	_asm cli
	_asm push ds
	_asm push es
	_asm pusha
	_asm mov ax, word ptr flag
	_asm call dword ptr switch_func
	_asm popa
	_asm pop es
	_asm pop ds
		
}

void win_term_notify(WORD mode)
{
    _asm mov dx, mode
    _asm mov ax, 1606h
    _asm int 2fh
}

void win_init_complete_notify(void)
{
    _asm mov ax, 1608h
    _asm int 2fh
}

void win_begin_exit_notify(void)
{
    _asm mov ax, 1609h
    _asm int 2fh
}

static int do_hex_dump = 0;

void print_startup_info(FPWININFO winfo)
{
    Instance_Item_Struc far *inst;
    if (winfo->SIS_Version >= 3)	// try to be forward compatible
    {
        printf("\nWin386_Startup_Info_Struc at %Fp %s\n", 
			winfo, owner(winfo));
        if (winfo->SIS_Virt_Dev_File_Ptr != 0)
        {
            printf("VxD name: %Fs   ", winfo->SIS_Virt_Dev_File_Ptr);
            printf("(Reference data: %08lX)\n", winfo->SIS_Reference_Data);
        }
        if ((inst = winfo->SIS_Instance_Data_Ptr) != 0)
        {
            printf("Instance_Item_Struc at %Fp %s\n", 
				inst, owner(inst));
            while (inst->IIS_Ptr)
            {
                printf("   %Fp   %04X %s\n", 
					inst->IIS_Ptr, inst->IIS_Size, 
					owner(inst->IIS_Ptr));
                if (do_hex_dump)
                    dump((unsigned char far *) inst->IIS_Ptr, 
                         (unsigned) inst->IIS_Size, "%Fp",
                         (unsigned long) inst->IIS_Ptr, 16);
                inst++;
            }
        }
        else
            printf("No instance data\n");
        
        // tail-recursively walk chain of startup info blocks
        if (winfo->SIS_Next_Dev_Ptr != 0)
            print_startup_info((FPWININFO) winfo->SIS_Next_Dev_Ptr);
    }
    else
        printf("%Fp not a valid Win386 startup info structure!\n", winfo);
}

void fail(const char *s, ...) { puts(s); exit(1); }

static char *usage_msg = 
    "usage: fakewin [-standard] [-win30 | -vers xxxx] [-dump]";

void usage(void) { fail(usage_msg); }

main(int argc, char *argv[])
{
    WORD mode = ENHANCED_MODE;  // default
    WORD vers = WIN31;          // default
    char *s;
    int dummy;
    int i;
        
fputs("FAKEWIN 1.0 -- Simulate Windows Initialization\n", stderr);
fputs("Displays instance data, VxD startup, DOSMGR interface\n", stderr);
fputs("Copyright (c) Andrew Schulman 1993.  All rights reserved\n\n", stderr);
        
    for (i=1; i<argc; i++)
        if ((argv[i][0] == '-') || (argv[i][0] == '/'))
        {
            s = strupr(argv[i]) + 1;
            if (strcmp(s, "STANDARD") == 0)        mode = STANDARD_MODE;
            else if (strcmp(s, "WIN30") == 0)      vers = WIN30;
            else if (strcmp(s, "VERS") == 0) sscanf(argv[++i], "%04X", &vers);
            else if (strcmp(s, "DUMP") == 0)       do_hex_dump++;
            else                                   usage();
        }
        else
            usage();
    
    if (is_win(&dummy, &dummy, &dummy))
        fail("Already running under Windows\n"
             "Exit back to DOS before running FAKEWIN");
        
    printf("FAKEWIN pretending to be Windows %u.%02u %s mode\n",
        (vers >> 8), (vers & 0xFF), 
        (mode == ENHANCED_MODE) ? "Enhanced" : "Standard");
         
    // tell DOS/whoever that Windows is starting up
    if (win_init_notify(vers, mode) == 0)
    {
        // start recursive walk of info chain
        if (fpinfo == 0)
            printf("\nNo Windows startup info\n");
		else
            print_startup_info(fpinfo);
		
		if (switch_func)
		{
			printf("\nV86 Enable/Disable function: %Fp %s\n", 
				switch_func, owner(switch_func));
			
			check_emm_import();
		}
    }
    else	// someone refused the 2F/1605 call
        fail("Not allowed to start Windows");
    
    // do device callouts -- VxD init order defined in VMM.INC
    fake_dosmgr_callouts();

	// do TSR identify call (2F/160B)
	tsr_identify();
        
    // tell DOS/whoever that Windows is done initializing
    win_init_complete_notify();
        
#ifdef TEST_1684
{
	extern void test_1684(void);	// in TEST1684.C
	test_1684();
}
#endif

    // tell DOS/whoever that Windows is beginning termination
    win_begin_exit_notify();
        
    // tell DOS/whoever that Windows is terminating
    win_term_notify(mode);
        
    return 0;
}
