/*
FAT.C -- Given drive and cluster number, print FAT chain 
Andrew Schulman, July 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <dos.h>
#include "diskstuf.h"

void fail(const char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    DPB far *dpb;
    WORD prev_cluster, cluster, num_clusters = 0, num_groups = 0;
	WORD start_cluster;
    int drive;
    
    if (argc < 3)
        fail("usage: fat [drive] [cluster]");

    drive = toupper(argv[1][0]) - 'A';
    cluster = atoi(argv[2]);
    
    if (! (dpb = get_dpb(drive + 1)))
        fail("can't get DPB");
	
	start_cluster = cluster;
    while (cluster < 0xFFF0)    // EOF
    {
        // printf("%u ", cluster);
        num_clusters++;
        prev_cluster = cluster;
        cluster = get_fat_entry(drive, dpb, cluster);
		if (cluster != prev_cluster + 1)
        {
			int num_clust = prev_cluster + 1 - start_cluster;
            num_groups++;
			if (num_clust > 1)
				printf("%u-%u (%u)\n", 
					start_cluster, prev_cluster, num_clust);
			else	// only one cluster
				printf("%u\n", start_cluster);
			start_cluster = cluster;
        }
    }
    printf("\n%u clusters in %u groups\n", num_clusters, num_groups);
    if (num_clusters > 2 && num_groups > 1)
        printf("(%u%% fragmentation)\n", (num_groups * 100) / num_clusters);
    return 0;
}

