/* 
FHANDLE.C
Alternative to using INT 21h Function 67h (added in DOS 3.3)
Andrew Schulman, revised July 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)

bcc fhandle.c
fhandle 40
fhandle 40 -testing 

NOTES:
-- Generally need FILES= greater than 20 in CONFIG.SYS to have any
   effect. Actually, this isn't strictly necessary.  Certainly you
   could have a larger JFT than your SFT, and have multiple JFT
   entries pointing to the same SFT entry.  This occurs already,
   where normally file handles 0, 1, and 2 all correspond to
   SFT entry 0 (CON).  But this generally isn't why programmers want
   to increase their JFT size!  So, if your intended JFT size is
   larger than the current SFT size, you can:
-- Change FILES= in CONFIG.SYS and reboot.
-- Or, can use Quarterdeck FILES.COM program to increase SFT size.
-- Or, see XLASTDRV program in Chapter 2, which expands CDS, and adopt
   it to expand the SFT chain.

-- If you want to open >20 files simultaneously with high-level function
   like fopen(), probably have to increase run-time library tables. For
   example, in Microsoft C the size of table for FILE* is hard-wired
   to 20 in /msc/source/startup/_file.c (#define _NFILE_ 20). So
   would have to increase size of this table as well (and recompile
   startup code) if wanted to use >20 fopen() at once, even after
   using the code below.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned WORD;
typedef unsigned long DWORD;
typedef BYTE far *FP;

#ifndef MK_FP
#define MK_FP(seg,ofs)  ((FP)(((DWORD)(seg) << 16) | (ofs)))
#endif

extern unsigned files(void);    // in COUNTF.C

void fail(char *s) { puts(s); exit(1); }

typedef struct _sft {
    struct _sft far *next;
    WORD num;
    // other stuff not used here
    } SFT;

WORD files(void)
{
    SFT far * far *sysvars;	// treat SysVars as array of far ptrs
    SFT far *sft;
    WORD files = 0;
    
	_asm mov ah, 52h
	_asm int 21h
	_asm mov word ptr sysvars+2, es
	_asm mov word ptr sysvars, bx
		
	if (! sysvars) return 0;
	for (sft = sysvars[1]; FP_OFF(sft) != 0xFFFF; sft = sft->next)
		files += sft->num;
	return files;
}

main(int argc, char *argv[])
{
    WORD far *pmax = (WORD far *) MK_FP(_psp, 0x32);
    BYTE far * far *pjft = (BYTE far * far *) MK_FP(_psp, 0x34);
    BYTE *new_jft;
    WORD max = *pmax;
    WORD new_max = atoi(argv[1]);
    int f, i;
    
    printf("Currently %u max JFT file handles\n", max);
	printf("And SFT FILES=%u\n", files());
    
    if (new_max <= max) 		
		fail("nothing to do");
    else if (new_max > files()) 
		fail("FILES= too low: edit CONFIG.SYS and reboot\n"
		     "or run a program like QEMM FILES.COM to grow the SFT");

    if (! (new_jft = (BYTE *) malloc(new_max))) 
		fail("insufficient memory");
	_fmemcpy(new_jft, *pjft, max);			    // copy over old entries
	_fmemset(new_jft+max, 0xFF, new_max - max); // fill in new entries
    *pmax = new_max;						    // set new max file handles
    *pjft = (BYTE far *) new_jft;			    // set new JFT!
    
    printf("Max file handles increased to %u\n", new_max);

    // now test how many files we can open by opening ourselves (argv[0])
	#define OPEN_SHARE_DENY_NONE 0x40
    for (i=0; ; i++)
        if (_dos_open(argv[0], OPEN_SHARE_DENY_NONE, &f) != 0)
            break;
    printf("Opened %d files\n", --i);

	if (argc > 2 && strcmp(strupr(argv[2]), "-TESTING") == 0)
	{
		BYTE cmd[16];
		printf("Closing %d ==> %d\n", f, new_jft[f]);
		_dos_close(f);  // close last one so we can spawn shell!
		sprintf(cmd, "H2NAME %04X", _psp);
		system(cmd);
		system("FILES");
	}
    
    return 0;
}

