/* 
FILES.C -- list all files in system file table
Version No. 3 (23 March 1991)
Author: Andrew Schulman (CIS 76320,302)

Microsoft C/C++:
    cl files.c

Borland C++:
    bcc files.c
		
protected-mode Windows:
	// show instructions!
        
This version is substantially different from that published in 
the first edition of UNDOCUMENTED DOS.
    -- includes DOS 3.0 file structure (USHORT dir_entry, not BYTE)
    -- allows PSP +- 1 (some DOS versions store PSPowner incorrectly!)
    -- fixes bug in for loops
    -- fixes test for AUX, CON, PRN (e.g. check for "AUX " not "AUX")
    -- split filename[11] into filename[8], ext[3]
    -- DOS 2.x, 3.0, 3.1+ structs combined into union, with access
       macros. This points up the relative inflexibility of structures:
       struct->field just turns into an offset, but at compile-time,
       with no control over changing field offsets at run-time.
    -- on suggestion of Geoff Chappell, got rid of AUX-CON "sanity check," 
	   which was insane 
            
To do:
	-- Print out directory!!!!
	-- Print out open/sharing modes
	-- Print drive number (inside dev_info field in DOS 3+)
	-- Find extra SFTs in Windows!! (Pat Byrne code)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#ifdef WINIO
#define WINDOWS
#include "windows.h"
#include "winio.h"
#include "prot.h"
#endif

typedef unsigned char BYTE;
typedef unsigned short USHORT;
typedef unsigned long ULONG;
typedef BYTE far *FP;

#pragma pack(1)

typedef struct file2 {   
    BYTE num_handles, open_mode;
    BYTE fattr;     // might not be accurate??
    BYTE drive;
    BYTE filename[8], ext[3];
    USHORT unknown1, unknown2;
    ULONG fsize;
    // ...
    // NOTE! no owner_psp!
    } file2; // for DOS 2.x
    
typedef struct file30 {   
    USHORT num_handles, open_mode;
    BYTE fattr;
    USHORT dev_info;    // includes drive number
    FP ptr;
    USHORT start_cluster, time, date;
    ULONG fsize, offset;
    USHORT rel_cluster, abs_cluster, dir_sector;
    USHORT dir_entry;    // only difference from file31: USHORT, not BYTE
    BYTE filename[8], ext[3];
    ULONG share_prev_sft;
    USHORT share_net_machine;
    USHORT owner_psp;
    // ...
    } file30; // for DOS 3.0 only
    
typedef struct file31 {   
    USHORT num_handles, open_mode;
    BYTE fattr;
    USHORT dev_info;    // includes drive number
    FP ptr;
    USHORT start_cluster, time, date;
    ULONG fsize, offset;
    USHORT rel_cluster, abs_cluster, dir_sector;
    BYTE dir_entry;
    BYTE filename[8], ext[3];
    ULONG share_prev_sft;
    USHORT share_net_machine;
    USHORT owner_psp;
    // ...
    } file31; // for DOS 3.1+
        
typedef union file {
    file2 f2;
    file30 f30;
    file31 f31;
    } file;
    
#define DOSVER_FIELD(pf,x) \
    (_osmajor == 2) ?                   pf->f2.x : \
    (_osmajor == 3 && _osminor == 0) ?  pf->f30.x : \
    /* DOS 3.1+ */                      pf->f31.x

// access macros
#define FILENAME(pf)    DOSVER_FIELD(pf,filename)
#define EXT(pf)         DOSVER_FIELD(pf,ext)
#define FSIZE(pf)       DOSVER_FIELD(pf,fsize)
#define FATTR(pf)       DOSVER_FIELD(pf,fattr)
#define NUM_HANDLES(pf) DOSVER_FIELD(pf,num_handles)

#define NUM_HANDLES_LVALUE(pf)  *((USHORT far *) pf)

// treat OWNER_PSP separately in DOS2, because no owner_psp field
#define OWNER_PSP(pf)   \
    (_osmajor == 2) ?                   -1 : \
    (_osmajor == 3 && _osminor == 0) ?  pf->f30.owner_psp : \
    /* DOS 3.1+ */                      pf->f31.owner_psp
        
typedef struct sysftab {
    struct sysftab far *next;
    USHORT num_files;
    file f[1];
    } SYS_FTAB;
    
typedef struct {
    BYTE type;      
    USHORT owner;    /* PSP of the owner */
    USHORT size;
    BYTE unused[3];
    BYTE dos4[8];
    } MCB;

void fail(char *s) { puts(s); exit(1); }    

#ifdef __TURBOC__
#define GETVECT(intno)  getvect(intno)
#define ASM             asm
#else
#define GETVECT(intno)  _dos_getvect(intno)
#define ASM             _asm
#endif

#ifndef MK_FP
#define MK_FP(seg,ofs)  ((FP)(((ULONG)(seg) << 16) | (ofs)))
#endif

#ifdef WINDOWS
#define MAP(ptr, bytes)		map_real((ptr), (bytes))
#define FREE_MAP(ptr)		free_mapped_linear(ptr)
#define GET_REAL(ptr)		get_real_addr(ptr)
#else
#define MAP(ptr, bytes)		(ptr)
#define FREE_MAP(ptr)		/**/
#define GET_REAL(ptr)		(ptr)
#endif

main(void)
{
    SYS_FTAB far *sys_filetab;
	SYS_FTAB far *next;
    file far *ff;
    int size;
    int i;

#ifdef WINDOWS
	RMODE_CALL r;
	memset(&r, 0, sizeof(r));
	r.eax = 0x5200;
	if (dpmi_rmode_intr(0x21, 0, 0, &r))
	{
		SYS_FTAB far * far *tmp = 
			(SYS_FTAB far * far *) MAP(MK_FP(r.es, (WORD) r.ebx+4), 4);
		SYS_FTAB far *psysftab = *tmp;
		FREE_MAP(tmp);
		sys_filetab = (SYS_FTAB far *) MAP(psysftab, 0xFFFF);
	}
	else
		fail("Could generate real-mode 21/52!");
#else
    ASM mov ah, 52h
    ASM int 21h
    ASM les bx, dword ptr es:[bx+4] /* ptr to list of DOS file tables */
    ASM mov word ptr sys_filetab, bx
    ASM mov word ptr sys_filetab+2, es
#endif		

    /* DOS box of OS/2 1.x doesn't provide system file tbl */
    if (sys_filetab == (SYS_FTAB far *) -1L)
        fail("system file table not supported");

    switch (_osmajor)
    {
        case 2:             size = 0x28; break;
        case 3:             size = 0x35; break;
        default:            size = 0x3b; break;
    }

    printf("Filename              Size      Attr Handles    Owner\n");
    printf("--------              ----      ---- -------    -----\n");

    do { /* FOR EACH SFT */
		
		printf("[SFT @ %Fp -- %d files]\n", 
			GET_REAL(sys_filetab), sys_filetab->num_files);

        /* FOR EACH ENTRY IN THIS SFT */
		i = sys_filetab->num_files;
		ff = (file far *) sys_filetab->f;		// DON'T MAP!!
		for (; i--; ((FP) ff) += size)
        if (NUM_HANDLES(ff))	// don't print unused entries
        {
#ifdef __BORLANDC__			
{
			// bcc pmode printf GP faults on non-null-terminated strings
			char buf[9];
			_fmemcpy(buf, FILENAME(ff), 8);
			buf[8] = '\0';
			printf("%.8s.", buf);
			_fmemcpy(buf, EXT(ff), 3);
			buf[3] = '\0';
			printf("%.3s\t", buf);
}
#else
            printf("%.8Fs.",   FILENAME(ff));
            printf("%.3Fs\t",  EXT(ff));
#endif			
            printf("%10lu\t",  FSIZE(ff));
            printf("%04X\t",   FATTR(ff));
            printf("%d\t",     NUM_HANDLES(ff));
			printf("%04X\n",   OWNER_PSP(ff));
        }
		
		/* FOLLOW LINKED LIST... */
		next = sys_filetab->next;		// get next...
		FREE_MAP(sys_filetab);			// then free old
		if (next && (FP_OFF(next) != 0xFFFF))
			sys_filetab = (SYS_FTAB far *) MAP(next, 0xFFFF);
		else
			break;

    } while (FP_SEG(sys_filetab) && 
             (FP_OFF(sys_filetab) != 0xFFFF)); /* ...UNTIL END */
		 
#ifdef WINDOWS
{
	WORD mapped = get_mapped();
	if (mapped != 0)
		printf("ERROR!  %u mapped selectors remaining!\n", mapped);
}
#endif
    return 0;
}

