/*
IMPDUMP.C -- dump import library

00000000 | F0 0D 00 00 C4 00 00 28 00 00 00 00 00 00 00 00 | .......(........
00000010 | 80 11 00 0F 41 44 44 46 4F 4E 54 52 45 53 4F 55 | ....ADDFONTRESOU
00000020 | 52 43 45 F8 88 1B 00 00 A0 01 01 0F 41 44 44 46 | RCE.........ADDF
00000030 | 4F 4E 54 52 45 53 4F 55 52 43 45 03 47 44 49 77 | ONTRESOURCE.GDIw
00000040 | 00 F6 8A 02 00 00 74 00 00 00 00 00 00 00 00 00 | ......t.........
00000050 | 80 10 00 0E 41 4E 49 4D 41 54 45 50 41 4C 45 54 | ....ANIMATEPALET
00000060 | 54 45 54 88 1A 00 00 A0 01 01 0E 41 4E 49 4D 41 | TET........ANIMA
00000070 | 54 45 50 41 4C 45 54 54 45 03 47 44 49 6F 01 59 | TEPALETTE.GDIo.Y

cf. Microsoft C Developer's Toolkit Reference
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

#define LIBMAGIC	0xF0
#define OBJMAGIC	0x80
#define ENDMAGIC	0xF1
#define COMENT		0x88
#define OMFEXTENS	0xA0
#define IMP			0x01
#define EXP			0x02
#define LIBMOD		0xA3

#pragma pack(1)

typedef struct {
	unsigned char magic;		// 0x80
	unsigned short reclen;
} OBJMOD;

typedef struct {
	unsigned char rectype;		// COMENT = 0x88
	unsigned short reclen;
	unsigned char zero;			// for Intel compatibility
	unsigned char class;		// OMF extension = 0xA0
	unsigned char subtype;		// IMPDEF=1 EXPDEF=2
	unsigned char fOrd;			// nonzero means import by ordinal
} IMPDEF;

void impdump(char *s);

void usage(void)
{
	puts("usage: impdump <libfile>");
	exit(1);
}

void fail(char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
	int i;
	
	fputs("Import Library Dumper version 1.0\n", stderr);
	fputs("Copyright (c) 1991 Andrew Schulman.  All rights reserved.\n\n",
		stderr);
	
	if (argc < 2)
		usage();
	
	for (i=1; i<argc; i++)
		if ((argv[i][0] == '-') || (argv[i][0] == '/'))
		{
			if (argv[i][1] == '?')
				usage();
			else
				usage();		// no switches right now
		}
		else	
			impdump(argv[i]);
	return 0;
}

void ignore_obj(FILE *f)	// discard rest of an obj module
{
	int c;
	while ((c = getc(f)) != OBJMAGIC)
		if (c == EOF)
			return;
	ungetc(c, f);
}
				
void freadcstr(char *s, FILE *f)	// read counted string
{
	unsigned char len;
	len = getc(f);
	fread(s, 1, len, f);
	s[len] = '\0';
}

void impdump(char *filename)
{
	unsigned char *impname, *modname, *intname;
	unsigned char page[16];
	unsigned short page_len;
	unsigned char len;
	unsigned char junk;
	unsigned char junk2[5];		// five unknown bytes: 8a 02 00 00 74
	unsigned short ord;
	OBJMOD obj;
	IMPDEF imp;
	FILE *f;
	int pad;
	
    if ((f = fopen(filename, "rb")) == NULL)
		fail("can't find file");
    fread(page, 1, 16, f);
    if (page[0] != LIBMAGIC)
		fail("not a library");
	page_len = *((unsigned short *) &page[1]) + 3;
	
	intname = malloc(256);
	modname = malloc(256);
	impname = malloc(256);
	
	while (fread(&obj, 1, sizeof(OBJMOD), f))
	{
		if (obj.magic == ENDMAGIC)
			break;	// end of object modules
		else if (obj.magic != OBJMAGIC)
			fail("not an object module");
		freadcstr(intname, f);		// function name: discard
		junk = getc(f);
		fread(&imp, 1, sizeof(IMPDEF), f);	// assume IMPDEF
			
		// if turns out wasn't an IMPDEF, compensate
			
		if ((imp.rectype != COMENT) || (imp.zero != 0))
		{
			ignore_obj(f);
			continue;
		}
		
		if (imp.class != OMFEXTENS)
		{
			ignore_obj(f);
			continue;
		}

		freadcstr(intname, f);		// internal name
		freadcstr(modname, f);		// module name
		if (imp.fOrd == 0)
		{
			freadcstr(impname, f);
			if (*impname)
				printf("; %s = %s . %s\n", intname, modname, impname);
			else
				printf("; %s . %s\n", modname, intname);
		}
		else
		{
			ord = getw(f);
			printf("%s . %u\t%s\n",	modname, ord, intname);
		}
		junk = getc(f);
		fread(&junk2, 1, 5, f);
		
		// have to skip over pad bytes - skip to next page
		if ((pad = ftell(f) % page_len) != 0)
			fread(page, 1, page_len - pad, f);
	}
	
	fclose(f);
}


