/* 
INSTDATA.C -- Show Windows Local, Global, Instance data
Andrew Schulman, November 1992
from "Undocumented Windows", 2nd edition (Addison-Wesley, 1993)

Borland C++: bcc instdata.c vxdcalls.c
Requires VXD.386 (generic Virtual Device Driver)

Sample output:
Address     Pages (Bytes)   Type
00000000      3 (00003000)  Instance
00003000     12 (0000C000)  Global
0000F000      5 (00005000)  Instance
00014000      1 (00001000)  Global
00015000      2 (00002000)  Instance
00017000      2 (00002000)  Global
00019000    134 (00086000)  Local
0009F000     33 (00021000)  Global
000C0000      2 (00002000)  Instance
000C2000     42 (0002A000)  Global
000EC000      1 (00001000)  Instance
000ED000     35 (00023000)  Global
*/

#include <stdlib.h>
#include <stdio.h>

typedef int BOOL;
typedef unsigned short WORD;
typedef unsigned long DWORD;

#include "vxdcalls.h"

int TestGlobalV86Mem(DWORD VMLinAddr, DWORD nBytes);

char *type_name(int type)
{
    switch (type)
    {
        case 0: return "Local";
        case 1: return "Global";
        case 2: return "Local and global";  // shouldn't happen
        case 3: return "Instance";
        default: return "Unknown";
    }
}

int iswin3e(void)
{
    WORD maj=0, min=0;
    _asm {
        mov ax, 1600h
        int 2fh
        mov byte ptr maj, al
        mov byte ptr min, ah
        }
    return ((maj > 1) && (maj != 0x80));
}

void fail(const char *s) { puts(s); exit(1); }

#define PAGE_SIZE   4096

main()
{
    DWORD base = 0;
    int pages;
    char *s;
    DWORD save_base;
    int size;
    int type;
    
    fputs("INSTDATA -- Show Windows Instance Data v. 1.0\n", stderr);
    fputs("by Andrew Schulman, November 1992\n", stderr);
    fputs("from \"Undocumented DOS\", 2nd edition (Addison-Wesley)\n\n", 
        stderr);
    
    if (! iswin3e())
        fail("This program requires Windows 3.x Enhanced mode");
    
    if (! GenericVxDInstalled())
        fail("This program requires VXD.386 (Generic Virtual Device Driver)");
        
    puts("Address\t\tPages (Bytes)\tType");
        
    for (base = 0; base < 0x100000L; )
    {
        type = TestGlobalV86Mem(base, PAGE_SIZE);
        pages = 1;
        save_base = base;
        base += PAGE_SIZE;
        while (TestGlobalV86Mem(base, PAGE_SIZE) == type)
        {
            base += PAGE_SIZE;
            pages++;
        }
        printf("%08lX\t%3d (%08lX)\t%s\n", 
            save_base, 
            pages,
            (DWORD) pages * PAGE_SIZE,
            type_name(type));
    }
    return 0;
}

int TestGlobalV86Mem(DWORD VMLinAddr, DWORD nBytes)
{
    VxDPushParams p;
    p.CallNum = _TestGlobalV86Mem;
    p.NumP = 3;
    p.P[0] = VMLinAddr;
    p.P[1] = nBytes;
    p.P[2] = 0;
    return (VxDPushCall(&p)) ? p.OutEAX : -1;
}
