//
//     iopal.c
//
//     (c)Copyright 1993 Jeffrey M. Cogswell
//
//     Based on the winio library,
//        (c)copyright 1991, 1992 David Maxey, Andrew Schulman
//
//     Portions of this code borrowed from:
//        gdiwalk.c (c)copyright 1992 Dave Maxey
//        winwalk.c (c)copyright 1992 Matt Pietrek
//
//     Note that this program has very little error checking!


#include <windows.h>
#include <dos.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "toolhelp.h"
#include "winio.h"       //used with winio library
#include "gdiobj.h"      //this came with Undocumented Windows
#include "trapdll.h"     //used with trapdll library


//Types
typedef struct tagCOLORS {
   WORD peForeIndex;
   WORD peCurIndex;
   WORD pePrevIndex;
   PALETTEENTRY peColor;
} COLORS;

typedef struct tagPALGLOBAL {
   WORD   phNumEntries;
   WORD   phCurRealTime;
   COLORS phColors[1];
} PALGLOBAL;

typedef struct tagNewPALETTEOBJ {
   GDIOBJHDR    iPaletteHead;
   HANDLE       ilphPal;
   WORD         ilphUseCount;
   HANDLE       ilphLDevice;
   WORD         ilpFlags;
   WORD         ilphMetaList;
} NewPALETTEOBJ;

typedef struct tagRANGEITEM {
   int x;
   int y;
   int type;
   //where type is one of:
   //  JUST_X;
   //  BOTHXY;
} RANGEITEM;
#define   JUST_X      1
#define   BOTHXY      2



//Globals
HWND thisHwnd;          //This program's window's hwnd
char buf[256]={0};      //buffer
char *p;                //pointer into buffer to current token.  It's Global
                        //to make life a little easier.
char prompt[25] = "iopal>";
WORD GDIHeap;           //Segment selector for GDI's heap.
SYSHEAPINFO shi;        //Structure for getting GDI heap from TOOLHELP.



//Function prototypes
void process(void);
void GetWinSegs(void);
void DoDumpPal(void);
void DoListPals(void);
void DoSetPalette(void);
void DumpPaletteObject(void);
void DumpHeader(GDIOBJHDR FAR *goh);
void DumpGlobal (HANDLE ilphPal);
void DumpSysPal(void);



//macros
#define CMD_NEXT   p = strtok (NULL, " ")
#define FREEZE     winio_setbusy();winio_setpaint(thisHwnd,FALSE);
#define UNFREEZE   winio_setpaint(thisHwnd,TRUE);winio_resetbusy();


void main()
{
   thisHwnd = winio_current();
   GetWinSegs();
   printf ("All values must be entered in Hex.\n");
   while (TRUE)
   {
      printf ("%s",prompt);
      gets(buf);
      process();
   }
}


//Get the segment of the GDI heap.
void GetWinSegs(void)
{
   shi.dwSize = sizeof(shi);
   SystemHeapInfo(&shi);
   GDIHeap = shi.hGDISegment;
   //GDIHeap &= 0xfffc;     //This is the code which was
   //GDIHeap |= 1;          //incorrectly included in GDIWalk.
   printf ("Default GDI heap: %04x\n",GDIHeap);

}

//Proccess each input line.
void process(void)
{

   p = strupr(strtok(buf," "));
   if (strcmp (p,"LIST") == 0) DoListPals();
   else if (strcmp (p,"DUMP") == 0) DoDumpPal();
   else if (strcmp (p,"SET") == 0) DoSetPalette();
   else if (strcmp (p,"GDIDUMP") == 0) DumpPaletteObject();
   else if (strcmp (p,"SYSTEM") == 0) DumpSysPal();
   else if (strcmp (p,"SETTRAP") == 0) SetTrap();
   else if (strcmp (p,"ENDTRAP") == 0) EndTrap();
   else printf ("Commands are:\n"
      "\tLIST\n\tDUMP <palette-handle-in-hex>\n"
         "\tSET <palette-handle-in-hex> <hwnd-in-hex>\n"
         "\tGDIDUMP <palette-handle-in-hex>\n"
         "\tSYSTEM\n"
         "\tSETTRAP\n"
         "\tENDTRAP\n");

}



//List the handles & addresses of all palettes in the system
//by walking through GDI's heap.  (LIST command)
void DoListPals(void)
{
   LOCALENTRY le;
   BOOL ok;

   FREEZE;

   le.dwSize = sizeof(le);
   for (ok = LocalFirst(&le, GDIHeap); ok; ok = LocalNext(&le))
       if (le.wType == LT_GDI_PALETTE)
       {
           printf ("Handle %04x Address %04x.\n",le.hHandle,le.wAddress);
       }
   UNFREEZE;
}


//Dump the RGB values for a palette.  (DUMP command)
void DoDumpPal(void)
{
   HPALETTE hpal;
   PALETTEENTRY MyEntry;
   int i;

   FREEZE;
   CMD_NEXT;

   sscanf (p,"%x",(HPALETTE)&hpal);
   printf ("Palette %04x\n",hpal);

   for (i=0;; i++)
   {
      if (GetPaletteEntries(hpal, i, 1, &MyEntry))
      {
         printf ("%02x:%02x %02x %02x %02x\n",i, MyEntry.peRed,
            MyEntry.peGreen, MyEntry.peBlue, MyEntry.peFlags);
      }
      else
      {
         break;
      }
   }
   UNFREEZE;
}



//change a given palette.  (SET command)
void DoSetPalette(void)
{
   HPALETTE hpal;
   RANGEITEM indexrange,redrange,greenrange,bluerange;
   PALETTEENTRY pe;
   HWND hwnd, hwndDest;
   char tempchar;

   hwndDest = winio_current();

   CMD_NEXT;
   sscanf (p,"%x",(HPALETTE)&hpal);
   CMD_NEXT;
   sscanf (p,"%x",(HPALETTE)&hwndDest);


   printf ("Valid ranges are in form n or n-n,\n"
      "where n is a hex value between 0 and ff.\n");
   //Get index range
   printf (">>Index:");
   gets(buf);
   sscanf (buf,"%x%c%x",&indexrange.x,&tempchar,&indexrange.y);
   if (tempchar == '-') indexrange.type = BOTHXY; else indexrange.type = JUST_X;


   //Get red range
   printf (">>Red:");
   gets(buf);
   if (sscanf (buf,"%x-%x",&redrange.x,&redrange.y) == 1)
      redrange.y = redrange.x;

   //Get green range
   printf (">>green:");
   gets(buf);
   if (sscanf (buf,"%x-%x",&greenrange.x,&greenrange.y) == 1)
     greenrange.y = greenrange.x;

   //Get blue range
   printf (">>blue:");
   gets(buf);
   if (sscanf (buf,"%x%c%x",&bluerange.x,&tempchar,&bluerange.y) == 1)
      bluerange.y = bluerange.x;


   //Set the palette according to the specifications
   if (indexrange.type == BOTHXY)
   {
      //Color gradient
      int range, redstep, greenstep, bluestep;
      int index,red, green, blue;

      range = indexrange.y - indexrange.x;
      redstep = (redrange.y - redrange.x)/range;
      greenstep = (greenrange.y - greenrange.x)/range;
      bluestep = (bluerange.y - bluerange.x)/range;

      for (index=indexrange.x, red=redrange.x, green=greenrange.x, blue=bluerange.x;
      index<=indexrange.y; index++, red+=redstep, green+=greenstep, blue+=bluestep)
      {
         pe.peRed = red;
         pe.peGreen = green;
         pe.peBlue = blue;
         pe.peFlags = NULL;
         SetPaletteEntries(hpal, index, 1, &pe);
         //note that SetPaletteEntries can modify several entries at once, but
         //this way is fast enough this time.
      }
      ShowWindow(hwndDest,SW_MINIMIZE);
      ShowWindow(hwndDest,SW_RESTORE);
   }
   else
   {
      //Just one entry to change
      pe.peRed = redrange.x;
      pe.peGreen = greenrange.x;
      pe.peBlue = bluerange.x;
      pe.peFlags = NULL;
      printf ("Mapped %d entries\n",SetPaletteEntries(hpal, indexrange.x, 1, &pe));
      ShowWindow(hwndDest,SW_MINIMIZE);
      ShowWindow(hwndDest,SW_RESTORE);
   }
}




//Dump the actual internal structures representing the palette.
// (GDIDUMP command)
void DumpPaletteObject(void)
{
   HPALETTE hpal;
   NewPALETTEOBJ FAR *MyPal;
   LOCALENTRY le;
   LPGDIOBJHDR lp;
   int ok = TRUE, i, column;

   FREEZE;
   CMD_NEXT;

   sscanf (p,"%x",(HPALETTE)&hpal);
   printf ("Palette %04x\n",hpal);

   //Find the palette in GDI's heap.
   le.dwSize = sizeof(le);

   for (ok = LocalFirst(&le, GDIHeap); ok; ok = LocalNext(&le))
   {
       lp = (LPGDIOBJHDR) MK_FP(GDIHeap, le.wAddress);
       if (le.hHandle == hpal) break;
   }
   MyPal = (NewPALETTEOBJ FAR *)lp;

   //Dump the structure inside GDI's heap.
   DumpHeader(&(MyPal->iPaletteHead));
   printf ("ilphPal      : %04x\n",MyPal->ilphPal);
   printf ("ilphUseCount : %04x\n",MyPal->ilphUseCount);
   printf ("ilphLDevice  : %04x\n",MyPal->ilphLDevice);
   printf ("ilpFlags     : %04x\n",MyPal->ilpFlags);
   printf ("ilphMetaList : %04x\n",MyPal->ilphMetaList);
   printf ("\n");

   //Dump the structure inside the global heap.
   DumpGlobal(MyPal->ilphPal);

   UNFREEZE;
}

//Dump the header structure.
void DumpHeader(GDIOBJHDR FAR *goh)
{
   printf ("GDI Object Header:\n"
      "\thNext\t\t: %04X\n"
      "\twMagic\t\t: %04X (\"%c%c\")\n"
      "\tdwCount\t\t: %ld\n"
      "\twMetaList\t: %04X\n",
      goh->hNext, goh->wMagic,
      goh->wMagic & 0xff, goh->wMagic >> 8,
      goh->dwCount, goh->wMetaList);
}

//Dump the structure in the global heap representing
//the palette.
void DumpGlobal (HANDLE ilphPal)
{
   BOOL ok;
   GLOBALENTRY ge;
   WORD wSel;
   WORD i, j, c;
   LPSTR p, addr;
   PALGLOBAL FAR*fp;

   ge.dwSize = sizeof(ge);
   ok = GlobalEntryHandle(&ge, ilphPal);
   if (!ok)
   {
      printf ("Can't find global heap!\n");
      return;
   }
   printf ("Global Object %04x (actual RGB data):\n",ilphPal);
   wSel = GlobalHandleToSel(ge.hBlock);

   fp = (PALGLOBAL FAR*) MK_FP(wSel,0);
   printf ("phNumEntries = %d\n"
           "phCurRealTime = %04x\n",
           fp->phNumEntries, fp->phCurRealTime);
   printf ("Palette:\n"
           "\tFore Cur  Prev  R    G    B   Flag\n"
           "\t--   --   --   --   --   --   --\n");
   for (i=0; i<fp->phNumEntries; i++)
   {
      printf ("\t%04x %04x %04x %02x   %02x   %02x   %02x\n",
         fp->phColors[i].peForeIndex,
         fp->phColors[i].peCurIndex,
         fp->phColors[i].pePrevIndex,
         fp->phColors[i].peColor.peRed,
         fp->phColors[i].peColor.peGreen,
         fp->phColors[i].peColor.peBlue,
         fp->phColors[i].peColor.peFlags);
   }
}


//Dump the system palette (SYSTEM command)
void DumpSysPal(void)
{
   HPALETTE hpal;
   PALETTEENTRY MyEntry;
   HDC hDC;
   int i;

   FREEZE;

   hDC = CreateDC("DISPLAY",NULL,NULL,NULL);

   for (i=0; i<GetDeviceCaps(hDC,SIZEPALETTE); i++)
   {
      if (GetSystemPaletteEntries(hDC, i, 1, &MyEntry))
      {
         printf ("%02x:%02x %02x %02x %02x\n",i, MyEntry.peRed,
            MyEntry.peGreen, MyEntry.peBlue, MyEntry.peFlags);
      }
      else
      {
         break;
      }
   }
   DeleteDC(hDC);
   UNFREEZE;

}
