/*
IS_DRDOS.C
Determine if DR-DOS running
Andrew Schulman
February 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

Digital Research OS version codes:
10h (single tasking, e.g., DR-DOS)
14h (multitasking, e.g., Concurrent DOS)
*/

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

unsigned short is_drdos(void)
{
	_asm mov ax, 4452h
	_asm int 21h
	_asm jnc have_drdos
	return 0;
have_drdos:;
	// return value in AX
}

unsigned short is_concdos(void)
{
	_asm mov ax, 4451h
	_asm int 21h
	_asm jnc have_concdos
	return 0;
have_concdos:;
	// return value in AX
}

main()
{
	unsigned short dr_vers = is_drdos();
	unsigned short conc_vers = is_concdos();
	
	if (dr_vers)
	{
		if (dr_vers < 0x1060)
			printf("Unknown old version of DR-DOS: %04Xh\n", dr_vers);
		else switch (dr_vers)
		{
			case 0x1060:	printf("DOS Plus\n"); break;
			case 0x1063:	printf("DR-DOS 3.41\n"); break;
			case 0x1064:	printf("DR-DOS 3.42\n"); break;
			case 0x1065:	printf("DR-DOS 5.0\n"); break;
			case 0x1067:	printf("DR-DOS 6.0\n"); break;
			case 0x1070:	printf("DR-DOS Palm-DOS\n"); break;
			default:		printf("Unknown version of DR-DOS\n"); break;
		}
		printf("Providing DOS interface %u.%02u\n", _osmajor, _osminor);
	}
	else if (conc_vers)
	{
		if (conc_vers < 0x1432)
			printf("Unknown old version of Concurrent DOS: %04Xh\n", 
				conc_vers);
		else switch (conc_vers)
		{
			case 0x1432:	printf("Concurrent PC-DOS 3.2\n"); break;
			case 0x1441:	printf("Concurrent DOS 4.1\n"); break;
			case 0x1450:	printf("Concurrent DOS/XM 5.0 or "
								   "Concurrent DOS/386 1.1\n"); break;
			case 0x1460:	printf("Concurrent DOS/XM 6.0 or "
								   "Concurrent DOS/386 2.0\n"); break;
			case 0x1462:	printf("Concurrent DOS/XM 6.2 or "
								   "Concurrent DOS/386 3.0\n"); break;
			case 0x1466:	printf("DR Multiuser DOS 5.1\n"); break;
			case 0x1467:	printf("Concurrent DOS 5.1\n"); break;
			default:		printf("Unknown version of Concurrent DOS\n");
							break;
		}
	}
	else
		printf("Neither DR-DOS nor Concurrent DOS running\n");

	return (dr_vers & 0xFF);
}
