/* LDHI.C */

#include <stdlib.h>
#include <stdio.h>
#include <process.h>
#include <dos.h>

typedef unsigned short WORD;

#define UMB_LINKED          1
#define FIRST_FIT_HIGH      0x80

WORD get_true_dos_version(void);
WORD get_alloc_strategy(void);
WORD set_alloc_strategy(WORD strat);
WORD get_umb_link(void);
WORD set_umb_link(WORD flag);

void fail(char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    WORD prev_strat;
    WORD prev_link = 0;
    WORD dos_vers;
    WORD err;
	
	#define PROGNAME	"LDHI"
	#define DESC		"Simulate LOADHIGH"
	#define AUTHOR		"Andrew Schulman"
	#include "..\banner"
    
    if (argc < 2)
        fail("LDHI [command]");
    
    dos_vers = get_true_dos_version();
    if ((dos_vers == 0) || ((dos_vers & 0xFF) < 5))
        fail("This program requires DOS 5.0 or higher");

    prev_strat = get_alloc_strategy();
    prev_link = get_umb_link();
    
    if (err = set_umb_link(UMB_LINKED))
        printf("Can't set UMB link -- loading low: %04X\n", err);
    if (err = set_alloc_strategy(FIRST_FIT_HIGH))
        printf("Can't change strategy -- may load low:  %04X\n", err);
        
    if (spawnvp(P_WAIT, argv[1], &argv[1]) == -1)
        printf("Couldn't spawn command\n");

    if (err = set_umb_link(prev_link))
        printf("Can't restore previous UMB link: %04X\n", err);
    if (err = set_alloc_strategy(prev_strat))
        printf("Can't restore previous alloc strategy: %04X\n", err);

    return err;
}

WORD get_true_dos_version(void)
{
    _asm xor bx, bx
    _asm mov ax, 3306h
    _asm int 21h
    _asm mov ax, bx
    // return value in AX
}

WORD get_alloc_strategy(void)
{
    _asm mov ax, 5800h
    _asm int 21h
    // return value in AX
}

WORD set_alloc_strategy(WORD strat)
{
    _asm mov ax, 5801h
    _asm mov bx, word ptr strat
    _asm int 21h
    _asm jc error
    _asm xor ax, ax
error:;
    // return 0 or error code in AX
}

WORD get_umb_link(void)
{
    _asm mov ax, 5802h
    _asm int 21h
    _asm xor ah, ah
    // return value in AX
}

WORD set_umb_link(WORD flag)
{
    _asm mov ax, 5803h
    _asm mov bx, flag
    _asm int 21h
    _asm jc error
    _asm xor ax, ax
error:;
    // return 0 or error code in AX
}

