/* LEDUMP.C -- linear executable dumper */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <dos.h>

/* lifted from MSoft newexe.h */
#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC		0x454C			/* linear executable */
#define LXMAGIC		0x584C			/* IBM OS/2 2.0 linear executable */
#define PEMAGIC		0x4550			/* NT portable executable */
#define W3MAGIC		0x3357			/* WIN386.EXE signature */

void fail(char *s) { fputs("ledump fail: ", stdout); puts(s); exit(1); }

void usage()
{
	puts("usage: ledump [-srdef] <exe>");
	exit(1);
}
	
void le(char *name);
void srdef(char *name);
void le2(unsigned long offset);
void le2_srdef(unsigned long offset);
int exe_read(int f, unsigned long ne_offset, void *p, unsigned bytes);
void *allocate(size_t x, size_t y);

static unsigned long align;
int f;

main(int argc, char *argv[])
{
	int i;
	
	fputs("Linear Executable Dumper version .3\n", stderr);
	fputs("Copyright (c) 1991 Andrew Schulman.  All rights reserved.\n\n", 
		stderr);
	
	if (argc < 2)
		usage();
	
	for (i=1; i<argc; i++)
		if ((argv[i][0] == '-') || (argv[i][0] == '/'))
		{
			// later: add -srdef option
			if (argv[i][1] == '?')
				usage();
			else if (strcmp(&argv[i][1], "srdef") == 0)
			{
				srdef(argv[i+1]);
				i++;
			}
		}
		else
			le(argv[i]);
	return 0;
}

void le(char *name)
{
	unsigned long start_newexe;
	unsigned long pos;
    unsigned short w;
	unsigned i;
	
    if ((f = open(name, O_RDONLY | O_BINARY)) == -1)
		fail("can't find file");
    read(f, &w, 2);
    if (w != EMAGIC)
		fail("not an executable");
	exe_read(f, ENEWHDR, &start_newexe, 4);
	exe_read(f, start_newexe, &w, 2);
	if (w != LEMAGIC)
		fail("not an LE (linear executable)");
	le2(start_newexe);
	close(f);
}

void srdef(char *name)
{
	unsigned long start_newexe;
	unsigned long pos;
    unsigned short w;
	unsigned i;
	
    if ((f = open(name, O_RDONLY | O_BINARY)) == -1)
		fail("can't find file");
    read(f, &w, 2);
    if (w != EMAGIC)
		fail("not an executable");
	exe_read(f, ENEWHDR, &start_newexe, 4);
	exe_read(f, start_newexe, &w, 2);
	if (w != LEMAGIC)
		fail("not an LE (linear executable)");

	printf("uP = P386\n");
	printf("Code style = Windows VXD\n");
	printf("Zap load bytes = VXD %04X:%04X\n",
		(unsigned) (start_newexe >> 16),
		(unsigned) (start_newexe & 0xFFFF));
	printf("Input filename = %s\n", name);
	printf("Header = %s\n", name);
	printf("Passes = 5\n");
	printf("Xref = ON\n");

	le2_srdef(start_newexe);
	close(f);
}

/**********************************************************************/

#define DO_NRESTAB
#include "le2.c"

/**********************************************************************/

/* read at offset from beginning of file */
int exe_read(int f, unsigned long exe_offset, void *p, unsigned bytes)
{
    lseek(f, exe_offset, SEEK_SET);
	return read(f, p, bytes);
}

void *allocate(size_t x, size_t y)
{
	void *p;
//	printf("allocate %u * %u\n", x, y);
	if (! x) x = 1;
	if (! y) y = 1;
	if (((long) x * (long) y) > 0xFFFFUL)
	{
		printf("\n\nx=%04x y=%04x\n", x, y);
		fail("request > 64k");
	}
	if ((p = calloc(x, y)) == 0)
	{
		printf("\n\nx=%04x y=%04x\n", x, y);
		fail("insufficient memory");
	}
	return p;
}

