/* 
MEMLOOP.C
How much memory can be allocated through DOS (INT 21h AH=48h)?
Can be compiled for real mode or protected mode (DPMI)
Andrew Schulman
February 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

real mode: bcc memloop.c
prot mode: bcc -2 -DDPMI_APP memloop.c dpmi_app.lib

On machine with 5 megabytes:
Real mode:                  453K
DPMI app, under 386MAX     4032K
DPMI app, under DOSX       3922K
DPMI app, under WIN386     4866K (a little virtual memory)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#ifdef DPMI_APP
#include "dpmishel.h"
#include "protmode.h"
#endif

void fail(const char *s) { puts(s); _dos_exit(1); }

void mem_loop(void)
{
    unsigned long kb = 0;
    unsigned blocks = 0;
    unsigned blocksize = 2048;
    unsigned kbytes = 32;
    unsigned segsel;
    char far *fp;
    while (kbytes)
    {
        blocks = 0;
        while (_dos_allocmem(blocksize, &segsel) == 0)	// INT 21h AH=48h
        {
            fp = (char far *) MK_FP(segsel, 0);
            _fmemset(fp, 'x', blocksize * 16);  // touch every byte!
            blocks++;
        }
        kb += (kbytes * blocks);
        blocksize >>= 1;
        kbytes >>= 1;
    }
    printf("Allocated %lu Kb\n", kb);
}

#ifdef DPMI_APP
int v86_main(int argc, char *argv[])    // bad name: could be real mode
{
    return 0;   // okay to switch to protected mode
}

int pmode_main(int argc, char *argv[])
{
    mem_loop();
    return 0;
}
#else
int main()
{
    mem_loop();
    return 0;
}
#endif
