/*
 * MSTENVP.C
 * test bed for different methods to find master environment
 * Andrew Schulman, July 1990
 * modified July 1992 to use "undocdos.h" typedefs - Jim Kyle
 */

#include <stdlib.h>
#include <dos.h>

#include "undocdos.h"   /* defines types and structs used here  */

extern LPMCB firsthi( void );

/****************************************************************/

LPMCB get_mcb(void)     /* returns far pointer to first MCB     */
{
    ASM mov ah, 52h
    ASM int 21h
    ASM mov dx, es:[bx-2]
    ASM xor ax, ax
    /* in both Microsoft C and Turbo C, far* returned in DX:AX  */
}

char far *env(LPMCB mcb) /* returns far ptr to env, or NULL     */
{
    char far *e;
    unsigned env_mcb;
    unsigned env_owner;
// printf("env for %Fp\n", mcb );

    if (! IS_PSP(mcb))
        return (char far *) 0;

    e = MK_FP(ENV_FM_PSP(mcb->owner), 0);
    env_mcb = MCB_FM_SEG(FP_SEG(e));
    env_owner = ((MCB far *) MK_FP(env_mcb, 0))->owner;
// printf("env_mcb %Fp, env_owner %Fp\n", env_mcb, env_owner );
// (void)getchar();
//  return (env_owner == mcb->owner) ? e : (char far *) 0;
    return e;
}

typedef BOOL (*WALKFUNC)(LPMCB mcb);

/*  General-purpose MCB walker.
 *  The second parameter to walk() is a function that expects an
 *  MCB pointer, and that returns TRUE to indicate that walk()
 *  should keep going, and FALSE to indicate that walk() should
 *  stop.
 */
BOOL walk(LPMCB mcb, WALKFUNC walker)
{
// printf("Walking MCBs from %Fp\n", mcb );
    for (;;)
        switch (mcb->type)
        {
            case 'M':
                if (! walker(mcb))
                    return FALSE;
                mcb = (LPMCB)MK_FP(FP_SEG(mcb) + mcb->size + 1, 0);
                break;
            case 'Z':
                walker(mcb);
                if( FP_SEG( mcb ) < 0xA000 )
                  { mcb = firsthi();    /* look for UMBs        */
                    if( mcb == ( FP ) 0L )
                      return FALSE;     /* none found           */
                  }
                else
                  return FALSE;         /* UMBs all done too    */
                break;
            default:
                return FALSE;   /* error in MCB chain! */
        }
}

/****************************************************************/

/* using the GETVECT 2Eh technique (ENVPKG.ASM) */
void far *mstenvp1(void)
{
    ASM mov ax, 352eh       /* get INT 2Eh vector */
    ASM int 21h
    ASM mov dx, es:[002Ch]  /* environment segment */
    ASM xor ax, ax          /* return far ptr in DX:AX */
}

/****************************************************************/

/* walk MCB chain, looking for very first environment */
void far *env2 = (void far *) 0;

BOOL walk2(LPMCB mcb)
{
    env2 = env(mcb);
//     if (env2)
//     {
//         unsigned psp = FP_SEG(mcb)+1;
// printf( "walk2, testing %Fp at %Fp, ", env2, mcb );
// printf( " PSP %04X has %04X as parent\n", psp, PARENT(psp));
//         return (PARENT(psp) == psp) ? FALSE /*found it!*/ : TRUE;
//     }
//     else
//         return TRUE;    /* keep going */
    return env2 ? FALSE : TRUE;
}

void far *mstenvp2(void)
{
    walk(get_mcb(), walk2);
    return env2;
}

/****************************************************************/

/* walk MCB chain, looking for very LAST env addr > PSP addr */
void far *env3 = (void far *) 0;

#define NORMALIZE(fp)   (FP_SEG(fp) + (FP_OFF(fp) >> 4))

BOOL walk3(LPMCB mcb)
{
    void far *fp;
    /* if env seg at greater address than PSP, then
       candidate for master env -- we'll take last */
    fp = env(mcb);
    if (fp)
        if (NORMALIZE(fp) > (FP_SEG(mcb)+1))
            env3 = fp;
// printf( "walk3, testing %Fp at %Fp\n", fp, mcb );
    return TRUE;
}

void far *mstenvp3(void)
{
    walk(get_mcb(), walk3);
    return env3;
}

/****************************************************************/

/* walk MCB chain, looking for very first environment belonging
   to PSP which is its own parent */
void far *env4 = (void far *) 0;

BOOL walk4(LPMCB mcb)
{
    env4 = env(mcb);
    if (env4)
    {
        unsigned psp = FP_SEG(mcb) + 1;
// printf( "walk4, testing %Fp at %Fp, ", env4, mcb );
// printf( " PSP %04X has %04X as parent\n", psp, PARENT(psp));
        return (PARENT(psp) == psp) ? FALSE /*found it!*/ : TRUE;
    }
    else
        return TRUE;    /* keep going */
}

void far *mstenvp4(void)
{
    walk(get_mcb(), walk4);
    return env4;
}

/****************************************************************/

void main( void )
{
//  ASM mov ax, 5803h       /* activate UMB link */
//  ASM mov bx, 0001h
//  ASM int 21h
    printf("GETVECT 2Eh method; mstenvp1   = %Fp\n", mstenvp1());
    printf("WALK MCB method; mstenvp2      = %Fp\n", mstenvp2());
    printf("WALK MCB/LAST method; mstenvp3 = %Fp\n", mstenvp3());
    printf("WALK MCB/OWN PARENT; mstenvp4  = %Fp\n", mstenvp4());
//  ASM mov ax, 5803h       /* inactivate UMB link */
//  ASM mov bx, 0000h
//  ASM int 21h
}
