/*
PE2.C -- take over INT 21h, count calls in V86 mode
Andrew Schulman
February 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

bcc -2 -P- pe2.c
pe2 \win31\win.com
Before Windows started:	53 DOS calls, 0 in V86 mode
During Windows init:	1123 DOS calls, 177 in V86 mode
While Windows running:	37859 DOS calls, 37859 in V86 mode
During Windows exit:	7 DOS calls, 4 in V86 mode
After Windows exited:	12 DOS calls, 0 in V86 mode
*/

#include <stdlib.h>
#include <stdio.h>
#include <process.h>
#include <dos.h>

#pragma pack(1)

typedef struct {
#ifdef __TURBOC__
    unsigned bp,di,si,ds,es,dx,cx,bx,ax;
#else
    unsigned es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    unsigned ip,cs,flags;
    } REG_PARAMS;

void interrupt far dos(REG_PARAMS r);
void interrupt far int2f(REG_PARAMS r);

void (interrupt far *old)();
void (interrupt far *old_2F)();

typedef enum { NO_WIN, WIN_INIT_BEGIN, WIN_INIT_DONE, WIN_FINI_BEGIN,
	WIN_FINI_DONE, NUM_STATES } STATE;

static STATE state = NO_WIN;
static unsigned long calls[NUM_STATES] = {0} ;
static unsigned long v86_calls[NUM_STATES] = {0};

static char *state_str[NUM_STATES] = {
	"Before Windows started", "During Windows init",
	"While Windows running", "During Windows exit",	"After Windows exited",
	} ;

void fail(char *s) { puts(s); exit(1); }

int pe(void)
{
	unsigned msw;
	_asm smsw ax		/* Store Machine Status Word: low CR0 */
	_asm mov msw, ax
	return (msw & 1);
}

main(int argc, char *argv[])
{
	STATE i;
	
    if (argc < 2)
        fail("usage: pe2 [win.com] <args...>");
    
    /* hook INT 21h */
    old = _dos_getvect(0x21);
    _dos_setvect(0x21, dos);
	
	/* hook INT 2Fh */
	old_2F = _dos_getvect(0x2F);
	_dos_setvect(0x2f, int2f);

    /* run command */
    spawnvp(P_WAIT, argv[1], &argv[1]);
    
    /* unhook INT 21h, 2Fh */
	_dos_setvect(0x2F, old_2F);
    _dos_setvect(0x21, old);

	printf("\n");
	for (i=NO_WIN; i<NUM_STATES; i++)
		printf("%s:\t%lu DOS calls, %lu in V86 mode\n",
			state_str[i], calls[i], v86_calls[i]);
		
	return 0;
}

void check_pe(void)
{
	calls[state]++;
	if (pe())
		v86_calls[state]++;
}

void interrupt far dos(REG_PARAMS r)
{
	check_pe();
    _chain_intr(old);
}

#define WIN_INIT_NOTIFY		0x1605
#define WIN_EXIT_NOTIFY		0x1606
#define WIN_INIT_COMPLETE	0x1608
#define WIN_EXIT_BEGIN		0x1609

void interrupt far int2f(REG_PARAMS r)
{
	switch (r.ax)
	{
		case WIN_INIT_NOTIFY:	state = WIN_INIT_BEGIN; break;
		case WIN_INIT_COMPLETE:	state = WIN_INIT_DONE; break;
		case WIN_EXIT_BEGIN:	state = WIN_FINI_BEGIN; break;
		case WIN_EXIT_NOTIFY:	state = WIN_FINI_DONE; break;
	}
	check_pe();
	_chain_intr(old_2F);
}
