/* PORTMAP.C */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include "dpmishel.h"
#include "protmode.h"

void fail(char *s) 
{ 
    fputs(s, stderr); 
    fputs("\n", stderr);
    _dos_exit(1);
}

#pragma pack(1)

typedef struct {
    WORD backlink;  /* Back link selector to TSS */
    WORD fl1;       /* Filler */
    DWORD esp0;     /* SP for CPL 0 */
    WORD ss0;       /* SS for CPL 0 */
    WORD fl2;       /* Filler */
    DWORD esp1;     /* SP for CPL 1 */
    WORD ss1;       /* SS for CPL 1 */
    WORD fl3;       /* Filler */
    DWORD esp2;     /* SP for CPL 2 */
    WORD ss2;       /* SS for CPL 2 */
    WORD fl4;       /* Filler */
    DWORD cr3;      /* CR3 */
    DWORD eip;      /* EIP */
    DWORD eflags;   /* Extended flags */
    DWORD eax;      /* EAX */
    DWORD ecx;      /* ECX */
    DWORD edx;      /* EDX */
    DWORD ebx;      /* EBX */
    DWORD esp;      /* ESP */
    DWORD ebp;      /* EBP */
    DWORD esi;      /* ESI */
    DWORD edi;      /* EDI */
    WORD es;        /* ES */
    WORD fl5;       /* Filler */
    WORD cs;        /* CS */
    WORD fl6;       /* Filler */
    WORD ss;        /* SS */
    WORD fl7;       /* Filler */
    WORD ds;        /* DS */
    WORD fl8;       /* Filler */
    WORD fs;        /* FS */
    WORD fl9;       /* Filler */
    WORD gs;        /* GS */
    WORD fl10;  /* Filler */
    WORD ldtr;  /* LDTR */
    WORD fl11;  /* Filler */
    WORD dflags;    /* Debugger flags */
    WORD iomapb;    /* I/O map base */
} TSS386;

// get task register (selector of TSS)
WORD str(void)
{
    _asm str ax
}

#define X(s) fputs(s ## "\n", stderr)

int v86_main(int argc, char *argv[])
{
    X("PORTMAP -- Display Hooked Ports");
    X("Copyright (c) 1992 Andrew Schulman.  All rights reserved.\n");
    return 0;
}

int pmode_main(int argc, char *argv[])
{
    WORD tss_sel;
    DESCRIPTOR far *gdt;
    DESCRIPTOR far *tss_desc;
    DWORD tss_base, tss_limit;
    TSS386 far *tss;
    BYTE far *iopbits;
    BYTE far *p;
    WORD i, j;

    tss_sel = str();
	printf("TSS = %04Xh\n", tss_sel);
    gdt = get_gdt();
    tss_desc = &gdt[tss_sel >> 3];
    tss_base = get_base(tss_desc);
    tss_limit = get_limit(tss_desc);
	printf("TSS base = %08lX, limit = %08lX\n", tss_base, tss_limit);
    free_mapped_linear(gdt);
    
    tss = (TSS386 far *) map_linear(tss_base, tss_limit + 1);
    iopbits = ((BYTE far *) tss) + tss->iomapb;
	printf("IO bitmap @ %04Xh\n", offsetof(TSS386, iomapb));
    
    // print hooked ports; ignore last BYTE (FFh)
    printf("\nHooked ports:\n");
    for (i=0, p=iopbits; i< (tss_limit - 1 - tss->iomapb); i++, p++)
        if (*p)
        {
            for (j=0; j<8; j++)
                if (*p & (1 << j))
                    printf("%04X ", (i << 3) + j);
            printf("\n");
        }
    free_mapped_linear(tss);
    return 0;
}

