/* 
PRIMES.C
Author: Andrew Schulman, February 1990
*/

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <limits.h>
#include <dos.h>

typedef enum { FALSE, TRUE } BOOL;

void fail(char *s) { puts(s); exit(1); }

#include "btmap386.h"

#define N(x)              ((x) / 3)  // exclude multiples of 2 and 3

// don't test double for equality: DBL_EPSILON in <float.h> 
#define EQ(x,y)           (((x) - (y)) < DBL_EPSILON)

void help(void);
BOOL is_prime(double x);
void prime_factors(double x);
void vm_stats(void);

BITMAP map;
ULONG size;
double dsize;

main()
{
    char buf[80], *s=buf;
    double d;
    FILE *f;
    ULONG map_size;
    
    if (! (f = fopen("primes.dat", "rb")))
        fail("requires PRIMES.DAT");
    fread(&size, sizeof(ULONG), 1, f);
    dsize = size;
    fread(&map_size, sizeof(ULONG), 1, f);
    if (! (map = make_bitmap(size)))
        fail("Insufficient memory");
    fseek(f, 4, SEEK_SET);
    fread(map, 1, bytes(map_size), f);
    for (;;)
    {
        printf("> ");
        gets(s);
        if (strlen(s+2) > DBL_DIG)
        {
            printf("Number too large\n");
            continue;
        }
        switch (toupper(*s))
        {
            case '?' : 
                d = strtod(s+2,0);
                if (is_prime(d)) printf("%.0f is prime\n", d);
                else
                {
                    printf("%.0f is not prime\n", d);
                    prime_factors(d);
                }
                break;
            case 'Q' : fclose(f); exit(1);
            case 'V' : vm_stats(); break;
            default  : help(); break;
        }
    }
}

void help(void)
{
    puts("? [x] -- is x prime? if not, show prime factors");
    puts("Q -- quit");
    puts("V -- virtual memory stats");
}

BOOL is_prime(double x)
{
    ULONG i, n, lx, sqrt_x;
    int incr;
    
    if (x <= dsize)
    {
        lx = x;
        if ((lx == 2) || (lx == 3)) return TRUE;
        if ((lx % 2) && (lx % 3) && BIT_OFF(map, N(lx))) return TRUE;
        else return FALSE;
    }
    else
    {
        if (EQ(fmod(x,2),0) || EQ(fmod(x,3),0)) return FALSE;
        sqrt_x = sqrt(x);
        for (i=5, incr=4, n=1; i<=sqrt_x; i+=(incr=6-incr), n++)
            if (BIT_OFF(map, n) && EQ(fmod(x,i),0)) return FALSE;
        // still here -- primes are residue
        return TRUE;
    }
}

void prime_factors(double x)
{
    ULONG i, n;
    int incr;
    printf("Prime factors: ");
    while (! EQ(x,1))
    {
        if (is_prime(x))     { printf("%.0f\n", x); return; }
        if (EQ(fmod(x,2),0)) { printf("2 "); x /= 2; }
        if (EQ(fmod(x,3),0)) { printf("3 "); x /= 3; }
        for (i=5, incr=4, n=1; i <= x; i+=(incr=6-incr), n++)
            if (BIT_OFF(map, n) && EQ(fmod(x,i),0))
            {
                printf("%lu ", i);
                x /= i;
            }
    }
    printf("\n");
}

void vm_stats(void)
{
    ULONG buf[25];
    union REGS r;
#ifdef __WATCOMC__
    r.x.edx = buf;
    r.x.ebx = 0;        // don't reset VM stats
    r.x.eax = 0x2520;
#else
    r.x.dx = (unsigned) (void *) buf;
    r.x.bx = 0;
    r.x.ax = 0x2520;
#endif
    intdos(&r, &r);
    if (buf[0])         // VM is present
    {
        printf("VM active for %lu seconds\n", buf[11]);
        printf("Page faults: %lu\n", buf[12]);
        printf("Pages written to swap file: %lu\n", buf[13]);
        printf("Reclaimed pages: %lu\n", buf[14]);
        printf("Application physical pages: %lu\n", buf[5]);
        printf("Application VM pages: %lu\n", buf[15]);
        printf("Swap file pages: %lu\n", buf[16]);
    }
    else
        puts("VM not present");
}
