/* 
PROTDUMP.C -- protected-mode memory dump
Andrew Schulman, October 1993

bcc -DDPMI_APP -2 -B -c protdump.c
bcc -DDPMI_APP -2 protdump.obj dpmish.c prot.c vxdcalls.c vmwalk.c ctrl_c.asm 
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#define DPMI_APP
#include "dpmish.h"
#include "prot.h"
#include "vmwalk.h"

#define FAR_PTR     7       // magic

unsigned long axton(char *s);
void do_test(void);
void do_vm_list(void);
void do_all_vms(DWORD lin, WORD bytes);
DWORD do_lin_dump(DWORD lin, WORD bytes, char *mask, 
    int width, int size, int fudge);
void install_handlers(void);
void deinstall_handlers(void);
DWORD get_vm_high_linear(int vm);
WORD get_vm_ldt(int vm);
DWORD my_get_selector_base(int vm, WORD seg);
DWORD dpmi_phys_to_lin(DWORD phys_addr, DWORD num_bytes);
void dump(BYTE far *fp, WORD bytes, char *mask, DWORD addr, 
    int width, int size);

static char *usage = 
"usage: protdump [options] [#vm] <lin_addr or DOS:ofs or seg:ofs> [bytes]\n"
"options:\n"
"-all\tDump address in all VMs (requires VXD.386)\n"
"-ldt xx\tChange VMCB LDT offset to xx (default 0x114 in Windows 3.1)\n"
"-list\tTreat lin_addr as VMM linked list\n"
"-prot\tTreat seg:ofs as protected mode ptr\n"
"-phys\tUse physical, not linear address\n"
"-ptr\tTreat lin_addr as linear ptr to lin_addr (can be repeated)\n"
"-test\tMake Windows DOS extender allocate a selector\n"
"-vm\tList all virtual machines (requires VXD.386)\n"
"#vm\tVirtual Machine ID (requires VXD.386)\n"
"-n\tWidth for hex dump (default: 16)\n"
"-word, -dword, -fp  Dump words, dwords, or far pointers (default: bytes)\n\n"
"example (treat 804C1048 as ptr to VMM linked list; dump 8 bytes per item):\n"
"\tprotdump -ptr -list -dword 804c1048 8\n\n"
"example (in DOS 4+, show current PSP in all VMs):\n"
"\tprotdump -all -word DOS:330 2" ;

static int width = 16;
static int size = sizeof(BYTE);
static WORD dos_ds;
static DWORD phys_addr = 0;
static WORD ldt_ofs = 0x114;	// default for Windows 3.1

static void far *last_mapped_linear = (void far *) 0;

#define MY_MAP_LINEAR(base, limit)  \
    (last_mapped_linear = map_linear((base), (limit)))
        
#define MY_FREE_MAPPED_LINEAR(fp)   \
    { if ((fp) == last_mapped_linear) last_mapped_linear = 0; \
      free_mapped_linear(fp); }

void check_mapped(void)
{
    if (get_mapped()) 
    {
        if (last_mapped_linear) 
            MY_FREE_MAPPED_LINEAR(last_mapped_linear);
        if (get_mapped())
            printf("%d selectors remaining!\n", get_mapped());
    }
}
    
void fail(const char *s, ...)
{
    deinstall_handlers();
    puts(s); 
    check_mapped();
    _dos_exit(1); 
}

void panic(const char *s) { fail(s); }      // for Kelly code

int real_main(int argc, char *argv[])
{ 
	puts("PROTDUMP 1.0 -- Protected Mode Memory Hex Dump");
	puts("Copyright (c) Andrew Schulman 1993.  All rights reserved.\n");
	
    if (argc < 2) { puts(usage); return 1; }
    
    _asm mov ah, 52h
    _asm int 21h
    _asm mov word ptr dos_ds, es
    return 0; 
}

#define DO_LIST		1
#define DO_PROT		4
#define DO_STRINGS	8
#define DO_ALL		0x10
#define DO_PHYS		0x20
#define DO_SYS		0x40

int pmode_main(int argc, char *argv[])
{
    DWORD lin = (DWORD) -1L;
    void far *fp;
	char *cmd;
    WORD bytes = 128;
	int do_flags = 0;
	int do_ptr = 0;
    int vm = 0;
    int i;
	
    install_handlers(); // handle GP fault, page not present
    
    for (i=1; i<argc; i++)
    {
        int c = argv[i][0];
        char *s = strupr(&argv[i][1]);
        if (c == '-' || c == '/')
        {
            #define CMP(s1, s2) (strcmp((s1), (s2)) == 0)
            if (CMP(s, "LIST"))         { do_flags |= DO_LIST; do_ptr++; }
            else if (CMP(s, "PTR"))     do_ptr++;
            else if (CMP(s, "PROT"))    do_flags |= DO_PROT;
            else if (CMP(s, "PHYS"))    do_flags |= DO_PHYS;
            else if (CMP(s, "STRINGS")) do_flags |= DO_STRINGS;
            else if (CMP(s, "TEST"))    do_test();
            else if (CMP(s, "ALL"))     do_flags |= DO_ALL; 
            else if (CMP(s, "WORD"))    size = sizeof(WORD);
            else if (CMP(s, "DWORD"))   size = sizeof(DWORD);
            else if (CMP(s, "FP"))      size = FAR_PTR;
            else if (CMP(s, "VM"))      { do_vm_list(); return 0; }
			else if (CMP(s, "LDT"))	    { ldt_ofs = axton(argv[i+1]); i++; }
			else if (CMP(s, "SYS"))	{do_flags |= DO_SYS; cmd=argv[i+1];i++;}
            else if (isdigit(*s))       width = atoi(s);
            else                        fail(usage);
        }
        else if (c == '#')              vm = atoi(s);
        else if (strchr("0123456789abcdefABCDEF", c))   // ishexchar(c)
        {
            if (strchr(argv[i], ':'))   // segment:offset
            {
				DWORD ofs;	// allow 32-bit offsets for Win32 apps, etc.
                WORD seg;
                if (strncmp(argv[i], "DOS", 3) == 0)
                {
                    sscanf(argv[i], "DOS:%04X", &ofs);
                    lin = (((DWORD) dos_ds) << 4L) + ofs;
                }
                else
                {
                    sscanf(argv[i], "%04X:%08lX", &seg, &ofs);
                    if (do_flags & DO_PROT) 
                    {
                        lin = my_get_selector_base(vm, seg) + ofs;
                        if (lin > 0x80000000L)  
                            vm = 0; // address isn't VM relative
						// should also get selector limit and use that
						// as max #bytes to map!
                    }
                    else         
                        lin = (((DWORD) seg) << 4L) + ofs;
                }
            }
            else
            {   // first number is lin_addr; second is optional #bytes
                if (lin == (DWORD) -1L) sscanf(argv[i], "%08lX", &lin);
                else                    bytes = axton(argv[i]);
            }
        }
        else                    fail(usage);
    }
    if (size > width)           width=size;
    
    if (vm != 0)        // treat lin_addr or seg:ofs as VM relative
    {
        DWORD high_lin = get_vm_high_linear(vm);
        if (! high_lin) fail("Can't get VM high linear address");
        else lin += high_lin;
    }
    
    while (do_ptr)      // treat lin_addr as lin ptr to lin_addr
    {
        DWORD far *fpd = (DWORD far *) MY_MAP_LINEAR(lin, sizeof(DWORD));
        if (! fpd) fail("can't get pointer");
        lin = *fpd;
        MY_FREE_MAPPED_LINEAR(fpd);
        if ((do_flags & DO_LIST) && lin == 0) fail("list empty");
        do_ptr--;
    }
    
    if (do_flags & DO_PHYS)
    {
        phys_addr = lin;
        if (! (lin = dpmi_phys_to_lin(phys_addr, bytes)))
            fail("Can't map physical address");
    }
    
    if (do_flags & DO_ALL) 
        do_all_vms(lin, bytes);
    else if (do_flags & DO_LIST)       // linked list (link at addr-4)
        while (lin != 0)
            lin = do_lin_dump(lin, bytes, "%08lX", width, size, 4);
    else
        do_lin_dump(lin, bytes, "%08lX", width, size, 0);
	
	if (do_flags & DO_SYS)
		system(cmd);	// fix!! need to call 21/4B in real mode

    check_mapped();
    return 0;
}

void do_test(void)  // make Windows DOS extender allocate a selector
{
    WORD dos_seg;
    _asm mov ah, 52h
    _asm int 21h
    _asm mov dos_seg, es
    printf("DOS=%04X base=%08lX\n", dos_seg, GetSelectorBase(dos_seg));
	printf("%u para DPMI host data at seg %04xh\n",
		hostdata_para, hostdata_seg);	// see DPMISH.C
}

/***********************************************************************/

static DWORD high_lin, rel_lin;
static WORD num_bytes, want_vm, vm_ldt;

BOOL find_vm_id(DWORD vm, VM_CB far *vm_cb)
{
    if (vm_cb->CB_VMID == want_vm)
    {
        high_lin = vm_cb->CB_High_Linear;
        vm_ldt = *((WORD far *) (((BYTE far *) vm_cb) + ldt_ofs));
        return 0;   // can stop looking
    }
    return 1;
}

DWORD get_vm_high_linear(int vm)
{
    high_lin = 0;
    want_vm = vm;
    vmwalk(find_vm_id);
    return high_lin;
}

BOOL list_all_vms(DWORD vm, VM_CB far *vm_cb)
{
    printf("#%lu\tVM CB=%08lX\thigh lin=%08lX\n",
        vm_cb->CB_VMID, vm, vm_cb->CB_High_Linear);
    return 1;
}

void do_vm_list(void) { vmwalk(list_all_vms); }
            
BOOL dump_all_vms(DWORD vm, VM_CB far *vm_cb)
{
    void far *fp;
    DWORD lin = rel_lin + vm_cb->CB_High_Linear;
    if (num_bytes >= width) printf("\n#%lu\n", vm_cb->CB_VMID);
    else                    printf("#%lu ", vm_cb->CB_VMID);
    do_lin_dump(lin, num_bytes, "%08lX", width, size, 0);
    return 1;
}

void do_all_vms(DWORD lin, WORD bytes)
{
    rel_lin = lin; 
    num_bytes = bytes;
    vmwalk(dump_all_vms);
}

/***********************************************************************/

DWORD axton(char *s)
{
    DWORD ret;
    if (s[0] == '0' && s[1] == 'x')
        sscanf(&s[2], "%lx", &ret);
    else
        ret = atol(s);
    return ret;
}

DWORD do_lin_dump(DWORD lin, WORD bytes, char *mask, 
    int width, int size, int fudge)
{
    BYTE far *fp;
    DWORD ret;
    if (! (fp = (BYTE far *) MY_MAP_LINEAR(lin-fudge, 
        bytes+fudge+sizeof(DWORD))))
        fail("can't map linear");
    if (phys_addr) lin = phys_addr;
    dump(fp+fudge, bytes, mask, lin+fudge, width, size);
    ret = *((DWORD far *) fp);  // some callers want this
    MY_FREE_MAPPED_LINEAR(fp);
    return ret;
}
        
void dump(BYTE far *fp, WORD bytes, char *mask, DWORD addr, 
    int width, int size)
{
    DWORD far *pd;
    WORD far *pw;
    BYTE far *p;
    WORD i, j, c;
    
    for (i=0; i<bytes; i+=width)
    {
        if (ctrl_c_hit)
        {
            MY_FREE_MAPPED_LINEAR(fp);     // should first check if mapped!!
            puts("\nCtrl-C detected");
            _dos_exit(0);
        }
        c = ((bytes-i) > width) ? width : bytes-i;
        printf(mask, addr+i);
        printf(" | ");
        switch (size)
        {
            case sizeof(DWORD):
                for (j=0, pd=(DWORD far *) (fp+i); j<c; pd++, j+=4) 
                    printf("%08lX ", *pd);
                break;
            case FAR_PTR:
                for (j=0, pd=(DWORD far *) (fp+i); j<c; pd++, j+=4) 
                    printf("%Fp ", *pd);
                break;
            case sizeof(WORD):
                for (j=0, pw=(WORD far *) (fp+i); j<c; pw++, j+=2) 
                    printf("%04X ", *pw);
                break;
            default:    // sizeof(BYTE); also show ASCII on right
                for (j=c, p=fp+i; j--; p++) 
                    printf("%02X ", *p);
                for (j=width-c; j--; )  // pad out on last line
                    printf("   ");
                printf("| ");
                for (j=c, p=fp+i; j--; p++)
                    putchar( isprint(*p) ? *p : '.' );
                break;
        }
        putchar('\n');
    }
    if (bytes > width) putchar('\n');
}

/**********************************************************************/

typedef void (far *FUNCPTR)(void);

static FUNCPTR prev_page_fault = (FUNCPTR) 0;
static FUNCPTR prev_gp_fault = (FUNCPTR) 0;

void page_fault(void)   { fail("\nPage fault detected!"); }
void gp_fault(void)     { fail("\nGP fault detected!"); }

FUNCPTR _dpmi_get_exception_handler(int intno)
{
    _asm mov ax, 0202h
    _asm mov bl, byte ptr intno
    _asm int 31h
    _asm jnc ok
error:
    return (FUNCPTR) 0;
ok:     
    _asm mov ax, dx
    _asm mov dx, cx
}

int _dpmi_set_exception_handler(int intno, FUNCPTR handler)
{
    _asm mov ax, 0203h
    _asm mov cx, word ptr handler+2
    _asm mov dx, word ptr handler
    _asm mov bl, byte ptr intno
    _asm int 31h
    _asm jc error
ok:    return 1;
error: return 0;
}

void install_handlers(void)
{
    if (! (prev_gp_fault = _dpmi_get_exception_handler(0x0d)))
        fail("Can't get GP fault handler!");
    if (! (prev_page_fault = _dpmi_get_exception_handler(0x0e)))
        fail("Can't get page fault handler!");
    if (! _dpmi_set_exception_handler(0x0d, (void (far *)()) gp_fault))
        fail("Can't set GP fault handler!");
    if (! _dpmi_set_exception_handler(0x0e, (void (far *)()) page_fault))
        fail("Can't set page fault handler!");
}

void deinstall_handlers(void)
{
    if (prev_gp_fault) _dpmi_set_exception_handler(0x0d, prev_gp_fault);
    if (prev_page_fault) _dpmi_set_exception_handler(0x0e, prev_page_fault);
}

/**********************************************************************/

#pragma pack(1)

typedef struct { WORD limit; DWORD base; } GDTR;

typedef struct {
    unsigned rpl : 2;
    unsigned ti : 1;
    unsigned index : 13;
    } SELECTOR; 

typedef struct {
    WORD limit_lo, base_lo;
    BYTE base_hi, rts_lo, limitrts_hi, base_xhi;
    } DESCRIPTOR;

#if 1
extern void sgdt(GDTR far *pgdtr);
#else
/* C wrapper for the Intel SGDT instruction; must compile with 286
   instructions (-G2 in Microsoft C; -2 in Borland C++). Places the
   Global Descriptor Table (GDT) base and limit into the six-byte
   (FWORD PTR) structure pointed to by pgdtr. In Borland, compile
   via TASM with -B (built-in assembler doesn't like FWORD) */
void sgdt(GDTR far *pgdtr)
{
    _asm les bx, pgdtr
    _asm sgdt fword ptr es:[bx]
}
#endif

// when done, free with MY_FREE_MAPPED_LINEAR
DESCRIPTOR far *get_gdt(WORD *plimit)
{
    DESCRIPTOR far *gdt;
    GDTR gdtr;
    WORD sel;
    
    /* get the linear base address and size (limit) of the Global
       Descriptor Table (GDT), using the Intel SGDT instruction */
    sgdt(&gdtr);
    gdt = (DESCRIPTOR far *) MY_MAP_LINEAR(gdtr.base, gdtr.limit + 1);
    *plimit = gdtr.limit;
    return gdt;
}

BOOL is_system_seg(DESCRIPTOR far *desc)
{
    return (desc->rts_lo & 0x10);
}

// get_base and get_limit not valid for some SYSTEM descriptors

DWORD get_base(DESCRIPTOR far *desc)
{
    return ((DWORD) desc->base_xhi << 24L) + ((DWORD) desc->base_hi << 16L) +
       desc->base_lo;
}

DWORD get_limit(DESCRIPTOR far *desc)
{
    DWORD limit = ((DWORD) (desc->limitrts_hi & 0x0F) << 16L) + 
        (DWORD) desc->limit_lo;
    if (desc->limitrts_hi & 0x80)   // page granularity
        limit *= 4096;
    return limit;
}

WORD sldt(void)
{
    _asm sldt ax
}

WORD get_vm_ldt(int vm)
{
    if (vm == 0) return sldt();
    vm_ldt = 0;
    want_vm = vm;
    vmwalk(find_vm_id);
    return vm_ldt;
}

DESCRIPTOR far *get_ldt(int vm, WORD *plimit)
{
    DWORD limit, base;
    WORD gdt_limit;
    DESCRIPTOR far *gdt = get_gdt(&gdt_limit);
    WORD ldt_sel = (vm == 0) ? sldt() : get_vm_ldt(vm);
    DESCRIPTOR far *ldt_desc = &gdt[ldt_sel >> 3];  // LDT descr. is in GDT
    *plimit = limit = get_limit(ldt_desc);
    base = get_base(ldt_desc);
    MY_FREE_MAPPED_LINEAR(gdt);
    return (DESCRIPTOR far *) MY_MAP_LINEAR(base, limit + 1);
}

DWORD my_get_selector_base(int vm, WORD seg)
{
    SELECTOR sel = *((SELECTOR *) &seg);
    DWORD base;
    if (sel.ti == 0)    // GDT
    {
        WORD gdt_limit;
        DESCRIPTOR far *gdt = get_gdt(&gdt_limit);
        if ((seg & ~8) > gdt_limit)
            fail("Invalid GDT selector");
        base = get_base(&gdt[sel.index]);
        MY_FREE_MAPPED_LINEAR(gdt);
    }
    else                // LDT
    {
        WORD ldt_limit;
        DESCRIPTOR far *ldt = get_ldt(vm, &ldt_limit);
        if ((seg & ~8) > ldt_limit)
            fail("Invalid LDT selector");
        base = get_base(&ldt[sel.index]);
        MY_FREE_MAPPED_LINEAR(ldt);
    }
    return base;
}

/**********************************************************************/

// not valid for phys_addr under one megabyte?
// note that free call (31/0801) not supported in DPMI 0.9!!
DWORD dpmi_phys_to_lin(DWORD phys_addr, DWORD num_bytes)
{
    _asm push si
    _asm push di
    _asm mov ax, 0800h
    _asm mov bx, word ptr [phys_addr+2]
    _asm mov cx, word ptr [phys_addr]
    _asm mov si, word ptr [num_bytes+2]
    _asm mov di, word ptr [num_bytes]
    _asm int 31h
    _asm jc error
    _asm mov dx, bx
    _asm mov ax, cx
    _asm jmp short done
error:
    _asm xor ax, ax
    _asm xor dx, dx
done:
    _asm pop di
    _asm pop si
    // return value in DX:AX
}

