/* 
REGDUMP.C
Win32 Console app to dump registry
Includes HKEY_CURRENT_CONFIG and HKEY_DYN_DATA (which, in turn, includes
Config Manager and PerfStats)
            
To do:

Strip out duplicates: 
	HKEY_CURRENT_CONFIG -> HKEY_LOCAL_MACHINE\Config\<<current config>>
	<<current config>> = 
		HKEY_LOCAL_MACHINE\System\CurrentControlSet\Control\IDConfigDB\
		CurrentConfig
			
	HKEY_CLASSES_ROOT -> HKEY_LOCAL_MACHINE\SOFTWARE\Classes
		
	HKEY_DYN_DATA -> HKEY_LOCAL_MACHINE\???
		
	HKEY_CURRENT_USER -> HKEY_USERS

Decode/resolve CLSIDs!
Any way to enumerate .INI file contents?
Relate this to OLE?  Connect with various OLE browers
Do a real-mode DOS version?

Andrew Schulman
andrew@ora.com
August 1995

cl regdump.c advapi32.lib

Sample output:

HKEY_CLASSES_ROOT
  Word.Document.6 -> "Microsoft Word 6.0 Document"
    shell
      open
        command -> "C:\WINWORD\WINWORD.EXE /w"
        ddeexec -> '[FileOpen("%1")]'
      print
        command -> "C:\WINWORD\WINWORD.EXE /w"
        ddeexec -> '[FileOpen("%1")][FilePrint()][DocClose(2)]'
          ifexec -> '[FileOpen("%1")][FilePrint .Background = 0][FileExit(2)]'
    CLSID -> "{00020900-0000-0000-C000-000000000046}"
    Insertable
    protocol
      StdFileEditing
        verb
          0 -> "Edit"
        server -> "C:\WINWORD\WINWORD.EXE"
  Word.Picture -> "Microsoft Word Picture"
    CurVer -> "Word.Picture.6"
    CLSID -> "{00020901-0000-0000-C000-000000000046}"
  Word.Document -> "Microsoft Word Document"
    CurVer -> "Word.Document.6"
    CLSID -> "{00020900-0000-0000-C000-000000000046}"
  .RTF -> "Word.Document.6"
    ShellNew
  .DOT -> "Word.Document.6"
  .DOC -> "Word.Document.6"
    ShellNew      NullFile=""
    Wordpad.Document.1
      ShellNew        FileName="winword.doc"
    ... (output is over 13,000 lines) ...
        
Keys: 6366
Values: 10396

HKEY_DYN_DATA\Config Manager\Enum is hardware tree:     
        
HKEY_DYN_DATA
   Config Manager
      Enum
         C112D198
            HardWareKey="HTREE\RESERVED\0"
            Problem=[0h]
            Status=[84Eh]
            Allocation=BIN (12 bytes)
         C112D278
            HardWareKey="ROOT\*PNP0000\0000"
            Problem=[0h]
            Status=[4A4Fh]
            Allocation=BIN (68 bytes)
        ....
            
Wrote CLSID.AWK:
# CLSID.AWK

BEGIN {
    f = "CLSID.DAT";
    while (getline < f)
        clsid[$2] = $1;
    close(f);
    }

$2 == "!" {
    clsid[$5] = $1;
    }

/\{.*\}/ {
    for (i=1; i<=NF; i++)
        if ($i ~ /\{.*\}/) {
            if ($i in clsid) 
                sub($i, "<<" clsid[$i] ">>", $0);
            break;
            }
}

{
    print $0;
}

END {
    # change to add new entries to CLSID.DAT
    print "\nCLSIDs:";
    for (x in clsid)
        print clsid[x], x;
    }

Sample output from REGDUMP | CLSID:
      .lnk -> "lnkfile"
        ShellNew
          Command="C:\WINDOWS\rundll32.exe AppWiz.Cpl,NewLinkHere %1"
      lnkfile -> "Shortcut"
        EditFlags=[1]
        IsShortcut=""
        NeverShowExt=""
        shellex
          IconHandler -> <<lnkfile>>
          DropHandler -> <<lnkfile>>
          ContextMenuHandlers
            {00021401-0000-0000-C000-000000000046} -> ""
        lnkfile ! CLSID -> <<lnkfile>>

Where original REGDUMP output is:
  lnkfile -> "Shortcut"
    EditFlags=[1]
    IsShortcut=""
    NeverShowExt=""
    shellex
      IconHandler -> "{00021401-0000-0000-C000-000000000046}"
      DropHandler -> "{00021401-0000-0000-C000-000000000046}"
      ContextMenuHandlers
        {00021401-0000-0000-C000-000000000046} -> ""
    lnkfile ! CLSID -> "{00021401-0000-0000-C000-000000000046}"

Note that REGDUMP itself changes all:
    CLSID -> "{...}" 
to:
    name ! CLSID -> "{...}"
which makes life easier for CLSID.AWK

This still leaves many {...} in output, however; for example:
    {00028B00-0000-0000-C000-000000000046} -> "The Microsoft Network"
These are stored under Classes\CLSID:
      Classes ! CLSID -> ""
        {00030003-0000-0000-C000-000000000046}
          NotInsertable
          Ole1Class -> "WordDocument"
          ProgID -> "WordDocument"
          AutoConvertTo -> "{00020900-0000-0000-C000-000000000046}"
        ...
        {00020900-0000-0000-C000-000000000046} -> "Microsoft Word 6.0 Document"
          ProgID -> "Word.Document.6"
          InprocHandler -> "ole2.dll"
          LocalServer -> "C:\WINWORD\WINWORD.EXE"
          verb
          ....

Need to look for this in CLSID.AWK???             
    if str2 == CLSID"
        if prev[level-1] == "Classes" 
            in_classes = 1 
        else                          
            in_classes = 0
    else if str2 == "ProgId" 
        if in_classes == 1
            print prev[level-1] "!" 
Or something like that...               
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#define WIN32_LEAN_AND_MEAN
#include "windows.h"            // includes winnt.h and winreg.h

void indent(int level)
{
    int i;
    for (i=0; i<level; i++)
        fputs("  ", stdout);
}

void dump(unsigned char *fp, unsigned bytes, 
    char *mask, unsigned long addr, int width)
{
    unsigned char *p;
    unsigned i, j, c;
    
    for (i=0; i<bytes; i += width)
    {
        c = ((bytes-i) > width) ? width : bytes-i;
        printf(mask, addr+i);
        putchar(' '); putchar('|'); putchar(' ');
        for (j=c, p=fp+i; j--; p++)  printf("%02X ", *p);
        for (j=width-c; j--; )  fputs("   ", stdout); // pad out on last line
        putchar('|'); putchar(' ');
        for (j=c, p=fp+i; j--; p++) putchar( isprint(*p) ? *p : '.' );
        putchar('\n');
    }
}

static int do_dump = 0;

void printdword(DWORD dw, int brackets)
{
    if (brackets)   putchar('[');
    if (dw < 10)    printf("%d", dw);
    else            printf("%0Xh", dw);
    if (brackets)   putchar(']');
}

void printvalue(DWORD type, char *data, DWORD cb)
{
    switch (type)
    {
        case REG_EXPAND_SZ: 
        case REG_SZ: 
            if (strchr(data, '\"'))
                printf("\'%s\'", data); 
            else
                printf("\"%s\"", data); 
            break;
        case REG_DWORD:
            printdword(*((DWORD *) data), 0);
            break;
        case REG_BINARY:
            switch (cb)
            {
                case 1: printdword(*((BYTE *) data), 1); break;
                case 2: printdword(*((WORD *) data), 1); break;
                case 4: printdword(*((DWORD *) data), 1); break;
                default:
                    printf("BIN (%d bytes)", cb); 
                    if (do_dump)
                    {
                        putchar('\n');
                        dump(data, cb, "%04X", 0, 16);
                    }
                    break;
            }
            break;
        default:
            printf("<type %d>", type);
            break;
    }
}

static DWORD numvalue = 0, numkey = 0;

void enumvalue(int level, HKEY hkey, char *str)
{
    char name[256], data[1024];
    DWORD size, cb, type, iValue;
    for (iValue=0; ;iValue++)
    {
        size = 256;
        cb = 1024;
        if (RegEnumValue(hkey, iValue, name, &size, 0, &type, data, &cb) !=
            ERROR_SUCCESS)
            break;
        if (size && *name)
        {
            putchar('\n');
            indent(level);
            printf("%s=", name);
        }
        else    // no new line
            printf(" -> ");
        printvalue(type, data, cb);
        numvalue++;
    }
}

void queryinfo(HKEY hkey)   // what is this "class" stuff? NT only?
{
    char class[1024];
    DWORD cb = 1024;
    DWORD dummy;
    FILETIME ft;
    if (RegQueryInfoKey(hkey, class, &cb, NULL, &dummy, &dummy, &dummy,
        &dummy, &dummy, &dummy, &dummy, &ft) == ERROR_SUCCESS)
        if (cb && *class)
            printf("<%s> ", class);
}

static char prev[256][8];

void check_clsid(int level, const char *str)
{
#if 0
    static int in_classes = 0;
    char str2[256];
    strcpy(str2, str);
    strupr(str2);
    if (strcmp(str2, "CLSID") == 0)
    {
        if (strcmp(prev[level-1], "Classes") == 0)
        {
            fprintf(stderr, "Turned classes on\n");
            in_classes = 1;
        }
        else
        {
            in_classes = 0;
            printf("%s ! ", prev[level-1]);
        }
    }
    else
    {
        if (strcmp(str2, "PROGID") == 0)
        {
            fprintf(stderr, "prev <%s>\n", prev[level-1]);
        }
        sscanf(str, "%s ", prev[level]);    // get first token?
    }
#else
    char str2[256];
    strcpy(str2, str);
    strupr(str2);
    if (strcmp(str2, "CLSID") == 0)
        printf("%s ! ", prev[level-1]);
    else
        sscanf(str, "%s ", prev[level]);    // get first token?
#endif          
}

void enumkey(int level, HKEY hkey, char *str)
{
    char name[256];
    HKEY hkey2;
    DWORD iSubKey = 0;
    if (str == NULL)
    {
        if (RegOpenKey(hkey, NULL, &hkey2) == ERROR_SUCCESS)
        {
            // try to get name, using RegQueryInfoKey?
            sprintf(name, "[#%08lX]", hkey);
            str = name;
            RegCloseKey(hkey2);
        }
        else
            return;
    }
    putchar('\n');
    indent(level);
    queryinfo(hkey);
    check_clsid(level, str);
    fputs(str, stdout);
    enumvalue(level+1, hkey, str);
    while (RegEnumKey(hkey, iSubKey, name, 256) == ERROR_SUCCESS)
    {
        if (RegOpenKeyEx(hkey, name, 0, 
            KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS | KEY_READ,
            &hkey2) == ERROR_SUCCESS)
        {
            enumkey(level+1, hkey2, name);  // recursive call!!
            RegCloseKey(hkey2);
        }
        iSubKey++;
        numkey++;
    }
}

void fail(const char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    int i;
    int did_enum = 0;
    char *s;
    for (i=1; i<argc; i++)
    {
        s = argv[i];
        if (*s == '-' || *s == '/')
        {
            if (strcmp(strupr(s+1), "DUMP") == 0)
                do_dump++;
            else
                fail("usage: regdump [-dump] [key#]");
        }
        else
        {
            enumkey(0, (HKEY) ((DWORD) HKEY_CLASSES_ROOT + atoi(s)), 
                "??");
            did_enum++;
        }
    }
    
    if (! did_enum)
    {
        char *names[] = {
            "HKEY_CLASSES_ROOT", "HKEY_CURRENT_USER",
            "HKEY_LOCAL_MACHINE", "HKEY_USERS",
            "HKEY_PERFORMANCE_DATA", "HKEY_CURRENT_CONFIG",
            "HKEY_DYN_DATA",
            } ;
        #define NUM_NAMES (sizeof(names) / sizeof(char*))
        int i;
            
        // not in include file, but REGEDIT does it
        #define HKEY_CURRENT_CONFIG     (HKEY) ((DWORD)HKEY_CLASSES_ROOT+5)
        #define HKEY_DYN_DATA           (HKEY) ((DWORD)HKEY_CLASSES_ROOT+6)
            
        // for (i=0; i<0x10000; i++)
        for (i=0; i<NUM_NAMES; i++)
            enumkey(0, (HKEY) ((DWORD) HKEY_CLASSES_ROOT + i), 
                (i < NUM_NAMES) ? names[i] : NULL);
    }
    
    printf("\n\nKeys: %u\n", numkey);
    printf("Values: %u\n", numvalue);
}


