/*
relation.c
Andrew Schulman
18 April 1989

Assignment:
A relation R on a set A containing n elements is given by its matrix M`R.
Write a program to do the following:
    i)      Read in M`R.
    ii)     Print the relation using the ordered pair notation.
    iii)    Determine if the relation R is reflexive? symmetric?
            transitive? asymmetric? antisymmetric? an equivalence relation?
    iv)     Print the relations R V S, R ^ S, R' and R^-1 using the ordered
            pair notation and their matrices.
Use the following test data:
    A = {1,2,3,4}
    
            1 1 0 0
    M`R =   1 1 0 0
            0 0 1 1
            0 0 1 1
                
    M`S =   1 0 0 1
            0 0 1 0
            0 0 1 0
            1 0 0 1
                
Comments:
    i)      Given that the test data involves relations on a set of 4
            elements, we could hard-wire in [4][4] as the size of the
            matrices.  However, this would make the code worthless for
            other uses, or even for other test data.  A matrix is
            instead a dynamically-allocated array of pointers to rows,
            which, in turn, are dynamically-allocated arrays.  Pointers
            are decremented so that m[1][1] points to the first column
            of the first row (this idea comes from Press, Flannery,
            Teukolsky, and Vetterling, "Numerical Recipes in C").
            Finally, rather than have to pass around the dimensions
            of the matrix as function parameters (as is done in
            "Numerical Recipes"), here we instead store the dimensions
            away inside the matrix itself, in m[0][0] and m[0][1].
                
    ii)     Sets and matrices are read from stdin.  The members of the
            set should separated by commas, and enclosed within curly
            braces.  For the two matrices, individual elements within a
            row must be separated by whitespace, and rows must be
            separated by newlines.  A text files containing a matrix
            may be read in by redirecting stdin, e.g., C>relation < mat.dat
                
Sample data in MAT.DAT:
    { 1, 2, 3, 4 }

    1 1 0 0
    1 1 0 0
    0 0 1 1
    0 0 1 1

    1 0 0 1
    0 0 1 0
    0 0 1 0
    1 0 0 1

Output:
C:\HOMEWORK>relation < mat.dat
R
1 1 0 0 
1 1 0 0 
0 0 1 1 
0 0 1 1 
{(1,1), (1,2), (2,1), (2,2), (3,3), (3,4), (4,3), (4,4)}
Relation is Reflexive
Relation is Symmetric
Relation is Transitive
    asymmetric test failed: (1,1)
Relation is not Asymmetric
    antisymmetric test failed: (1,2) and (2,1), but 1 != 2
Relation is not Antisymmetric
Relation is an Equivalence Relation

S
1 0 0 1 
0 0 1 0 
0 0 1 0 
1 0 0 1 
{(1,1), (1,4), (2,3), (3,3), (4,1), (4,4)}
    reflexive test failed: no (2,2)
Relation is not Reflexive
    symmetric test failed: (2,3) but no (3,2)
Relation is not Symmetric
Relation is Transitive
    asymmetric test failed: (1,1)
Relation is not Asymmetric
    antisymmetric test failed: (1,4) and (4,1), but 1 != 4
Relation is not Antisymmetric
    reflexive test failed: no (2,2)
    equivalence relation test failed: not reflexive
Relation is not an Equivalence Relation

R V S
1 1 0 1 
1 1 1 0 
0 0 1 1 
1 0 1 1 
{(1,1), (1,2), (1,4), (2,1), (2,2), (2,3), (3,3), (3,4), (4,1), (4,3), (4,4)}

R ^ S
1 0 0 0 
0 0 0 0 
0 0 1 0 
0 0 0 1 
{(1,1), (3,3), (4,4)}

R'
0 0 1 1 
0 0 1 1 
1 1 0 0 
1 1 0 0 
{(1,3), (1,4), (2,3), (2,4), (3,1), (3,2), (4,1), (4,2)}

R^-1
1 1 0 0 
1 1 0 0 
0 0 1 1 
0 0 1 1 
{(1,1), (1,2), (2,1), (2,2), (3,3), (3,4), (4,3), (4,4)}
*/

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>

#define TRUE 1
#define FALSE 0

#define MAX_ELEM        32

/* rows and cols stored inside matrix */
#define HEADER(m)       ((m)[0])
#define ROWS(m)         (HEADER(m)[0])
#define COLS(m)         (HEADER(m)[1])

/* cardinality of set stored inside array of set_names */
#define N(s)            ((WORD) ((s)[0]))

typedef unsigned WORD;
typedef WORD BOOL;
typedef WORD *ROW;
typedef ROW *MATRIX;
typedef char **SET_NAMES;

MATRIX matrix(WORD rows, WORD cols);
MATRIX read_matrix(WORD rows, WORD cols);
MATRIX join(MATRIX r, MATRIX s);
MATRIX meet(MATRIX r, MATRIX s);
MATRIX inverse(MATRIX r);
MATRIX complement(MATRIX r);
BOOL reflexive(MATRIX m);
BOOL symmetric(MATRIX m);
BOOL transitive(MATRIX m);
BOOL asymmetric(MATRIX m);
BOOL antisymmetric(MATRIX m);
BOOL equiv_rel(MATRIX m);
SET_NAMES read_set_names(void);
void free_matrix(MATRIX m);
void fail(char *msg);
void print_matrix(MATRIX m);
void print_ordered_pairs(SET_NAMES s, MATRIX m);
void properties(MATRIX m);
void property(MATRIX m, BOOL (*pred)(MATRIX m), char *msg);
void show(MATRIX m, SET_NAMES s, char *msg, BOOL show_prop, BOOL free_m);

/**********************************************************************/

main(void)
{
    MATRIX r, s;
    SET_NAMES a;
    
    a = read_set_names();
    r = read_matrix(N(a), N(a));
    s = read_matrix(N(a), N(a));

    show(r, a, "R", TRUE, FALSE);
    show(s, a, "S", TRUE, FALSE);
    show(join(r, s), a, "R V S", FALSE, TRUE);
    show(meet(r, s), a, "R ^ S", FALSE, TRUE);
    show(complement(r), a, "R'", FALSE, TRUE);
    show(inverse(r), a, "R^-1", FALSE, TRUE);

    free_matrix(r);
    free_matrix(s);
}

/**********************************************************************/

/* read in names of set elements, in the form "{ A, B, C, D }" */
SET_NAMES read_set_names(void)
{
    SET_NAMES s;
    static char buf[256];
    WORD i;
    s = (SET_NAMES) calloc(MAX_ELEM, sizeof(char *));
    gets(buf);
    s[1] = strtok(&buf[2], ", \t");     /* yuk! hard-wired! */
    i = 2;
    while (s[i] = strtok(NULL, ", \t}"))
    {
        if (i >= MAX_ELEM)
            break;
        i++;
    }
    s[i] = NULL;
    N(s) = --i;
    return s;
}

/* reads text representation of a matrix from stdin; returns a MATRIX */
MATRIX read_matrix(WORD rows, WORD cols)
{
    char buf[256];
    char ch;
    MATRIX m;
    WORD r,c;
    char *s;
    
    m = matrix(rows,cols);
    r = 0;
    
    /* skip blank lines */
    while ((ch = getchar()) == '\n')
        ;
    ungetc(ch, stdin);

    while (gets(buf))       /* get next row */
    {
        r++;
        if (r > rows)
            break;
        if (s = strtok(buf, " \t,"))
        {
            c = 1;
            m[r][c] = atoi(s);
            while (s = strtok(NULL," \t,")) /* get next element within row */
            {
                c++;
                if (c > cols)
                    break;
                m[r][c] = atoi(s);
            }
        }
        else
            break;
    }

    if (! r)
        fail("read_matrix failure");
    
    return m;
}

void show(MATRIX m, SET_NAMES s, char *msg, BOOL show_prop, BOOL free_m)
{
    printf("\n%s\n", msg);
    print_matrix(m);
    print_ordered_pairs(s, m);
    if (show_prop) properties(m);
    if (free_m) free_matrix(m);
}

/**********************************************************************/
    
void print_matrix(MATRIX m)
{
    WORD i,j;
    for (i=1; i<=ROWS(m); i++)
    {
        for (j=1; j<=COLS(m); j++)
            printf("%u ", m[i][j]);
        printf("\n");
    }
}

void print_ordered_pairs(SET_NAMES s, MATRIX m)
{
    WORD i,j;
    BOOL first_time = TRUE;
    printf("{");
    for (i=1; i<=ROWS(m); i++)
        for (j=1; j<=COLS(m); j++)
            if (m[i][j])
            {
                printf("%s(%s,%s)", (first_time) ? "" : ", ", s[i], s[j]);
                first_time = FALSE;
            }
    printf("}\n");
}

void properties(MATRIX m)
{
    property(m, reflexive, "Reflexive");
    property(m, symmetric, "Symmetric");
    property(m, transitive, "Transitive");
    property(m, asymmetric, "Asymmetric");
    property(m, antisymmetric, "Antisymmetric");
    property(m, equiv_rel, "an Equivalence Relation");
}

/**********************************************************************/
    
void property(MATRIX m, BOOL (*pred)(MATRIX m), char *msg)
{
    if ((*pred)(m))
        printf("Relation is %s\n", msg);
    else
        printf("Relation is not %s\n", msg);
}

BOOL reflexive(MATRIX m)
{
    WORD i;
    for (i=1; i<=ROWS(m); i++)
        if (! m[i][i])
        {
            printf("\treflexive test failed: no (%u,%u)\n", i,i);
            return FALSE;
        }
    /* still here */
    return TRUE;
}

BOOL symmetric(MATRIX m)
{
    WORD i,j;
    for (i=1; i<=ROWS(m); i++)
        for (j=i; j<=COLS(m); j++)      /* start with j=i, not j=1 */
            if (m[i][j] && (! m[j][i]))
            {
                printf("\tsymmetric test failed: (%u,%u) but no (%u,%u)\n",
                    i, j, j, i);
                return FALSE;
            }
            else if (m[j][i] && (! m[i][j]))
            {
                printf("\tsymmetric test failed: (%u,%u) but no (%u,%u)\n",
                    j, i, i, j);
                return FALSE;
            }
    /* still here */
    return TRUE;
}

BOOL transitive(MATRIX m)
{
    WORD i,j,k;
    for (i=1; i<=ROWS(m); i++)
        for (j=1; j<=COLS(m); j++)
            if (m[i][j])
                for (k=1; k<=COLS(m); k++)
                    if (m[j][k] && (! m[i][k]))
                    {
                        printf("\ttransitive test failed: (%u,%u) and \
(%u,%u) but no (%u,%u)\n",
                            i, j, j, k, i, k);
                        return FALSE;
                    }
    /* still here */
    return TRUE;
}

BOOL asymmetric(MATRIX m)
{
    WORD i,j;
    for (i=1; i<=ROWS(m); i++)
        for (j=i; j<=COLS(m); j++)      /* start with j=i, not j=1 */
            if (m[i][j] && m[j][i])
            {
                if (i == j)
                    printf("\tasymmetric test failed: (%u,%u)\n", i, j);
                else
                    printf("\tasymmetric test failed: (%u,%u) and (%u,%u)\n",
                        i, j, j, i);
                return FALSE;
            }
    /* still here */
    return TRUE;
}

BOOL antisymmetric(MATRIX m)
{
    WORD i,j;
    for (i=1; i<=ROWS(m); i++)
        for (j=i; j<=COLS(m); j++)      /* start with j=i, not j=1 */
            if (m[i][j] && m[j][i] && (i != j))
            {
                printf("\tantisymmetric test failed: (%u,%u) and (%u,%u), \
but %u != %u\n",                    
                    i, j, j, i, i, j);
                return FALSE;
            }
    /* still here */
    return TRUE;
}

BOOL equiv_rel(MATRIX m)
{
    if (! reflexive(m))
    {
        printf("\tequivalence relation test failed: not reflexive\n");
        return FALSE;
    }
    if (! symmetric(m))
    {
        printf("\tequivalence relation test failed: not symmetric\n");
        return FALSE;
    }
    if (! transitive(m))
    {
        printf("\tequivalence relation test failed: not transitive\n");
        return FALSE;
    }
    /* still here */
    return TRUE;
}

/**********************************************************************/
    
MATRIX join(MATRIX r, MATRIX s)
{
    MATRIX m;
    WORD i,j;
    if ((ROWS(r) != ROWS(s)) || (COLS(r) != COLS(s)))
        return (MATRIX) 0;
    m = matrix(ROWS(r), COLS(r));
    for (i=1; i<=ROWS(m); i++)
        for (j=1; j<=COLS(m); j++)
            m[i][j] = r[i][j] | s[i][j];
    return m;
}

MATRIX meet(MATRIX r, MATRIX s)
{
    MATRIX m;
    WORD i,j;
    if ((ROWS(r) != ROWS(s)) || (COLS(r) != COLS(s)))
        return (MATRIX) 0;
    m = matrix(ROWS(r), COLS(r));
    for (i=1; i<=ROWS(m); i++)
        for (j=1; j<=COLS(m); j++)
            m[i][j] = r[i][j] & s[i][j];
    return m;
}

MATRIX inverse(MATRIX r)
{
    MATRIX m;
    WORD i,j;
    m = matrix(COLS(r), ROWS(r));
    for (i=1; i<=ROWS(m); i++)
        for (j=1; j<=COLS(m); j++)
            m[i][j] = r[j][i];
    return m;
}

MATRIX complement(MATRIX r)
{
    MATRIX m;
    WORD i,j;
    m = matrix(ROWS(r), COLS(r));
    for (i=1; i<=ROWS(r); i++)
        for (j=1; j<=COLS(r); j++)
            m[j][i] = !(r[i][j]);
    return m;
}

/**********************************************************************/
    
MATRIX matrix(WORD rows, WORD cols)
{
    WORD i;
    MATRIX m;
    
    /* allocate pointers to rows */
    m = (MATRIX) calloc(rows+1, sizeof(ROW));
    if (! m) 
        fail("matrix alloc failure #1");
    
    /* store header information */
    m--;
    HEADER(m) = calloc(2,2);
    ROWS(m) = rows;
    COLS(m) = cols;
    
    /* allocate rows and set pointers to them */
    for (i=1; i <= rows; i++)
    {
        m[i] = (ROW) calloc(cols+1, sizeof(WORD));
        if (! m[i]) 
            fail("matrix alloc failure #2");
        m[i]--;
    }
    
    return m;
}

void free_matrix(MATRIX m)
{
    WORD i;
    
    free(HEADER(m));
    for (i=1; i<=ROWS(m); i++)
        free(m[i]);
    free(m);
}

void fail(char *msg)
{
    puts(msg);
    exit(1);
}

