/* 
RESDMP32.C -- RESDUMP for Win32 PE executables
Andrew Schulman, March 1993
link with RESDUMP.C:  bcc -ml resdump.c resdmp32.c

see Ray Duncan, "Dissecting Resources in Windows NT Executable
Files," PC Magazine, 16 March 1993.
see NTIMAGE.H
see RESFMT.TXT

To do:
What is Language 1033?
Do string table numbers properly
*/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <dos.h>
#include "filelib.h"

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

#define UNALIGNED   /**/    // ???
typedef char CHAR;
typedef unsigned char UCHAR;
typedef UCHAR far *PUCHAR;      // far *??
typedef unsigned short WCHAR;   // ??
typedef short SHORT;
typedef unsigned short USHORT;
typedef USHORT far *PUSHORT;    // ??
typedef long LONG;
typedef unsigned long ULONG;
typedef ULONG far *PULONG;      // ??
typedef void far *PVOID;		// ??
typedef int BOOLEAN;
typedef int BOOL;

#define VOID void

#define FALSE		0
#define TRUE		(! FALSE)

#define FIELD_OFFSET(x,y)   offsetof((x),(y))

#define min(a,b)    ((a) < (b) ? (a) : (b))

#pragma pack(1)
#include "ntimage.h"    // from Win32 SDK

/* lifted from MSoft newexe.h */
#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC     0x454C          /* linear executable */
#define LXMAGIC     0x584C          /* IBM OS/2 2.0 linear executable */
#define PEMAGIC     0x4550          /* NT portable executable */
#define PLMAGIC		0x4C50			/* hacked NT port exe for TNT! */
#define W3MAGIC     0x3357          /* WIN386.EXE signature */

extern void fail(char *s);
extern void fail_perror(void);
extern void dump(BYTE far *fp, WORD bytes, char *mask, DWORD addr);

void pe2(ULONG start);
void do_sectiontab(void);
void do_resource_dir(IMAGE_SECTION_HEADER *sect_hdr);

#define PUBLIC /**/

static int did_resources = 0;
static ULONG align;
extern int f;
PUBLIC ULONG start_newexe = 0;
static char *file_name;

static IMAGE_NT_HEADERS hdr;
static ULONG resource_dir = 0;
static IMAGE_SECTION_HEADER *section_tab;

#define SECTION_FILEOFS(x)      (section_tab[x].PointerToRawData)
#define SECTION_SIZE(x)         (section_tab[x].SizeOfRawData)
#define SECTION_VIRTADDR(x)     (section_tab[x].VirtualAddress)
#define SECTION_END(x)          (SECTION_VIRTADDR(x) + SECTION_SIZE(x))
#define SECTION_NAME(x)         (section_tab[x].Name)

char *not_a_pe = "not a PE (portable executable) file" ;

#if 0
void pe(char *name)
{
    ULONG pos;
    USHORT w;
    unsigned i;
	
	file_name = name;
    
    if ((f = sopen(name, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
        fail_perror();
    exe_read(f, &w, 0, 2);  // read first two bytes of file
    if (w != EMAGIC)
        fail("not an executable");
    exe_read(f, &w, 0x18, 2);
    if (w < 0x40)
        fail(not_a_pe);
    exe_read(f, &start_newexe, ENEWHDR, 4);
    exe_read(f, &w, start_newexe, 2);
    if ((w != PEMAGIC) && (w != PLMAGIC))	// accept hacked PE too
        fail(not_a_pe);
    pe2(start_newexe);
    close(f);
}
#endif

#define GET_DIR(x)  (hdr.OptionalHeader.DataDirectory[x].VirtualAddress)
    
void pe2(ULONG start)
{
	printf("\n");
	start_newexe = start;	// called from RESDUMP
    exe_read(f, &hdr, start, sizeof(hdr));
	resource_dir = GET_DIR(IMAGE_DIRECTORY_ENTRY_RESOURCE);
    do_sectiontab();
}

void do_sectiontab(void)
{
	USHORT num_sect = 0;
    ULONG section_tab_ofs;
    USHORT i;
    
    section_tab_ofs = 
        start_newexe +
        sizeof(ULONG) +
        sizeof(IMAGE_FILE_HEADER) +
        hdr.FileHeader.SizeOfOptionalHeader;
    num_sect = hdr.FileHeader.NumberOfSections;
    section_tab = (IMAGE_SECTION_HEADER *)
        allocate(num_sect, sizeof(IMAGE_SECTION_HEADER));
    exe_read(f, section_tab, section_tab_ofs, 
        num_sect * sizeof(IMAGE_SECTION_HEADER));
    for (i=0; i<num_sect; i++)
		if (SECTION_VIRTADDR(i) == resource_dir)
			do_resource_dir(&section_tab[i]);
    // don't free section_tab: used throughout program
	if (! did_resources)
		fail("no resources");
}

/**********************************************************************/

static ULONG rsrc_base = 0;
static ULONG rsrc_virt = 0;
static ULONG rsrc_name = 0;
static int level = 0;

/* Predefined resource types -- from resfmt.txt */
#define     RT_NEWRESOURCE     0x2000
#define     RT_ERROR           0x7fff
#define     RT_CURSOR          1
#define     RT_BITMAP          2
#define     RT_ICON            3
#define     RT_MENU            4
#define     RT_DIALOG          5
#define     RT_STRING          6
#define     RT_FONTDIR         7
#define     RT_FONT            8
#define     RT_ACCELERATORS    9
#define     RT_RCDATA          10
#define     RT_MESSAGETABLE    11
#define     RT_GROUP_CURSOR    12
#define     RT_GROUP_ICON      14
#define     RT_VERSION         16
#define     RT_NEWBITMAP       (RT_BITMAP|RT_NEWRESOURCE)
#define     RT_NEWMENU         (RT_MENU|RT_NEWRESOURCE)
#define     RT_NEWDIALOG       (RT_DIALOG|RT_NEWRESOURCE)

#define DIR_SIZE	(sizeof(IMAGE_RESOURCE_DIRECTORY))
#define ENTRY_SIZE	(sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))
#define MASK		(0x80000000L)
#define NOT_MASK	(~(MASK))

static char *resname[] = { "UNKNOWN_0", "CURSOR", "BITMAP", "ICON", "MENU",
	"DIALOG", "STRINGTABLE", "FONTDIR", "FONT", "ACCELERATORS", "RCDATA",
	"MESSAGETABLE", "GROUP_CURSOR", "UNKNOWN_0Dh", "GROUP_ICON",
	"UNKNOWN_0Fh", "VERSION" };

/**********************************************************************/

void print_unicode_str(WORD size, WORD *str)
{
	USHORT i;
	USHORT *p;
	putchar('\"');
	for (i=0, p=str; i<size; i++, p++)
		putchar( (*p < 256) ? *p : '?' );
	putchar('\"');
}

void print_unicode_string(ULONG ofs)
{
	USHORT *str;
	USHORT size;
	exe_read(f, &size, ofs, sizeof(USHORT));
	str = (USHORT *) allocate(size, sizeof(USHORT));
	exe_read(f, str, ofs+sizeof(USHORT), size * sizeof(USHORT));
	print_unicode_str(size, str);
	free(str);
}

WORD print_null_unicode_str(WORD *str)	// return length
{
	USHORT i;
	USHORT *p;
	putchar('\"');
	for (i=0, p=str; *p; i++, p++)
		putchar( (*p < 256) ? *p : '?' );
	putchar('\"');
	return i;
}

/**********************************************************************/

#define     GRAYED         0x0001     // 'GRAYED' keyword
#define     INACTIVE       0x0002     // 'INACTIVE' keyword
#define     BITMAP         0x0004     // 'BITMAP' keyword
#define     OWNERDRAW      0x0100     // 'OWNERDRAW' keyword
#define     CHECKED        0x0008     // 'CHECKED' keyword
#define     POPUP          0x0010     // Used internally
#define     MENUBARBREAK   0x0020     // 'MENUBARBREAK' keyword
#define     MENUBREAK      0x0040     // 'MENUBREAK' keyword
#define     ENDMENU        0x0080     // Used internally

void do_menu(BYTE *data, ULONG size)
{
	WORD *wdata = (WORD *) (data+4);	// skip past unused header
	WORD *end = (WORD *) (data+size);
	while (wdata < end)
	{
		WORD flags = *wdata;
		wdata++;
		
		if (flags & POPUP)
			printf("POPUP\t");
		else
		{
			printf("   %u (%04Xh)\t", *wdata, *wdata);	// ID
			wdata++;
		}
		wdata += (1 + print_null_unicode_str(wdata));	// skip past str
		if (flags & GRAYED)		printf(" GRAYED");
		if (flags & BITMAP)		printf(" BITMAP");
		if (flags & OWNERDRAW)	printf(" OWNERDRAW");
		if (flags & CHECKED)	printf(" CHECKED");
		if (flags & MENUBARBREAK)	printf(" MENUBARBREAK");
		if (flags & MENUBREAK)	printf(" MENUBREAK");
		printf("\n");
		if (flags & ENDMENU)	printf("   ENDMENU\n");
	}
}

/**********************************************************************/

void do_stringtab(BYTE *data, ULONG size)
{
	int i;
	WORD *p = (WORD *) data;
	for (i=0; i<16; i++)
		if (*p)
		{
			// use (ordinal # - 1) as top 12 bits of string ID???
			printf("%d\t", i);
			print_unicode_str(*p, &p[1]);
			printf("\n");
			p += (1 + *p);
		}
		else
			p++;
}

/**********************************************************************/

/*
struct DialogBoxHeader {
	 DWORD     lStyle;
	 DWORD     lExtendedStyle;    // New for Windows NT
	 WORD      NumberOfItems;
	 WORD      x;
	 WORD      y;
	 WORD      cx;
	 WORD      cy;
	 [Name or Ordinal] MenuName;
	 [Name or Ordinal] ClassName;
	 WCHAR     szCaption[];
	 WORD      wPointSize;       // Only here if FONT set for dialog
	 WCHAR     szFontName[];     // This too
	 };

struct ControlData {
	 DWORD     lStyle;
	 DWORD     lExtendedStyle;
	 WORD      x;
	 WORD      y;
	 WORD      cx;
	 WORD      cy;
	 WORD      wId;
	 [Name or Ordinal] ClassId;
	 [Name or Ordinal] Text;
	 WORD      nExtraStuff;
	 };
*/	 

typedef struct {
	 DWORD     lStyle;
	 DWORD     lExtendedStyle;    // New for Windows NT
	 WORD      NumberOfItems;
	 WORD      x;
	 WORD      y;
	 WORD      cx;
	 WORD      cy;
 } DLGBOXHDR1;
 
typedef struct {
	 DWORD     lStyle;
	 DWORD     lExtendedStyle;
	 WORD      x;
	 WORD      y;
	 WORD      cx;
	 WORD      cy;
	 WORD      wId;
 } CTRLHDR1;
 
#define     BUTTON     0x80
#define     EDIT       0x81
#define     STATIC     0x82
#define     LISTBOX    0x83
#define     SCROLLBAR  0x84
#define     COMBOBOX   0x85

/* Window Styles */
#define WS_OVERLAPPED       0x00000000L
#define WS_POPUP            0x80000000L
#define WS_CHILD            0x40000000L
#define WS_MINIMIZE         0x20000000L
#define WS_VISIBLE          0x10000000L
#define WS_DISABLED         0x08000000L
#define WS_CLIPSIBLINGS     0x04000000L
#define WS_CLIPCHILDREN     0x02000000L
#define WS_MAXIMIZE         0x01000000L
#define WS_CAPTION          0x00C00000L     /* WS_BORDER | WS_DLGFRAME  */
#define WS_BORDER           0x00800000L
#define WS_DLGFRAME         0x00400000L
#define WS_VSCROLL          0x00200000L
#define WS_HSCROLL          0x00100000L
#define WS_SYSMENU          0x00080000L
#define WS_THICKFRAME       0x00040000L
#define WS_GROUP            0x00020000L
#define WS_TABSTOP          0x00010000L

#define WS_MINIMIZEBOX      0x00020000L
#define WS_MAXIMIZEBOX      0x00010000L

#define WS_TILED            WS_OVERLAPPED
#define WS_ICONIC           WS_MINIMIZE
#define WS_SIZEBOX          WS_THICKFRAME
#define WS_TILEDWINDOW      WS_OVERLAPPEDWINDOW

/* Common Window Styles */
#define WS_OVERLAPPEDWINDOW (WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX)
#define WS_POPUPWINDOW      (WS_POPUP | WS_BORDER | WS_SYSMENU)
#define WS_CHILDWINDOW      (WS_CHILD)

/* Extended Window Styles */
#define WS_EX_DLGMODALFRAME  0x00000001L
#define WS_EX_NOPARENTNOTIFY 0x00000004L
#define WS_EX_TOPMOST        0x00000008L
#define WS_EX_ACCEPTFILES    0x00000010L
#define WS_EX_TRANSPARENT    0x00000020L

// from winuser.h
#define DS_ABSALIGN         0x01L
#define DS_SYSMODAL         0x02L
#define DS_LOCALEDIT        0x20L   /* Edit items get Local storage. */
#define DS_SETFONT          0x40L   /* User specified font for Dlg controls */
#define DS_MODALFRAME       0x80L   /* Can be combined with WS_CAPTION  */
#define DS_NOIDLEMSG        0x100L /* WM_ENTERIDLE message will not be sent */
#define DS_SETFOREGROUND    0x200L  //not in win3.1

/* Dialog Box Command IDs */
#define IDOK                1
#define IDCANCEL            2
#define IDABORT             3
#define IDRETRY             4
#define IDIGNORE            5
#define IDYES               6
#define IDNO                7

/* Edit Control Styles */
#define ES_LEFT             0x0000L
#define ES_CENTER           0x0001L
#define ES_RIGHT            0x0002L
#define ES_MULTILINE        0x0004L
#define ES_UPPERCASE        0x0008L
#define ES_LOWERCASE        0x0010L
#define ES_PASSWORD         0x0020L
#define ES_AUTOVSCROLL      0x0040L
#define ES_AUTOHSCROLL      0x0080L
#define ES_NOHIDESEL        0x0100L
#define ES_OEMCONVERT       0x0400L
#define ES_READONLY         0x0800L
#define ES_WANTRETURN       0x1000L

void print_edit(ULONG style)
{
	if (style && (ES_LEFT | WS_BORDER | WS_TABSTOP))
		printf("EDITTEXT");
	else
	{
		printf("EDIT ( ");
		if (style & ES_LEFT) printf("LEFT ");
		if (style & ES_CENTER) printf("CENTER ");
		if (style & ES_RIGHT) printf("RIGHT ");
		if (style & ES_MULTILINE) printf("MULTILINE ");
		if (style & ES_UPPERCASE) printf("UPPERCASE ");
		if (style & ES_LOWERCASE) printf("LOWERCASE ");
		if (style & ES_PASSWORD) printf("PASSWORD ");
		if (style & ES_AUTOVSCROLL) printf("AUTOVSCROLL ");
		if (style & ES_AUTOHSCROLL) printf("AUTOHSCROLL ");
		if (style & ES_NOHIDESEL) printf("NOHIDESEL ");
		if (style & ES_OEMCONVERT) printf("OEMCONVERT ");
		if (style & ES_READONLY) printf("READONLY ");
		if (style & ES_WANTRETURN) printf("WANTRETURN ");
		printf(")");
	}
}

/* Button Control Styles */
#define BS_PUSHBUTTON      0x00L
#define BS_DEFPUSHBUTTON   0x01L
#define BS_CHECKBOX        0x02L
#define BS_AUTOCHECKBOX    0x03L
#define BS_RADIOBUTTON     0x04L
#define BS_3STATE          0x05L
#define BS_AUTO3STATE      0x06L
#define BS_GROUPBOX        0x07L
#define BS_USERBUTTON      0x08L
#define BS_AUTORADIOBUTTON 0x09L
#define BS_PUSHBOX         0x0AL
#define BS_OWNERDRAW       0x0BL
#define BS_LEFTTEXT        0x20L

void print_button(ULONG style)
{
	if (style & (BS_CHECKBOX | WS_TABSTOP)) printf("CHECKBOX");
	else if (style & (BS_PUSHBUTTON | WS_TABSTOP)) printf("PUSHBUTTON");
	else if (style & (BS_GROUPBOX)) printf("GROUPBOX");
	else if (style & (BS_DEFPUSHBUTTON | WS_TABSTOP)) printf("DEFPUSHBUTTON");
	else if (style & (BS_RADIOBUTTON)) printf("RADIOBUTTON");
	else if (style & (BS_AUTOCHECKBOX)) printf("AUTOCHECKBOX");
	else if (style & (BS_AUTO3STATE)) printf("AUTO3STATE");
	else if (style & (BS_AUTORADIOBUTTON)) printf("AUTORADIOBUTTON");
	else if (style & (BS_3STATE)) printf("STATE3");
	else
	{
		printf("BUTTON");
	}
}

/* Static Control Constants */
#define SS_LEFT            0x00L
#define SS_CENTER          0x01L
#define SS_RIGHT           0x02L
#define SS_ICON            0x03L
#define SS_BLACKRECT       0x04L
#define SS_GRAYRECT        0x05L
#define SS_WHITERECT       0x06L
#define SS_BLACKFRAME      0x07L
#define SS_GRAYFRAME       0x08L
#define SS_WHITEFRAME      0x09L
#define SS_USERITEM        0x0AL
#define SS_SIMPLE          0x0BL
#define SS_LEFTNOWORDWRAP  0x0CL
#define SS_NOPREFIX        0x80L    /* Don't do "&" character translation */

void print_static(ULONG style)
{
	if (style & SS_LEFT) printf("LTEXT");
	else if (style & SS_RIGHT) printf("RTEXT");
	else if (style & SS_CENTER) printf("CTEXT");
	else if (style & SS_ICON) printf("ICON");
	else printf("STATIC");	// others not interesting??
}

/* Listbox Styles */
#define LBS_NOTIFY            0x0001L
#define LBS_SORT              0x0002L
#define LBS_NOREDRAW          0x0004L
#define LBS_MULTIPLESEL       0x0008L
#define LBS_OWNERDRAWFIXED    0x0010L
#define LBS_OWNERDRAWVARIABLE 0x0020L
#define LBS_HASSTRINGS        0x0040L
#define LBS_USETABSTOPS       0x0080L
#define LBS_NOINTEGRALHEIGHT  0x0100L
#define LBS_MULTICOLUMN       0x0200L
#define LBS_WANTKEYBOARDINPUT 0x0400L
#define LBS_EXTENDEDSEL       0x0800L
#define LBS_DISABLENOSCROLL   0x1000L
#define LBS_NODATA            0x2000L
#define LBS_STANDARD          (LBS_NOTIFY | LBS_SORT | WS_VSCROLL | WS_BORDER)

void print_listbox(ULONG style)
{
	if (style & (WS_BORDER | LBS_NOTIFY))
		printf("LISTBOX");
	else
	{
		printf("LISTBOX (%08lXh)", style);
	}
}

/* Scroll Bar Styles */
#define SBS_HORZ                    0x0000L
#define SBS_VERT                    0x0001L
#define SBS_TOPALIGN                0x0002L
#define SBS_LEFTALIGN               0x0002L
#define SBS_BOTTOMALIGN             0x0004L
#define SBS_RIGHTALIGN              0x0004L
#define SBS_SIZEBOXTOPLEFTALIGN     0x0002L
#define SBS_SIZEBOXBOTTOMRIGHTALIGN 0x0004L
#define SBS_SIZEBOX                 0x0008L

void print_scrollbar(ULONG style)
{
	printf("SCROLLBAR (%08lXh)", style);
}

/* Combo Box styles */
#define CBS_SIMPLE            0x0001L
#define CBS_DROPDOWN          0x0002L
#define CBS_DROPDOWNLIST      0x0003L
#define CBS_OWNERDRAWFIXED    0x0010L
#define CBS_OWNERDRAWVARIABLE 0x0020L
#define CBS_AUTOHSCROLL       0x0040L
#define CBS_OEMCONVERT        0x0080L
#define CBS_SORT              0x0100L
#define CBS_HASSTRINGS        0x0200L
#define CBS_NOINTEGRALHEIGHT  0x0400L
#define CBS_DISABLENOSCROLL   0x0800L

void print_combobox(ULONG style)
{
	printf("COMBOBOX (%08lXh)", style);
}

WORD print_name_or_ord(WORD *data, BOOL do_control, ULONG style)
{
	if (*data == 0xFFFF) 
	{
		if (do_control)
		{
			switch (data[1])
			{
				case BUTTON:	print_button(style); break;
				case EDIT:		print_edit(style); break;
				case STATIC:	print_static(style); break;
				case LISTBOX:	print_listbox(style); break;
				case SCROLLBAR: print_scrollbar(style); break;
				case COMBOBOX:	print_combobox(style); break;
				default:		printf("%04Xh (Style %08lXh)", 
					data[1], style); break;
			}
		}
		else
			printf("%04Xh", data[1]);
		return 2;
	}
	else
	{
		return (1 + print_null_unicode_str(data));
	}
}
 
void do_dialog(BYTE *data, ULONG size)
{
	DLGBOXHDR1 *hdr = (DLGBOXHDR1 *) data;
	CTRLHDR1 *ctrl;
	BYTE *p2;
	WORD *p = (WORD *) (data + sizeof(DLGBOXHDR1));
	WORD num_items = hdr->NumberOfItems;
	WORD pt_size;
	WORD i;
	if (hdr->lStyle)
	{
		ULONG style = hdr->lStyle;
		printf("Style: ");
		if (style & DS_ABSALIGN) printf("ABSALIGN ");
		if (style & DS_SYSMODAL) printf("SYSMODAL ");
		if (style & DS_LOCALEDIT) printf("LOCALEDIT ");
		if (style & DS_SETFONT) printf("SETFONT ");
		if (style & DS_MODALFRAME) printf("MODALFRAME ");
		if (style & DS_NOIDLEMSG) printf("NOIDLEMSG ");
		if (style & DS_SETFOREGROUND) printf("SETFRGRND ");
		printf("\n");
	}
	printf("Menu: "); p += print_name_or_ord(p,0,0); printf("\n");
	printf("Class: "); p += print_name_or_ord(p,0,0); printf("\n");
	printf("Caption: "); p += (1 + print_null_unicode_str(p)); printf("\n");
	if (hdr->lStyle & DS_SETFONT)
	{
		pt_size = *p++;	// skip past pt. size
		printf("Font: ");
		p += (1 + print_null_unicode_str(p));
		printf(" (%u Pt.)\n", pt_size);
	}
	// now at end of header; do each control
	for (i=0; i<num_items; i++)
	{
		p2 = (BYTE *) p;
		while (((WORD) p2) & 3) p2++;	// make DWORD-aligned!
		ctrl = (CTRLHDR1 *) p2;
		printf("%u (%04Xh)\t", ctrl->wId, ctrl->wId);
		p = (WORD *) (p2 + sizeof(CTRLHDR1));
		p += print_name_or_ord(p,1, ctrl->lStyle); printf("\t");
		p += print_name_or_ord(p,0,0); printf("\n");
		p++;	// skip nExtraStuff
	}
}

/**********************************************************************/

void do_res_data(ULONG base)
{
	IMAGE_RESOURCE_DATA_ENTRY *hdr;
	ULONG ofs;
	BYTE *data;
	hdr = (IMAGE_RESOURCE_DATA_ENTRY *) 
		allocate(sizeof(IMAGE_RESOURCE_DATA_ENTRY), 1);
	exe_read(f, hdr, base, sizeof(IMAGE_RESOURCE_DATA_ENTRY));
	data = (BYTE *) allocate(hdr->Size, 1);
	ofs = hdr->OffsetToData - rsrc_virt + rsrc_base;
	exe_read(f, data, ofs, hdr->Size);
	switch (rsrc_name)
	{
		case RT_MENU:	do_menu(data, hdr->Size); break;
		case RT_STRING: do_stringtab(data, hdr->Size); break;
		case RT_DIALOG: do_dialog(data, hdr->Size); break;
#if 0		
		default:		dump(data, hdr->Size, "%04X", 0);
#endif	
	}
	free(data);
	free(hdr);
}

void do_res_dir(ULONG base)
{
	IMAGE_RESOURCE_DIRECTORY *resdir;
	IMAGE_RESOURCE_DIRECTORY_ENTRY *resentry;
	USHORT num_res;
	USHORT i;

	level++;
	resdir = (IMAGE_RESOURCE_DIRECTORY *) allocate(DIR_SIZE, 1);
    exe_read(f, resdir, base, DIR_SIZE);
	num_res = resdir->NumberOfNamedEntries + resdir->NumberOfIdEntries;
	resentry = (IMAGE_RESOURCE_DIRECTORY_ENTRY *)
		allocate(num_res, ENTRY_SIZE);
	exe_read(f, resentry, base + DIR_SIZE, num_res * ENTRY_SIZE);
	for (i=0; i<num_res; i++)
	{
		ULONG name = resentry[i].Name;
		ULONG ofs = resentry[i].OffsetToData;
		if (level == 1)	// at level 1, name is resource type
		{
			rsrc_name = name;		// set global!
		}
		else if (level == 2)	// at level 2, name is ID # or name
		{
			if (rsrc_name <= RT_VERSION)
				printf("%s", resname[rsrc_name]);
			else switch (rsrc_name)
			{
				case RT_NEWBITMAP: printf("NEWBITMAP"); break;
				case RT_NEWMENU:   printf("NEWMENU"); break;
				case RT_NEWDIALOG: printf("NEWDIALOG"); break;
				default:           
					if (rsrc_name & MASK)
						print_unicode_string(rsrc_base + 
							(rsrc_name & NOT_MASK));
					else
						printf("#%08lXh", rsrc_name);
			}
			printf("   ");
			if (name & MASK)
				print_unicode_string(rsrc_base + (name & NOT_MASK));
			else
				printf("#%08lXh", name);
			printf("\n");
		}
		else if (level == 3)	// at level 3, name is language (0 = English)
		{
			if (name != 0)
				printf("Language %u\n", name);
		}
		if (ofs & MASK)
			do_res_dir(rsrc_base + (ofs & NOT_MASK));	// recursive
		else
			do_res_data(rsrc_base + ofs);
	}
	free(resentry);
	free(resdir);
	level--;
	if (level == 2) printf("\n");
}

void do_resource_dir(IMAGE_SECTION_HEADER *sect_hdr)
{
	rsrc_base = sect_hdr->PointerToRawData;		// set global!
	rsrc_virt = sect_hdr->VirtualAddress;		// set global!
	do_res_dir(rsrc_base);
	printf("\n");
	did_resources++;
}

