/* RESDUMP.C */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

/* lifted from MSoft newexe.h */
#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC		0x454C			/* linear executable */
#define LXMAGIC		0x584C			/* IBM OS/2 2.0 linear executable */
#define W3MAGIC		0x3357			/* WIN386.EXE signature */

#define DLL_FLAG    0x8000

#define WIND_FLAG	2

#define	NERESBYTES	9

/* lifted from WINDOWS.H */
/* Predefined Resource Types */
#define MAKEINTRESOURCE(x)	((x) + 0x8000)		/* for file only!!!! */
#define RT_CURSOR           MAKEINTRESOURCE(1)
#define RT_BITMAP           MAKEINTRESOURCE(2)
#define RT_ICON             MAKEINTRESOURCE(3)
#define RT_MENU             MAKEINTRESOURCE(4)
#define RT_DIALOG           MAKEINTRESOURCE(5)
#define RT_STRING           MAKEINTRESOURCE(6)
#define RT_FONTDIR          MAKEINTRESOURCE(7)
#define RT_FONT             MAKEINTRESOURCE(8)
#define RT_ACCELERATOR      MAKEINTRESOURCE(9)
#define RT_RCDATA           MAKEINTRESOURCE(10)
#define RT_ERRORTABLE		MAKEINTRESOURCE(11)	/* from Duncan Murdoch */
#define RT_GROUPCURSOR		MAKEINTRESOURCE(12)	/* windows.inc */
#define RT_GROUPICON		MAKEINTRESOURCE(14)	/* windows.inc */
#define RT_NAMETABLE		MAKEINTRESOURCE(15)	/* from Open Tools */
#define RT_VERSIONINFO		MAKEINTRESOURCE(16) /* some sort version info */

/* have also seen #400 used */

#define DS_SETFONT          0x40L   /* User specified font for Dlg controls */

#define MF_POPUP           0x0010
#define MF_END             0x0080
#define MF_STRING          0x0000
#define MF_BITMAP          0x0004
#define MF_OWNERDRAW       0x0100
#define MF_SEPARATOR       0x0800

#define BS_PUSHBUTTON      0x00L
#define BS_DEFPUSHBUTTON   0x01L
#define BS_CHECKBOX        0x02L
#define BS_AUTOCHECKBOX    0x03L
#define BS_RADIOBUTTON     0x04L
#define BS_3STATE          0x05L
#define BS_AUTO3STATE      0x06L
#define BS_GROUPBOX        0x07L
#define BS_USERBUTTON      0x08L
#define BS_AUTORADIOBUTTON 0x09L
#define BS_PUSHBOX         0x0AL
#define BS_OWNERDRAW       0x0BL
#define BS_LEFTTEXT        0x20L

#define SS_ICON            0x03L

#define ES_MULTILINE       0x0004L

#pragma pack(1)

typedef struct
{
    WORD    ne_magic;
    char        ne_ver;
    char        ne_rev;
    WORD    ne_enttab;
    WORD    ne_cbenttab;
    long        ne_crc;
    WORD    ne_flags;
    WORD    ne_autodata;			/* dgroup */
    WORD    ne_heap;
    WORD    ne_stack;
    void far *  ne_csip;
    void far *  ne_sssp;
    WORD    ne_cseg;
    WORD    ne_cmod;
    WORD    ne_cbnrestab;           /* size of nonres name table */
    WORD    ne_segtab;
    WORD    ne_rsrctab;             /* offset of resource table */
    WORD    ne_restab;              /* offset of resident name table */
    WORD    ne_modref;
    WORD    ne_impname;
    long        ne_nonrestab;           /* offset of nres table from begin */
	WORD	ne_cmovent;
	WORD	ne_align;				/* segment align shift count */
	WORD	ne_cres;
	BYTE	ne_exetyp;
	char		ne_res[NERESBYTES];		/* pad struct to 64 bytes */
	/* more stuff, but we don't care */
} NEWEXE;

/* following from MSJ, September 1991, p. 65 */

typedef struct {
	WORD rt_type;
	WORD rt_count;
	DWORD rt_proc;
} RESTYPEINFO;

typedef struct {
	WORD rn_offset;
	WORD rn_length;
	WORD rn_flags;
	WORD rn_id;
	WORD rn_handle;
	WORD rn_usage;
} RESNAMEINFO;

typedef RESNAMEINFO *PRESNAMEINFO;

typedef struct {
	RESTYPEINFO rt;
	RESNAMEINFO rn[1];
} RESTAB;

typedef RESTAB *PRESTAB;

void fail(char *s) { fputs("resdump fail: ", stdout); puts(s); exit(1); }

void usage()
{
	static char *msg = "usage: resdump [-verbose or -hex] <exe>" ;
	puts(msg);
	exit(1);
}

void newexe(char *name);
int ne_read(int f, void *p, DWORD ne_offset, WORD bytes);
int exe_read(int f, void *p, DWORD exe_offset, WORD bytes);
void *allocate(size_t x, size_t y);
void count2null(char *s, int len);
void dump(BYTE far *fp, WORD bytes, 
	char *mask, DWORD addr);

void do_restab(WORD offset);
void do_dialog(DWORD offset, DWORD length, WORD id);
void do_menu(DWORD offset, DWORD length, WORD id);
void do_string(DWORD offset, DWORD length, WORD id);
void do_accel(DWORD offset, DWORD length, WORD id);
void dump_generic(DWORD offset, DWORD length);

static WORD start_newexe = 0;
static NEWEXE hdr = {0} ;
static DWORD align;
static int f;
static int verbose = 0;
static int do_dump = 0;

main(int argc, char *argv[])
{
	int i;
	
	fputs("Windows Resource Dumper version 1.0\n", stderr);
	fputs("Copyright (c) 1991 Andrew Schulman.  All rights reserved.\n\n",
		stderr);
	
	if (argc < 2)
		usage();
	
	for (i=1; i<argc; i++)
		if ((argv[i][0] == '-') || (argv[i][0] == '/'))
		{
			if (argv[i][1] == '?')
				usage();
			else if (strcmp(&argv[i][1], "verbose") == 0)
				verbose = 1;
			else if (strcmp(&argv[1][1], "hex") == 0)
				do_dump = 1;
			else
				usage();
		}
		else	
			newexe(argv[i]);
	return 0;
}

void newexe(char *name)
{
    WORD w;
	WORD i;
	
    if ((f = open(name, O_RDONLY | O_BINARY)) == -1)
		fail("can't find file");
    read(f, &w, 2);
    if (w != EMAGIC)
		fail("not an executable");
    lseek(f, ENEWHDR, SEEK_SET);
    read(f, &start_newexe, 2);
    lseek(f, (long) start_newexe, SEEK_SET);
    read(f, &hdr, sizeof(NEWEXE));
	switch (hdr.ne_magic)
	{
		case LEMAGIC:
		case LXMAGIC:
		case W3MAGIC:
			fail("this is a linear executable, not a segmented executable");
			break;
		case NEMAGIC:
			break;	// okay
		default:
			fail("not a segmented (new) executable");
	}
	
	if (hdr.ne_exetyp != WIND_FLAG)		/* NOT a bit flag */
		fail("not a Windows program");
	
	align = 1 << hdr.ne_align;
	
	do_restab(hdr.ne_rsrctab);
	
	close(f);
}

#define NOTIMP(s, id) \
{ \
	printf("%s %u (%02Xh) (ofs=%lu, len=%lu)\n", \
		(s), (id), (id), (res_offset), (res_length)); \
}

char *type_name(WORD offset)
{
	static char buf[128];
	ne_read(f, buf, offset, 128);
	return &buf[6];		// name table?
}

void do_restab(WORD offset)
{
	RESTYPEINFO *rt = allocate(1, sizeof(RESTYPEINFO));
	RESNAMEINFO *rn, *rn2;
	DWORD res_offset, res_length;
	WORD align_shift;
	WORD id;
	int i;
	
	ne_read(f, &align_shift, offset, sizeof(short));
	offset += sizeof(short);
	
	for (;;)
	{
		ne_read(f, rt, offset, sizeof(RESTYPEINFO));
		if (rt->rt_type == 0)
			break;
		if (rt->rt_count == 0)
			fail("no resources?");
		if (((long) rt->rt_count * sizeof(RESNAMEINFO)) > 0xFFFFUL)
			fail("no resources??");
		rn = allocate(rt->rt_count, sizeof(RESNAMEINFO));
		read(f, rn, sizeof(RESNAMEINFO) * rt->rt_count);
		for (i=rt->rt_count, rn2=rn; i--; rn2++)
		{
			res_offset = (DWORD) rn2->rn_offset << align_shift;
			res_length = (DWORD) rn2->rn_length << align_shift;
			if (rn2->rn_id & 0x8000)
				id = rn2->rn_id & ~0x8000;		// for now!!!
			
			switch (rt->rt_type)
			{
				case RT_DIALOG:
					do_dialog(res_offset, res_length, id);
					break;
				case RT_MENU:
					do_menu(res_offset, res_length, id);
					break;
				case RT_STRING:
					do_string(res_offset, res_length, id);
					break;
				case RT_ACCELERATOR:
					do_accel(res_offset, res_length, id);
					break;
				case RT_FONT:		NOTIMP("FONT", id); break;
				case RT_CURSOR:		NOTIMP("CURSOR", id); break;
				case RT_BITMAP:		NOTIMP("BITMAP", id); break;
				case RT_ICON:		NOTIMP("ICON", id); break;
				case RT_FONTDIR:	NOTIMP("FONTDIR", id); break;
				case RT_RCDATA:		NOTIMP("USERDATA", id); break;
				case RT_GROUPCURSOR:NOTIMP("CURSOR DIRECTORY", id); break;
				case RT_ERRORTABLE: NOTIMP("ERRORTABLE", id); break;
				case RT_GROUPICON:	NOTIMP("ICON DIRECTORY", id); break;
				case RT_NAMETABLE:	NOTIMP("NAMETABLE", id); break;
				case RT_VERSIONINFO:NOTIMP("VERSION", id); break;
				default:
					if (rt->rt_type & 0x8000)
						printf("#%u", rt->rt_type & ~0x8000);
					else
						printf("%02X", rt->rt_type); 
					NOTIMP("", id); 
					break;
			}
			if (do_dump)
				dump_generic(res_offset, res_length);
			putchar('\n');
		}
end:
		offset += (sizeof(RESTYPEINFO) + 
				  (rt->rt_count * sizeof(RESNAMEINFO)));
		free(rn);
	}
	
	free(rt);
}

typedef struct {
	DWORD dlg_style;
	BYTE num_items;
	WORD x, y, cx, cy;
	// rest is variable length!
} DLGHEADER;

typedef struct {
	WORD x, y, cx, cy, id;
	DWORD ctrl_style;
	// rest is variable length!
} CTRLHEADER;

void do_dialog(DWORD offset, DWORD length, WORD id)
{
	DWORD style;
	BYTE *dlg = allocate(1, length);
	DLGHEADER *dlghdr = dlg;
	BYTE *control;
	BYTE *ctrl;
	BYTE *text;
	BYTE *menu_name = 0;
	BYTE *class_name = 0;
	BYTE *class_type;
	int num_controls;
	int control_type;
	int typ_len, text_len;
	int len;
	int i;
	
	exe_read(f, dlg, offset, length);
	num_controls = dlghdr->num_items;
	text = dlg + sizeof(DLGHEADER);
	if (*text == 0) 
		text++;
	else if (*text == 0xFF) 
	{
		menu_name = text;
		text += 3;
	}
	else
	{
		menu_name = text;
		text += (strlen(text) + 1);
	}
	if (*text)
		class_name = text;
	text += (strlen(text) + 1);
	printf("DIALOG %u (%02Xh), \"%s\"",
		id, id, text);
	if (verbose)
		printf(" [%u,%u,%u,%u]", 
			dlghdr->x, dlghdr->y, dlghdr->cx, dlghdr->cy);
	putchar('\n');
	if (menu_name)
	{
		if (*menu_name == 0xFF)
		{
			WORD ord = *((WORD *) &menu_name[1]);
			printf("MENU #%u (%02Xh)\n", ord);
		}
		else
			printf("MENU \"%s\"\n", menu_name);
	}
	if (class_name)
		printf("CLASS \"%s\"\n", class_name);
	ctrl = text + strlen(text) + 1;
	if (dlghdr->dlg_style & DS_SETFONT)
	{
		ctrl += 2;
		ctrl += (strlen(ctrl) + 1);
	}
	for (i=num_controls; i--; )
	{
		// this code is ridiculous -- fix it!
		CTRLHEADER *ctrlhdr = ctrl;
		control = ctrl;
		id = ctrlhdr->id;
		style = ctrlhdr->ctrl_style;
		if (control[sizeof(CTRLHEADER)] & 0x80)
		{
			switch (control[sizeof(CTRLHEADER)])
			{
				case 0x80: // BUTTON
					switch ((WORD) style)	// not bit mask???
					{
						case BS_PUSHBUTTON:
							class_type = "PUSHBUTTON"; break;
						case BS_DEFPUSHBUTTON:
							class_type = "DEFPUSHBUTTON"; break;
						case BS_CHECKBOX:
							class_type = "CHECKBOX"; break;
						case BS_AUTOCHECKBOX:
							class_type = "AUTOCHECKBOX"; break;
						case BS_RADIOBUTTON:
							class_type = "RADIOBUTTON"; break;
						case BS_3STATE:
							class_type = "3STATE"; break;
						case BS_GROUPBOX:
							class_type = "GROUPBOX"; break;
						case BS_USERBUTTON:
							class_type = "USERBUTTON"; break;
						case BS_AUTORADIOBUTTON:
							class_type = "AUTORADIOBUTTON"; break;
						case BS_PUSHBOX:
							class_type = "PUSHBOX"; break;
						case BS_OWNERDRAW:
							class_type = "OWNERDRAW"; break;
						case BS_LEFTTEXT:
							class_type = "LEFTTEXT"; break;
						default:
							class_type = "BUTTON"; break;
					}
					break;
				case 0x81: 
					// should do other ES_ styles
					if (style & ES_MULTILINE)
						class_type = "EDITMULTILINE";
					else
						class_type = "EDIT"; 
					break;
				case 0x82:
					// should do other SS_ styles
					class_type = "STATIC"; 
					break;
				case 0x83: 
					// should do LBS_ styles
					class_type = "LISTBOX"; break;
				case 0x84: 
					// should do SBS_ styles
					class_type = "SCROLLBAR"; break;
				case 0x85: 
					// should do CBS_ styles
					class_type = "COMBOBOX"; break;
				default:   class_type = "????"; break;
			}
			ctrl++;
			text = &control[sizeof(CTRLHEADER)+1];
			printf("    %s %u (%02Xh), \"%s\"",
				class_type, id, id, text);
		}
		else
		{
			class_type = &control[sizeof(CTRLHEADER)];
			len = strlen(class_type);
			ctrl += (len + 1);
			text = &control[sizeof(CTRLHEADER)] + (len + 1);
			printf("    \"%s\" %u (%02Xh), \"%s\"",
				class_type, id, id, text);
		}
		if (verbose)
			printf(" [%u,%u,%u,%u]", 
				ctrlhdr->x, ctrlhdr->y, ctrlhdr->cx, ctrlhdr->cy);
		putchar('\n');
		ctrl += (sizeof(CTRLHEADER) + strlen(text) + 2);
	}
	free(dlg);
}

void do_menu(DWORD offset, DWORD length, WORD id)
{
	BYTE *m, *p;
	BYTE *menu = allocate(1, length);
	WORD flags;
	int saw_end = 0;
	exe_read(f, menu, offset, length);
	printf("MENU %u (%02Xh)\n", id, id);
	m = menu + 4;	// skip header
	while (m < (menu + length))
	{
		flags = *((WORD *) m);
		p = &m[2];
		if (flags & MF_POPUP)
		{
			printf("    POPUP \"%s\"\n", p);
			m += (3 + strlen(p));
			saw_end = 0;
		}
		else
		{
			if (saw_end < 2)
			{
				printf("        %u (%02Xh) \"%s\"\n", 
					*((WORD *) p),
					*((WORD *) p),
					&m[4]);
			}
			m += (5 + strlen(&m[4]));
		}
		if (flags & MF_END)
		{
			printf("        END\n");
			saw_end++;
		}
	}
	free(menu);
}

void do_string(DWORD offset, DWORD length, WORD id)
{
	BYTE buf[1024];	// workaround Borland problem with %*s
	BYTE *str = allocate(1, length);
	BYTE *s;
	WORD ord;
	exe_read(f, str, offset, length);
	printf("STRINGTABLE %u (%02Xh)\n", id, id);
	// check string numbering scheme
	for (s=str, ord=(id-1)<<4; s<(str+length); ord++)
	{
		if (s[0])
		{
			memcpy(buf, &s[1], s[0]);
			buf[s[0]] = '\0';
			printf("    %u (%02Xh) \"%s\"\n", ord, ord, buf);
		}
		s += (s[0] + 1);
	}
	free(str);
}

typedef struct {
	BYTE fFlags;
	WORD wEvent;
	WORD wId;
	} ACCEL;

/* lifted from WINDOWS.H */
#define VK_LBUTTON          0x01
#define VK_RBUTTON          0x02
#define VK_CANCEL           0x03
#define VK_MBUTTON          0x04
#define VK_BACK             0x08
#define VK_TAB              0x09
#define VK_CLEAR            0x0C
#define VK_RETURN           0x0D
#define VK_SHIFT            0x10
#define VK_CONTROL          0x11
#define VK_MENU             0x12
#define VK_PAUSE            0x13
#define VK_CAPITAL          0x14
#define VK_ESCAPE           0x1B
#define VK_SPACE            0x20
#define VK_PRIOR            0x21
#define VK_NEXT             0x22
#define VK_END              0x23
#define VK_HOME             0x24
#define VK_LEFT             0x25
#define VK_UP               0x26
#define VK_RIGHT            0x27
#define VK_DOWN             0x28
#define VK_SELECT           0x29
#define VK_PRINT            0x2A
#define VK_EXECUTE          0x2B
#define VK_SNAPSHOT         0x2C
#define VK_COPY				0x2C
#define VK_INSERT           0x2D
#define VK_DELETE           0x2E
#define VK_HELP             0x2F
#define VK_A				'A'
#define VK_Z				'Z'
#define VK_0				'0'
#define VK_9				'9'
#define VK_NUMPAD0          0x60
#define VK_NUMPAD9          0x69
#define VK_MULTIPLY         0x6A
#define VK_ADD              0x6B
#define VK_SEPARATOR        0x6C
#define VK_SUBTRACT         0x6D
#define VK_DECIMAL          0x6E
#define VK_DIVIDE           0x6F
#define VK_F1               0x70
#define VK_F16              0x7F
#define VK_NUMLOCK          0x90

void do_accel(DWORD offset, DWORD length, WORD id)
{
	ACCEL *accel = allocate(1, length);
	WORD count = length / sizeof(ACCEL);
	ACCEL *pa;
	WORD i;
	
	printf("ACCELERATOR %u (%02Xh)\n", id, id);
	exe_read(f, accel, offset, length);
	for (i=count, pa=accel; i--; pa++)
	{
		printf("\t%d (%04xh) ", pa->wId, pa->wId);
		if (pa->fFlags & 0x04)		printf("SHIFT ");
		else if (pa->fFlags & 0x08)	printf("CTRL ");
		else if (pa->fFlags & 0x10) printf("ALT ");
		if (pa->wEvent >= 'A' && pa->wEvent <= 'Z')
			printf("VK_%c\n", pa->wEvent);
		else if (pa->wEvent >= '0' && pa->wEvent <= '9')
			printf("VK_%c\n", pa->wEvent);
		else if (pa->wEvent >= VK_NUMPAD0 && pa->wEvent <= VK_NUMPAD9)
			printf("VK_NUMPAD%d\n", pa->wEvent - VK_NUMPAD0);
		else if (pa->wEvent >= VK_F1 && pa->wEvent <= VK_F16)
			printf("VK_F%d\n", 1 + pa->wEvent - VK_F1);
		else switch (pa->wEvent)
		{
			case VK_LBUTTON: puts("VK_LBUTTON"); break;
			case VK_RBUTTON: puts("VK_RBUTTON"); break;
			case VK_CANCEL: puts("VK_CANCEL"); break;
			case VK_MBUTTON: puts("VK_MBUTTON"); break;
			case VK_BACK: puts("VK_BACK"); break;
			case VK_TAB: puts("VK_TAB"); break;
			case VK_CLEAR: puts("VK_CLEAR"); break;
			case VK_RETURN: puts("VK_RETURN"); break;
			case VK_SHIFT: puts("VK_SHIFT"); break;
			case VK_CONTROL: puts("VK_CONTROL"); break;
			case VK_MENU: puts("VK_MENU"); break;
			case VK_PAUSE: puts("VK_PAUSE"); break;
			case VK_CAPITAL: puts("VK_CAPITAL"); break;
			case VK_ESCAPE: puts("VK_ESCAPE"); break;
			case VK_SPACE: puts("VK_SPACE"); break;
			case VK_PRIOR: puts("VK_PRIOR"); break;
			case VK_NEXT: puts("VK_NEXT"); break;
			case VK_END: puts("VK_END"); break;
			case VK_HOME: puts("VK_HOME"); break;
			case VK_LEFT: puts("VK_LEFT"); break;
			case VK_UP: puts("VK_UP"); break;
			case VK_RIGHT: puts("VK_RIGHT"); break;
			case VK_DOWN: puts("VK_DOWN"); break;
			case VK_SELECT: puts("VK_SELECT"); break;
			case VK_PRINT: puts("VK_PRINT"); break;
			case VK_EXECUTE: puts("VK_EXECUTE"); break;
			case VK_SNAPSHOT: puts("VK_SNAPSHOT [VK_COPY]"); break;
			case VK_INSERT: puts("VK_INSERT"); break;
			case VK_DELETE: puts("VK_DELETE"); break;
			case VK_HELP: puts("VK_HELP"); break;
			case VK_MULTIPLY: puts("VK_MULTIPLY"); break;
			case VK_ADD: puts("VK_ADD"); break;
			case VK_SEPARATOR: puts("VK_SEPARATOR"); break;
			case VK_SUBTRACT: puts("VK_SUBTRACT"); break;
			case VK_DECIMAL: puts("VK_DECIMAL"); break;
			case VK_DIVIDE: puts("VK_DIVIDE"); break;
			case VK_NUMLOCK: puts("VK_NUMLOCK"); break;
			default:
				printf("%04x\n", pa->wEvent); break;	// ASCII char of VK_
		}
		if (pa->fFlags & 0x80)	// END
			break;
	}
	free(accel);
}

void dump_generic(DWORD offset, DWORD length)
{
	BYTE *res = allocate(1, length);
	exe_read(f, res, offset, length);
	dump(res, length, "%08lx", 0);
	free(res);
}

/**********************************************************************/

/* read at offset from NE */
int ne_read(int f, void *p, DWORD ne_offset, WORD bytes)
{
    lseek(f, (long) start_newexe + ne_offset, SEEK_SET);
	return read(f, p, bytes);
}

/* read at offset from beginning of file */
int exe_read(int f, void *p, DWORD exe_offset, WORD bytes)
{
    lseek(f, exe_offset, SEEK_SET);
	return read(f, p, bytes);
}

void *allocate(size_t x, size_t y)
{
	void *p;
	if (! x) x = 1;
	if (! y) y = 1;
	if (((long) x * (long) y) > 0xFFFFUL)
	{
		printf("\n\nx=%02X y=%02X\n", x, y);
		fail("request > 64k");
	}
	if ((p = calloc(x, y)) == 0)
	{
		printf("\n\nx=%02X y=%02X\n", x, y);
		fail("insufficient memory");
	}
	return p;
}

#define WIDTH		16

void dump(BYTE far *fp, WORD bytes, 
	char *mask, DWORD addr)
{
	BYTE far *p;
	WORD i, j, c;
	
    for (i=0; i<bytes; i += WIDTH)
    {
		c = ((bytes-i) > WIDTH) ? WIDTH : bytes-i;
		printf(mask, addr+i);
		putchar(' '); putchar('|'); putchar(' ');
        for (j=c, p=fp+i; j--; p++)
            printf("%02X ", *p);
		for (j=WIDTH-c; j--; )	// pad out on last line
			printf("   ");
		putchar('|'); putchar(' ');
        for (j=c, p=fp+i; j--; p++)
			putchar( isprint(*p) ? *p : '.' );
		putchar('\n');
    }
}

