/* 
sect13.C -- dump sectors
Andrew Schulman, September 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <bios.h>

unsigned long axton(char *s);
void dump(unsigned char far *fp, unsigned bytes, 
    char *mask, unsigned long addr, int width);
int read_sectors(int drive, int head, int track, 
	unsigned first_sector, int sectors, unsigned char far *buf);

void fail(const char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
	int drive;
	unsigned head, track;
	unsigned start_sector, sect;
	int num_sectors, i;
	unsigned char *buf;
	
	if (argc < 3)
		fail("usage: sect13 <drive> <head> <track> <sector> [# sectors]");
	
	// drive number
	drive = axton(argv[1]);
	head = axton(argv[2]);
	track = axton(argv[3]);
	start_sector = axton(argv[4]);
	num_sectors = (argc > 5) ? axton(argv[5]) : 1;
	
	if (! (buf = (unsigned char *) malloc(1024)))
		fail("insufficient memory");
	
	for (i=0, sect=start_sector; i<num_sectors; i++, sect++)
	{
		printf("\nDrive %02Xh, Head %u, Track %u, Sector %u\n",
			drive, head, track, sect);
		if (! read_sectors(drive, head, track, sect, 1, buf))
			fail("can't read sector");
		dump(buf, 512, "%04X", 0, 16);
	}
	return 0;
}

unsigned long axton(char *s)
{
    if ((s[0] == '0') && (s[1] == 'x'))
    {
        unsigned long ret;
        sscanf(&s[2], "%lx", &ret);
        return ret;
    }
    else
        return atol(s);
}

void dump(unsigned char far *fp, unsigned bytes, 
    char *mask, unsigned long addr, int width)
{
    unsigned char far *p;
    unsigned i, j, c;
    
    for (i=0; i<bytes; i += width)
    {
        c = ((bytes-i) > width) ? width : bytes-i;
        printf(mask, addr+i);
        printf(" | ");
        for (j=c, p=fp+i; j--; p++)
            printf("%02X ", *p);
        for (j=width-c; j--; )  // pad out on last line
            printf("   ");
        printf("| ");
        for (j=c, p=fp+i; j--; p++)
            putchar( isprint(*p) ? *p : '.' );
        putchar('\n');
    }
}

int read_sectors(int drive, int head, int track, 
	unsigned first_sector, int sectors, unsigned char far *buf) 
{
#if 1
	int track_hi = track & ~63;
	int track_lo = track & 63;
	_asm mov ch, byte ptr track_lo
	_asm mov cl, byte ptr track_hi
	_asm or cl, byte ptr first_sector
	_asm mov dh, byte ptr head
	_asm mov dl, byte ptr drive
	_asm mov al, byte ptr sectors
	_asm mov ah, 02h
	_asm les bx, dword ptr buf
	_asm int 13h
	_asm jc error
	return 1;
error:
	return 0;
#else
	struct diskinfo_t diskinfo;
	diskinfo.drive = drive;
	diskinfo.track = track;
	diskinfo.track = track;
	diskinfo.sector = first_sector;
	diskinfo.nsectors = sectors;
	diskinfo.buffer = buf;
	return (_bios_disk(_DISK_READ, &diskinfo) == 0);
#endif	
}

