/* sieve.c 
Andrew Schulman, July 1997
andrew@ora.com; http://www.sonic.net/~undoc
for Web Review article "Java on the Fly"*/

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <limits.h>
#include <time.h>

#if 1
typedef unsigned long ULONG;
typedef unsigned char BYTE;

#define ALLOC(x,y)        calloc(x,y)
#define FREE(x)           free(x)
typedef ULONG *BITMAP;

#define index(c)        ((c) >> 5)

#ifndef DID_MASKS
#define DID_MASKS
static ULONG masks[32] = {1<<0, 1<<1, 1<<2, 1<<3, 1<<4, 1<<5, 1<<6, 1<<7,
1<<8, 1<<9, 1<<10, 1<<11, 1<<12, 1<<13, 1<<14, 1<<15, 1<<16, 1<<17, 1<<18, 
1<<19, 1<<20, 1<<21, 1<<22, 1<<23, 1<<24, 1<<25, 1<<26, 1<<27, 1<<28, 
1<<29, 1<<30, 1<<31, } ;
#endif

#ifdef DID_MASKS
#define mask(c)              masks[c & 31]
#else
#define mask(c)           (1 << ((c) & 31))
#endif

#define BIT_ON(map,c)     (map[index(c)] & mask(c))
#define BIT_OFF(map,c)    (! BIT_ON(map,c)) /* any single-operation equiv??*/
#define SET_TRUE(map,c)   map[index(c)] |= mask(c)

#define SET_BIT_TRUE_IF_FALSE(map, c) { \
    int ndx = index(c); \
    int mask = mask(c); \
    if (! (map[ndx] & mask)) \
        map[ndx] |= mask; \
    }

#define free_bitmap(map)  FREE(map)

/* just to make it easy to find code in ASM output*/
unsigned INDEX(int x)               { return index(x); }
unsigned MASK(int x)                { return mask(x); }
void set_true(BITMAP map, ULONG c)  { SET_BIT_TRUE_IF_FALSE(map, c); }

ULONG bytes(ULONG size)
{
    /*return index(size) + ((size & 0x07) ? 1 : 0);*/
    return sizeof(ULONG) * (index(size) + ((size & 31) ? 1 : 0));
}

BITMAP make_bitmap(ULONG size) { return (BITMAP) ALLOC(bytes(size), 1); }
#else
#include "bitmap.h"
#endif

#define N(x)        ((x) / 3)  /* exclude multiples of 2 and 3*/

void fail(char *s)  { puts(s); exit(1); }     

main(int argc, char *argv[])
{
    BITMAP map;
    double dsize;
    clock_t t1, t2;
    float runtime;
    unsigned i, n, primes, size, sqrt_size, map_size;
    int incr, top;
    int ii, incr2, nn, nn_incr;
    ULONG ndx, msk, prev;
    
    if (argc < 2)
        fail("syntax: sieve <x>");
    if ((dsize = strtod(argv[1], 0)) < 5)
        return 1;

    /* estimate number of primes, using Legendre formula (1778)*/
    printf("Prime Number Theorem estimates: %.0f primes <= %.0f\n",
        floor(dsize / (log(dsize) - 1.08366)), dsize);
    
    if (dsize > UINT_MAX)
        fail("number too large");
    size = (ULONG) dsize;
    top = N(size) + (size % 2);
    map_size = N(size) + 1;
    if (! (map = make_bitmap(map_size))) 
        fail("Insufficient memory");
    printf("bitmap: %lu bytes\n", bytes(map_size));
    
    sqrt_size = sqrt(dsize);
    t1 = clock();
    primes = 2;
    nn = nn_incr = 8;

    /* set composites: MARK*/
    for (i=5, incr=4, n=1; i<=sqrt_size; i+=(incr=6-incr), n++)
    {
        if (BIT_OFF(map, n))            /* bit clear -> prime*/
        {
            primes++;
            nn = N(i * i);
            ii = i + i;
            incr2 = ii - (N(i*(i+(6-incr))) - nn); /* ii - (next-nn)*/

#if 1
            for (; nn < top; nn += (incr2=ii-incr2))
                SET_BIT_TRUE_IF_FALSE(map, nn);
#else
            /* hmm, following is slightly slower right now*/
            /* any way to get ndx, msk in loop??*/
            /* any way to do without if-else??*/

#if 0
            Sample output from printfs below:
            nn  ndx mask      prev  msk
            8   0   100         0   0000
            11  0   800         0   0100
            18  0   40000       0   0900
            21  0   200000      0   40900
            28  0   10000000    0   240900
            31  0   80000000    0   10240900
            38  1   40          0   90240900
            map 0 |= 90240900
            41  1   200         1   0040
            48  1   10000       1   0240
            51  1   80000       1   10240
            58  1   4000000     1   90240
            61  1   20000000    1   4090240
            68  2   10          1   24090240
            map 1 |= 24090240
            71  2   80          2   0010
            78  2   4000        2   0090
            81  2   20000       2   4090
            88  2   1000000     2   24090
            91  2   8000000     2   1024090
            98  3   4           2   9024090
            map 2 |= 9024090
            101 3   20          3   0004
            108 3   1000        3   0024
            111 3   8000        3   1024
            118 3   400000      3   9024
            121 3   2000000     3   409024
            128 4   1           3   2409024
            map 3 |= 2409024
#endif                   
            
            for (prev=index(nn),msk=0; nn < top; nn += (incr2=ii-incr2))
            {
                /* printf("%u\t%u\t%0X\t%u\t%04X\n", */
                /*  nn, index(nn), mask(nn), prev, msk);*/
                if ((ndx = index(nn)) == prev)
                    msk |= mask(nn);
                else
                {   /* rack*/
                    /* if (! map[prev] & msk)*/
                        map[prev] |= msk;
                    /* printf("map %u |= %0X\n", prev, msk);*/
                    prev = ndx;
                    msk = mask(nn);
                }
            }
            /* need to handle end case*/
            map[prev] |= msk;
#endif
        }
    }

    /* count primes: SWEEP (primes are residue)*/
    for ( ; n < top; n++)
        if (BIT_OFF(map, n))
            primes++;

    runtime = (t2 = clock()) - t1;
    
    printf("\n%lu primes <= %lu\n", primes, size);
#ifdef CLOCKS_PER_SEC
    printf("%.2f seconds\n", runtime/CLOCKS_PER_SEC);
#else
    printf("runtime: %lu\n", runtime);
#endif	
}


