/*
SIEVE.C
Author: Andrew Schulman, February 1990
*/

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <limits.h>
#include <time.h>

#if defined(__WATCOMC__) && defined(BT_386)
#include "btmap386.h"
#else
#include "bitmap.h"
#endif

#define N(x)              ((x) / 3)  // exclude multiples of 2 and 3

void fail(char *s)        { puts(s); exit(1); }

main(int argc, char *argv[])
{
    BITMAP map;
	BYTE *p;
    FILE *f;
    double dsize;
    time_t t1, t2;
    ULONG i, j, n, primes, size, map_size;
    int incr, jincr;
    
    if (argc < 2)
        fail("syntax: <run386> sieve <x>");
    if ((dsize = strtod(argv[1], 0)) < 5)
        return 1;

    // estimate number of primes, using Legendre formula (1778)
    printf("Prime Number Theorem estimates: %.0f primes <= %.0f\n",
        floor(dsize / (log(dsize) - 1.08366)), dsize);
    
    if (dsize > ULONG_MAX)
        fail("number too large");
    size = (ULONG) dsize;
    map_size = N(size) + 1;
    if (! (map = make_bitmap(map_size)))
        fail("Insufficient memory");
    printf("bitmap: %lu bytes\n", bytes(SIZE(map)));

    time(&t1);

    // set composites: mark
    for (i=5, incr=4, n=1; i*i<=size; i+=(incr=6-incr), n++)
        if (BIT_OFF(map, n))            // bit clear -> prime
            for (j=i, jincr=incr; j<=size/i; j+=(jincr=6-jincr))
                SET_TRUE(map, N(i*j));  // bit set -> composite

    // count primes: sweep
    // printf("2 3 ");
    for (i=5, incr=4, n=1, primes=2; i <= size; i+=(incr=6-incr), n++)
        if (BIT_OFF(map, n))
        {
            primes++;
            // printf("%lu ", i);
        }

    time(&t2);
    
    printf("\n%lu primes <= %lu\n", primes, size);
    printf("%f seconds\n", difftime(t2, t1));

    puts("Saving bitmap file PRIMES.DAT");
    f = fopen("primes.dat", "wb");
    fwrite(&size, sizeof(ULONG), 1, f);
    fwrite(map, 1, bytes(SIZE(map)), f);  // write out entire bitmap
    fclose(f);
}

