/*
spanning.c
Andrew Schulman
3 May 1989

Assignment:
Write a program to read in a graph with n vertices, labeled 1,2,...n,
and construct a dfs [depth-first search] and bfs [breadth-first search]
spanning tree.  (Print the vertices in order.)

Input:
    12 nodes
    1 2 5 7 8
    2 3 4 5 6
    3 4
    4 6 12
    7 9 11
    9 10 11
        
Output:
    [1] 2 5 7 8 
    [2] 1 3 4 5 6 
    [3] 2 4 
    [4] 2 3 6 12 
    [5] 1 2 
    [6] 2 4 
    [7] 1 9 11 
    [8] 1 
    [9] 7 10 11 
    [10] 9 
    [11] 7 9 
    [12] 4 
    DFS: {1 2 3 4 6 12 5 7 9 10 11 8 }
    BFS: {1 2 5 7 8 3 4 6 9 11 12 10 }
*/

#include <stdio.h>
#include "graph.h"

void dfs(LIST graph);
void bfs(LIST graph);
void visit_dfs(NODE *p);

main(void)
{
    MATRIX tmp;
    LIST graph;
    
    while (tmp = read_graph())
    {
        graph = mat_to_list(tmp);
        free_matrix(tmp);
        print_list(graph);
        dfs(graph);
        bfs(graph);
        /* free_graph(graph); */
        putchar('\n');
    }
}

LIST base;

void dfs(LIST graph)
{
    NODE *p;
    WORD i, size;

    if (! graph) return;
    base = graph;
    size = SIZE(graph);
    for (p=graph+1, i=1; i<= size; i++, p++)
        p->visited = FALSE;
    printf("DFS: {");
    for (p=graph+1, i=1; i<= size; i++, p++)
        visit_dfs(p);
    printf("}\n");
}

void visit_dfs(NODE *p)
{
    if (p->visited) return;
    p->visited = TRUE;
    printf("%d ", p-base);
    do
    {
        if (p->info)
            visit_dfs(&base[p->info]);
    }
    while (p = p->ptr);
}

QUEUE q;

void bfs(LIST graph)
{
    NODE *p;
    WORD i, size, x;

    if (! graph) return;
    size = SIZE(graph);
    base = graph;
    q = queue(size);
    
    for (i=1, p=&graph[1]; i<=size; i++, p++)           
        p->visited = FALSE;
    printf("BFS: {");
    put(q, 1);
    while (! empty(q))
    {
        x = get(q);
        p = &graph[x];
        if (! p->visited)
        {
            p->visited = TRUE;
            printf("%d ", x);
            do
            {
                if (! graph[p->info].visited)
                    put(q, p->info);
            }
            while (p = p->ptr);
        }
    }
    printf("}\n");
}

