/* 
SYMDUMP.C
Andrew Schulman
December 1992

Based on information in Matt Pietrek, "A Windows assert() with
Symbolic Stack Trace," _Windows/DOS Developer's Journal_, July 1992,
and in Windows 3.1 SDK, Vol. 4 (Resources), Chapter 12.

1.01 -- 3/1/93
	Don't check if first is last MAPDEF -- wrong!!
*/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <dos.h>
#include "filelib.h"

typedef int BOOL;
typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

#pragma pack(1)

typedef struct {
	WORD ppNextMap;
	BYTE bFlags, bRes1;
	WORD pSegEntry, cConsts, pConstDef, cSegs, ppSegDef;
	BYTE cbMaxSym, cbModName;
	// BYTE achModName[1];
	} MAPDEF;
	
typedef struct {
	WORD ppNextSeg;
	WORD cSymbols, pSymDef, segNumber, wRes2, wRes3, wRes4;
	BYTE bFlags, bRes1;
	WORD ppLineDef;
	BYTE bRes2, bRes3, cbSegName;
	// BYTE achSegName[1];
	} SEGDEF;
	
typedef struct {
	WORD offset;	// DWORD if 32-bit symbols?
	BYTE cbSymName;
	// BYTE achSymName[1];
	} SYMDEF;

typedef struct {
	DWORD offset;	// DWORD if 32-bit symbols?
	BYTE cbSymName;
	// BYTE achSymName[1];
	} SYMDEF32;

void symdump(char *s);

void usage(void)
{
	puts("usage: symdump <symfile>");
	exit(1);
}

void fail(char *s) { fputs("symdump fail: ", stdout); puts(s); exit(1); }

void fail_perror(void)  { perror("symdump fail"); exit(1); }

main(int argc, char *argv[])
{
	int i;

	fputs("Windows Source Symbol File Dumper version 1.01\n", 
		stderr);
	fputs("Copyright (c) 1992 V Communications Inc.  All rights reserved.\n",
		stderr);
	fputs("Copyright (c) 1992 Andrew Schulman.  All rights reserved.\n\n",
		stderr);
	
	if (argc < 2)
		usage();
	
	for (i=1; i<argc; i++)
		if ((argv[i][0] == '-') || (argv[i][0] == '/'))
		{
			if (argv[i][1] == '?')
				usage();
			else
				usage();		// no switches right now
		}
		else if (strstr(strupr(argv[i]), ".SYM"))
			symdump(argv[i]);
		else
			fail("file doesn't have a .SYM extension");
	return 0;
}

#define file_read(f, p, ofs, bytes)	exe_read((f),(p),(ofs),(bytes))

DWORD start_newexe = 0;

static int f = 0;

char *getcstr(DWORD ofs, int len)	// already got len
{
	char *str = (char *) allocate(1, len+1);
	file_read(f, str, ofs, len);
	str[len] = '\0';
	return str;	// caller must deallocate
}

void symdump(char *filename)
{
	MAPDEF md, md2;
	SEGDEF sd;
	SYMDEF sym;
	SYMDEF32 sym32;
	WORD cSegs;
	DWORD ppSegDef;
	WORD cSym;
	DWORD ppSymDef;
	WORD *symofs;
	char *str;
	int i;
	
    if ((f = sopen(filename, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
        fail_perror();
	file_read(f, &md, 0, sizeof(MAPDEF));
	if ((md.pSegEntry > 255) ||
		(md.cSegs > 255) ||
		(md.bFlags > 2))
		fail("not a Windows .SYM file");
	file_read(f, &md2, md.ppNextMap << 4L, sizeof(MAPDEF));

#if 0	
	if (md2.ppNextMap != 0)	// should be last MAPDEF
		fail("not a .SYM file");
#endif	
	
	str = getcstr(sizeof(MAPDEF), md.cbModName);
	printf("Module %s\n", str);
	free(str);

	cSegs = md.cSegs;
	ppSegDef = md.ppSegDef;
	while (cSegs--)
	{
		file_read(f, &sd, ppSegDef << 4L, sizeof(SEGDEF));
		str = getcstr((ppSegDef << 4L) + sizeof(SEGDEF), sd.cbSegName);
		printf("%u 0000    %s (SEG)\n", sd.segNumber, str);
		free(str);
		cSym = sd.cSymbols;
		ppSymDef = (ppSegDef << 4L) + sd.pSymDef;
		symofs = (WORD *) allocate(cSym, sizeof(WORD));
		file_read(f, symofs, ppSymDef, cSym * sizeof(WORD)); 
		if (sd.bFlags == 1)	// 32-bit symbols
		{
			for (i=0; i<cSym; i++)
			{
				DWORD ofs = (ppSegDef << 4L) + symofs[i];
				file_read(f, &sym32, ofs, sizeof(SYMDEF32));
				str = getcstr(ofs + sizeof(SYMDEF32), sym32.cbSymName);
				printf("   %u %08lx    %s\n", 
					sd.segNumber, sym32.offset, str);
				free(str);
			}
		}
		else
		{
			for (i=0; i<cSym; i++)
			{
				DWORD ofs = (ppSegDef << 4L) + symofs[i];
				file_read(f, &sym, ofs, sizeof(SYMDEF));
				str = getcstr(ofs + sizeof(SYMDEF), sym.cbSymName);
				printf("   %u %04x    %s\n", 
					sd.segNumber, sym.offset, str);
				free(str);
			}
		}
		free(symofs);
		ppSegDef = sd.ppNextSeg;
	}
	close(f);
}


