/* 
tautology.c

Write a program to determine if (p->q)^(q->r)->(p->r) is a tautology.
Display the truth table.
Note: p, q, and r are arbitrary propositions.

Andrew Schulman
3 February 1989

compile with: cl tautolog.c
*/

#include <stdio.h>

#pragma pack(1)

/*
precedence rules:
BICOND, IMPLIC, OR, AND, NOT
*/

/*
because p, q, and r are arbitrary propositions, we can't assign a
simple T/F value to them.  So the truth values needed in the construction
of a truth table are encoded in a 4-byte quantity.  This way, we handle
all possible truth values at once.

Note that operands and operators are stored in the same format.

Note that usual TTFF truth table must be turned around to FFTT:
F must precede T because 0 must precede 1 for array indexing.

problem:  must be inconsistent in order!
*/

/* for representing three-way truth tables */
typedef struct {
	unsigned char t[8];
	} BOOL;

typedef struct {
	unsigned char t[4];
	} BOOLOP;
	
BOOLOP boolop[16] = {
	0, 0, 0, 0,			/* 0 */
	0, 0, 0, 1,			/* 1 */
	0, 0, 1, 0,			/* 2 */
	0, 0, 1, 1,			/* 3 */
	0, 1, 0, 0,			/* 4 */
	0, 1, 0, 1,			/* 5 */
	0, 1, 1, 0,			/* 6 */
	0, 1, 1, 1,			/* 7 */
	1, 0, 0, 0,			/* 8 */
	1, 0, 0, 1,			/* 9 */
	1, 0, 1, 0,			/* 10 */
	1, 0, 1, 1,			/* 11 */
	1, 1, 0, 0,			/* 12 */
	1, 1, 0, 1,			/* 13 */
	1, 1, 1, 0,			/* 14 */
	1, 1, 1, 1,			/* 15 */
	};

#define FFFF	bool[0]				/* contradiction */
#define FFFT	bool[1]				/* AND: p ^ q */
#define FFTF	bool[2]
#define FFTT	bool[3]				/* p */
#define FTFF	bool[4]
#define FTFT	bool[5]				/* q */
#define FTTF	bool[6]				/* XOR: ~(p=q) */
#define FTTT	bool[7]				/* OR: p V q */
/* -------------------- */
#define TFFF	bool[8]				/* NOR: ~(p V q) */
#define TFFT	bool[9]				/* biconditional: p=q */
#define TFTF	bool[10]			/* NOT q */
#define TFTT	bool[11]			/* converse: q->p */
#define TTFF	bool[12]			/* NOT p */
#define TTFT	bool[13]			/* implication: p->q */
#define TTTF	bool[14]			/* NAND, Sheffer stroke: ~(p ^ q) */
#define TTTT	bool[15]			/* tautology */

/* of the 16 "raster ops", we're only using two */
#define IMPLIC  TTFT
#define AND     FFFT

BOOL f(char *msg, BOOL x, BOOL op, BOOL y);
void print_truth_row(char c, BOOL p);

main()
{
    BOOL p,q,r,result;
	
	/* for three way truth table, must use 8 places for each var ! */
	p = FFFFTTTT;
    q = FFTTFFTT;
    r = FTFTFTFT;

#if 1
    /* put intermediary results into variables so we don't crash the stack */
    /* perhaps print out truth table in here */
    result = f("p->q", p,IMPLIC,q);
    result = f( "(p->q)^(q->r)", result, AND, f("q->r", q,IMPLIC,r) );
    result = f( "(p->q)^(q->r)->(p->r)", result, IMPLIC, f("p->r", p,IMPLIC,r) );
#else
    result = 
        f(
            f(
                f(p,IMPLIC,q),
                AND,
                f(q,IMPLIC,r)
            ),
            IMPLIC,
            f(p,IMPLIC,r)
        );
#endif  

#if 0
    if (result == TTTT)
        puts("Tautology!\n");
    else
        puts("Not tautology!\n");
#endif	
}

BOOL f(char *msg, BOOL p, BOOL op, BOOL q)
{
	BOOL ret;
    register int i;
    
	/* op is a look up table */
    for (i=0; i<4; i++)
		ret.t[i] = op.t[(p.t[i] << 1) + q.t[i]];

	puts(msg);
    print_truth_row('p', p);
    print_truth_row('q', q);
    print_truth_row('o', op);
    print_truth_row('f', ret);
    putchar('\n');
    
    return ret;    
}

void print_truth_row(char c, BOOL p)
{
    register int i;
    
    putchar(c); putchar(':'); putchar(' ');
    
    for (i=0; i<4;i++)
        putchar((p.t[i] == 1) ? 'T' : (p.t[i] == 0) ? 'F' : '?');
    
    putchar('\n');
}
    
