/* 
tautology.c

Write a program to determine if (p->q)^(q->r)->(p->r) is a tautology.
Display the truth table.
Note: p, q, and r are arbitrary propositions.

Andrew Schulman
3 February 1989

compile with: cl tautolog.c
*/

#include <stdio.h>

#if 0
#define AND(x,y)            ((x) && (y))
#define IMPLIC(x,y)         ((!(x)) || (y))
#else
/* would like to do this, but compiler runs out of macro preprocessor space */
/* assume we only have NAND, and build all other ops from that */
#define NAND(x,y)           (! ((x) && (y)))
#define NOT(x)              NAND(x,x)
#define AND(x,y)            NOT(NAND(x,y))
#define OR(x,y)             NAND(NAND(x,x),NAND(y,y))
#define IMPLIC(x,y)         OR(NOT(x),(y))
#endif

#define FALSE               0
#define TRUE                1

#define TF(x)               ((x) ? 'T' : 'F')

typedef unsigned BOOL;

main()
{
    BOOL p, q, r, s;
	BOOL tautology = TRUE;		/* may be set to FALSE within loop */

    printf("p \tq \tr \t(p->q)^(q->r)->(p->r)\n");
    printf("- \t- \t- \t---------------------\n");
    
    for (p = FALSE; p <= TRUE; p++)
        for (q = FALSE; q <= TRUE; q++)
            for (r = FALSE; r <= TRUE; r++)
            {
                s = IMPLIC(AND(IMPLIC(p,q),IMPLIC(q,r)),IMPLIC(p,r));
                if (s == FALSE)
                    tautology = FALSE;
                printf("%c \t%c \t%c \t\t%c\n", TF(p), TF(q), TF(r), TF(s));
            }
    
    puts(tautology ? "\nTautology" : "\nNot tautology");
}

/*
output:

p   q   r   (p->q)^(q->r)->(p->r)
-   -   -   ---------------------
F   F   F       T
F   F   T       T
F   T   F       T
F   T   T       T
T   F   F       T
T   F   T       T
T   T   F       T
T   T   T       T

Tautology
*/
