/* 
TEST21.C -- See if INT 21h calls are reflected to V86 mode
from Schulman, Unauthorized Windows, 1994
bcc -2 -P- test21.c
bcc -2 -P- -DTESTFILE="test" test21.c

test21
test21 -mysetvect
test21 > tmp.tmp
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <dos.h>
#include <fcntl.h>

typedef unsigned short WORD;
typedef unsigned long DWORD;

#pragma pack(1)

typedef struct {
#ifdef __TURBOC__
    WORD bp,di,si,ds,es,dx,cx,bx,ax;
#else
    WORD es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    WORD ip,cs,flags;
    } REG_PARAMS;

static void interrupt far int21(REG_PARAMS r);
static void interrupt far ctrl_c(REG_PARAMS r);
static void interrupt far crit_err(REG_PARAMS r);
static int failed = 0;

typedef void (interrupt far *INTPROC)();
static INTPROC old_21 = (INTPROC) 0;

static void (*setv)(unsigned intno, INTPROC proc) = _dos_setvect;
static INTPROC (*getv)(unsigned intno) = _dos_getvect;

static DWORD total_called = 0;
static DWORD called[0x100] = {0};
static DWORD total_received = 0;
static DWORD received[0x100] = {0};

void fail(const char *s) { puts(s); exit(1); }

void my_setvect(unsigned intno, INTPROC proc)
{
    INTPROC far *ivt = (INTPROC far *) 0L;
    ivt[intno] = proc;
}

INTPROC my_getvect(unsigned intno)
{
    INTPROC far *ivt = (INTPROC far *) 0L;
    return ivt[intno];
}

main(int argc, char *argv[])
{
	char *filename = argv[0];
    time_t t1, t2;
    WORD num_iter = 100;
    WORD i;
    int use_my_setvect = 0;
    int received_less = 0;
    int received_more = 0;
    
    for (i=1; i<argc; i++)
        if (argv[i][0] == '-' || argv[i][0] == '/')
        {
            if ((strncmp(strupr(&argv[i][1]), "MYSETVECT", 2)) == 0)
            {
                use_my_setvect++;
                setv = my_setvect;
                getv = my_getvect;
                printf("Using my_setvect\n");
            }
            else
                fail("usage: test21 [-mysetvect] [num_iter]");
        }
        else if (atoi(argv[i]))
			num_iter = atoi(argv[i]);
		else
			filename = argv[i];
    
    time(&t1);
    
    /* hook INT INT 21h */
    old_21 = (*getv)(0x21);
	(*setv)(0x23, ctrl_c);
	(*setv)(0x24, crit_err);
    (*setv)(0x21, int21);
    
    /* issue INT 21h calls */
    for (i=0; i<num_iter; i++)
    {
#ifdef   TESTFILE
#include TESTFILE
#else
        unsigned n;
        char buf;
        int f;

        // we don't care if these calls succeed or fail
        _dos_open(filename, O_RDWR, &f); called[0x3d]++;  // or 0x6c
        _dos_read(f, &buf, 1, &n); called[0x3f]++;
        _dos_close(f); called[0x3e]++;
        _dos_open(filename, O_RDWR, &f); called[0x3d]++;  // or 0x6c
        _dos_write(f, &buf, 1, &n); called[0x40]++;
        _dos_close(f); called[0x3e]++;
        buf = '.';
        #define STDOUT 1
        _dos_write(STDOUT, &buf, 1, &n); called[0x40]++;
        total_called += 7;
#endif
		if (failed)
		{
			(*setv)(0x21, old_21);
			fail("\nCritical error!");
		}
		if (my_getvect(0x21) != int21)
			fail("\nSomeone else grabbed INT 21h!");
    }
            
    /* unhook INT 21h: don't forget to include this call too
       (unless if using my_setvect rather than _dos_setvect)! */
    (*setv)(0x21, old_21);
    if (! use_my_setvect) { called[0x25]++; total_called++; }
	// INT 23h and 24h automatically restored on exit

    time(&t2);
    
    /* display results */
    printf("\n%lu seconds elapsed\n\n", t2 - t1);
    printf("Generated %lu calls\tReceived %lu calls\n",
        total_called, total_received);
    for (i=0; i<0x100; i++)
        if (called[i] || received[i])
        {
            printf("21/%02X\t%lu called\t%lu received\t\t",
                i, called[i], received[i]);
            if (received[i] < called[i])
            {
                received_less++;
                printf("-");
            }
            else if (received[i] > called[i])
            {
                received_more++;
                printf("+");
            }
            printf("\n");
        }
    
    printf("\n");
    if (received_less)
        printf("Some INT 21h are handled without calling DOS!\n");
    if (received_more)
        printf("Some extra INT 21h calls are occurring!\n");
    if (! (received_less || received_more))
        printf("INT 21h appears to be handled in the normal way\n");

    return 0;
}

void interrupt far int21(REG_PARAMS r)
{
    total_received++;
    received[r.ax >> 8]++;
    _chain_intr(old_21);
}

void interrupt far ctrl_c(REG_PARAMS r)
{
	(*setv)(0x21, old_21);
	fail("Ctrl-C detected!");
}

void interrupt far crit_err(REG_PARAMS r)
{
	failed++;
	r.ax = 3;	// fail the operation
}


