/*
TH32.C
Very simple demo of some Toolhelp32 functions in Chicago

from Andrew Schulman, Unauthorized Windows 95, p. 530-531

Copyright (c) Andrew Schulman 1994.
But may be freely distributed, so long as you head this header intact.

Unauthorized Windows: Developers Resource Kit by Andrew Schulman
Price: $39.99 USA/$54.99 Canada
ISBN:  1-56884-305-4; Publication Date: November 1994; Pages: 608

IDG Books are available at all major bookstore chains or direct from
the publisher at (800) 762-2974, 8 a.m. to 5 p.m. Central.  
*/

#include <stdlib.h>
#include <stdio.h>
#define  WIN32_LEAN_AND_MEAN
#include "windows.h"

#if 0
#include "tlhelp32.h"
#else
#define TH32CS_SNAPHEAPLIST 0x00000001
#define TH32CS_SNAPPROCESS  0x00000002
#define TH32CS_SNAPTHREAD   0x00000004
#define TH32CS_SNAPMODULE   0x00000008
#define TH32CS_SNAPALL      (TH32CS_SNAPHEAPLIST | TH32CS_SNAPPROCESS | TH32CS_SNAPTHREAD | TH32CS_SNAPMODULE)

typedef struct tagPROCESSENTRY32
{
    DWORD   dwSize;
    DWORD   cntUsage;
    DWORD   th32ProcessID;      // this process
    DWORD   th32DefaultHeapID;
    DWORD   th32ModuleID;       // associated exe
    DWORD   cntThreads;
    DWORD   th32ParentProcessID;    // this process's parent process
    LONG    pcPriClassBase;     // Base priority of process's threads
    DWORD   dwFlags;
} PROCESSENTRY32;
typedef PROCESSENTRY32 *  PPROCESSENTRY32;
typedef PROCESSENTRY32 *  LPPROCESSENTRY32;

#define MAX_MODULE_NAME32 255

typedef struct tagMODULEENTRY32
{
    DWORD   dwSize;
    DWORD   th32ModuleID;   // This module
    DWORD   th32ProcessID;  // owning process
    DWORD   GlblcntUsage;   // Global usage count on the module
    DWORD   ProccntUsage;   // Module usage count in th32ProcessID's context
    BYTE  * modBaseAddr;    // Base address of module in th32ProcessID's context
    DWORD   modBaseSize;    // Size in bytes of module starting at modBaseAddr
    HMODULE hModule;        // The hModule of this module in th32ProcessID's context
    char    szModule[MAX_MODULE_NAME32 + 1];
    char    szExePath[MAX_PATH];
} MODULEENTRY32;
typedef MODULEENTRY32 *  PMODULEENTRY32;
typedef MODULEENTRY32 *  LPMODULEENTRY32;
#endif

void fail(const char *s) { puts(s); exit(1); }

main()
{
    PROCESSENTRY32 pe32;
    MODULEENTRY32 me32;
    BOOL ok, ok2;
    HANDLE snap;
    char *name;

    HANDLE (WINAPI *CreateToolhelp32Snapshot)(DWORD dwFlags, 
        DWORD th32ProcessID);
    BOOL (WINAPI *Process32First)(HANDLE hSnapshot, LPPROCESSENTRY32 lppe);
    BOOL (WINAPI *Process32Next)(HANDLE hSnapshot, LPPROCESSENTRY32 lppe);
    BOOL (WINAPI *Module32First)(HANDLE hSnapshot, LPMODULEENTRY32 lpme);
    BOOL (WINAPI *Module32Next)(HANDLE hSnapshot, LPMODULEENTRY32 lpme);
    DWORD (WINAPI *VxDCall0)(DWORD srvc, DWORD eax, DWORD ecx);

    HANDLE k32 = GetModuleHandle("KERNEL32");
    #define GET(func) func = GetProcAddress(k32, #func)
    
    // didn't seem to be in import library
    GET(CreateToolhelp32Snapshot);
    GET(Process32First);
    GET(Process32Next);
    GET(Module32First);
    GET(Module32Next);
    GET(VxDCall0);
        
    #define VWIN32_INT31_CALL   0x2A0029
    #define DPMICall(eax, ecx)  VxDCall0(VWIN32_INT31_CALL, (eax), (ecx))
        
    if (! (snap = CreateToolhelp32Snapshot(TH32CS_SNAPALL, 0)))
        fail("Can't create Toolhelp32 snapshot");

    printf("Process    Module     Parent     PSP (#Files)   #Threads\n");
    printf("-------    ------     ------     ------------   --------\n");
    
    pe32.dwSize = sizeof(pe32);
    for (ok=Process32First(snap, &pe32); ok; ok=Process32Next(snap, &pe32))
    {
        // PSP is currently at offset 28h in Process
        WORD psp = *((WORD *) (pe32.th32ProcessID + 0x28));
        DWORD psp_base, jft;
        WORD max;
        _asm mov bx, psp
        DPMICall(0x0006, 0);
        _asm mov word ptr psp_base+2, cx
        _asm mov word ptr psp_base, dx
        max = *((WORD *) (psp_base + 0x32));
        jft = *((DWORD *) (psp_base + 0x34));

        name = "";
        me32.dwSize = sizeof(me32);
        for (ok2=Module32First(snap, &me32); ok2; 
             ok2=Module32Next(snap, &me32))
            if (me32.th32ModuleID == pe32.th32ModuleID)
            {
                if (! (name = me32.szModule))
                    name = me32.szExePath;
                break;
            }
        
        printf("%08lX   %08lX   %08lX   %04X (%d)\t%d\t%s\n",
            pe32.th32ProcessID,
            pe32.th32ModuleID,
            pe32.th32ParentProcessID,
            (DWORD) psp, (DWORD) max,
            pe32.cntThreads,
            name);
    }
    
    CloseHandle(snap);
}


