/* 
TRUENAME.C -- for Windows
Andrew Schulman, February 1993
From "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

bcc -W -DWINDOWS -2 truename.c win_app.lib
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include "windows.h"
#include "printf.h"
#include "prot.h"

void fail(const char *s) { MessageBox(NULL, s, "TRUENAME", MB_OK); exit(1); }

char far *truename(char far *src, char far *dest)
{
    RMODE_CALL r;
    char far *s2, far *fp;
    WORD para, sel;
    
    /* INT 21h AH=60h doesn't like trailing or leading blanks */
    while (isspace(*src)) src++;
    s2 = src;
    while (*s2) s2++;
    s2--;
    while (isspace(*s2)) *s2-- = '\0';
    
    /* Alloc 256 bytes conventional memory:  first 128 to transfer in src;
       second half to transfer back dest */
    if (! alloc_real_seg(256, &para, &sel))
        fail("Couldn't allocate conventional memory!");
    
    /* Generate real-mode 21/60 */
    memset(&r, 0, sizeof(r));
    r.eax = 0x6000;
    r.ds = r.es = para;  /* pass real-mode para addr to DOS */
    r.esi = 0;
    r.edi = 128;
    fp = (char far *) MK_FP(sel, 0);
    _fstrcpy(fp, src);              /* use prot-mode sel addr in program */
    if (! dpmi_rmode_intr(0x21, 0, 0, &r))
        fail("DPMI real-mode interrupt failed!");
    
    if (r.flags & 1)    /* If carry set */
        _fstrcpy(dest, "<Invalid>");
    else
        _fstrcpy(dest, (char far *) MK_FP(sel, 128));
    
    /* copied to caller's dest; don't need conventional-memory
       transfer buffer anymore */
    if (! free_real_seg(sel))
        fail("Couldn't free conventional memory!");
    
    return dest;
}

int PASCAL WinMain(HANDLE hInst, HANDLE hPrevInst,
    LPSTR lpszCmdLine, int nCmdShow)
{
    char buf[128], dest[128];
    
    if (! (GetWinFlags() & WF_PMODE))
        fail("This program requires Windows Enhanced or Standard mode");
    
    if (! lpszCmdLine && *lpszCmdLine)
        fail("syntax: truename <pathname>");
    
    sprintf(buf, "TRUENAME %Fs", lpszCmdLine);
    open_display(buf);
    printf("%Fs\n", truename(lpszCmdLine, dest));
    show_display();
    return 0;
}
