/*
 * UDMEM.C -- walks DOS MCB chain(s): detailed version
 * Andrew Schulman and Jim Kyle, July 1990 (MEM.C)
 * Modified July 1992 to include UMBs and subsegments and change name
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <dos.h>

#include "undocdos.h"

#ifdef __TURBOC__
#define GETVECT(x)	getvect(x)
#else
#define GETVECT(x)	_dos_getvect(x)
#endif


//extern LPMCB firsthi( void );		/* locates start of UMB MCBs	*/

LPMCB firsthi( void);

BOOL belongs( FP vec, WORD start, WORD size );
void display( MCB far * mcb );
char far * env( MCB far * mcb );
void display_progname( MCB far * mcb );
void display_cmdline( MCB far * mcb );
void display_vectors( MCB far * mcb );
WORD fstrlen( char far * s );

LPMCB firsthi( void)
{
	return(0);
}

#ifdef TSR
void fail( char * s)
{
	printf("%s\r\n", s);
	return;
}
#else
void fail( char * s )			/* print message and get out	*/
{ puts( s );
  exit( 1 );
}
#endif

LPMCB get_mcb( void )			/* get MCB start from LofL	*/
{ ASM mov ah,52h
  ASM int 21h
  ASM mov dx,es:[bx - 2]
  ASM xor ax,ax
/* in both Microsoft C and Turbo C, far* returned in DX:AX */
}

mcb_chk( LPMCB mcb )			/* verify integrity of chain	*/
{ for( ; ; )
    if( mcb->type == 'M' )
      mcb = MK_FP( FP_SEG( mcb ) + mcb->size + 1, 0 );
    else
      return ( mcb->type == 'Z' );
}

void walk( LPMCB mcb )			/* walks chain displaying data	*/
{ 
#ifdef TSR
  printf( "Set     Owner   Size\r\n");
#else
  printf( "Seg     Owner   Size\n" );
#endif
  for( ; ; )
    switch( mcb->type )
      {
      case 'M' :                        /* Mark : belongs to MCB chain  */
	display( mcb );
	mcb = MK_FP( FP_SEG( mcb ) + mcb->size + 1, 0 );
	break;

      case 'Z' :                        /* Zbikowski : end of MCB chain */
	display( mcb );
	printf( "%04X                            ",
		FP_SEG( mcb ) + mcb->size + 1 );
	if( FP_SEG( mcb ) < 0xA000 )
	  { 
#ifdef TSR
       printf( "End of conventional RAM\r\n" );
#else
       printf( "End of conventional RAM\n" );
#endif


	    mcb = firsthi();		/* try to link to UMB area	*/
	    if( mcb == ( FP ) 0L )	/* No UMB MCB's found           */
	      return;
#ifdef TSR
	    printf( "        UMB Chain\r\n" );
#else
	    printf( "        UMB Chain\n" );
#endif
	  }
	else
	  { 
#ifdef TSR
       printf( "End of UMB Chain\r\n" );
#else
       printf( "End of UMB Chain\n" );
#endif

	    return;
	  }
	break;

      default :
	fail( "error in MCB chain" );
       return;
      }
}

static BOOL ssdone = FALSE;

void display_subsegs( LPMCB mcb, WORD subsegsz )
{ char tmp[ 10 ];
  char far * src;
  int i;
  if( ssdone )				/* if not first DOS area	*/
    { 
#ifdef TSR
      printf( "Code area\r\n" );          /* cannot be data segment       */
#else
      printf( "Code area\n" );          /* cannot be data segment       */
#endif

      return;
    }
#ifdef TSR
  printf( "Data Segment\r\n" );
#else
  printf( "Data Segment\n" );
#endif
  if( _osmajor < 4 )			/* subsegments here only in V4+ */
    return;

#ifdef TSR
  printf( "\t  Seg  Size Type \r\n\t  ---- ---- ----------------\r\n" );
#else
  printf( "\t  Seg  Size Type \n\t  ---- ---- ----------------\n" );
#endif

  mcb = MK_FP( FP_SEG( mcb ) + 1, 0 );
  while( ( int ) subsegsz > 0 ) 	/* process each subsegment	*/
    { printf( "\t  %04X %04X ", FP_SEG( mcb ), mcb->size );
      src = &( mcb->dos4[ 0 ]); 	/* copy filename for D, I	*/
      for( i = 0; i < 8 && * src > ' '; i++ )
	tmp[ i ] = *src++;
      tmp[ i ] = 0;
      switch( mcb->type )		/* translate type codes 	*/
	{
	case 'D' :
	  printf( "Device Driver (%s) ", tmp );
	  break;
	case 'E' :
	  printf( "Device Driver appendage " );
	  break;
	case 'I' :
	  printf( "IFS Driver (%s) ", tmp );
	  break;
	case 'F' :
	  printf( "System File Tables " );
	  break;
	case 'X' :
	  printf( "FCBs " );
	  break;
	case 'C' :
	  printf( "Buffer Workspace " );
	  break;
	case 'B' :
	  printf( "Buffers " );
	  break;
	case 'L' :
	  printf( "CDS Table " );
	  break;
	case 'S' :
	  printf( "Stacks " );
	  break;
	case 'T' :
	  printf( "Transient Code " );
	  break;
	default :
	  printf( "Unknown subsegment " );
	}
      display_vectors( mcb );		/* show vectors in this area	*/
#ifdef TSR
      printf( "\r\n" );
#else
      printf( "\n" );
#endif
      subsegsz -= ( mcb->size + 1 );
      mcb = MK_FP( FP_SEG( mcb ) + mcb->size + 1, 0 );
    }
  ssdone = TRUE;
}

void display( LPMCB mcb )
{ static FP vect_2e = ( FP ) 0;
  WORD env_seg;
  printf( "%04X    %04X    %04X (%6lu)   ",
	  FP_SEG( mcb ), mcb->owner,
	  mcb->size, ( long ) mcb->size << 4 );

  if( IS_PSP( mcb ))
    { FP e = env( mcb );		/* MSC wants lvalue */
      if( ( env_seg = FP_SEG( e )) != 0 )
	printf( "Env at %04X    ", env_seg );
      else
	printf( "No Env Segment " );
    }

  if( ! vect_2e )
    vect_2e = GETVECT( 0x2e );		/* do just once */

/* INT 2Eh belongs to master COMMAND.COM (or other shell) */
  if( belongs( vect_2e, FP_SEG( mcb ), mcb->size ))
    printf( "%s ", getenv( "COMSPEC" ));
  else
    switch( mcb->owner )		/* decode special stuff */
      {
      case 0 :
	printf( "free " );
	break;
      case 6 :
	printf( "DR-DOS XMS UMB " );
	break;
      case 7 :
	printf( "DR-DOS hole " );
	break;
      case 8 :
	printf( "DOS " );
	display_subsegs( mcb, mcb->size );
	return; 			/* display_subsegs cleans up */
      case 0xFFFA :
	printf( "386MAX UMB control block " );
	break;
      case 0xFFFD :
	printf( "386MAX locked-out area " );
	break;
      case 0xFFFE :
	printf( "386MAX UMB " );
	break;
      case 0xFFFF :
	printf( "386LOADed Driver " );
	break;
      }
  display_progname( mcb );		// moved from original location
  if( IS_PSP( mcb ))
    display_cmdline( mcb );
  display_vectors( mcb );
#ifdef TSR
  printf( "\r\n" );
#else
  printf( "\n" );
#endif
}

char far * env( LPMCB mcb )
{ char far * e;
  WORD env_mcb;
  WORD env_owner;

  /*  if the MCB owner is one more than the MCB segment then
   *	  psp := MCB owner
   *	  env_seg := make_far_pointer(psp, 2Ch)
   *	  e := make_far_pointer(env_seg, 0)
   *  else
   *	  return NULL
   */
  if( IS_PSP( mcb ))
    e = MK_FP( ENV_FM_PSP( mcb->owner ), 0 );
  else
    return ( char far *) 0;

  /* Does this environment really belong to this PSP? An
   * environment is just another memory block, so its MCB is
   * located in the preceding paragraph. Make sure the env
   * MCB's owner is equal to the PSP whose environment this
   * supposedly is! Thanks to Rob Adams of Phar Lap Software
   * for pointing out the need for this check; this is a
   * good example of the sort of consistency check one must
   * do when working with undocumented DOS.
   *
   * Note that with DOS5, this test had to be changed to just
   * reject free MCBs, because 386MAX added special codes in
   * the owner field that caused our original test to fail!
   */
  env_mcb = MCB_FM_SEG( FP_SEG( e ));
  env_owner = (( MCB far *) MK_FP( env_mcb, 0 ))->owner;
/*return ( env_owner == mcb->owner ) ? e :( char far *) 0;  changed! */
  return ( env_owner ) ? e :( char far *) 0;
}

char far * progname_fm_psp( WORD psp )
{ char far * e;
  WORD len;

/* is there an environment? */
  if( ! ( e = env( MK_FP( MCB_FM_SEG( psp ), 0 ))))
    return ( char far *) 0;

/* program name only available in DOS 3+ */
  if( _osmajor >= 3 )
    { do				/* skip environment variables	*/
	e += ( len = fstrlen( e )) + 1;
      while( len );
      if( !*e ) 	/* test for empty env in UMB, created by 386MAX */
	e++;

      /*  e now points to WORD containing number of strings following
       *  environment; check for reasonable value: signed because
       *  could be FFFFh; will normally be 1
       */
      if(( *(( signed far *) e ) >= 1 ) && ( *(( signed far *) e ) < 10 ))
	{ e += sizeof( signed );
	  if( isalpha( * e ))
	    return e;
	}
    }
  return ( char far *) 0;
}

void display_progname( LPMCB mcb )
{ char far * s;
  if( IS_PSP( mcb ))
    if(( s = progname_fm_psp(( FP_SEG( mcb ) + 1 ))) != 0 )
      printf( "%Fs ", s );
}

BOOL belongs( FP vec, WORD start, WORD size )
{ WORD seg = FP_SEG( vec ) + ( FP_OFF( vec ) >> 4 );/* normalize */
  return ( seg >= start ) && ( seg <= ( start + size ));
}

void display_cmdline( LPMCB mcb )
{ /*	psp := MCB owner
   *	cmdline_len := psp[80h]
   *	cmdline := psp[81h]
   *	print cmdline(display width := cmdline_len)
   */
  int len =* (( BYTE far *) MK_FP( mcb->owner, 0x80 ));
  char far * cmdline = MK_FP( mcb->owner, 0x81 );
  printf( "%.*Fs ", len, cmdline );
}

void display_vectors( LPMCB mcb )
{ 
#ifdef TSR
  void far *vec;
  int i;
  int did_one = 0;


    for (i=0; i<256; i++)
        if ((vec = GETVECT(i)) && 
            belongs(vec, FP_SEG(mcb), mcb->size))
        {
            if (! did_one) { did_one++; printf("["); }
            printf("%02X ", i);
        }
    if (did_one) printf("]");

#else

  static FP * vec = ( FP *) 0;
  WORD vec_seg;
  int i;
  int did_one = 0;

  if( ! vec )
    { if( !( vec = calloc( 256, sizeof( void far *))))
	fail( "insufficient memory" );
      for( i = 0; i < 256; i ++ )
	vec[ i ] = GETVECT( i );
    }

  for( i = 0; i < 256; i ++ )
    if( vec[ i ] && belongs( vec[ i ], FP_SEG( mcb ), mcb->size ))
      { if( ! did_one )
	  { did_one ++ ;
	    printf( "[" );
	  }
	printf( "%02X ", i );
	vec[ i ] = 0;
      }

  if( did_one )
    printf( "]" );
#endif
}



#ifdef TSR
void application(void)
{
    ssdone = FALSE;
    walk(get_mcb());
}
#else

main( int argc, char * argv[])
{ if( argc < 2 )
    walk( get_mcb());			/* walk "normal" MCB chain */
  else
    { unsigned seg;
      sscanf( argv[ 1 ], "%04X", &seg );
      walk( MK_FP( seg, 0 ));		/* walk arbitrary MCB chain */
    }
  return 0;
}
#endif


