/*
UNINIT.C
Win32 app, sees if uninitialized pointers are writeable
Andrew Schulman, April 1995
76320.302@compuserve.com

Sample output:
005201B5    Can read/write  
00000001    Can't access    
00530000    Can't access    
81550D84    Can read/write  
8154B5C4    Can read/write  
8154B5C4    Can read/write  
0040119A    Can read        
0063FDFC    Can read/write  
00000002    Can't access    
00000002    Can't access    
0063FE40    Can read/write  
0063FD9C    Can read/write  
0063FBC8    Can read/write  
0063FDD0    Can read/write  
FFFFFFFF    Can't access    
00406000    Can read        
004011AC    Can read        
0063FE1C    Can read/write  
00530000    Can't access    
81550D84    Can read/write  
8154B5C4    Can read/write  
00407078    Can read/write  
00401122    Can read        
00000000    Can't access    

If run UNINIT 200, start seeing KERNEL32 access:
BFF9C147        Can read        KERNEL32
BFF991B1        Can read        KERNEL32
BFF72058        Can read        KERNEL32
BFFC177B        Can read        KERNEL32
BFF8CD9C        Can read        KERNEL32
BFF8DB5A        Can read        KERNEL32
BFF8DB4E        Can read        KERNEL32
But notice all are read-only.
*/

#include <stdlib.h>
#include <stdio.h>
#define WIN32_LEAN_AND_MEAN
#include "windows.h"
#include "excpt.h"

typedef enum { NO_ACCESS, DID_READ, DID_WRITE } STATUS;
char *status_str[] = { "Can't access", "Can read     ", "Can read/write" } ;

int danger = 0;

void attempt(void *p)
{
    STATUS status = NO_ACCESS;
    _try { 
        unsigned long *pul = (unsigned long *) p;
        unsigned long ul = *pul;
        status = DID_READ;
        *pul = (danger) ? ~ul : ul;
        status = DID_WRITE;
    }
    _except (EXCEPTION_EXECUTE_HANDLER) { ; }
    printf("%08X\t%s\t", p, status_str[status]);
    if (status > NO_ACCESS)
    {
        #define GET_PROC(mod, func) GetProcAddress(GetModuleHandle(mod), func)
        static unsigned k32 = 0;
        if (k32 == 0) 
            k32 = ((unsigned) GET_PROC("KERNEL32", 
                "VirtualAlloc")) & 0xFFF00000;
        if ((((unsigned) p) & 0xFFF00000) == k32) printf("KERNEL32");
    }
    printf("\n");
}

main(int argc, char *argv[])
{
    unsigned dummy;
    unsigned *stk = &dummy;
    int i;
    unsigned *p;
    int iter = (argc < 2) ? 24 : atoi(argv[1]);
    for (i=0, p=stk; i < iter; i++, p--)
        attempt(*p);
}

