comment~
VXD.ASM
Generic Virtual Device Driver for Windows 3.x 386 Enhanced mode
from Microsoft Systems Journal
by Andrew Schulman and David S. Maxey, 1991-1992

Provides applications running under Windows Enhanced mode (Windows
apps, real-mode DOS apps, and protected-mode DPMI DOS apps) with
access to the VMM and VxD functions documented in the Windows DDK.
A program calls VXD.386 using its V86 or PM API; the API entry
point is retrieved with INT 2Fh AX=1684h BX=28C0h (VXD.386's device
ID).  The entry point is called with AX set to a subfunction
number:
	0 -- Get VXD.386 version number (ret AX = version)
	1 -- VxDCall (ES:BX -> VxDParams struct; ret carry clear = success)
	2 -- MapFlat (ES:DI = seg:ofs address; ret DX:AX = FLAT32 address)
	3 -- VxDPushCall (ES:BX -> VxDPushParams; ret carry clear = success)

Possible future enhancements:
	Allocate callbacks
	Ability to hook Control events
	Call code in another VM

The files VXDCALLS.C and VXDCALLS.H provide C/C++ wrappers.

To build:
   set include=\ddk\include
   masm5 -p -w2 vxd;
   link386 vxd,vxd.386,,,vxd.def
   addhdr vxd.386

Add device=vxd.386 to SYSTEM.INI [386Enh]

To do:
Get rid of reg map
Get rid of FS, GS, EFLAGS as params!
Why is DPMI_MEM broken?
~

.386p

INCLUDE VMM.INC

;*****************************************************************************
;		 	S T R U C T U R E S   A N D   E Q U A T E S
;*****************************************************************************

VxDParams 		STRUC
	CallNum		dd	?
	InRegMap	dd	?
	InEAX		dd	?
	InEBX		dd	?
	InECX		dd	?
	InEDX		dd	?
	InEBP		dd	?
	InESI		dd	?
	InEDI		dd	?
	InFS		dw	?
	InGS		dw	?
	InEFLAGS	dd	?
	OutEAX		dd	?
	OutEBX		dd	?
	OutECX		dd	?
	OutEDX		dd	?
	OutEBP		dd	?
	OutESI		dd	?
	OutEDI		dd	?
	OutFS		dw	?
	OutGS		dw	?
	OutEFLAGS	dd	?
VxDParams		ENDS

VxDPushParams 	STRUC
	CallNum_2	dd	?
	NumP		dd	?
	; PushCParams macro in VMM.INC allows 10 params
	P			dd	10 dup	(?)
	OutEAX_2	dd	?
	OutEDX_2	dd	?
VxDPushParams	ENDS

VXD_V_HI		EQU	1
VXD_V_LO		EQU 0

; VxD ID assigned by vxdid@microsoft.com
VxD_ID     		EQU 28C0h

CFlag			EQU 1

RegMap_EAX		EQU	0000000000000001b
RegMap_EBX		EQU	0000000000000010b
RegMap_ECX		EQU	0000000000000100b
RegMap_EDX		EQU	0000000000001000b
RegMap_EBP		EQU	0000000000010000b
RegMap_ESI		EQU	0000000000100000b
RegMap_EDI		EQU	0000000001000000b
RegMap_FS		EQU	0000001000000000b
RegMap_GS		EQU	0000010000000000b
RegMap_EFLAGS	EQU	0000100000000000b

;*****************************************************************************
;		 	V I R T U A L	 D E V I C E   D E C L A R A T I O N
;*****************************************************************************

Declare_Virtual_Device VXD, VXD_V_HI, VXD_V_LO, VxD_Control, \
	VxD_ID, Undefined_Init_Order, \
	API_Handler, API_Handler,

;*****************************************************************************
;			S T U B   C O N T R O L   P R O C
;*****************************************************************************

VxD_LOCKED_CODE_SEG

BeginProc 	VxD_Control
	clc
	ret
EndProc 	VxD_Control

VxD_LOCKED_CODE_ENDS

;*****************************************************************************
;			D A T A   S E G M E N T
;*****************************************************************************

VxD_DATA_SEG

API_Call	label dword			; table used by API_Handler
	dd	offset32 VxDVersion
	dd	offset32 MyVxdCall
	dd	offset32 MyMapFlat
	dd	offset32 VxDPushCall

API_MaxCall 	EQU ( $ - API_Call ) / 4

; Miscellaneous data used by MyVxDCall and VxDPushCall
Store_ESI		dd	?
Store_EFlags	dd	?
Save_Call_Num	dd	0		; was call same as last time?
Sem				dd	0		; semaphore -- protect self-modifying code
Save_Call_Num_2	dd	0		; was call same as last time?
Stack_Restore	dd	0		; number of bytes to add to esp
Sem2			dd	0		; semaphore -- protect self-modifying code

VxD_DATA_ENDS

;*****************************************************************************
;          	C O D E   S E G M E N T
;*****************************************************************************

VxD_CODE_SEG

;*****************************************************************************
;          	API_Handler
;*****************************************************************************

BeginProc	API_Handler
	movzx	eax, [ebp.Client_AX]
	cmp		eax, API_MaxCall				; Is this a valid function?
	jae		SHORT API_Failed
	and		[ebp.Client_EFlags], NOT CFlag	; Set up for success
	call	API_Call[eax * 4]
	ret
API_Failed:
	or		[ebp.Client_EFlags], CFlag
	ret
EndProc		API_Handler

;*****************************************************************************
;          	VxDVersion
;			Input: None
;			Output: Version number in AX
;*****************************************************************************

BeginProc	VxDVersion
	mov		[ebp.Client_AX], ((VXD_V_HI SHL 8) OR VXD_V_LO)
	ret
EndProc		VxDVersion

;*****************************************************************************
;          	MyVxDCall
;			Input: Ptr to VxDParams struct in ES:BX
;			Output: VxDParams OUT regs changed; carry flag
;*****************************************************************************

IF 0
TryReg	MACRO	reg, next
	mov		reg, [esi.In&reg]
	ENDM
ELSE
TryReg	MACRO  reg, next
Do_In&reg:
	test	edi, RegMap_&reg
	jz		short Do_In&next
	mov		reg, [esi.In&reg]
	xor		edi, RegMap_&reg	; clear bit
	jz		Do_Int20			; if no parameters left
	ENDM
ENDIF

BeginProc	MyVxDCall
	mov		ax, (Client_ES shl 8) + Client_BX
	VMMcall	Map_Flat		; Get FLAT address for caller's parameter block
	pushad					; Save our registers
	pushfd					; ... and flags
	pushfd					; ... twice - we're going to need them
	pop		[Store_EFlags]	; one push left on the stack
	mov		esi, eax
	mov		[Store_ESI], esi

	; self-modifying code below: only one caller at a time
	cmp		Sem, 0
	jnz		Busy
	inc		Sem

	mov		eax, [esi.CallNum]	; get requested function call number
	cmp		eax, ds:[Save_Call_Num]
	je		short Do_InRegs		; function hasn't changed since last time

SetUp_Function:
	mov		ds:[Save_Call_Num], eax
	mov		ds:[Call_Num], eax
	; the following is crucial:  after first call, Windows changes
	; the INT 20h/DD into a CALL instruction; change it back!
	mov		word ptr ds:[Int_20], 020cdh

Do_InRegs:
	mov		edi, [esi.InRegMap]
	or		edi, edi			; see if any parameters
	jz		Do_Int20

Do_InEFLAGS:
	test	edi, RegMap_EFLAGS	; Does caller want to set flags?
	jz		short Do_PushFlags
	push	[esi.InEFLAGS]
	pop		[Store_EFlags]

Do_PushFlags:
	push	[Store_EFlags]	; Push either our flags or caller's ...

	TryReg	EAX, EBX
	TryReg	EBX, ECX
	TryReg	ECX, EDX
	TryReg	EDX, EBP
	TryReg	EBP, FS
	TryReg	FS, GS
	TryReg	GS, ESI
	TryReg	EDI, ESI
	; InRegMap is no longer in EDI

Do_InESI:
	test	[esi.InRegMap], RegMap_ESI	; do ESI last: need for client regs
	jz		short Do_Int20
	mov		esi, [esi.inESI]

Do_Int20:
	popfd							; Get EFlags set up from earlier push
Int_20:
; following will get changed to 32-bit near call; breakpoints will be lost
	int		20h
Call_Num	dd		0		; This is the most important line in VXD.ASM

	xchg	esi, [Store_ESI]		; we need ESI restored immediately
	pushfd							; and we don't want to damage the flags
	pop		[Store_EFlags]
	dec		Sem						; don't want to affect flags
	mov		[esi.OutEAX], eax
	mov		[esi.OutEBX], ebx
	mov		[esi.OutECX], ecx
	mov		[esi.OutEDX], edx
	mov		[esi.OutEBP], ebp
	mov		[esi.OutEDI], edi
	mov		[esi.OutFS], fs
	mov		[esi.OutGS], gs
	mov		eax, [Store_EFlags]
	mov		[esi.OutEFLAGS], eax
	mov		eax, [Store_ESI]
	mov		[esi.OutEAX], eax
Done:
	popfd
	popad
	ret
Busy:
	or		[ebp.Client_EFlags], CFlag
	jmp		short Done
EndProc		MyVxDCall

;*****************************************************************************
;          	MyMapFlat
;			Input: Seg:ofs address (PM or V86) in ES:DI
;			Output: Equivalent FLAT32 address in DX:AX
;*****************************************************************************

BeginProc	MyMapFlat
	mov		ax, (Client_ES shl 8) + Client_DI	; uses structure offsets!
	VMMcall	Map_Flat
	mov		[ebp.Client_AX], ax
	shr		eax, 16
	mov		[ebp.Client_DX], ax
	ret
EndProc		MyMapFlat

;*****************************************************************************
;          	VxDPushCall
;			Input: VxDPushParams struct in ES:BX
;			Output: VxDPushParams OUT regs changed; carry flag
;*****************************************************************************

BeginProc	VxDPushCall
	mov		ax, (Client_ES shl 8) + Client_BX
	VMMcall	Map_Flat		; Get FLAT address for caller's parameter block
	mov		esi, eax

	cmp		Sem2, 0
	jnz		short Busy2
	inc		Sem2

	mov		eax, [esi.CallNum_2]	; get requested function call number
	cmp		eax, ds:[Save_Call_Num_2]
	je		short DoPush			; function hasn't changed since last time

SetUp_Function_2:
	mov		ds:[Save_Call_Num_2], eax
	mov		ds:[Call_Num_2], eax
	; the following is crucial:  after first call, Windows changes
	; the INT 20h/DD into a CALL instruction; change it back!
	mov		word ptr ds:[Int_20_2], 020cdh

DoPush:
	mov 	ecx, [esi.NumP]					; number of arguments
	mov		eax, ecx
	shl		eax, 2							; number of bytes
	mov		ds:[Stack_Restore], eax			; save to add later to ESP
	or		ecx, ecx
	jz		short Int_20_2					; no params
	lea		ebx, [esi.P]					; EBX = offset into regs struct
	add		ebx, eax						; going to push in reverse order
PushLoop:
	sub		ebx, 4							; get offset to next arg
	push	[ebx]							; push next arg
	dec		ecx
	jnz		PushLoop

Int_20_2:
; following will get changed to 32-bit near call; breakpoints will be lost
	int		20h
Call_Num_2	dd		0
	add		esp,	ds:[Stack_Restore]		; C calling convention: clean up
	dec		Sem2
	mov		[esi.OutEAX_2], eax
	mov		[esi.OutEDX_2], edx
	ret
Busy2:
	or		[ebp.Client_EFlags], CFlag
	ret
EndProc		VxDPushCall

VxD_CODE_ENDS

	END
