/*
V86INT.C -- Walk Windows V86 int chain
Andrew Schulman, February 1994
bcc -DDPMI_APP -2 v86int.c dpmish.c prot.c ctrl_c.asm
*/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include "dpmish.h"
#include "prot.h"
#include "vxdcalls.h"

#pragma pack(1)
typedef struct {
    BYTE  call;     // must be E8
    DWORD func;     // relative to offset of next instruction
    BYTE mov;       // must be B8
    BYTE intno;     // for sanity checking
    BYTE pad[9];
    WORD jc;        // must be 0F 82
    DWORD next;     // relative to offset of next instruction
    BYTE ret;       // must be C3
    } INTCHAIN;
#pragma pack()

#define IS_INTCHAIN(ic)     ((ic->call == 0xe8) && \
                             (ic->mov == 0xb8) && \
                             (ic->jc == 0x820f) && \
                             (ic->ret == 0xc3))

DWORD axtol(char *s)
{
    DWORD ret;
    #define OX      0x7830      // '0x'
    sscanf((*((WORD *) s) == OX) ? &s[2] : s, "%lx", &ret);
    return ret;
}

#define PUT(s)                { fputs(s, stderr); fputs("\n", stderr); }

void fail(const char *s, ...) { PUT(s); _dos_exit(1); }

int real_main(int argc, char *argv[])   
{ 
    PUT("V86INT -- Walk Windows V86 Int Chain");
    PUT("From \"Unauthorized Windows\" (IDG Books, 1994)");
    PUT("Copyright (c) 1994 Andrew Schulman.  All rights reserved.\n");
    if (argc < 2)
        fail("usage: v86int <v86 int chain table linear> [default handler]");
    return 0; 
}

int pmode_main(int argc, char *argv[])
{
    INTCHAIN far *ic;
    DWORD v86_int_chain_lin;
    DWORD far *v86_int_chain_tab;
    DWORD far *fp;
    DWORD func, next;
    DWORD default_handler;
    int show_header = 0;
    int i;
    
    if (strcmp(strupr(argv[1]), "-HEADER") == 0)
    {
        show_header++;
        argv++; argc--;
    }
    
    v86_int_chain_lin = axtol(argv[1]);
    default_handler = (argc > 2) ? axtol(argv[2]) : 0;
    
    printf("Possible V86 int chain table at %08lXh\n",
        v86_int_chain_lin);
    
    v86_int_chain_tab = (DWORD far *) map_linear(v86_int_chain_lin, 
        0x100 * sizeof(DWORD));
    for (i=0, fp=v86_int_chain_tab; i<0x100; i++, fp++)
    {
        if (*fp < 0x80000000L)
            fail("Something wrong!");
        if ((default_handler == 0) || (*fp != default_handler))
        {
            printf("%04X\t%08lX", i, *fp);
            ic = (INTCHAIN far *) map_linear(*fp, sizeof(INTCHAIN));
            next = *fp;
            while (IS_INTCHAIN(ic))
            {
                if (ic->intno != i)
                    fail("Something wrong!");
                func = ic->func + next + offsetof(INTCHAIN, mov);
                next = ic->next + next + offsetof(INTCHAIN, ret);
                printf(" -> %08lX", show_header ? next : func);
                free_mapped_linear(ic);
                ic = (INTCHAIN far *) map_linear(next, sizeof(INTCHAIN));
            }
            free_mapped_linear(ic);
            if ((next != *fp) && (next != default_handler))
                printf(" -> %08lX", next);  // shouldn't happen
            printf("\n");
        }
    }
    free_mapped_linear(v86_int_chain_tab);
    if (default_handler)
        printf("Other entries go to default handler at %08lXh\n",
            default_handler);
    return 0;
}

