/*
VMWALK.H -- See VMWALK.C
Andrew Schulman, February 1993
From "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)
*/

// the first four fields are documented in the DDK vmm.inc
typedef struct {
    DWORD CB_VM_Status;
    DWORD CB_High_Linear;
    DWORD CB_Client_Pointer;
    DWORD CB_VMID;
	BYTE other[0x100];	// e.g., vm->next at VM+68h
	// actual VM_CB is larger than this: about 1B40h bytes!  For details,
	// see article on VM CB by Kelly Zytaruk in Dr. Dobb's Journal
	// "Undocumented Corner," probably December 1993 and/or January 1994
    } VM_CB;			// Virtual Machine Control Block
		
// Known VM_CB offsets in Windows 3.1:
// For more details, see article on VM CB in Dr. Dobb's Journal "Undocumented
// Corner," probably in December 1993, January 1994
// 00h	status
// 04h  high linear
// 08h  client pointer
// 0Ch  VM ID
// 14h  Get_Last_Updated_VM_Exec_Time return
// 18h  page tables
// 64h  Get_Cur_PM_App_CB return
// 68h  Get_Next_VM_Handle return
// BCh  instance data struct
// 10Ch A20 wrap flag

// Not 100% sure about PGTAB_VM and INST_VM
#define PGTAB_VM(vm_cb)	*((DWORD far *) (((BYTE far *) vm_cb)+0x18))
#define NEXT_VM(vm_cb)	*((DWORD far *) (((BYTE far *) vm_cb)+0x68))
#define INST_VM(vm_cb)	*((DWORD far *) (((BYTE far *) vm_cb)+0xBC))
        
typedef BOOL (*WALKFUNC)(DWORD vm, VM_CB far *);        
int vmwalk(WALKFUNC walkfunc);
extern BOOL walkfunc(DWORD vm, VM_CB far *vm_cb);   // app to provide

#ifndef DPMI_APP
// for Windows-based WVMWALK (without generic VxD; 3.1 Win apps only!)
typedef BOOL (*WWALKFUNC)(HWND hwnd, char *title, DWORD vm, VM_CB far *vm_cb);
int wvmwalk(WWALKFUNC walkfunc);
extern BOOL wwalkfunc(HWND hwnd, char *title, DWORD vm, VM_CB far *vm_cb);
#endif

DWORD GetSysVMHandle(void);
DWORD GetCurVMHandle(void);
DWORD GetNextVMHandle(DWORD vm);

void set_vm(DWORD vm);  // sets base address for map_linear
void far *map_real_vm(DWORD vm, void far *fp, WORD bytes);
void far *map_linear_vm(DWORD vm, DWORD lin_addr, WORD bytes);
