/*
W16LOCK.C
Win32 program that grabs Win16Lock (Win16Mutex) in Chicago

from Andrew Schulman, Unauthorized Windows 95, p. 553

Copyright (c) Andrew Schulman 1994.
But may be freely distributed, so long as you keep this header intact.

Unauthorized Windows: Developers Resource Kit by Andrew Schulman
Price: $39.99 USA/$54.99 Canada
ISBN:  1-56884-305-4; Publication Date: November 1994; Pages: 608

IDG Books are available at all major bookstore chains or direct from
the publisher at (800) 762-2974, 8 a.m. to 5 p.m. Central.  
*/

#include <stdlib.h>
#define  WIN32_LEAN_AND_MEAN
#include "windows.h"

#define MSG(s)           MessageBox(0, s, "W16LOCK", MB_OK)

void fail(const char *s) { MSG(s); exit(1); }

void (WINAPI *GetpWin16Lock)(DWORD *pWin16Lock);
int (WINAPI *_ConfirmWin16Lock)(void);  // returns current lock count
void (WINAPI *_EnterSysLevel)(DWORD lock);
void (WINAPI *_LeaveSysLevel)(DWORD lock);

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
    LPSTR lpszCmdLine, int nCmdShow)
{
    // Microsoft C code per MSJ, May 1991, pp. 135-6
    #define argc __argc
    #define argv __argv
    extern int __argc;
    extern char **__argv;
    
    DWORD Win16Lock = 0;
    int iter = (argc < 2) ? 10000 : atoi(argv[1]);
    int i;
    
    #define GET(mod, func) { \
        if (! (func = GetProcAddress(GetModuleHandle(mod), #func))) \
            fail("Can't link to " mod "!" #func); \
        }
        
    GET("KERNEL32", GetpWin16Lock);
    GET("KERNEL32", _EnterSysLevel);
    GET("KERNEL32", _LeaveSysLevel);
    GET("KERNEL32", _ConfirmWin16Lock);

    GetpWin16Lock(&Win16Lock);
    if (Win16Lock == 0)
        fail("GetpWin16Lock didn't work");
    
    _EnterSysLevel(Win16Lock);
    for (i=0; i<iter; i++)
        (void) GetVersion();    // or any call that doesn't use Win16
    _LeaveSysLevel(Win16Lock);
    return 0;
}
