/* 
W32DUMP.C -- portable executable (PE) dumper
Andrew Schulman, May 1994
Copyright (c) Andrew Schulman 1993-94.  All rights reserved.
Copyright (c) V Communications 1993-94.  All rights reserved.
*/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <dos.h>
#include "filelib.h"

#define UNALIGNED   /**/    // ???
typedef char CHAR;
typedef unsigned char UCHAR;
typedef UCHAR far *PUCHAR;      // far *??
typedef unsigned short WCHAR;   // ??
typedef short SHORT;
typedef unsigned short USHORT;
typedef USHORT far *PUSHORT;    // ??
typedef long LONG;
typedef unsigned long ULONG;
typedef ULONG far *PULONG;      // ??

typedef int BOOL;

#define FALSE		0
#define TRUE		(! FALSE)

#define FIELD_OFFSET(x,y)   offsetof((x),(y))

#define min(a,b)    ((a) < (b) ? (a) : (b))

#pragma pack(1)
#include "ntimage.h"    // from Win32 SDK

/* lifted from MSoft newexe.h */
#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC     0x454C          /* linear executable */
#define LXMAGIC     0x584C          /* IBM OS/2 2.0 linear executable */
#define PEMAGIC     0x4550          /* NT portable executable */
#define PLMAGIC		0x4C50			/* hacked NT port exe for TNT! */
#define W3MAGIC     0x3357          /* WIN386.EXE signature */

void fail(char *s) { fputs("w32dump fail: ", stdout); puts(s); exit(1); }

void fail_perror(void)  { perror("w32dump fail"); exit(1); }

void usage()
{
    puts("W32DUMP displays information about a Win32");
    puts("portable executable (PE) file.");
    puts("");
    puts("To display information about a Win32 PE file:");
    puts("         W32DUMP file");
    puts("example: w32dump \\win32\\bin\\ntdll.dll");
	puts("");
	puts("To disassemble a Win32 file, run W32P and Sourcer");
    exit(1);
}
    
void pe(char *name);
void srdef(char *name);
void pe2(ULONG start);
void srdef2(ULONG start);
void do_sectiontab(void);
void do_import_dir(IMAGE_SECTION_HEADER *sect_hdr);
void do_export_dir(IMAGE_SECTION_HEADER *sect_hdr);
void do_reloc_dir(IMAGE_SECTION_HEADER *sect_hdr);
void do_debug_dir(IMAGE_SECTION_HEADER *sect_hdr);

#define PUBLIC /**/

static ULONG align;
PUBLIC int f;
PUBLIC ULONG start_newexe;
static char *file_name;
static BOOL saw_reloc = FALSE;
static BOOL no_reloc = FALSE;

static IMAGE_NT_HEADERS hdr;
static ULONG hdr_size = 0;
static ULONG import_dir = 0;
static ULONG export_dir = 0;
static ULONG reloc_dir = 0;
static ULONG debug_dir = 0;
static IMAGE_SECTION_HEADER *section_tab;
static USHORT num_sect = 0;
static ULONG image_base = 0;
static ULONG code_virt = 0;

#define SECTION_FILEOFS(x)      (section_tab[x].PointerToRawData)
#define SECTION_SIZE(x)         (section_tab[x].SizeOfRawData)
#define SECTION_VIRTADDR(x)     (section_tab[x].VirtualAddress)
#define SECTION_END(x)          (SECTION_VIRTADDR(x) + SECTION_SIZE(x))
#define SECTION_NAME(x)         (section_tab[x].Name)

main(int argc, char *argv[])
{
    char buf[128];
    int i;

    fputs("Windows Source Win32 Portable Executable Dumper v. 1.00\n", stderr);
    fputs("Copyright (c) 1993-4 V Communications Inc.  All rights reserved.\n",
        stderr);
    fputs("Copyright (c) 1993-4 Andrew Schulman.  All rights reserved.\n\n", 
        stderr);
    
    if (argc < 2)
        usage();
    
    for (i=1; i<argc; i++)
        if ((argv[i][0] == '-') || (argv[i][0] == '/'))
        {
            if (argv[i][1] == '?')
                usage();
            else if (strcmp(&argv[i][1], "srdef") == 0)
            {
                strcpy(buf, argv[i+1]);
                if (! strchr(buf, '.'))
                    strcat(buf, ".exe");
                srdef(buf);
                i++;
            }
			else if (strcmp(&argv[i][1], "noreloc") == 0)
			{
				no_reloc++;
			}
        }
        else
        {
            strcpy(buf, argv[i]);
            if (! strchr(buf, '.'))
                strcat(buf, ".exe");
            pe(buf);
        }
    return 0;
}

char *not_a_pe = "not a PE (portable executable) file" ;

void pe(char *name)
{
    ULONG pos;
    USHORT w;
    unsigned i;
	
	file_name = name;
    
    if ((f = sopen(name, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
        fail_perror();
    exe_read(f, &w, 0, 2);  // read first two bytes of file
    if (w != EMAGIC)
        fail("not an executable");
    exe_read(f, &w, 0x18, 2);
	printf("w == %04X\n", w);
#if 0	
    if (w < 0x40)
        fail(not_a_pe);
#endif	
    exe_read(f, &start_newexe, ENEWHDR, 4);
	printf("start_newexe = %08lX\n", start_newexe);
    exe_read(f, &w, start_newexe, 2);
	printf("w = %04X\n", w);
    if ((w != PEMAGIC) && (w != PLMAGIC))	// accept hacked PE too
        fail(not_a_pe);
    pe2(start_newexe);
    close(f);
}

void srdef(char *name)
{
    ULONG start_newexe;
    ULONG pos;
    USHORT w;
    unsigned i;
    
    if ((f = open(name, O_RDONLY | O_BINARY)) == -1)
        fail("can't find file");
    read(f, &w, 2);
    if (w != EMAGIC)
        fail("not an executable");
    exe_read(f, &w, 0x18, 2);
    if (w < 0x40)
        fail(not_a_pe);
    exe_read(f, &start_newexe, ENEWHDR, 4);
    exe_read(f, &w, start_newexe, 2);
    if (w != PEMAGIC)
        fail(not_a_pe);
    srdef2(start_newexe);
    close(f);
}

#define GET_DIR(x)  (hdr.OptionalHeader.DataDirectory[x].VirtualAddress)
    
void print_dir(char *s, int dir)
{
    printf("%s: %lx\n", s, GET_DIR(dir));
}

char *get_machine(USHORT x)
{
	switch (x)
	{
		case IMAGE_FILE_MACHINE_UNKNOWN:	return "Unknown";
		case IMAGE_FILE_MACHINE_I860:		return "Intel 860";
		case IMAGE_FILE_MACHINE_I386:		return "Intel 386";
		case IMAGE_FILE_MACHINE_R3000:		return "MIPS R3000";
		case IMAGE_FILE_MACHINE_R4000:		return "MIPS R4000";
		default:							return "Unknown??";
	}
}

char *get_subsystem(int x)
{
	switch (x)
	{
		case IMAGE_SUBSYSTEM_UNKNOWN:		return "Unknown";
		case IMAGE_SUBSYSTEM_NATIVE:		return "NT Native";
		case IMAGE_SUBSYSTEM_WINDOWS_GUI:	return "Windows";
		case IMAGE_SUBSYSTEM_WINDOWS_CUI:	return "Console";
		case IMAGE_SUBSYSTEM_OS2_CUI:		return "OS/2";
		case IMAGE_SUBSYSTEM_POSIX_CUI:		return "Posix";
		default:							return "Unknown?";
	}
}

void pe2(ULONG start)
{
    exe_read(f, &hdr, start, sizeof(hdr));
	
	printf("file %s\n", file_name);
	printf("subsystem %s\n", get_subsystem(hdr.OptionalHeader.Subsystem));
	printf("machine %s\n", get_machine(hdr.FileHeader.Machine));
	if (hdr.FileHeader.Characteristics & IMAGE_FILE_DLL)
		printf("dll\n");
	if (hdr.FileHeader.Characteristics & IMAGE_FILE_SYSTEM)
		printf("system file\n");
	image_base = hdr.OptionalHeader.ImageBase;
	printf("image_base %lx\n", image_base);
	printf("start %lx\n", hdr.OptionalHeader.AddressOfEntryPoint);
    printf("\n");
	
#if 0	
    printf("PE at %lx\n", start_newexe);
    printf("Machine:  %x\n", hdr.FileHeader.Machine);
    printf("Sections: %x\n", hdr.FileHeader.NumberOfSections);
    printf("SymTab:   %lx\n", hdr.FileHeader.PointerToSymbolTable);
    printf("Num Sym:  %lx\n", hdr.FileHeader.NumberOfSymbols);
    printf("Opt Hdr:  %x\n",  hdr.FileHeader.SizeOfOptionalHeader);
    printf("Charac:   %x\n",  hdr.FileHeader.Characteristics);
    printf("Image Base: %lx\n", hdr.OptionalHeader.ImageBase);
    printf("Sect Align: %lx\n", hdr.OptionalHeader.SectionAlignment);
    printf("File Align: %lx\n", hdr.OptionalHeader.FileAlignment);
    printf("Image Size: %lx\n", hdr.OptionalHeader.SizeOfImage);
    printf("Hdr Size: %lx\n", hdr.OptionalHeader.SizeOfHeaders);
    print_dir("Export Directory", IMAGE_DIRECTORY_ENTRY_EXPORT); 
    print_dir("Import Directory", IMAGE_DIRECTORY_ENTRY_IMPORT); 
    print_dir("Resource Directory", IMAGE_DIRECTORY_ENTRY_RESOURCE);
    print_dir("Exception Directory", IMAGE_DIRECTORY_ENTRY_EXCEPTION);
    print_dir("Security Directory", IMAGE_DIRECTORY_ENTRY_SECURITY); 
    print_dir("Base Relocation Table", IMAGE_DIRECTORY_ENTRY_BASERELOC); 
    print_dir("Debug Directory", IMAGE_DIRECTORY_ENTRY_DEBUG);  
    print_dir("Description String", IMAGE_DIRECTORY_ENTRY_COPYRIGHT); 
    print_dir("Machine Value (MIPS GP)", IMAGE_DIRECTORY_ENTRY_GLOBALPTR);
    print_dir("Thread Local Storage", IMAGE_DIRECTORY_ENTRY_THREAD_SPACE); 
    print_dir("Callbacks", IMAGE_DIRECTORY_ENTRY_CALLBACKS); 
#endif	

    hdr_size = hdr.OptionalHeader.SizeOfHeaders;
    export_dir = GET_DIR(IMAGE_DIRECTORY_ENTRY_EXPORT);
    import_dir = GET_DIR(IMAGE_DIRECTORY_ENTRY_IMPORT);
    reloc_dir = GET_DIR(IMAGE_DIRECTORY_ENTRY_BASERELOC);
    debug_dir = GET_DIR(IMAGE_DIRECTORY_ENTRY_DEBUG);  
    
    do_sectiontab();
}

void do_sectiontab(void)
{
	char name[9];
    ULONG section_tab_ofs;
    USHORT i;
    
    section_tab_ofs = 
        start_newexe +
        sizeof(ULONG) +
        sizeof(IMAGE_FILE_HEADER) +
        hdr.FileHeader.SizeOfOptionalHeader;
    // printf("Section table: %lx\n", section_tab_ofs);
    num_sect = hdr.FileHeader.NumberOfSections;
    section_tab = (IMAGE_SECTION_HEADER *)
        allocate(num_sect, sizeof(IMAGE_SECTION_HEADER));
    exe_read(f, section_tab, section_tab_ofs, 
        num_sect * sizeof(IMAGE_SECTION_HEADER));
    printf("begin section\n");
    for (i=0; i<num_sect; i++)
    {
		strncpy(name, SECTION_NAME(i), 8);
		name[8] = 0;
        printf("   %08lx\t%08lx\t%08lx\t%s\n", 
            SECTION_FILEOFS(i),
            SECTION_SIZE(i),
            SECTION_VIRTADDR(i),
			name);
		if (strcmp(name, ".text") == 0)
			code_virt = SECTION_VIRTADDR(i);
    }
    printf("\n");
    for (i=0; i<num_sect; i++)
    {
        if (SECTION_VIRTADDR(i) == import_dir)
            do_import_dir(&section_tab[i]);
        else if (SECTION_VIRTADDR(i) == export_dir)
            do_export_dir(&section_tab[i]);
        else if (SECTION_VIRTADDR(i) == reloc_dir)
		{
			if (! no_reloc)
				do_reloc_dir(&section_tab[i]);
		}
		else if (SECTION_VIRTADDR(i) == debug_dir)
			do_debug_dir(&section_tab[i]);
    }
    // don't free section_tab: used throughout program
		
	if (! saw_reloc)
		if (hdr.FileHeader.Characteristics & IMAGE_FILE_RELOCS_STRIPPED)
			printf("reloc table stripped\n");
}

// get section number to which virtual address belongs
USHORT get_section(ULONG virtaddr)
{
    int i;
    for (i=0; i<num_sect; i++)
        if (virtaddr >= SECTION_VIRTADDR(i))
            if (SECTION_END(i) > virtaddr)
                return i;
    /* still here */
    return -1;
}
        
#if 0
char            64 00 07 00
timedate        00 00 00 00
maj, min        00 00 00 00
name!           CA 02 07 00 = 0702CA - 070000 = 02CA (KERNEL32.dll)
FirstThunk      90 01 07 00 = 070190 - 070000 = 0190

0190            BC 02 07 00 = 0702BC - 070000 = 02BC -> 26 00 'CloseHandle'

00010000 | 64 00 07 00 00 00 00 00 00 00 00 00 CA 02 07 00 | d...............
00010010 | 90 01 07 00 E8 00 07 00 00 00 00 00 00 00 00 00 | ................
...
00010190 | BC 02 07 00 D8 02 07 00 E8 02 07 00 FE 02 07 00 | ................
000101a0 | 0E 03 07 00 22 03 07 00 30 03 07 00 3E 03 07 00 | ...."...0...>...

// array of IMAGE_IMPORT_DESCRIPTORS
// PIMAGE_THUNK_DATA FirstThunk field is virtual address,
// so subtract virtual address for directory!
// what is size?  number of entries?    
#endif

typedef struct _MY_IMAGE_IMPORT_DESCRIPTOR {
    ULONG   Characteristics;
    ULONG   TimeDateStamp;
    USHORT  MajorVersion;
    USHORT  MinorVersion;
    ULONG   Name;
    ULONG   FirstThunk;     //    PIMAGE_THUNK_DATA FirstThunk;
} MY_IMAGE_IMPORT_DESCRIPTOR;

// Have to rewrite do_import_dir:  see Matt P., MSJ, March 1994

void do_import_dir(IMAGE_SECTION_HEADER *sect_hdr)
{
    UCHAR *impdata = (UCHAR *) allocate(1, sect_hdr->SizeOfRawData);
    MY_IMAGE_IMPORT_DESCRIPTOR *imp = (MY_IMAGE_IMPORT_DESCRIPTOR *) impdata;
    ULONG virtaddr = sect_hdr->VirtualAddress;
    ULONG *pthunk;
    int i, j;
    
    exe_read(f, impdata, sect_hdr->PointerToRawData, sect_hdr->SizeOfRawData);
    printf("begin import\n");
    for (i=0; ; i++)    // don't know size of table
    {                   // ends with 0 entry
        if (! (imp[i].Name && imp[i].FirstThunk))
            break;
		printf("   %s   %lx\n", 
			&impdata[imp[i].Name - virtaddr],
			imp[i].FirstThunk - virtaddr);
        pthunk = (ULONG *) &impdata[imp[i].FirstThunk - virtaddr];
		j = 0;
use_characteristics:		
		if ((pthunk[j] - virtaddr) >= 0x0000FFFFUL)
		{
			if (imp[i].Characteristics == 0)
				goto done;	// Borland not have Characteristics?
			pthunk = (ULONG *) &impdata[imp[i].Characteristics - virtaddr];
			for (; ; j++)    // don't know size of table; ends with 0 thunk
			{
				if (! pthunk[j])
					break;

				// print so can find relocations to import directory
				printf("      %08lx", 
					imp[i].FirstThunk + (j * sizeof(ULONG)));
				
				if (pthunk[j] & IMAGE_ORDINAL_FLAG)
					printf("   ord %u\n", (USHORT) IMAGE_ORDINAL(pthunk[j]));
				else if (pthunk[j] - virtaddr < 0xFFFFU)
				{
					IMAGE_IMPORT_BY_NAME *impname = (IMAGE_IMPORT_BY_NAME *)
						&impdata[pthunk[j] - virtaddr];
					printf("   %5u   %s\n", impname->Hint, impname->Name);
				}
				else	// should never happen in here
				{
					printf("   %08lx ??? \n", pthunk[j] - virtaddr);
				}
			}
		}
		else
		{
			for (j=0; ; j++)    // don't know size of table; ends with 0 thunk
			{
				if (! pthunk[j])
					break;

				// print so can find relocations to import directory
				printf("      %08lx", 
					imp[i].FirstThunk + (j * sizeof(ULONG)));

				if (pthunk[j] & IMAGE_ORDINAL_FLAG)
					printf("   ord %u\n", (USHORT) IMAGE_ORDINAL(pthunk[j]));
				else if (pthunk[j] - virtaddr < 0xFFFFU)
				{
					IMAGE_IMPORT_BY_NAME *impname = (IMAGE_IMPORT_BY_NAME *)
						&impdata[pthunk[j] - virtaddr];
					printf("   %5u   %s\n", impname->Hint, impname->Name);
				}
				else
				{
					printf("\nOops, should have used Characteristics, "
						   "not FirstThunk\n\n");
					goto use_characteristics;
				}
			}
        }
    }
done:
    free(impdata);
    printf("\n");
}

typedef struct _MY_IMAGE_EXPORT_DIRECTORY {
    ULONG   Characteristics;
    ULONG   TimeDateStamp;
    USHORT  MajorVersion;
    USHORT  MinorVersion;
    ULONG   Name;
    ULONG   Base;
    ULONG   NumberOfFunctions;
    ULONG   NumberOfNames;
    ULONG   AddressOfFunctions;
    ULONG   AddressOfNames;
    ULONG   AddressOfNameOrdinals;  // array of SHORTS
} MY_IMAGE_EXPORT_DIRECTORY;

#define CLIVE

#ifdef CLIVE
/*
#: 3840 S0/CompuServe Mail  [MAIL]
    05-Feb-95 11:09 PST
Sb: W32DUMP, Export Problem
Fm: Clive Turvey [74011,1732]


Hi Andrew,

Looked at WSOCK32.DLL & KERNEL32.DLL, and the problem is with the
export processing. With KERNEL32.DLL all the exports are named and
defined in ASCII order, but WSOCK32.DLL they are really mixed up.

When displaying addresses and names you have to double index, using
the ordinal number to reference into the address table.

I mark the entries printed by name, so that they don't occur as an
explicit ordinal by number. If you are generating a file like W32.DAT
that translates ordinal numbers to export names beware of the double
indexing explained above.

An export address of 0 seems like a bad thing, I'll look to see if
there is any use for it.  
*/

/*
	This disagrees with output of DUMPBIN /EXPORTS, but Clive is
	right!  (Tested with HELLO.C and HELLO.DEF, with deliberately
	out-of-order exports.)
*/

void do_export_dir(IMAGE_SECTION_HEADER *sect_hdr)
{
  MY_IMAGE_EXPORT_DIRECTORY *exp;
  UCHAR *expdata = (UCHAR *) allocate(sect_hdr->SizeOfRawData, 1);
  ULONG base = sect_hdr->PointerToRawData;
  ULONG size = sect_hdr->SizeOfRawData;
  ULONG virtaddr = sect_hdr->VirtualAddress;
  USHORT baseord, *ordtab;
  ULONG *functab, *nametab;
  int i, j;

  exe_read(f, expdata, base, size);
  exp = (MY_IMAGE_EXPORT_DIRECTORY *) expdata;
  functab = (ULONG *) &expdata[exp->AddressOfFunctions - virtaddr];
  nametab = (ULONG *) &expdata[exp->AddressOfNames - virtaddr];
  ordtab = (USHORT *) &expdata[exp->AddressOfNameOrdinals - virtaddr];
  baseord = (USHORT) exp->Base;
  
  printf("begin export\n");

  if (baseord != 1) // Is this a problem?
    printf("; Warning: Ordinal base is not 1!\n");

  // The name table & ordinal table are only NumberOfNames in size.
  // Also the values in the ordinal table can be in a random order
  // anywhere in the domain of NumberOfFunctions. Actually the
  // order is determined by the ASCII sorting of the the name table.

  j = 0; // Export #, is this important?
  for (i=0; i<exp->NumberOfNames; i++)
  {
    printf("  %5u   %5u   %08lx   %s\n",
      j++,
      baseord + ordtab[i],
      functab[ordtab[i]], // This is an index to an index
      &expdata[(nametab[i] - virtaddr)]);

    // Mark Ordinal already displayed explictly by name
    functab[ordtab[i]] = 0xFFFFFFFF; // 4GB-1, very low chance of this
  }

  for (i=0; i<exp->NumberOfFunctions; i++)
  {
    if (functab[i] == 0xFFFFFFFF) // This one had a name
      continue;

    // functab[i] == 0, Sucks too but let ignore that for now

	  if (functab[i] != 0)		// don't show 0 entries: AS
    printf("  %5u   %5u   %08lx\n",
      j++,
      baseord + i, // No Ordinal Table for this
      functab[i]);
  }

  free(expdata);
  printf("\n");
}
#else
void do_export_dir(IMAGE_SECTION_HEADER *sect_hdr)
{
    MY_IMAGE_EXPORT_DIRECTORY *exp;
    UCHAR *expdata = (UCHAR *) allocate(sect_hdr->SizeOfRawData, 1);
    ULONG base = sect_hdr->PointerToRawData;
    ULONG size = sect_hdr->SizeOfRawData;
    ULONG virtaddr = sect_hdr->VirtualAddress;
	USHORT baseord;
    ULONG *functab;
    ULONG *nametab;
    USHORT *ordtab;
    int i;
    
    exe_read(f, expdata, base, size);
    exp = (MY_IMAGE_EXPORT_DIRECTORY *) expdata;
    functab = (ULONG *) &expdata[exp->AddressOfFunctions - virtaddr];
    nametab = (ULONG *) &expdata[exp->AddressOfNames - virtaddr];
    ordtab = (USHORT *) &expdata[exp->AddressOfNameOrdinals - virtaddr];
	baseord = (USHORT) exp->Base;
    
    printf("begin export\n");
	if (baseord != 1)
		printf("; Warning: Ordinal base is not 1!\n");
	
	// is this right???
	for (i=0; i<exp->NumberOfNames; i++)
	{
		printf("  %5u   %5u   %08lx   %s\n",
			i,
			baseord + ordtab[i],
			functab[i],
			&expdata[(nametab[i] - virtaddr)]);
	}

#if 1
	for ( ; i<exp->NumberOfFunctions; i++)
	{
		printf("  %5u   %5u   %08lx\n",
			i,
			baseord + ordtab[i],
			functab[i]);
	}
#else
    for (i=0; i<exp->NumberOfFunctions; i++)
	{
		if (! functab[i])
		{
			printf("   ; Warning: Invalid function %u of %u\n",
				i,
				exp->NumberOfFunctions);
			// break;
		}
		else if ((i < exp->NumberOfNames) && nametab[i])
			printf("   %08lx   %5u   %s\n",
				functab[i],     // address of code -- print as sect:off
				ordtab[i],      // ordinal #
				&expdata[(nametab[i] - virtaddr)]);     // ASCIIZ name
		else
			printf("   %08lx   %5u   (%u)\n",
				functab[i], ordtab[i], i);
	}
#endif

    free(expdata);
    printf("\n");
}
#endif

char *reloc_type(int x)
{
    switch (x)
    {
        case IMAGE_REL_BASED_ABSOLUTE:      return "ABSOLUTE";
        case IMAGE_REL_BASED_HIGH:          return "HIGH";
        case IMAGE_REL_BASED_LOW:           return "LOW";
        case IMAGE_REL_BASED_HIGHLOW:       return "HL";
        case IMAGE_REL_BASED_HIGHADJ:       return "HIGHADJ";
        case IMAGE_REL_BASED_MIPS_JMPADDR:  return "MIPS_JMPADDR";
        case IMAGE_REL_BASED_I860_BRADDR:   return "I860_BRADDR";
        case IMAGE_REL_BASED_I860_SPLIT:    return "I860_SPLIT";
        default:                            return "UNKNOWN??";
    }
}

#define CLIVE

#ifdef CLIVE
/*
#: 3824 S0/CompuServe Mail  [MAIL]
    03-Feb-95 17:02 PST
Sb: W32DUMP
Fm: Clive Turvey [74011,1732]

i)  basevirt / baserel->VirtualAddress can be zero and valid. It's an
    offset into the file which the 12bit relocation address is added.
    You could check that basevirt + relofs[] are within a section, by
    putting the sect = get_section() code in the relocation printing
    loop, but that's probably overkill. You should be able to get rid
    of the code_virt kludge. If baserel->SizeOfBlock is zero then you
    are definately in dead space.

ii) baserel->SizeOfBlock must be a multiple of 4, ie end on a DWORD
    boundary. The relocation list can be padded with one zero WORD to
    ensure this, and this must be special cased. The relocations seem
    to be in numerical order, although that might be arbitary, so a zero
    WORD in any entry except the first would be odd. If there were no
    relocations the linker would probably pitch a relocation block for
    that 4k page rather than generate a null relocation block.

iii)Relocations that fall on the 4K boundary, like a 32bit fixup at
    0xFFD-0xFFF. To this end I read additional 16 bytes of slop, 4
    would probably do but I want to find an example first. The 16 bytes
    are only used on sections that exceed 4K, I also fixed the min() to
    account for the delta. The fix should be OK, but I've noticed 'LX'
    files back reference on boundary conditions like this, they do this
    probably because the code that caused the page fault which loads the
    next page did an indivisable read or executed across the page
    boundary and thus the preceding page IS in memory and the rerunning
    of the faulting instruction will clear the prefetch queue allowing
    any fixed up data to be utilized. This means I'm not 100% confident
    that my solution to this problem is valid, there's bound to be an
    example in the Win95 Beta 2, so I'll go look. I've pointed this out
    mainly to plant a seed in your mind that this might be a problem.
*/

void do_reloc_dir(IMAGE_SECTION_HEADER *sect_hdr)
{
  char name[9];
  UCHAR *reldata = (UCHAR *) allocate(sect_hdr->SizeOfRawData, 1);
  UCHAR *reldata2 = reldata;
  IMAGE_BASE_RELOCATION *baserel;
  USHORT *relofs;
  ULONG base = sect_hdr->PointerToRawData;
  ULONG size = sect_hdr->SizeOfRawData;
  ULONG virtaddr = sect_hdr->VirtualAddress;
  ULONG basevirt, delta;
  UCHAR *segdata;
  USHORT num_reloc, sect, i;

  saw_reloc = TRUE;
  baserel = (IMAGE_BASE_RELOCATION *) reldata;
  exe_read(f, reldata, base, size);
  printf("begin reloc\n");

  // printf("; %08lX %08lX %08lX\n", base, size, virtaddr);

  for (;;)
  {
    basevirt = baserel->VirtualAddress;

#if 0
    // is this right?????? (NO it isn't! CTurvey)
    if (basevirt == 0)
      basevirt = code_virt; // You can pitch this kludge. CTurvey

    if (! (basevirt && baserel->SizeOfBlock))
      break;  // don't know table size; ends with 0 entry
#else
    if (! (basevirt || baserel->SizeOfBlock)) // Both Base & Size 0. CTurvey
      break;  // don't know table size; ends with 0 entry
#endif

    num_reloc = (baserel->SizeOfBlock - sizeof(IMAGE_BASE_RELOCATION)) /
		sizeof(USHORT);
    relofs = (USHORT *) (reldata2 + sizeof(IMAGE_BASE_RELOCATION));

    if (basevirt == 0) // I'll special case this so change localized. CTurvey
      sect = 0;        // should really find closest.
    else if ((sect = get_section(basevirt)) == -1)
        fail("Invalid section!");

    strncpy(name, SECTION_NAME(sect), 8);
    name[8] = 0;
    printf("   %08lx   %08lx   %u   %s\n",
      basevirt,
      baserel->SizeOfBlock,
      num_reloc,
      name);

    // 4k page? (What about boundary relocs at ffd thru fff? CTurvey)
    // Haven't found an example yet, but Sod's Law suggests that i'll
    // find an example if I look! So lets get 16bytes of slop. CTurvey

    segdata = (UCHAR *) allocate(1, ((4 * 1024)+16));

    delta = basevirt - SECTION_VIRTADDR(sect);

    exe_read(f, segdata, SECTION_FILEOFS(sect) + delta,
      min((SECTION_SIZE(sect) - delta), ((4 * 1024)+16))); // 16byte of slop. 
														   // CTurvey

    for (i=0; i<num_reloc; i++)
    {
      // Special Case, relocs must end on DWORD boundary, possibly one
      // WORD of 0 as filler. CTurvey

      if ((i != 0) && (relofs[i] == 0))
        break;

      printf("     %04x   %08lx   %s\n",
        relofs[i] & 0xfff,  // top 4 bits used for reloc type
        (*((ULONG *) &segdata[relofs[i] & 0xfff])) - image_base,
        reloc_type(relofs[i] >> 12));
    }

    reldata2 += baserel->SizeOfBlock;
    baserel = (IMAGE_BASE_RELOCATION *) reldata2;
    free(segdata);
  }

  free(reldata);
  printf("\n");
}
#else
void do_reloc_dir(IMAGE_SECTION_HEADER *sect_hdr)
{
	char name[9];
    UCHAR *reldata = (UCHAR *) allocate(sect_hdr->SizeOfRawData, 1);
    UCHAR *reldata2 = reldata;
    IMAGE_BASE_RELOCATION *baserel;
    USHORT *relofs;
    ULONG base = sect_hdr->PointerToRawData;
    ULONG size = sect_hdr->SizeOfRawData;
    ULONG virtaddr = sect_hdr->VirtualAddress;
	ULONG basevirt, delta;
    UCHAR *segdata;
    USHORT num_reloc, sect, i;
	
	saw_reloc = TRUE;
	baserel = (IMAGE_BASE_RELOCATION *) reldata;
	exe_read(f, reldata, base, size);
    printf("begin reloc\n");
	// printf("; %08lX %08lX %08lX\n", base, size, virtaddr);
    for (;;)
    {
		basevirt = baserel->VirtualAddress;
#if 1	
		// is this right??????
		if (basevirt == 0) 
			basevirt = code_virt;
#endif		
        if (! (basevirt && baserel->SizeOfBlock))
            break;  // don't know table size; ends with 0 entry
        num_reloc = (baserel->SizeOfBlock - sizeof(IMAGE_BASE_RELOCATION)) / 
            sizeof(USHORT);
        relofs = (USHORT *) (reldata2 + sizeof(IMAGE_BASE_RELOCATION));
        
        if ((sect = get_section(basevirt)) == -1)
            fail("Invalid section!");
                
		strncpy(name, SECTION_NAME(sect), 8);
		name[8] = 0;
        printf("   %08lx   %08lx   %u   %s\n", 
			basevirt,
            baserel->SizeOfBlock,
            num_reloc,
			name);
        
        segdata = (UCHAR *) allocate(1, 4 * 1024); // 4k page?
            
        delta = basevirt - SECTION_VIRTADDR(sect);
    
        exe_read(f, segdata, SECTION_FILEOFS(sect) + delta,
            min(SECTION_SIZE(sect), 4 * 1024));
        
        for (i=0; i<num_reloc; i++)
        {
            printf("     %04x   %08lx   %s\n", 
                relofs[i] & 0xfff,  // top 4 bits used for reloc type
                (*((ULONG *) &segdata[relofs[i] & 0xfff])) - image_base,
                reloc_type(relofs[i] >> 12));
        }
        reldata2 += baserel->SizeOfBlock;
        baserel = (IMAGE_BASE_RELOCATION *) reldata2;
        free(segdata);
    }
    free(reldata);
    printf("\n");
}
#endif

void do_debug_dir(IMAGE_SECTION_HEADER *sect_hdr)
{
    ULONG base = sect_hdr->PointerToRawData;
    ULONG size = sect_hdr->SizeOfRawData;
    ULONG virtaddr = sect_hdr->VirtualAddress;
	printf("Debug directory: %08lX %08lX %08lX\n\n",
		base, size, virtaddr);
}

void srdef2(ULONG start)
{
    fail("To generate a Sourcer .DEF file, run W32P");
}
