/* 
W386MAP.C -- WIN386/DOS386 mapper
Andrew Schulman

Version for book; does not have option to generate Sourcer definition file
Last version of W3MAP was 2.01; make this 2.01a

note: embedded LE have no nonresident names table, so no description
string or DDB name
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <dos.h>
#include "filelib.h"

/* lifted from MSoft newexe.h */
#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC     0x454C          /* linear executable */
#define LXMAGIC     0x584C          /* IBM OS/2 2.0 linear executable */
#define PEMAGIC     0x4550          /* NT portable executable */
#define W3MAGIC     0x3357          /* WIN386.EXE signature */
#define SZMAGIC		0x5A53			/* SZ -- compressed file */

#pragma pack(1)

typedef struct {
    unsigned char sign[8];
    unsigned long offset;
    unsigned char pad[4];
    } W3REC;

void fail(char *s) { fputs("W386MAP fail: ", stdout); puts(s); exit(1); }

void fail_perror(void)	{ perror("W386MAP fail"); exit(1); }

void usage()
{
	puts("W386MAP displays information about the Virtual Device Drivers");
	puts("(VxDs) in a Windows 3+ Enhanced mode WIN386 or DOS386 (W3) file.");
	puts("");
	puts("To display a list of all VxDs in the WIN386 or DOS386 file:");
	puts("         W386MAP win386_file");
	puts("example: W386MAP \\windows\\system\\win386.exe");
	puts("");
	puts("To display details of an individual VxD in the WIN386 or DOS386 file:");
	puts("         W386MAP -VXD vxd_name win386_file");
	puts("example: W386MAP -vxd dosmgr \\windows\\system\\win386.exe");
	puts("");
	puts("To display details of each VxD in the WIN386 or DOS386 file:");
	puts("         W386MAP -VERBOSE win386_file");
	puts("example: W386MAP -verbose \\windows\\system\\win386.exe");
    exit(1);
}
    
void le2(unsigned long start_le);
void le2_srdef(unsigned long start_le);
void w3(char *name);
void srdef(char *name, char *w3name);

unsigned long start_newexe = 0;

static unsigned long align;
static int f;
static int paranoid = 0;
static int verbose = 0;
static char *one_vxd = 0;

main(int argc, char *argv[])
{
	char *arg;
    int i;
    
    fputs("WIN386 (W3) Dumper version 2.01a\n", stderr);
    fputs("Copyright (c) 1992-93 Andrew Schulman.  All rights reserved.\n\n", 
        stderr);
    
    if (argc < 2)
        usage();
    
    for (i=1; i<argc; i++)
        if ((argv[i][0] == '-') || (argv[i][0] == '/'))
        {
            if (argv[i][1] == '?')
                usage();
			
			arg = strupr(&argv[i][1]);
			
            if (strncmp(arg, "VXD", 3) == 0)
            {
				char buf[128];
				one_vxd = strupr(argv[i+1]);
				if (strcmp(one_vxd, "VMM") == 0)
					one_vxd = "WIN386";
                i ++;
            }
            else if (strcmp(arg, "PARANOID") == 0)  // undoc
                paranoid = 1;
            else if (strncmp(arg, "VERBOSE", 4) == 0)
                verbose = 1;
            else
                usage();
        }
        else
		{
			char buf[128];
			strcpy(buf, argv[i]);
			if (! strchr(buf, '.'))
				strcat(buf, ".exe");
            w3(buf);
		}
    return 0;
}

void w3(char *name)
{
    W3REC rec;
    unsigned char le[2];
    unsigned long pos;
    unsigned short w;
    unsigned i;
    
	if ((f = sopen(name, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
		fail_perror();
	exe_read(f, &w, 0, 2);	// read first two bytes of file
    if (w != EMAGIC)
	{
		if (w == SZMAGIC)
			fail("compressed file (SZ) -- use EXPAND.EXE");
		else
			fail("not an executable");
	}
	exe_read(f, &w, 0x18, 2);
	if (w < 0x40)
		fail("not a W3 (WIN386) executable");
	exe_read(f, &start_newexe, ENEWHDR, 4);
	exe_read(f, &w, start_newexe, 2);
	if (w == LEMAGIC)
		fail("use LEDUMP to examine this VxD");
    if (w != W3MAGIC)
        fail("not a W3 (WIN386) executable");
	if (! one_vxd)
	{
		printf("W3      \t%08lx\n", start_newexe);
		if (verbose)
			putchar('\n');
	}
    pos = start_newexe + sizeof(W3REC);
    for (;;)
    {
        exe_read(f, &rec, pos, sizeof(W3REC));
        pos = tell(f);
        if (rec.sign[0] == 0)
            break;

		if (one_vxd)
		{
			if (strncmp(strupr(rec.sign), one_vxd, 
				strlen(one_vxd)) == 0)
			{
				printf("File offset %08lx\n", rec.offset);
				le2(rec.offset);
				putchar('\n');
			}
		}
		else
			printf("%-8s\t%08lx\n", rec.sign, rec.offset);
        
        if (verbose)
        {
            le2(rec.offset);
            putchar('\n');
        }
        else if (paranoid)
        {
            // ensure there is an 'LE' file at designated offset
            exe_read(f, &w, rec.offset, sizeof(w));
            if (w != LEMAGIC)
                fail("WIN386 file corrupt?");
        }
        
    }
    close(f);
}

/**********************************************************************/

#ifdef DO_NRESTAB
#undef DO_NRESTAB
#endif
#include "le2.c"

/**********************************************************************/

