/*
WDUMPMEM.C -- simple Windows protected-mode memory dump
Andrew Schulman, May 1993

bcc -W -DWINDOWS wdumpmem.c swindos.lib
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <dos.h>

#ifdef WINDOWS
#include "windows.h"
#include "winio.h"
#endif

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

void fail(char *s) { puts(s); exit(1); }

#ifdef WINDOWS
#define MK_LIN(fp)		(((DWORD) FP_SEG(fp) << 4) + FP_OFF(fp))

void far *map_linear(DWORD lin_addr, DWORD num_bytes)
{
    WORD sel;
    
    /* allocate a selector similar to our current DS 
       (i.e., a data selector) */
    _asm mov sel, ds
    if ((sel = AllocSelector(sel)) == 0)
        fail("Cannot allocate a selector!");
    
    /* set the base and limit of the new selector */
    SetSelectorBase(sel, lin_addr);
    SetSelectorLimit(sel, num_bytes - 1);   
    
    /* turn into a far pointer */
    return MK_FP(sel, 0);
}

void free_mapped_linear(void far *fp)
{
    FreeSelector(FP_SEG(fp));
}
#endif

main(int argc, char *argv[])
{
	BYTE far *real_addr;
	BYTE far *fp;
	WORD bytes, seg, ofs, rows;
	int i, j;
	
	if (argc < 2)
		fail("usage: dumpmem seg:ofs [bytes]");
	sscanf(argv[1], "%04X:%04X", &seg, &ofs);
	real_addr = (BYTE far *) MK_FP(seg, ofs);
	if (argc > 2)
		sscanf(argv[2], "%u", &bytes);
	else
		bytes = 256;
	rows = (bytes >> 4) + 1;	// kludge
		
#ifdef WINDOWS
	real_addr = (BYTE far *) map_linear(MK_LIN(real_addr), rows * 16);
	winio_setpaint(__hMainWnd, FALSE);
#endif
		
	for (i=0; i<rows; i++)
	{
		printf("%04X   ", i << 4);
		for (j=0, fp=&real_addr[i*16]; j<16; j++, fp++)
			printf("%02X ", *fp);
		printf("  ");
		for (j=0, fp=&real_addr[i*16]; j<16; j++, fp++)
			if (isalnum(*fp))
				putchar(*fp);
			else
				putchar('.');
		printf("\n");
	}
	
#ifdef WINDOWS
	winio_setpaint(__hMainWnd, TRUE);
	free_mapped_linear(real_addr);
#endif

	return 0;
}
