/* 
WIDTMAP.C -- Displays Interrupt Descriptor Table
Andrew Schulman, February 1994
bcc -2 -B -W widtmap.c prot.c
Need to compile with -B because Borland C++ doesn't like FWORD
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include "prot.h"

#pragma pack(1)
typedef struct {
    WORD ofs_lo, sel;
    BYTE reserved, flags;
    WORD ofs_hi;
    } GATE;
#pragma pack()

typedef enum { 
    UNDEFINED=0, AVAIL_286_TSS=1, LDT=2, BUSY_286_TSS=3, 
    CALLGATE_286=4, TASK_GATE=5, INTGATE_286=6, TRAPGATE_286=7,
    UNDEFINED2=8, AVAIL_386_TSS=9, UNDEFINED3=10, BUSY_386_TSS=11,
    CALLGATE=12, UNDEFINED4=13, INTERRUPT_GATE=14, TRAP_GATE=15,
    } TYPES;
    
char *type_names[] = { 
    "UNDEF0", "TSS16 ", "LDT   ", "BUSYT2",
    "CALL16", "TASK  ", "INT16 ", "TRAP16",
    "UNDEF2", "TSS   ", "UNDEF3", "BUSYT3",
    "CALL  ", "UNDEF4", "INTR  ", "TRAP  ",
    } ;

#define P(flags)            ((flags) >> 7)
#define DPL(flags)          (((flags) >> 5) & 3)
#define TYPE(flags)         ((TYPES) ((flags) & 0x0f))

#define MAKELONG(hi, lo)    ((((DWORD) hi) << 16L) + (lo))

typedef struct { WORD limit; DWORD base; } IDTR;

static IDTR pm_idtr;

void sidt(IDTR far *pidtr)
{
    _asm les bx, pidtr
    _asm sidt fword ptr es:[bx]
}

#define PUT(s)                { fputs(s, stderr); fputs("\n", stderr); }

void fail(const char *s, ...) { PUT(s); exit(1); }

void show_gate(GATE far *fpgate)
{
    printf("%s ", type_names[TYPE(fpgate->flags)]);
    
    printf(" %04X:%08lX (%d) %c",
        fpgate->sel, 
        MAKELONG(fpgate->ofs_hi, fpgate->ofs_lo),
        DPL(fpgate->flags),
        P(fpgate->flags) ? 'P' : ' ');
}

int main(int argc, char *argv[])  
{
    GATE far *pm_idt;
    GATE far *pm;
    WORD i, pm_size;
    
    sidt(&pm_idtr);
    printf("PM IDT base=%08lX size=%04X\n",
        pm_idtr.base, pm_idtr.limit + 1);
    
    pm_idt = (GATE far *) map_linear(pm_idtr.base,
        pm_idtr.limit + 1);
    pm_size = (pm_idtr.limit + 1) / 8;

	if (argc < 2)
	{
		for (i=0, pm=pm_idt; i<pm_size; i++, pm++)
		{
			printf("%04X\t", i);
			show_gate(pm);
			printf("\n");
		}
	}
	else
	{
		for (i=1; i<argc; i++)
		{
			int intno;
			sscanf(argv[i], "%04X", &intno);
			printf("%04X\t", intno);
			show_gate(&pm_idt[intno]);
			printf("\n");
		}
	}
        
    free_mapped_linear(pm_idt);
    return 0;
}

