/*
WIN32PSP.C
Win32 app accesses its own real-mode DOS PSP
Using VxDCall0 to make INT 21h (DOS) and INT 31h (DPMI) calls

from Schulman, Unauthorized Windows, 1994
*/

#include <stdlib.h>
#include <stdio.h>
#define  WIN32_LEAN_AND_MEAN
#include "windows.h"

#define MSG(s)			 MessageBox(0, s, "WIN32PSP", MB_OK)

void fail(const char *s) { MSG(s); exit(1); }

DWORD (WINAPI *VxDCall)(DWORD srvc, DWORD eax, DWORD ecx);

#define GET_PROC(mod, func) GetProcAddress(GetModuleHandle(mod), (func))

DWORD lsl(WORD sel)
{
	if (! sel) return 0;
	_asm lsl eax, sel
	// retval in EAX
}

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
    LPSTR lpszCmdLine, int nCmdShow)
{
	char buf[512];
	MEMORY_BASIC_INFORMATION info;
	WORD *pmax;
	DWORD *pjft;
	DWORD base;
	unsigned short psp;
	int len;
	
	// init
	if (! (VxDCall = GET_PROC("KERNEL32", "VxDCall0")))
		fail("Cannot link to VxDCall");
	
	#define VWIN32_INT21_CALL		0x2A0010
	#define VWIN32_INT31_CALL		0x2A0029
	#define DosCall(eax, ecx)		VxDCall(VWIN32_INT21_CALL, (eax), (ecx))
	#define DPMICall(eax, ecx)		VxDCall(VWIN32_INT31_CALL, (eax), (ecx))

	// Call DOS INT 21h function 62h (Get PSP)
	DosCall(0x6200, 0);
	_asm mov psp, bx
	// returns protected-mode selector to PSP
	len = sprintf(buf, "Win32 app's PSP (prot mode): %04Xh\n", psp);

	// Pass prot-mode PSP to DPMI INT 31h function 6 (Get Sel Base)
	_asm mov bx, psp
	DPMICall(0x0006, 0);
#if 1
	// we're really a 16-bit DPMI client!!
	_asm mov word ptr base+2, cx
	_asm mov word ptr base, dx
#else
	// this is wrong on the disk!!
	_asm mov base, edx
#endif		
	len += sprintf(buf+len, "PSP base address: %08lXh\n", base);
	// if base is C0FFE780, then real-mode PSP is 3E78h
		
	// Pass prot-mode PSP to LSL (Load Selector Limit) instruction
	len += sprintf(buf+len, "PSP length: %04Xh bytes\n", lsl(psp) + 1);
	// Win32 app PSP is 120h bytes, not 100h!
		
	// Locate JFT, #files: this isn't in WIN32PSP.C version on disk
	pmax = (WORD *) (base + 0x32);
	pjft = (DWORD *) (base + 0x34);
	len += sprintf(buf+len, "JFT @ %08lXh (%u files)\n", 
		*pjft, (DWORD) *pmax);
		
	MSG(buf);
		
	return 0;
}

