/* 
WINDASM.C
*/

#include <stdlib.h>
#ifdef WINDOWS
#include "windows.h"
#include "winio.h"
#else
#include <stdio.h>
#endif
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <dos.h>

#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC     0x454C          /* linear executable */
#define LXMAGIC     0x584C          /* IBM OS/2 2.0 linear executable */
#define PEMAGIC     0x4550          /* NT portable executable */
#define W3MAGIC     0x3357          /* WIN386.EXE signature */

#define DLL_FLAG    0x8000
#define NEI386      0x0040
#define NEI286      0x0020
#define NEI086      0x0010

#define NSTYPE      0x0007
#define NSCODE      0x0000
#define NSDATA      0x0001
#define NSRELOC     0x0100
#define NS32BIT     0x2000

#define NERESBYTES  9

#pragma pack(1)

typedef struct
{
    unsigned    ne_magic;
    char        ne_ver;
    char        ne_rev;
    unsigned    ne_enttab;
    unsigned    ne_cbenttab;
    long        ne_crc;
    unsigned    ne_flags;
    unsigned    ne_autodata;            /* dgroup */
    unsigned    ne_heap;
    unsigned    ne_stack;
    void far *  ne_csip;
    void far *  ne_sssp;
    unsigned    ne_cseg;
    unsigned    ne_cmod;
    unsigned    ne_cbnrestab;           /* size of nonres name table */
    unsigned    ne_segtab;
    unsigned    ne_rsrctab;             /* offset of resource table */
    unsigned    ne_restab;              /* offset of resident name table */
    unsigned    ne_modref;
    unsigned    ne_impname;
    long        ne_nonrestab;           /* offset of nres table from begin */
    unsigned    ne_cmovent;
    unsigned    ne_align;               /* segment align shift count */
    unsigned    ne_cres;
    unsigned char   ne_exetyp;
    char        ne_res[NERESBYTES];     /* pad struct to 64 bytes */
    /* more stuff, but we don't care */
} NEWEXE;

typedef struct {
    unsigned short ns_sector;
    unsigned short ns_cbseg;
    unsigned short ns_flags;
    unsigned short ns_minalloc;
    } SEG; 
    
typedef struct {
    unsigned char ep_flags;
    unsigned short ep_int_3f;
    unsigned char ep_segnum;
    unsigned short ep_offset;
    } ENTRY_MOVEABLE;

typedef struct {
    unsigned char ep_flags;
    unsigned short ep_offset;
    } ENTRY_FIXED;
            
typedef struct {
    union {
        ENTRY_MOVEABLE ep_moveable;
        ENTRY_FIXED ep_fixed;
        } ep_union;
    } ENTRY_POINT;
    
typedef struct {
    unsigned char eb_nentry;
    unsigned char eb_segnum;
    } BUNDLE_HEADER;
    
typedef struct {
    BUNDLE_HEADER eb_hdr;
    ENTRY_POINT eb_epoint [1] ;
    } ENTRY_BUNDLE;
    
typedef struct {
    char nr_stype;
    char nr_flags;
    unsigned short nr_soff;
    union {
        struct {
            char nr_segno;
            char nr_res;
            unsigned short nr_entry;
            } nr_intref;
        struct {
            unsigned short nr_mod;
            unsigned short nr_proc;
            } nr_import;
        struct {
            unsigned short nr_ostype;
            unsigned short nr_osres;
            } nr_osfix;
        } nr_union;
    } RELOC_ITEM;
    
typedef struct {
    unsigned short seg, ofs;
    char *name;
    } ENTRY_SEGOFS;

#define NRSTYP      0x0f
#define NRSBYT      0x00
#define NRSSEG      0x02
#define NRSPTR      0x03
#define NRSOFF      0x05
#define NRSPTR48    0x0b
#define NRSOFF32    0x0d

#define NRADD       0x04
#define NRRTYP      0x03
#define NRRINT      0x00
#define NRRORD      0x01
#define NRRNAM      0x02
#define NRROSF      0x03

#define MOVEABLE    0xFF

#define MODNAME(x)  (strupr(IMPNAME(modref[x-1])))
#define FUNCNAME(x) (entry_segofs[x].name)
    
void failret(int ret, char *s)
{ 
    printf("windasm fail: %s\n", s);
    exit(ret); 
}

void fail(char *s) { failret(1, s); }

void banner(void)
{
#ifndef WINDOWS
    fputs("Windows Quicky Disassembler version 0.1\n", stderr);
    fputs("Copyright (c) 1992 Andrew Schulman.  All rights reserved.\n\n", 
        stderr);

	fputs("80386 Disassembler Library Copyright (c) 1986-92 "
		  "Phar Lap Software, Inc.\n\n", stderr);
#endif  
}
    
void usage()
{
    banner();
	puts("usage: windasm [-func funcname] [-seg segnum] <exe>");
    exit(1);
}
    
extern void disasm(unsigned segno, void far *code, unsigned bytes, int use);
void newexe(char *name);
void do_one_reloc(unsigned seg, unsigned ofs, unsigned type,
    unsigned p1, unsigned p2, RELOC_ITEM *r, unsigned char *code);
void do_one_reloc_2(unsigned ofs, RELOC_ITEM *r, unsigned char *code);
void do_seg(int segnum, SEG *seg);
void dump(unsigned segnum, char *data, unsigned bytes);
void do_reloc(int segnum, unsigned char *code, RELOC_ITEM *pr, 
	unsigned nreloc);
void do_segtab(void);
void do_enttab(void);
void do_modref(void);
void do_impnam(void);
void do_cv_sym(void);
void do_nres(void);
void do_resnam(void);
int ne_read(int f, void *p, unsigned long ne_offset, unsigned bytes);
int exe_read(int f, void *p, unsigned long exe_offset, unsigned bytes);
void *allocate(size_t x, size_t y);
void count2null(char *s, int len);
unsigned long find_cvsym_byname(char *name);

/* key global structures and variables */
static NEWEXE hdr = {0} ;
ENTRY_SEGOFS *entry_segofs = 0;  
ENTRY_SEGOFS *entrytab = 0;
static SEG *segtab;
static unsigned long start_newexe = 0;
unsigned *modref;
unsigned char *impnam;
char *this_modname;
static unsigned max_ord;
static unsigned long align;
static int f;
char *do_one_func = 0;
static int do_one_seg = 0;
static void far *func_ptr = 0;

/* command-line options */
static int no_banner = 0;

/* arbitrary limits */
#define MAX_IMPNAME     8092
#define MAX_ENTRY       4096

main(int argc, char *argv[])
{
    int i;
    
    if (argc < 2)
        usage();
    
    for (i=1; i<argc; i++)
        if ((argv[i][0] == '-') || (argv[i][0] == '/'))
        {
            if (argv[i][1] == '?')
                usage();
			else if (strcmp(&argv[i][1], "func") == 0)
			{
				i++;
				do_one_func = argv[i];
			}
			else if (strcmp(&argv[i][1], "seg") == 0)
			{
				i++;
				do_one_seg = atoi(argv[i]);
			}
            else if (strcmp(&argv[i][1], "nobanner") == 0)
                no_banner = 1;
            else
                usage();
        }
        else
        {
            if (! no_banner)
                banner();
            newexe(argv[i]);
            exit(0);
        }
    return 0;
}

void newexe(char *name)
{
    SEG *seg;
    unsigned w;
    unsigned i;
    
    if ((f = open(name, O_RDONLY | O_BINARY)) == -1)
        fail("can't find file");
    read(f, &w, 2);
    if (w != EMAGIC)
        fail("not an executable");
    lseek(f, ENEWHDR, SEEK_SET);
    read(f, &start_newexe, 4);
    lseek(f, start_newexe, SEEK_SET);
    read(f, &hdr, sizeof(NEWEXE));
    
    switch (hdr.ne_magic)
    {
        case LEMAGIC:
            failret(2, "Use LEDUMP to examine this file");
            break;
        case LXMAGIC:
            fail("sorry, this is an OS/2 2.0 linear executable, not "
                 "a segmented executable");
            break;
        case W3MAGIC:
            failret(3, "Use W3MAP to examine this file");
            break;
        case PEMAGIC:
            fail("use COFF -DUMP to examine this NT portable executable "
                 "(PE)");
            break;
        case NEMAGIC:
            // this is the one we want
            break;
        default:
            fail("not a segmented (new) executable");
    }
    
    align = 1 << hdr.ne_align;
    
	do_enttab();
	do_segtab();
	do_modref();
	do_impnam();
	do_cv_sym();
	
	if (do_one_func)
	{
		int i;
		ENTRY_SEGOFS *eso;
		unsigned segnum;
		
		strupr(do_one_func);
		if (! (func_ptr = find_cvsym_byname(do_one_func)))
			for (i=1, eso=entry_segofs+1; i<=max_ord; i++, eso++)
				if (strcmp(eso->name, do_one_func) == 0)
				{
					func_ptr = MK_FP(eso->seg, eso->ofs);
					break;
				}
		if (! func_ptr)
		{
			if (strcmp(do_one_func, "START") == 0)
				func_ptr = hdr.ne_csip;	// initial entry point
			else	
				fail("can't find function");
		}
		segnum = FP_SEG(func_ptr);
		do_seg(segnum, &segtab[segnum-1]);
	}
	else if (do_one_seg)
	{
		if (do_one_seg <= hdr.ne_cseg)
			do_seg(do_one_seg, &segtab[do_one_seg-1]);
		else
			fail("segment number too large");
	}
	else	// do entire program
	{
		for (i=0, seg=segtab; i<hdr.ne_cseg; i++, seg++)
		{
			do_seg(i+1, seg);
			printf("\n");
		}
	}

    close(f);
}

void do_modref(void)
{
    modref = allocate(hdr.ne_cmod, sizeof(unsigned));
    ne_read(f, modref, hdr.ne_modref, hdr.ne_cmod * sizeof(unsigned));
}

void do_impnam(void)
{
    impnam = allocate(MAX_IMPNAME, 1);  
    ne_read(f, impnam, hdr.ne_impname, MAX_IMPNAME);
    count2null(impnam, MAX_IMPNAME);
}
        
void do_segtab(void)
{
    SEG *seg;
    unsigned i;
    
    segtab = allocate(hdr.ne_cseg, sizeof(SEG));
    ne_read(f, segtab, hdr.ne_segtab, hdr.ne_cseg * sizeof(SEG));
}

/**********************************************************************/
/* from Microsoft C Developer's Toolkit Reference, Chapter 3, 
   "Extended .EXE Format for Debug Information" */

typedef struct {
    unsigned char sign[4];
    unsigned long offset;
    } CV_HDR;
    
typedef struct {
    unsigned short SubSection;
    unsigned short iMod;
    signed long lfo;
    unsigned short cb;
    } DNT;
    
typedef struct {
    unsigned short seg;
    unsigned short ofs;
    unsigned short len;
    unsigned short overlay;
    unsigned short index;
    unsigned char  cseg;
    unsigned char  reserved;
    unsigned char  str[1];
    } MOD;
    
typedef struct {
    unsigned short ofs;
    unsigned short seg;
    unsigned short typ;
    unsigned char str[1];
    } PUB;
    
#define SSTMODULES      0x101
#define SSTPUBLIC       0x102
#define SSTTYPES        0x103
#define SSTSYMBOLS      0x104
#define SSTSRCLINES     0x105
#define SSTLIBRARIES    0x106
#define SSTCOMPACTED    0x107
#define SSTSRCLNSEG     0x108

typedef struct {
	unsigned seg, ofs;
	char *str;
} CVSYM;

#define MAX_CVSYM		4096
#define MAX_OBJNAME		512
	
CVSYM *cvsym = 0;
int num_cvsym = 0;
CVSYM *objname = 0;
int num_objname = 0;

signed int cvsym_cmp(const void *s1, const void *s2)
{
	CVSYM *sym1 = s1;
	CVSYM *sym2 = s2;
	if (sym1->seg == sym2->seg)
		return (sym1->ofs - sym2->ofs);
	else
		return (sym1->seg - sym2->seg);
}

unsigned long find_cvsym_byname(char *name)
{
	int i;
	CVSYM *pcv;
	
	if (! (cvsym && num_cvsym))
		return 0;
	
	// cvsym not sorted by name, so do linear search
	
	for (i=0, pcv=cvsym; i<num_cvsym; i++, pcv++)
		if (strcmp(name, pcv->str) == 0)
			return MK_FP(pcv->seg, pcv->ofs);
	/* still here */
	return 0;
}

char *find_cvsym(unsigned seg, unsigned ofs)
{
	CVSYM *pcv;
	CVSYM sym;
	
	if (! (cvsym && num_cvsym))
		return (char *) 0;
	
	sym.seg = seg;
	sym.ofs = ofs;
	if (pcv = bsearch(&sym, cvsym, num_cvsym, sizeof(CVSYM), cvsym_cmp))
	{
		if (strstr(pcv->str, "__"))	// probably a constant
			return (char *) 0;
		else
			return pcv->str;
	}
	else
		return (char *) 0;
}

char *find_objname(unsigned seg, unsigned ofs)
{
	CVSYM *pcv;
	CVSYM sym;
	
	if (! (objname && num_objname))
		return (char *) 0;
	
	sym.seg = seg;
	sym.ofs = ofs;
	if (pcv = bsearch(&sym, objname, num_objname, sizeof(CVSYM), cvsym_cmp))
		return pcv->str;
	else
		return (char *) 0;
}

void do_subsection(DNT *pdnt);

static unsigned long base;

void do_cv_sym(void)        // CodeView symbols
{
    CV_HDR cv_hdr;
    DNT *pdnt, *p;
    unsigned long filesize;
    unsigned long IfoDIR;
    unsigned short cdnt;
    int i;
    
    // go to last eight bytes of file, look for NB0 signature
    lseek(f, filelength(f) - 8, SEEK_SET);  // problems with SEEK_END
    read(f, &cv_hdr, sizeof(CV_HDR));
    if (memcmp(cv_hdr.sign, "NB0", 3) != 0)
        return; // no CodeView debug information
    
    // go to CV info base address, look for NB0 signature again
    base = filelength(f) - cv_hdr.offset;
    lseek(f, base, SEEK_SET);
    read(f, &cv_hdr, sizeof(CV_HDR));
    if (memcmp(cv_hdr.sign, "NB0", 3) != 0)
        return; // error finding second CV debug signature
    if (cv_hdr.offset == -1L)
        return; // not CodeView:  Turbo Debug?
    
	cvsym = allocate(MAX_CVSYM, sizeof(CVSYM));
	objname = allocate(MAX_OBJNAME, sizeof(CVSYM));

    // find the SubSection directory
    IfoDIR = base + cv_hdr.offset;
    lseek(f, IfoDIR, SEEK_SET);
    read(f, &cdnt, sizeof(cdnt));   // number of directory entries
        
    // read the dnt
    pdnt = allocate(cdnt, sizeof(DNT));
    read(f, pdnt, cdnt * sizeof(DNT));
    for (i=0, p=pdnt; i<cdnt; i++, p++)
        do_subsection(p);
	
	qsort(cvsym, num_cvsym, sizeof(CVSYM), cvsym_cmp);
	qsort(objname, num_objname, sizeof(CVSYM), cvsym_cmp);

#ifdef DEBUG
{
	int i;
	CVSYM *psym;
	for (i=num_cvsym, psym=cvsym; i--; psym++)
		printf("%d:%04x %s\n", psym->seg, psym->ofs, psym->str);
	for (i=num_objname, psym=objname; i--; psym++)
		printf("%d:%04x %s\n", psym->seg, psym->ofs, psym->str);
}
#endif
		
    free(pdnt);
}

int allspace(char *s)
{
    while (*s)
        if (*s == ' ')
            s++;
        else
            return 0;
    /* still here */
    return 1;
}

void do_subsection(DNT *pdnt)
{
    MOD *pmod;
    PUB *ppub, *psav;
    char name[128];
    unsigned short offset;
    
    if (pdnt->SubSection == SSTMODULES)
    {
        pmod = allocate(1, pdnt->cb);
        lseek(f, base + pdnt->lfo, SEEK_SET);
        read(f, pmod, pdnt->cb);
        if (pmod->cseg > 1)
            ;   // ok?
        strncpy(name, &pmod->str[1], pmod->str[0]);
        name[pmod->str[0]] = '\0';

		objname[num_objname].seg = pmod->seg;
		objname[num_objname].ofs = pmod->ofs;
		objname[num_objname].str = allocate(1, strlen(name)+1);
		strcpy(objname[num_objname].str, name);
		num_objname++;
			
        free(pmod);
    }
    else if (pdnt->SubSection == SSTPUBLIC)
    {
        ppub = psav = allocate(1, pdnt->cb);
        offset = 0;
        lseek(f, base + pdnt->lfo, SEEK_SET);
        read(f, ppub, pdnt->cb);
        while (offset < pdnt->cb)
        {
            strncpy(name, &ppub->str[1], ppub->str[0]);
            name[ppub->str[0]] = '\0';
            if (! allspace(name))   // don't print blank symbols!
			{
				if (num_cvsym == MAX_CVSYM)
					printf("Error: CVSYM table overflow\n");
				else
				{
					cvsym[num_cvsym].seg = ppub->seg;
					cvsym[num_cvsym].ofs = ppub->ofs;
					cvsym[num_cvsym].str = allocate(1, strlen(name)+1);
					strcpy(cvsym[num_cvsym].str, name);
					num_cvsym++;
				}
			}
            offset += (sizeof(PUB) + ppub->str[0]);
            ((unsigned char *) ppub) += (sizeof(PUB) + ppub->str[0]);
        }
        free(psav);
    }
}

/**********************************************************************/
    
int mangled_name(char *s)   // look for mangled names like "mmmmmm"
{
    int c = *s;
    char *s2 = s;
    int len = strlen(s);
    if (len < 3)
        return 0;
    while (*s2 == c)
        s2++;
    return (s + len == s2);
}
                    
int eso_cmp(const void *s1, const void *s2)
{
	ENTRY_SEGOFS *sym1 = s1;
	ENTRY_SEGOFS *sym2 = s2;
	
	if (sym1->seg == sym2->seg)
		return (sym1->ofs - sym2->ofs);
	else
		return (sym1->seg - sym2->seg);
}

void do_enttab(void)
{
    ENTRY_BUNDLE *enttab = 0;
    ENTRY_BUNDLE *ep;
    ENTRY_SEGOFS *eso;
    unsigned eb;
    int ord = 0;
    int i;
    
    entry_segofs = allocate(MAX_ENTRY, sizeof(ENTRY_SEGOFS));
    eso = entry_segofs;

    enttab = allocate(hdr.ne_cbenttab, 1);  /* number BYTES in entry table */ 
    ne_read(f, enttab, hdr.ne_enttab, hdr.ne_cbenttab);
    for (eb=0, ep=enttab; ; )
    {
        BUNDLE_HEADER bhdr = ep->eb_hdr;
        if (ord >= MAX_ENTRY)
            fail("too many entry points");
        if (bhdr.eb_nentry == 0)        // end of table
            break;
        else if (bhdr.eb_segnum == 0)   // null entry
        {
            ((unsigned char *) ep) += sizeof(BUNDLE_HEADER);
            ord += bhdr.eb_nentry;      // for ordinal numbers
            continue;
        }
        eb++;
        ((unsigned char *) ep) += sizeof(BUNDLE_HEADER);
        if (bhdr.eb_segnum == MOVEABLE)
        {
            unsigned i;
            ENTRY_MOVEABLE *em;
            for (i=0, em=ep; i<bhdr.eb_nentry; i++, em++)
            {
                ord++;
                eso[ord].seg = em->ep_segnum;
                eso[ord].ofs = em->ep_offset;
            }
            ((unsigned char *) ep) += 
                (bhdr.eb_nentry * sizeof(ENTRY_MOVEABLE));
        }
        else                            // fixed
        {
            unsigned i;
            ENTRY_FIXED *ef;
            for (i=0, ef=ep; i<bhdr.eb_nentry; i++, ef++)
            {
                ord++;
                eso[ord].seg = bhdr.eb_segnum;
                eso[ord].ofs = ef->ep_offset;
            }
            ((unsigned char *) ep) += 
                (bhdr.eb_nentry * sizeof(ENTRY_FIXED));
        }
    }
    max_ord = ord;
	
    free(enttab);
    
    eso = entry_segofs;

    /* walk non-resident name table */
    do_nres();
    
    /* walk resident names table */
    /* get module name from first entry */
    /* don't know size of resident names table! */
    do_resnam();
    
#if 1
	entrytab = allocate(max_ord, sizeof(ENTRY_SEGOFS));
	memcpy(entrytab, entry_segofs, max_ord * sizeof(ENTRY_SEGOFS));
	// do NOT sort entry_segofs:  indexed by ordinal!!!!
	qsort(entrytab, max_ord, sizeof(ENTRY_SEGOFS), eso_cmp);
#endif
}

char *find_entryname(unsigned seg, unsigned ofs)
{
#if 0
	int i;
	ENTRY_SEGOFS *eso;
	
	for (i=1, eso=entry_segofs+1; i<=max_ord; i++, eso++)
		if ((eso->seg == seg) && (eso->ofs == ofs))
		{
			if (strstr(eso->name, "__"))	// probably a constant
				return (char *) 0;
			else
				return eso->name;
		}

	/* still here */
	return (char *) 0;
#else
	ENTRY_SEGOFS *pcv;
	ENTRY_SEGOFS sym;
	
	if (MK_FP(seg, ofs) == hdr.ne_csip)
		return "start";
	
	if (! (entrytab && max_ord))
		return (char *) 0;
	
	sym.seg = seg;
	sym.ofs = ofs;
	if (pcv = bsearch(&sym, entrytab, 
		max_ord, sizeof(ENTRY_SEGOFS), eso_cmp))
	{
		if (strstr(pcv->name, "__"))	// probably a constant
			return (char *) 0;
		else
			return pcv->name;
	}
	else
		return (char *) 0;
#endif	
}

void do_nres(void)
{
    ENTRY_SEGOFS *eso = entry_segofs;
    unsigned char *nres = allocate(hdr.ne_cbnrestab, 1);
    unsigned char *nr;
    unsigned i;
    unsigned char c;
    unsigned ndx;

    exe_read(f, nres, hdr.ne_nonrestab, hdr.ne_cbnrestab);
    
    for (i=0, nr=nres; i<hdr.ne_cbnrestab; )
    {
        if (! (c = nr[0]))
            break;
        
        ndx = *((unsigned *) &nr[c+1]);     // ordinal number
        nr[c+1] = '\0';     // terminate string, replacing ord#
            
        if ((ndx == 0) && nr[1])
			;	// do nothing
        else if (ndx < MAX_ENTRY)
        {
            if (eso[ndx].name && eso[ndx].name[0])  // already there!
                if (ndx > 1)
                    printf(";; conflict on %u - %s\n", ndx, eso[ndx].name);
            if (! isprint(nr[1]))
                fail("internal error: bad nres name!");
            eso[ndx].name = &nr[1];
        }
        else if (&nr[1] && nr[1])   // internal error: ordnum too big
            printf(";; discarding nres %u - %s\n", ndx, &nr[1]);
        
        nr += (c+3);
        i += (c+3);
    }
}

void do_resnam(void)
{
    ENTRY_SEGOFS *eso = entry_segofs;
    unsigned res;
    unsigned char c;
    unsigned char *s;
    unsigned ndx;
    
    for (res=hdr.ne_restab;;)
    {
        ne_read(f, &c, res, 1);
        if (c == 0)
            break;
        s = allocate(c+2, 1);
        ne_read(f, s, ++res, c+2);
        ndx = *((unsigned *) &s[c]);        // WORD, not a BYTE!
        s[c] = '\0';            // now turn ord# into NULL terminate
        if (ndx == 0)           // module name
            this_modname = s;
        else if (ndx < MAX_ENTRY)
        {
            if (eso[ndx].name && eso[ndx].name[0])  // if slot already used
                printf(";; conflict on %u - %s\n", ndx, eso[ndx].name);
            eso[ndx].name = s;
        }
        else if (s && *s)   // ordnum too big
            printf(";; discarding resnam %u - %s\n", ndx, s);
        res += (c+2);
    }
}

/**********************************************************************/

void do_seg(int segnum, SEG *seg)
{
    /* multiply seg->ns_sector by hdr.ne_align to get _file_ offset; 
       _not_ NE offset */
    unsigned long reloc = (align * seg->ns_sector) + seg->ns_cbseg;
    unsigned char *code;
    RELOC_ITEM *rel;
    unsigned short nr_nreloc;
	unsigned code_seg;
 
    exe_read(f, &nr_nreloc, reloc, sizeof(unsigned short));
	/* crucial that code segment be zero-aligned; can't use malloc */
    if (_dos_allocmem(1 + (seg->ns_cbseg >> 4), &code_seg) != 0)
		fail("insufficient memory for code");
	code = MK_FP(code_seg, 0);
    exe_read(f, code, align * seg->ns_sector, seg->ns_cbseg);
    rel = allocate(nr_nreloc, sizeof(RELOC_ITEM));
    exe_read(f, rel, reloc+2, nr_nreloc * sizeof(RELOC_ITEM));

	if (nr_nreloc && (nr_nreloc < (0xFFFFU / sizeof(RELOC_ITEM))))
		do_reloc(segnum, code, rel, nr_nreloc);

	if (do_one_func)
	{
		disasm(segnum, code + FP_OFF(func_ptr), 
			seg->ns_cbseg,	/* don't subtract ofs */
			(seg->ns_flags & NS32BIT) ? 32 : 16);
	}
	else
	{
		if ((seg->ns_flags & NSTYPE) == NSCODE)
			disasm(segnum, code, seg->ns_cbseg, 
				(seg->ns_flags & NS32BIT) ? 32 : 16);
		else
			dump(segnum, code, seg->ns_cbseg);
	}

    _dos_freemem(code_seg);
    free(rel);
}

#define SPECIAL_SEG		0x8000
#define THIS_MOD		0x800
#define BY_NAME			0x400

void do_reloc(int segnum, unsigned char *code, RELOC_ITEM *rel, 
	unsigned nr_nreloc)
{
	int i;
	for (i=0; i<nr_nreloc; i++)
	{
		RELOC_ITEM *r = &rel[i];
		unsigned next;
		int ordnum, segno;

		unsigned tmp = *((unsigned *) &code[r->nr_soff]);
		
		/*
			relocation prefix:
				[i]		module.funcname
				[m] 	module.ordinal	
				[n]		funcname // thismod.ord
				[s] 	segment:offset
		*/

		switch (r->nr_flags & 0x03)
		{
			case NRRINT:
				ordnum = r->nr_union.nr_intref.nr_entry;
				segno = r->nr_union.nr_intref.nr_segno;
				if (segno == (char) MOVEABLE)
				{
					if (FUNCNAME(ordnum))
						do_one_reloc(segnum, r->nr_soff, 
							'n', 0, ordnum,	r, code);
					else
						do_one_reloc(segnum, r->nr_soff, 
							'm', -1	/* this_modname */, ordnum, r, code);
				}
				else	// fixed
				{
					do_one_reloc(segnum, r->nr_soff, 
						's', segno, ordnum, r, code);
				}
				break;
			case NRRORD:
				do_one_reloc(segnum, r->nr_soff, 
					'm', 
					r->nr_union.nr_import.nr_mod,
					r->nr_union.nr_import.nr_proc,
					r, code);
				break;
			case NRRNAM:
				do_one_reloc(segnum, r->nr_soff, 
					'i', 
					r->nr_union.nr_import.nr_mod,
					r->nr_union.nr_import.nr_proc,
					r, code);
				break;
			case NRROSF:
				/* do nothing: ok as is */
				break;
		}

		/* follow possible relocation chain */
		if (! (r->nr_flags & NRADD))
		{
			next = tmp;
			while (next != 0xFFFF)
			{
				tmp = *((unsigned *) &code[next]);
				do_one_reloc(segnum, next, -1, 0, 0, 0, code);
				if ((next = tmp) == 0)
					break;
			}
		}
	}
}

unsigned in_entry_segofs(unsigned seg, unsigned ofs)
{
    unsigned i;
    ENTRY_SEGOFS *eso;
//  printout helps establish that only called when necessary
//	printf("in_entry_so %04x:%04x\n", seg, ofs);
    for (i=1, eso=entry_segofs+1; i<=max_ord; i++, eso++)
        if ((seg == eso->seg) && (ofs == eso->ofs))
            return i;
    /* still here */
    return 0;
}
        
void do_one_reloc(unsigned seg, unsigned ofs, unsigned type,
    unsigned p1, unsigned p2, RELOC_ITEM *r, unsigned char *code)
{
    static unsigned s_type, s_p1, s_p2;
    static RELOC_ITEM *s_r;
    unsigned ord;
	
#ifdef DEBUG	
	printf("reloc <%c> %d:%04x\n", type, seg, ofs);
#endif
	
    if (type == (unsigned) -1)
    {   /* use same as last time, except for seg and ofs */
        type=s_type; p1=s_p1; p2=s_p2; r=s_r;
    }
    else
    {   /* new relocation chain */  
        s_type=type; s_p1=p1; s_p2=p2; s_r=r;
    }
    
    if (type == 's')
    {
        if (ord = in_entry_segofs(p1, p2))
        {
            // turning [s] into [n]
            do_one_reloc(seg, ofs, 'n', 0, ord, r, code);
            return;     // NB! don't print reloc type again!
        }
        else
		{
			*((unsigned long *) &code[ofs]) =
				MK_FP(p1, p2);	// segment:offset
			return;
		}
    }

    if (type == 'i')
		*((unsigned long *) &code[ofs]) =
			MK_FP(SPECIAL_SEG + BY_NAME + p1, p2);
    else if (type == 'm')
	{
		if (p1 == 0xFFFF)
			*((unsigned long *) &code[ofs]) =
				MK_FP(SPECIAL_SEG + THIS_MOD, p2);
		else
			*((unsigned long *) &code[ofs]) =
				MK_FP(SPECIAL_SEG + p1, p2);
	}
    else if (type == 'n')
		*((unsigned long *) &code[ofs]) =
			MK_FP(SPECIAL_SEG + THIS_MOD + BY_NAME, p2);
    
#if 0		
    do_one_reloc_2(ofs, r, code);
#endif
}

void do_one_reloc_2(unsigned ofs, RELOC_ITEM *r, unsigned char *code)
{
#ifdef DEBUG	
    char *typ;
    
    // display relocation type -- crucial for WINPP!
    switch (r->nr_stype)
    {
        case NRSPTR:    typ = "PTR";    break;
        case NRSBYT:    typ = "BYT";    break;
        case NRSSEG:    typ = "SEG";    break;  // BASE
        case NRSOFF:    typ = "OFF";    break;
        case NRSPTR48:  typ = "P48";    break;
        case NRSOFF32:  typ = "O32";    break;
        default:        typ = "???";    break;
    }
    printf("%s", typ);
    
    if (r->nr_flags & NRADD)
    {
        void *cod = &code[ofs];
        switch (r->nr_stype)
        {
            case NRSPTR:    
            case NRSOFF32:
                printf(" ADD+%08lx", (*(unsigned long *) cod)); break;
            case NRSBYT:    
                printf(" ADD+%02x", (*(unsigned char *) cod)); break;
            case NRSSEG:
            case NRSOFF:
                if (*((unsigned *) cod) != 0)   // don't print if ADD+0000
                    printf(" ADD+%04x", (*(unsigned *) cod)); break;
            case NRSPTR48:
                printf(" ADD+%04x:", (*(unsigned *) cod));
                printf("%08lx", (*(unsigned long *) cod+2)); break;
            default:
                printf(" ADD????"); break;
        }
    }
#endif	
}
/**********************************************************************/

/* read at offset from NE */
int ne_read(int f, void *p, unsigned long ne_offset, unsigned bytes)
{
    lseek(f, start_newexe + ne_offset, SEEK_SET);
    return read(f, p, bytes);
}

/* read at offset from beginning of file */
int exe_read(int f, void *p, unsigned long exe_offset, unsigned bytes)
{
    lseek(f, exe_offset, SEEK_SET);
    return read(f, p, bytes);
}


void *allocate(size_t x, size_t y)
{
    void *p;
	static unsigned long total=0;
	total+=(x*y);
//	printf("allocate %u*%u=%u (%lu)\n", x, y, x*y, total);
    if (! x) x = 1;
    if (! y) y = 1;
    if (((long) x * (long) y) > 0xFFFFUL)
        fail("internal error");
    if ((p = calloc(x, y)) == 0)
        fail("insufficient memory");
    return p;
}

/* in place, turn counted strings into null-terminated strings */
void count2null(char *s, int len)
{
    unsigned char *s2; 
    int i;
    for (i=0, s2=s; i<len; )
    {
        int len = *s2;  // get string length
        *s2 = '\0';     // replace with NULL
        i += (len+1);   // increment ptr
        s2 += (len+1);
    }
    *s2 = '\0';         // get last one -- table must include room for it
}

/**********************************************************************/

#define WIDTH		16

void dump(unsigned segnum, char *data, unsigned bytes)
{
	unsigned char far *p;
	unsigned i, j, c;
	
    for (i=0; i<bytes; i += WIDTH)
    {
		c = ((bytes-i) > WIDTH) ? WIDTH : bytes-i;
		printf("%d:%04x | ", segnum, i);
        for (j=c, p=data+i; j--; p++)
            printf("%02X ", *p);
		for (j=WIDTH-c; j--; )	// pad out on last line
			printf("   ");
		putchar('|'); putchar(' ');
        for (j=c, p=data+i; j--; p++)
			putchar( isprint(*p) ? *p : '.' );
		putchar('\n');
    }
}

