/*
WLOG212F.C -- combination of WV86TEST, V86TEST, HOOKINT, and WISPY2
Hook INT 21h from Windows app, hook in V86 mode via DPMI, see which
INT 21h are passed down to DOS.

from Schulman, Unauthorized Windows, 1994

bcc -I..\include -L..\lib -W -2 -P- wlog212f.c unauthw.lib 
*/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#include <process.h>
#include <time.h>
#include <dos.h>
#include "windows.h"
#include "winio.h"
#include "wmhandlr.h"
#include "prot.h"

typedef unsigned short WORD;
typedef unsigned long DWORD;

#pragma pack(1)

typedef struct {
    DWORD calls;
    DWORD int21[0x100], int2144[0x100], int2f[0x100], int2f11[0x100];
    time_t start, end;
    } WSTATS, far *FPWSTATS;
    
static BOOL do_auto_refresh = 0;
static UINT htimer = 0;
static far FPWSTATS wstats;
static far FPWSTATS rstats;
static far FPWSTATS kstats;
static far FARPROC old_dos = (FARPROC) 0;
static HMENU gmenu;

static char *prot_mode = "Protected mode";
static char *kdos = "Protected mode (KernelDOSProc)";
static char *real_mode = "Real mode";

/**********************************************************************/

void display_line(int func, DWORD w, DWORD k, DWORD r)
{
    printf("%02Xh\t   ", func);
    if (w) printf("%6lu\t", w); else printf("      \t");
    if (k) printf("%6lu\t\t", k); else printf("      \t\t");
    if (r) printf("%6lu\t", r); else printf("      \t");
    if ((w || k) && (! r)) printf("Emulated\n");
    else if ((r >= w) && (r >= k)) printf("Passed down\n");
    else printf("\n"); // sometimes passed down
}
        
void display_results(void)
{
    static FPWSTATS ws = (FPWSTATS) 0;
    DWORD elapsed;
    DWORD far *p, far *p2, far *p3;
    int i;
    
    // one-time initialization
    if (! ws)
        if (! (ws = (FPWSTATS) _fmalloc(sizeof(WSTATS))))
            fail("Insufficient memory");
        
    wstats->end = time(NULL);
    _fmemcpy(ws, wstats, sizeof(WSTATS));
    
    elapsed = ws->end - ws->start;
    printf("Elapsed: %lu seconds\n", elapsed);
    printf("Calls: %lu\n", ws->calls);
    if (elapsed)
        printf("%lu calls/second\n", ws->calls / elapsed);

	#define PRINT_STATS(s, tab) { \
		printf("\n%s:\n", s); \
		printf("Func\t   Prot mode\tKernelDOSProc\tV86 mode\n"); \
		for (i=0, p=wstats->tab, p2=kstats->tab, p3=rstats->tab; \
			 i<0x100; \
			 i++, p++, p2++, p3++) \
			if (*p || *p2 || *p3) \
				display_line(i, *p, *p2, *p3); \
		}

	PRINT_STATS("INT 21h", int21);
	PRINT_STATS("IOCTL (INT 21h function 44h)", int2144);
	PRINT_STATS("INT 2Fh", int2f);
	PRINT_STATS("Network redirector (INT 2Fh AH=11h)", int2f11);
}

/**********************************************************************/

void refresh(HWND hwnd, int wID)
{
    winio_clear(hwnd);
    display_results(); 
}

void clear(HWND hwnd, int wID)
{
    _fmemset(wstats, 0, sizeof(WSTATS));
    _fmemset(kstats, 0, sizeof(WSTATS));
    _fmemset(rstats, 0, sizeof(WSTATS));
}

long on_time(HWND hwnd, unsigned message, WORD wParam, LONG lParam)
{
    winio_clear(hwnd);
    display_results();
    return 0;
}

void auto_refresh(HWND hwnd, int wID)
{
    if (do_auto_refresh)    // on, turn off
    {
        do_auto_refresh = 0;
        KillTimer(hwnd, htimer);
        CheckMenuItem(gmenu, 4, MF_UNCHECKED);
    }
    else    // off, turn on
    {
        do_auto_refresh = 1;
        wmhandler_set(hwnd, WM_TIMER, on_time);
        if (! (htimer = SetTimer(hwnd, 1, 5000, NULL))) // every 5 seconds
            fail("Can't create timer");
        CheckMenuItem(gmenu, 4, MF_CHECKED);
    }
}

/**********************************************************************/

FARPROC (FAR PASCAL *GetSetKernelDosProc)(FARPROC DosProc) = 0;

typedef struct {
#ifdef __TURBOC__
    WORD bp,di,si,ds,es,dx,cx,bx,ax;
#else
    WORD es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    WORD ip,cs,flags;
    } REG_PARAMS;
    
typedef void (interrupt far *INTRFUNC)(REG_PARAMS);

INTRFUNC _dpmi_get_pmode_vect(int intno)
{
    INTRFUNC iv;
    _asm mov ax, 0204h
    _asm mov bl, byte ptr intno
    _asm int 31h
    _asm jc error
    _asm mov word ptr iv+2, cx
    _asm mov word ptr iv, dx
    return iv;
error:
    return (INTRFUNC) 0;
}

INTRFUNC _dpmi_get_rmode_vect(int intno)
{
    INTRFUNC iv;
    _asm mov ax, 0200h
    _asm mov bl, byte ptr intno
    _asm int 31h
    _asm jc error
    _asm mov word ptr iv+2, cx
    _asm mov word ptr iv, dx
    return iv;
error:
    return (INTRFUNC) 0;
}

void _dpmi_set_pmode_vect(int intno, INTRFUNC iv)
{
    _asm mov ax, 0205h
    _asm mov bl, byte ptr intno
    _asm mov cx, word ptr iv+2
    _asm mov dx, word ptr iv
    _asm int 31h
    // _asm jc error
}

void _dpmi_set_rmode_vect(int intno, INTRFUNC iv)
{
    _asm mov ax, 0201h
    _asm mov bl, byte ptr intno
    _asm mov cx, word ptr iv+2
    _asm mov dx, word ptr iv
    _asm int 31h
    // _asm jc error
}

typedef struct {
    DWORD edi, esi, ebp, esp, ebx, edx, ecx, eax;
    WORD flags, es, ds, fs, gs, ip, cs, sp, ss;
    } RMODE_REGS;

INTRFUNC _dpmi_alloc_rmode_callback(INTRFUNC pmode, RMODE_REGS far *regs)
{
    INTRFUNC callback;
    _asm mov ax, 0303h
    _asm push si
    _asm push di
    _asm push ds
    _asm lds si, dword ptr pmode
    _asm les di, dword ptr regs
    _asm int 31h
    _asm pop ds
    _asm pop di
    _asm pop si
    _asm jc error
    _asm mov word ptr callback+2, cx
    _asm mov word ptr callback, dx
    return callback;
error:
    return (INTRFUNC) 0;
}

void _dpmi_free_rmode_callback(INTRFUNC callback, RMODE_REGS *regs)
{
    _asm mov ax, 0304h
    _asm mov cx, word ptr callback+2
    _asm mov dx, word ptr callback
    _asm int 31h
    // _asm jc error
    if (regs)
        free(regs);
}

INTRFUNC GetSetInt(int intno, INTRFUNC new)
{
    INTRFUNC old = _dpmi_get_pmode_vect(intno);
    if (old != 0)
        _dpmi_set_pmode_vect(intno, new);
    return old;
}

void RestoreOldRealInt(int intno, INTRFUNC old, INTRFUNC callback,
	RMODE_REGS *regs)
{
	_dpmi_set_rmode_vect(intno, old);
	_dpmi_free_rmode_callback(callback, regs);
}

INTRFUNC GetSetRealInt(int intno, INTRFUNC new, INTRFUNC *pcallback,
	RMODE_REGS **pregs)
{
    INTRFUNC old;
    if ((old = _dpmi_get_rmode_vect(intno)) != 0)
    {
        RMODE_REGS *regs;
        INTRFUNC callback;
        if (! (regs = (RMODE_REGS *) calloc(sizeof(RMODE_REGS), 1)))
			return (INTRFUNC) 0;
        if (! (callback = _dpmi_alloc_rmode_callback(new, regs)))
			return (INTRFUNC) 0;
        _dpmi_set_rmode_vect(intno, callback);
		*pcallback = callback;
		*pregs = regs;
    }
    return old;
}

static INTRFUNC prev_21, prev_2f;
static INTRFUNC real_prev_21, real_prev_2f;
static WORD real_prev_21_seg, real_prev_21_off;
static WORD real_prev_2f_seg, real_prev_2f_off;

void interrupt far int21(REG_PARAMS r)
{
    wstats->calls++;
    wstats->int21[r.ax >> 8]++;
	if ((r.ax >> 8) == 0x44)
		wstats->int2144[r.ax & 0xff]++;
    _chain_intr(prev_21);
}

void interrupt far new_dos(REG_PARAMS r)
{
    kstats->calls++;
    kstats->int21[r.ax >> 8]++;
	if ((r.ax >> 8) == 0x44)
		kstats->int2144[r.ax & 0xff]++;
    _chain_intr((INTRFUNC) old_dos);
}

void far interrupt real_int21(REG_PARAMS r)
{
    RMODE_REGS far *regs = (RMODE_REGS far *) MK_FP(r.es, r.di);
    rstats->calls++;
    rstats->int21[((WORD) regs->eax) >> 8]++;
	if ((r.ax >> 8) == 0x44)
		rstats->int2144[r.ax & 0xff]++;
    regs->cs = real_prev_21_seg;
    regs->ip = real_prev_21_off;
}

void far interrupt real_int2f(REG_PARAMS r)
{
    RMODE_REGS far *regs = (RMODE_REGS far *) MK_FP(r.es, r.di);
    rstats->calls++;
    rstats->int2f[((WORD) regs->eax) >> 8]++;
    if ((r.ax >> 8) == 0x11)
        rstats->int2f11[r.ax & 0xff]++;
    regs->cs = real_prev_2f_seg;
    regs->ip = real_prev_2f_off;
}

void interrupt far int2f(REG_PARAMS r)
{
    wstats->calls++;
    wstats->int2f[r.ax >> 8]++;
    if ((r.ax >> 8) == 0x11)
        wstats->int2f11[r.ax & 0xff]++;
    _chain_intr(prev_2f);
}

static INTRFUNC callback_21, callback_2f;
static RMODE_REGS *regs_21, *regs_2f;

void restore_prev(HWND hwnd)
{
    if (old_dos)
        (void) GetSetKernelDosProc(old_dos);
    (void) GetSetInt(0x21, prev_21);
    (void) GetSetInt(0x2f, prev_2f);
	RestoreOldRealInt(0x21, real_prev_21, callback_21, regs_21);
	RestoreOldRealInt(0x2f, real_prev_2f, callback_2f, regs_2f);
}

/**********************************************************************/

main(int argc, char *argv[])
{
    HWND hwnd;
    int i;
    
    if (__hPrevInst != 0)
        fail("WLOG212F already running");
    
    hwnd = winio_current();
    gmenu = CreateMenu();   // global
    #define MK_MENU(id, str, func) { \
        AppendMenu(gmenu, MF_STRING | MF_ENABLED, id, str); \
        winio_setmenufunc(hwnd, id, func); \
        }
    MK_MENU(1, "&Refresh", refresh);
    MK_MENU(2, "&Clear", clear);
    MK_MENU(4, "&Auto-Refresh", auto_refresh);
    InsertMenu(winio_hmenumain(hwnd), 1, 
        MF_STRING | MF_POPUP | MF_BYPOSITION, gmenu, "&Test");
    DrawMenuBar(hwnd);
    
    if ((! (wstats = calloc(sizeof(WSTATS), 1))) ||
        (! (rstats = calloc(sizeof(WSTATS), 1))) ||
        (! (kstats = calloc(sizeof(WSTATS), 1))))
        fail("Insufficient memory");

    #define GET_PROC(mod,func) GetProcAddress(GetModuleHandle(mod), (func))
    if (! (GetSetKernelDosProc = GET_PROC("KERNEL", "GETSETKERNELDOSPROC")))
        puts("Can't get KernelDosProc");
    else
        old_dos = GetSetKernelDosProc((FARPROC) new_dos);
    
    if (((prev_21 = GetSetInt(0x21, int21)) == 0) ||
        ((prev_2f = GetSetInt(0x2f, int2f)) == 0))
            fail("Can't set prot-mode interrupts");
        
    if (((real_prev_21 = GetSetRealInt(0x21, real_int21,
			&callback_21, &regs_21)) == 0) ||
        ((real_prev_2f = GetSetRealInt(0x2f, real_int2f,
			&callback_2f, &regs_2f)) == 0))
            fail("Can't set real-mode interrupts");
    real_prev_21_seg = FP_SEG(real_prev_21);
    real_prev_21_off = FP_OFF(real_prev_21);
    real_prev_2f_seg = FP_SEG(real_prev_2f);
    real_prev_2f_off = FP_OFF(real_prev_2f);
    winio_onclose(hwnd, restore_prev);
    
    kstats->start = rstats->start = wstats->start = time(NULL);
    wmhandler_yield();  // let someone else run a little
    display_results();
    
    return 0;
}

