/*
WMHANDLR.C
Event (WM_ message) handlers - implementation
Dave Maxey and Andrew Schulman - 1991
*/

#include <windows.h>
#include <string.h>
#include "wmhandlr.h"

long  defwmhandler(HWND, unsigned, WORD, LONG);

/* -----------------------------------------------------------------------  */
/* This is our event processor. It is the dispatcher for the handlers set   */
/* using SetHandler. An Application plugs this function into its            */
/* window, sets handlers for messages, and WndProc handles the rest.        */
/* This window procecedure should never need to be changed!                 */
/* -----------------------------------------------------------------------  */
long FAR PASCAL _export WndProc(HWND hWnd, WORD message, 
    WORD wParam, LONG lParam)
    {
	WMTAB wmTab;
	
	// On a CREATE, the message table needs to be installed
	// and a WM_GETMINMAXINFO is the first message received

	if (message == WM_CREATE)
		SetWindowLong(hWnd, 0, (DWORD) *(DWORD FAR *) lParam);

	if ((wmTab = (WMTAB) GetWindowLong(hWnd, 0)) == NULL)
		return DefWindowProc(hWnd, message, wParam, lParam);
    
	if (message < WM_USER)
        return (*((*wmTab)[message]))(hWnd, message, wParam, lParam);
    else
        return DefWindowProc(hWnd, message, wParam, lParam);
    }

/* ---------------------------------------------------------------- */
/* Routines to get and set the message handlers. Setting to NULL    */
/* uninstalls a handler, by setting the handler to the default      */
/* which calls Windows own DefWndProc.                              */
/* ---------------------------------------------------------------- */
WMHANDLER wmhandler_get(HWND hWnd, WORD message)
    {
    return (message < WM_USER) ?
		(* (WMTAB) GetWindowLong(hWnd, 0))[message] :
		0;
    }

WMHANDLER wmhandler_settab(WMTAB wmTab, WORD message, WMHANDLER f)
    {
    WMHANDLER oldf;

	if (message < WM_USER)
        {
        oldf = (* wmTab)[message];
        (* wmTab)[message] = f ? f : defwmhandler;
        return (oldf ? oldf : defwmhandler);
        }
    else
        return 0;
    }

WMHANDLER wmhandler_set(HWND hWnd, WORD message, WMHANDLER f)
    {
	WMTAB wmTab = (WMTAB) GetWindowLong(hWnd, 0);
	return wmhandler_settab(wmTab, message, f);
    }

/* ----------------------------------------------------------------------- */
/* This is a default handler so that an application chain on to a previous */
/* handler from their current one without having to worry what was there   */
/* before. All this default handler does is to call DefWindowProc.         */
/* ----------------------------------------------------------------------- */
long defwmhandler(HWND hwnd, unsigned message, WORD wParam, LONG lParam)
    {   
    return DefWindowProc(hwnd, message, wParam, lParam);
    }

/* -------------------------------------------------------------------  */
/* Create a new wm handler function disatch table.						*/
/* -------------------------------------------------------------------  */
WMTAB wmhandler_create(void)
    {
    int i;
	WMTAB wmTab;
	
	if (! (wmTab = (WMTAB) GlobalLock(
						GlobalAlloc(GMEM_MOVEABLE, sizeof(WMHANDLERTABLE)))))
		return NULL;
	
	for (i = 0; i < WM_USER; i++) 
        (* wmTab)[i] = defwmhandler;
	
	return wmTab;
    }


/* -------------------------------------------------------------------  */
/* Destroy a wm handler function disatch table.							*/
/* -------------------------------------------------------------------  */
void wmhandler_destroy(HWND hwnd)
    {
	GlobalFree(LOWORD(GlobalHandle(HIWORD(GetWindowLong(hwnd, 0)))));
	SetWindowLong(hwnd, 0, NULL);
    }


/* -------------------------------------------------------------------  */
/* Returns an HWND that can be used as a message recicipent. The 		*/
/* window is invisible.  Somewhat like "object windows" in PM.			*/
/* -------------------------------------------------------------------  */
HWND wmhandler_hwnd(char *name)
{
	extern HANDLE __hPrevInst;	// in ARGCARGV
	extern HANDLE __hInst;
	HWND hWnd;
	WMTAB wmTab;
		
	if (! __hPrevInst) // NOT VALID! ASSUMES LAST APP DID WMHANDLER_HWND()
		{
		WNDCLASS wndclass;
		memset(&wndclass, 0, sizeof(wndclass));
        wndclass.lpfnWndProc = WndProc;
        wndclass.hInstance = __hInst;
        wndclass.lpszClassName = "OBJECTWND";
		wndclass.cbWndExtra = 4;
        RegisterClass(&wndclass);
		}
	
	if (! (wmTab = wmhandler_create()))
		return NULL;
	
	/* MAYBE SHOULD HAVE ICON, WITH EXIT/ABOUT... */

    if (! (hWnd = CreateWindow("OBJECTWND", name, WS_DISABLED,
        0, 0, 0, 0, NULL, NULL, __hInst, (LPSTR) wmTab)))
		return NULL;
	
	SetWindowLong(hWnd, 0, (DWORD) wmTab);
	
	return hWnd;
	}

/* -------------------------------------------------------------------  */
/* This is where messages get received and dispatched. In cooperative   */
/* multitasking of Windows, this function MUST be called periodically.  */
/* Same as winio_yield(); probably ought to have been here instead.     */
/* -------------------------------------------------------------------  */
void wmhandler_yield(void)
	{
    MSG msg;
    while (PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
        {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
        }
	}
		
/* -------------------------------------------------------------------  */
/* This sets up an application's main event loop.  It will break out    */
/* when a WM_DESTROY message is processed.								*/
/* -------------------------------------------------------------------  */
void wmhandler_mainloop(void)
	{
	for (;;)
		wmhandler_yield();
	}
