/* 
WSPY21.C
Andrew Schulman, August 1994
Based loosely on WISPY.C from Undocumented Windows
*/

#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include "windows.h"
#include "wmhandlr.h"
#include "winio.h"

#pragma pack(1)

typedef struct {
#ifdef __BORLANDC__
    WORD bp,di,si,ds,es,dx,cx,bx,ax;
#else
    WORD es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    WORD ip,cs,flags;
    } REG_PARAMS;
	
/**********************************************************************/
    
#define BUF_SIZE    (0xFFFF / sizeof(REQUEST))
#define MAX_STR     40
#define MAX_REQU    (BUF_SIZE-1)

#define SET_NEXT()  { if (next==MAX_REQU) next=0; else next++; }

typedef struct {
    int type;
    REG_PARAMS r;
    char taskname[9];
    char dsdx[MAX_STR], dssi[MAX_STR];
    int flag;
    } REQUEST, FAR * LPREQUEST;

#define REQU_FREE       'f'     /* request block not in use */
#define REQU_USE        'u'     /* in use */

#define BEGIN_CRIT_SEC()        _asm cli
#define END_CRIT_SEC()          _asm sti

#define IS_REQUEST(buf) \
    (buf[next].type != REQU_FREE)
        
#define BUFFER_FULL(buf) \
    IS_REQUEST(buf)
    
/**********************************************************************/
    
#ifndef MK_FP       
#define MK_FP(a,b)  ((void far *)(((DWORD)(a) << 16) | (b)))
#endif

#define INT21           ' '
#define KERNELDOSPROC   '*'

#define COPY_DSDX       1
#define COPY_DSSI       2
#define CHECK_AL        3
#define FIRST_FUNC      0x3b
#define LAST_FUNC       0xa1

// Instructions for WRT_REQUEST on how to handle each INT 21h function
static char copyflags[LAST_FUNC - FIRST_FUNC] = { 
    /*3b*/ COPY_DSDX, COPY_DSDX, COPY_DSDX,
    /*3e*/ 0, 0, 0, 0, 0,
    /*43*/ COPY_DSDX,
    /*44*/ 0, 0, 0, 0, 0, 0, 0,
    /*4b*/ COPY_DSDX,
    /*4c*/ 0, 0, 
    /*4e*/ COPY_DSDX, 
    /*4f*/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    /*5a*/ COPY_DSDX, COPY_DSDX,
    /*5c*/ 0, 0, 0, 0,
    /*60*/ COPY_DSSI,
    /*61*/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*6c*/ COPY_DSSI,
    /*6d*/ 0, 0, 0, 0,
    /*71*/ CHECK_AL,
    /*72*/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    /*80*/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*90*/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*a0*/ COPY_DSDX,
    } ;

static int show_all = 0;
static DWORD calls = 0, lost = 0, already = 0, mine = 0;

#define sel_size(sel)       (GetSelectorLimit(sel)+1)

WORD verr(WORD short sel)
{   _asm {
        mov ax, 1
        verr word ptr sel
        je short okay
        dec ax
    }
    okay:;  
}

static REQUEST far *requ;

char far *GetTaskName(HANDLE hTask)
{
    static char none[2] = "";
    if (verr(hTask) && (sel_size(hTask) > (0xf2+8)))
        return (char far *) MK_FP(hTask, 0xf2);
    else
        return none;
}

int WRT_REQUEST(REQUEST far *buf, REG_PARAMS far *pr, HANDLE task, int flag)
{
    static int next = 0;
    unsigned char ah;
    void far *fp;

    if (BUFFER_FULL(buf))
        return 0;
    BEGIN_CRIT_SEC();
    _fstrncpy(&buf[next].taskname, GetTaskName(task), 8);
    buf[next].taskname[8] = '\0';
    _fmemcpy(&buf[next].r, pr, sizeof(REG_PARAMS));
    buf[next].type = REQU_USE;
    
#if 1
    #define MAYBE_COPY(str, r1, r2) { \
        _fmemset(str, 0, MAX_STR); \
        if (verr(r1) && (sel_size(r1) > r2)) \
            _fmemcpy(str, MK_FP(r1, r2), min(MAX_STR-1, sel_size(r1)-r2)); \
    }
#else
    // IsBadStringPtr causes GP faults which KRNL386 catches
    // But this seems to cause unexpected page faults in Win32s!
    #define MAYBE_COPY(str, r1, r2) { \
        _fmemset(str, 0, MAX_STR); \
        fp = MK_FP(r1, r2); \
        if (! IsBadStringPtr(fp, 256)) \
            _fstrncpy(str, fp, min(_fstrlen(fp), MAX_STR-1)); \
        }
#endif      

    ah = pr->ax >> 8;
    if (ah >= FIRST_FUNC && ah <= LAST_FUNC)
    {
        ah -= FIRST_FUNC;
    top:
        if (copyflags[ah] == COPY_DSDX)
            MAYBE_COPY(&buf[next].dsdx, pr->ds, pr->dx) //;
        else if (copyflags[ah] == COPY_DSSI)
            MAYBE_COPY(&buf[next].dssi, pr->ds, pr->si) //;
        else if (copyflags[ah] == CHECK_AL)
        {
            ah = (pr->ax & 0xff) - FIRST_FUNC;
            if (copyflags[ah] != CHECK_AL)
                goto top;
        }
    }
    
    buf[next].flag = flag;
    SET_NEXT();
    END_CRIT_SEC();
    PostMessage(__hMainWnd, WM_NULL, 0, 0); // just to keep moving!
    return 1;
}

int RD_REQUEST(REQUEST far *buf, REG_PARAMS *pr, LPSTR ptaskname, 
    LPSTR lpdsdx, LPSTR lpdssi, int far *pflag)
{
    static int next = 0;
    if (! IS_REQUEST(buf))
        return 0;
    BEGIN_CRIT_SEC();
    _fmemcpy(pr, &buf[next].r, sizeof(REG_PARAMS));
    _fmemcpy(ptaskname, &buf[next].taskname, 9);
    _fstrncpy(lpdsdx, &buf[next].dsdx, MAX_STR);
    _fstrncpy(lpdssi, &buf[next].dssi, MAX_STR);
    *pflag = buf[next].flag;
    buf[next].type = REQU_FREE;
    END_CRIT_SEC();
    SET_NEXT();
    return 1;
}

REQUEST far *INIT_REQUBUF(void)
{
    LPREQUEST buf, p;
    int i;
    if (! (buf =
        (LPREQUEST) GlobalLock(GlobalAlloc(GMEM_FIXED | GMEM_ZEROINIT, 
        BUF_SIZE * sizeof(REQUEST)))))
        return 0;
    for (i=BUF_SIZE, p=buf; i--; p++)
        p->type = REQU_FREE;
    return buf;
}
    
FARPROC (FAR PASCAL *GetSetKernelDosProc)(FARPROC DosProc) = 0;

typedef void (_interrupt _far *INTRFUNC)(); 

INTRFUNC _dpmi_get_pmode_vect(int intno)
{
    INTRFUNC iv;
    _asm {
        mov ax, 0204h
        mov bl, byte ptr intno
        int 31h
        jc error
        mov word ptr iv+2, cx
        mov word ptr iv, dx
    }
    return iv;
error:
    return (INTRFUNC) 0;
}

void _dpmi_set_pmode_vect(int intno, INTRFUNC iv)
{   _asm {
        mov ax, 0205h
        mov bl, byte ptr intno
        mov cx, word ptr iv+2
        mov dx, word ptr iv
        int 31h
    }
    // _asm jc error
}

#define get_vect(intno)     _dpmi_get_pmode_vect(intno)

int set_vect(WORD intno, INTRFUNC handler)
{
	// Don't use INT 21h function 25h (Set Interrupt Vector), because
	// KERNEL's INT 21h handler prevents setting INT 1Bh, 1Ch, 21h, 24h.
	// So use DPMI instead.  However, could also use undoc NoHookDOSCall, 
	// which bypasses most of KERNEL's special INT 21h handling.
    _dpmi_set_pmode_vect(intno, handler);
    return (_dpmi_get_pmode_vect(intno) == handler);
}

void _interrupt _far IntHandler(REG_PARAMS r);
void _interrupt _far IntHandler2(REG_PARAMS r);

static HANDLE wspy21_task;
static int intno = 0x21;    // INT 21h default 
static INTRFUNC old = 0;
static FARPROC old_dos = (FARPROC) 0;

void on_close(HWND hwnd)
{
    int i;

    winio_warn(FALSE, "WSPY21",
        "%08lu\tintercepted\n"
        "%08lu\talready seen in first handler\n"
        "%08lu\tabsorbed in first handler\n"
        "%08lu\tlost through buffer overruns\n"
        "%08lu\twere my own calls\n",
            calls, already, (intno == 0x21) ? calls - already : 0, 
            lost, mine);
    
    if (! set_vect(intno, old))
        winio_warn(FALSE, "WSPY21", "Couldn't restore Int %02Xh", intno);
    if ((intno == 0x21) && old_dos)
        GetSetKernelDosProc(old_dos);
    }

void display_int(REG_PARAMS *pr)
{
    printf("AX=%04x BX=%04x CX=%04x DX=%04x DS=%04x SI=%04x DI=%04x "
           "CS:IP=%04x:%04xh",
        pr->ax, pr->bx, pr->cx, pr->dx, 
        pr->ds, pr->si, pr->di, pr->cs, pr->ip);
}
    
#define PRINT_STR(s)            printf("%s", (s))
#define PRINT_2_STR(s1, s2)     printf("%s \'%s\'", (s1), (s2))
#define PRINT_STR_WORD(s, w)    printf("%s %u (%04Xh)", (s), (w), (w))
#define PRINT_2_WORD(s, w1, w2) printf("%s %u (%04Xh), %u (%04Xh)", \
                                    (s), (w1), (w1), (w2), (w2))
#define PRINT_STR_BYTE(s, b)    printf("%s %u (%02Xh)", (s), (b), (b))
#define PRINT_STR_FP(s, fp)     printf("%s %Fp", (s), (fp))

void print_ioctl(REG_PARAMS *pr)
{
    int al = pr->ax & 0xFF;
    int bl = pr->bx & 0xFF;
    printf("IOCTL (%02X) ", al);
    switch (al)
    {
        case 0x00: PRINT_STR_WORD("Get Device Data", pr->bx); break;
        case 0x01: PRINT_STR_WORD("Set Device Data", pr->bx); break;
        case 0x02: PRINT_STR_WORD("Get Char Dev Data", pr->bx); break;
        case 0x03: PRINT_STR_WORD("Set Char Dev Data", pr->bx); break;
        case 0x04: PRINT_STR_BYTE("Get Block Dev Data", bl); break;
        case 0x05: PRINT_STR_BYTE("Set Block Dev Data", bl); break;
        case 0x06: PRINT_STR_WORD("Check Input Status", pr->bx); break;
        case 0x07: PRINT_STR_WORD("Check Output Status", pr->bx); break;
        case 0x08: PRINT_STR_BYTE("Is Drv Removeable?", bl); break;
        case 0x09: PRINT_STR_BYTE("Is Drv Remote?", bl); break;
        case 0x0A: PRINT_STR_WORD("Is File/Dev Remote?", pr->bx); break;
        case 0x0C: PRINT_STR_BYTE("Char Minor Code", pr->cx & 0xFF); break;
        case 0x0D: printf("(%02X) ", pr->cx & 0xFF); 
            switch (pr->cx & 0xFF) {
            case 0x40: PRINT_STR_WORD("Set Dev Params", pr->bx); break;
            case 0x41: PRINT_STR_WORD("Write Track", pr->bx); break;
            case 0x42: PRINT_STR_WORD("Format Track", pr->bx); break;
            case 0x46: PRINT_STR_WORD("Set Media ID", pr->bx); break;
            case 0x47: PRINT_STR_WORD("Set Access Flag", pr->bx); break;
            case 0x49: PRINT_STR_WORD("Eject", pr->bx); break;
            case 0x4A: PRINT_STR_WORD("Lock Logical Vol", pr->bx); break;
            case 0x4B: PRINT_STR_WORD("Lock Phys Vol", pr->bx); break;
            case 0x60: PRINT_STR_WORD("Get Dev Params", pr->bx); break;
            case 0x61: PRINT_STR_WORD("Read Track", pr->bx); break;
            case 0x62: PRINT_STR_WORD("Verify Track", pr->bx); break;
            case 0x66: PRINT_STR_WORD("Get Media ID", pr->bx); break;
            case 0x67: PRINT_STR_WORD("Get Access Flag", pr->bx); break;
            case 0x68: PRINT_STR_WORD("Sense Media Flag", pr->bx); break;
            case 0x6A: PRINT_STR_WORD("Unlock Logical Vol", pr->bx); break;
            case 0x6B: PRINT_STR_WORD("Unlock Phys Vol", pr->bx); break;
            case 0x6C: PRINT_STR_BYTE("Write Poll", bl); break;
            case 0x6D: PRINT_STR_WORD("Enum Open Files", pr->bx); break;
            case 0x6E: PRINT_STR_WORD("Find Swap File", pr->bx); break;
            }
            break;
        case 0x0E: PRINT_STR_BYTE("Get Log Drv Map", bl); break;
        case 0x0F: PRINT_STR_BYTE("Set Log Drv Map", bl); break;
    }
}

void display_dos_int(REG_PARAMS *pr, char *dsdx, char *dssi)
{
    int ah;
top:                // IDoBelieveIUsedAGoToStatement
    ah = pr->ax >> 8;
    printf("(%02X) ", ah);
    switch (ah)     // ICantBelieveIUsedASwitchStatement
    {
        case 0x00:  PRINT_STR("Exit"); break;
        case 0x0E:  PRINT_STR_BYTE("Set Disk", pr->dx & 0xFF); break;
        case 0x19:  PRINT_STR("Get Disk"); break;
        case 0x1A:  PRINT_STR_FP("Set DTA", MK_FP(pr->ds, pr->dx)); break;
        case 0x25:  PRINT_STR_BYTE("Set Vect", pr->ax & 0xFF); break;
        case 0x2A:  PRINT_STR("Get Date"); break;
        case 0x2C:  PRINT_STR("Get Time"); break;
        case 0x2F:  PRINT_STR("Get DTA"); break;
        case 0x30:  PRINT_STR("Get DOS Vers"); break;
        case 0x32:  PRINT_STR_BYTE("Get DPB", pr->dx & 0xFF); break;
        case 0x35:  PRINT_STR_BYTE("Get Vect", pr->ax & 0xFF); break;
        case 0x36:  PRINT_STR_BYTE("Get Disk Space", pr->dx & 0xFF); break;
        case 0x3B:  PRINT_2_STR("ChDir", dsdx); break;
        case 0x3C:  PRINT_2_STR("Create", dsdx); break;
        case 0x3D:  PRINT_2_STR("Open", dsdx); break;
        case 0x3E:  PRINT_STR_WORD("Close", pr->bx); break;  
        case 0x3F:  PRINT_2_WORD("Read", pr->bx, pr->cx); break;
        case 0x40:  PRINT_2_WORD("Write", pr->bx, pr->cx); break;
        case 0x41:  PRINT_2_STR("Delete", dsdx); break;
        case 0x42:  printf("Lseek%d %u %04x%04x", 
                    pr->ax & 0xFF, pr->bx, pr->cx, pr->dx);
                    break;
        case 0x43:  PRINT_2_STR("Get/Set File Attr", dsdx); break;
        case 0x44:  print_ioctl(pr); break;
        case 0x47:  PRINT_STR_BYTE("Get Curr Dir", pr->dx & 0xff); break;
        case 0x4b:  PRINT_2_STR("Exec", dsdx); break;
        case 0x4c:  PRINT_STR_BYTE("Exit", pr->ax & 0xff); break;
        case 0x4e:  PRINT_2_STR("Find First", dsdx); break;
        case 0x4f:  PRINT_STR("Find Next"); break;
        case 0x50:  PRINT_STR_WORD("Set PSP", pr->bx); break;
        case 0x51: case 0x62: PRINT_STR("Get PSP"); break;
        case 0x55:  PRINT_2_WORD("(Undoc) Create PSP", pr->dx, pr->si); break;
        case 0x57:  PRINT_STR_WORD("Get/Set File Date/Time", pr->bx); break;
        case 0x59:  PRINT_STR("Get Extended Error Info"); break;
        case 0x5A:  PRINT_2_STR("Create Temp File", dsdx); break;
        case 0x5B:  PRINT_2_STR("Create New File", dsdx); break;
        case 0x60:  PRINT_2_STR("Truename", dssi); break;
        case 0x65:  PRINT_STR_BYTE("International", pr->ax & 0xFF); break;
        case 0x68:  PRINT_STR_WORD("Commit", pr->bx); break;
        case 0x6C:  PRINT_2_STR("Open/Create", dssi); break;
        // for 21/71XX, do what IFSMgr does:  mark as an LFN function
        // and then turn into 21/XX.  
        case 0x71:  printf("LFN "); pr->ax <<= 8; goto top;
        case 0xA0:  PRINT_2_STR("Get Volume Info", dsdx); break;
        case 0xA1:  PRINT_STR("Find Close"); break;
        default:    display_int(pr);
        }
}

void display_ints(void)
{
    REG_PARAMS r;
    static char modname[16];
    static char dsdx[MAX_STR];
    static char dssi[MAX_STR];
    int flag;
    
    // This loop looks for, extracts, and displays messages queued
    // up by the interrupt handler. The only way we know to 
    // terminate the function is that the window closes
    while (winio_openwindows()) 
    { 
        wmhandler_yield(); 
        while (RD_REQUEST(requ, &r, &modname, &dsdx, &dssi, &flag))
		{
			printf("<%s> %c", modname, flag);
			if (intno == 0x21)
				display_dos_int(&r, dsdx, dssi);
			else    // Could do formatted display of other INT like DPMI
				display_int(&r);
			printf("\n");
		}
    }
}

main(int argc, char *argv[])
{
    char buf[128];
    char num[8];
    
    if (__hPrevInst)
        fail("Sorry, only one WSPY21 at a time");
    winio_setbufsize(__hMainWnd, (WORD) 32768, TRUE);
    if (! (requ = INIT_REQUBUF()))
        fail("INIT_REQUBUF fail!");
    winio_about("WSPY21");
    winio_onclose(__hMainWnd, (DESTROY_FUNC) on_close);
    wspy21_task = GetCurrentTask();
	
	if (argc < 2)
		intno = 0x21;
	else
	{
		int i;
		for (i=1; i<argc; i++)
			if (argv[i][0] == '-')
			{
				if (strncmp(strupr(&argv[i][1]), "SHOWALL", 4) == 0)
					show_all++;
				else
					fail("usage: wspy21 [-showall] [intno]");
			}
			else
				sscanf(argv[i], "%X", &intno);
	}
    
    old = get_vect(intno);
    if (! set_vect(intno, (INTRFUNC) IntHandler))
        fail("SetVect failed!");

    if ((intno == 0x21) && (GetSetKernelDosProc = GetProcAddress(
            GetModuleHandle("KERNEL"), "GETSETKERNELDOSPROC")))
        old_dos = GetSetKernelDosProc((FARPROC) IntHandler2);
	
    sprintf(buf, "Windows Interrupt Spy: INT %02Xh", intno);
    winio_settitle(__hMainWnd, buf);
    
    display_ints();
    return 0;
}
    
static REG_PARAMS prev_r = {0} ;
static int was_prev = 0;

// CS:IP can be different
#define SAME_CALL(r1, r2)  \
    (((r1)->es == (r2)->es) && ((r1)->ds == (r2)->ds) && \
    ((r1)->di == (r2)->di) &&  ((r1)->si == (r2)->si) && \
    ((r1)->bx == (r2)->bx) &&  ((r1)->dx == (r2)->dx) && \
    ((r1)->cx == (r2)->cx) &&  ((r1)->ax == (r2)->ax))
    
void _interrupt _far IntHandler(REG_PARAMS r)
{
    HANDLE task = GetCurrentTask();
    
	if (! show_all) { prev_r = r; was_prev = 1; }
        
    if (task != wspy21_task)     /* don't show my own ints */
    {
        if (! WRT_REQUEST(requ, &r, task, INT21))
            lost++;     /* buffer overflow */
        calls++;
    }
    else
        mine++;
        
    _chain_intr(old);
}
    
void _interrupt _far IntHandler2(REG_PARAMS r)
{
    HANDLE task = GetCurrentTask();

	if ((! show_all) && was_prev-- && SAME_CALL(&r, &prev_r))
	{
		already++;	
		goto done;	/* saw this already in IntHandler */
	}
    
    if (task == wspy21_task)    /* don't show my own ints */
    {
        mine++;
        goto done;
    }
    
    if (! WRT_REQUEST(requ, &r, task, KERNELDOSPROC))
        lost++;     /* buffer overflow */
    calls++;

done:       
    _chain_intr((INTRFUNC) old_dos);
}
