/*
WVMWALK.C -- Version of VMWALK for Windows 3.1 
Andrew Schulman, April 1993
From "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

Borland C++ (EasyWin): bcc -DTESTING -W -2 wvmwalk.c prot.c
Microsoft C (QuickWin): cl -DTESTING -Mq -G2 wvmwalk.c prot.c
*/

#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include "windows.h"
#include "prot.h"
#include "vmwalk.h"

// see W/DDJ, December 1992 (this is Paul Bonneau's doing!)
// see also MSJ, February 1993, p. 30
#define VM_FROM_HWND(hWndDosBox) \
    *((LPDWORD) MK_FP(GetWindowWord(hWndDosBox, GWW_HINSTANCE), 0x0FC))

// Offset 68h in VM CB holds ptr to next VM;
// same as VMM function Get_Next_VM_Handle().
#define MY_GET_NEXT_VM_HANDLE(vm_cb)  \
	*((DWORD far *) (((BYTE far *) vm_cb)+0x68))

#if 1
// since undoc, missing from some import libraries, so do run-time link
static BOOL (FAR PASCAL *IsWinOldApTask)(HANDLE hTask) = 0;
#else
#ifdef __cplusplus
extern "C" BOOL FAR PASCAL IsWinOldApTask(HANDLE hTask);
#else
extern BOOL FAR PASCAL IsWinOldApTask(HANDLE hTask);
#endif
#endif

// from Undocumented Windows, p. 304
BOOL IsDosBox(HWND hWnd)
{
	if (! IsWinOldApTask) // one-time init
		if (! (IsWinOldApTask = (BOOL (FAR PASCAL *)(HANDLE)) 
			GetProcAddress(GetModuleHandle("KERNEL"), "ISWINOLDAPTASK")))
				fail("Can't find KERNEL.IsWinOldApTask");
    return (*IsWinOldApTask)(GetWindowTask(hWnd));
}

typedef BOOL (*WWALKFUNC)(HWND hwnd, char *title, DWORD vm, VM_CB far *vm_cb);

int wvmwalk(WWALKFUNC walkfunc)
{
    char buf[0x50];
    VM_CB dummy;
    HWND hwnd;
    int num_vm;
    
    if (((WORD) GetVersion()) != 0x0A03)
        return 0;       // only works in Windows 3.1
    if (! (GetWinFlags() & WF_ENHANCED))
        return 0;       // only works in Enhanced mode
    
    hwnd = GetActiveWindow();
    GetWindowText(hwnd, buf, 0x50);

    // Can't get handle for System VM, but that is VM the Windows
    // app is running in!  So make a dummy VM_CB and pass that:
    memset(&dummy, 0, sizeof(dummy));
    dummy.CB_High_Linear = 0;   // current VM!
    dummy.CB_VMID = 1;          // System VM is always VM 1
    if (! (*walkfunc)(hwnd, buf, 0, &dummy))
        return 0;
    num_vm = 1;

    hwnd = GetWindow(hwnd, GW_HWNDFIRST);
    while (hwnd)
    {
        if (IsDosBox(hwnd) && GetWindowText(hwnd, buf, 0x50))
        {
            DWORD vm = VM_FROM_HWND(hwnd);
            VM_CB my_vm_cb;
            VM_CB far *vm_cb = (VM_CB far *) map_linear(vm, sizeof(VM_CB));
			if (! vm_cb) fail("can't map VM");
            _fmemcpy(&my_vm_cb, vm_cb, sizeof(VM_CB));
            free_mapped_linear(vm_cb);
            if (! (*walkfunc)(hwnd, buf, vm, &my_vm_cb))
                return num_vm;
			set_base(0);	// in case app fooled with it
			// could use MY_GET_NEXT_VM_HANDLE() to do loop over VMs, 
			// given just one (any) VM handle.
        }

        hwnd = GetWindow(hwnd, GW_HWNDNEXT);
    }
    return num_vm;
}

#ifdef TESTING
#include <stdio.h>

void fail(const char *s, ...) { puts(s); exit(1); }

BOOL wwalkfunc(HWND hwnd, char *title, DWORD vm, VM_CB far *vm_cb)
{
    printf("%04Xh\t%08lXh\t\"%s\"\n", hwnd, vm, title);
    printf("   VM #%lu   st=%08lXh   lin=%08lXh   crs=%08lXh\n\n",
        vm_cb->CB_VMID,
        vm_cb->CB_VM_Status,
        vm_cb->CB_High_Linear,
        vm_cb->CB_Client_Pointer);
    return 1;
}

main()
{
	#define PROGNAME	"WVMWALK"
	#define DESC		"Walk Windows Enhanced mode virtual machines (VMs)"
	#define AUTHOR		"Andrew Schulman"
	#include			"..\banner"
	
	return wvmwalk(wwalkfunc);
}
#endif
