/*
XLASTDRV.C
Andrew Schulman, March 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

Straightforward clone of QEMM LASTDRIV.COM utility, except:
-- Checks for presence of Windows Enhanced mode (DOSMGR hangs onto original
   CDS pointer from SysVars); won't run in DOS box
-- Uses sleazy hack to try to free up previously-created XLASTDRV CDSes
-- Uses "instant TSR" technique from Prosise UMBFILES
-- Provides -RESTORE command to put back original CDS pointer and LASTDRIVE
   from SysVars; uses sleazy hack entry to save these away.  This necessary
   for MSCDEX and other utilities and hold onto original SysVars pointers.
Not such a straightforward clone after all!    
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <dos.h>
#ifdef __TURBOC__
#include <alloc.h>
#include <dir.h>
#else
#include <direct.h>
#endif

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

#ifndef min
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif

#ifndef MK_FP
#define MK_FP(seg,ofs) \
    ((void far *)(((DWORD)(seg) << 16) | (ofs)))
#endif

typedef struct {
    BYTE zero;
    BYTE far *orig_cds;
    BYTE orig_lastdrv;
    BYTE signature[];
    } SLEAZY_HACK;
    
#define GET_SLEAZY_HACK(cds, ld)    \
    ((SLEAZY_HACK far *) ((cds)+(((ld)-1)*sizeof_cds_entry)))

#define XLASTDRV_SIGN      "XLASTDRV sleazy hack"

void set_genuine_dos_vers(void);
BYTE far *get_sysvars(void);
void set_mcb_owner(WORD para, WORD owner);
void set_mcb_name(WORD para, char far *name);
int is_win3e(void);
int get_lastdrive_doc(void);

void fail(char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    SLEAZY_HACK far *old_hack, far *new_hack;
    BYTE far *sysvars;
    BYTE far *old_cds, far *new_cds;
    unsigned int para;
    int lastdrv_ofs, cdsptr_ofs;
    int new_lastdrv, old_lastdrv;
    int sizeof_cds_entry;
    int do_restore = 0;
    
    if (argc < 2)
        fail("xlastdrv [new value for lastdrv: A-Z] or -restore");
    
    if (is_win3e())
        fail("Sorry, xlastdrv can't run under Windows Enhanced mode");
    
    // get new LASTDRV (or -RESTORE) from command line
    if (strcmp(strupr(argv[1]), "-RESTORE") == 0)
        do_restore = 1;
    else
        new_lastdrv = 1 + (toupper(argv[1][0]) - 'A');

    // possibly change _osmajor, _osminor; see if ok
    set_genuine_dos_vers();
    if ((_osmajor < 3) || (_osmajor >= 10))
        fail("Unsupported DOS version");

    // get offsets within SysVars
    if ((_osmajor == 3) && (_osminor == 0))
    {
        cdsptr_ofs = 0x17;
        lastdrv_ofs = 0x1b;
    }
    else
    {
        cdsptr_ofs = 0x16;
        lastdrv_ofs = 0x21;
    }
    
    // old_lastdrv = sysvars[lastdrv_ofs];
    if (! (sysvars = get_sysvars()))
        fail("Can't get SysVars");
    old_lastdrv = sysvars[lastdrv_ofs];
    
    // see if SysVars looks ok
    // get_lastdrive_doc() is wrong for Novell:  see okay.c
    if (get_lastdrive_doc() != old_lastdrv)
        fail("Internal DOS data structures wrong?");
    if (old_lastdrv == 0xFF)
        fail("LASTDRIVE not supported");
    
    // old_cds = sysvars[cdsptr_ofs];
    old_cds = *((BYTE far * far *) &sysvars[cdsptr_ofs]);
    
    // sizeof_cds_entry depends on DOS version;
    if (_osmajor == 3)
        sizeof_cds_entry = 0x51;
    else
        sizeof_cds_entry = 0x58;
    
    if (do_restore)
    {
        old_hack = GET_SLEAZY_HACK(old_cds, old_lastdrv);
        if (_fstrcmp(old_hack->signature, XLASTDRV_SIGN) != 0)
            fail("Can't restore original CDS!");
        _disable();
        _fmemcpy(old_hack->orig_cds, old_cds,
            min(old_lastdrv, old_hack->orig_lastdrv) * sizeof_cds_entry);
        *((BYTE far * far *) &sysvars[cdsptr_ofs]) = old_hack->orig_cds;
        sysvars[lastdrv_ofs] = old_hack->orig_lastdrv;
        _enable();
        printf("Restored CDS to %Fp (LASTDRIVE=%c)\n",
            old_hack->orig_cds,
            (old_hack->orig_lastdrv - 1) + 'A');
        goto free_old;
    }
    
    if (new_lastdrv != old_lastdrv)
        printf("Changing LASTDRIVE from %c to %c\n",
            (old_lastdrv - 1) + 'A',
            (new_lastdrv - 1) + 'A');
    
    // temporarily bump up new_lasdtrv for signature entry
    new_lastdrv++;

    // new_cds = allocate(new_lastdrv * sizeof_cds_entry);
    if (_dos_allocmem(1+((new_lastdrv * sizeof_cds_entry) >> 4), &para) != 0)
        fail("Insufficient memory");
    new_cds = (BYTE far *) MK_FP(para, 0);
    
#if 1
    // This taken from Jeff Prosise, UMBFILES (PC MAG, Nov. 26, 1991)
    // Instant TSR!  (MEM /D thinks XLASTDRV is TSR)
    set_mcb_owner(para, para);
    set_mcb_name(para, "XLASTDRV");
#else
    // This is the tricky part:  have to allocate so that memory stays
    // around when program exits.  But, don't need a TSR.  So mark as
    // belonging to DOS. Unfortunately, comes up in MEM.EXE as IO, not MSDOS.
    set_mcb_owner(para, 8);
#endif
    
    // mark newly-formed CDS entries as invalid drives;
    _fmemset(new_cds, 0, new_lastdrv * sizeof_cds_entry);

    printf("Moving CDS from %Fp to %Fp\n", old_cds, new_cds);
    
    // disable interrupts;
    _disable();

    // copy old CDS entries over to new CDS table
    _fmemcpy(new_cds, old_cds, 
        sizeof_cds_entry * min(new_lastdrv, old_lastdrv));

    // put back actual new LASTDRIVE value requested by user
    new_lastdrv--;
    
    // use extra entry for signature, and to keep original SysVars to restore
    new_hack = GET_SLEAZY_HACK(new_cds, new_lastdrv);
    old_hack = GET_SLEAZY_HACK(old_cds, old_lastdrv);
    _fstrcpy(new_hack->signature, XLASTDRV_SIGN);
    if (_fstrcmp(old_hack->signature, XLASTDRV_SIGN) == 0)
    {
        // have done XLASTDRV before; copy orig values from old to new
        new_hack->orig_cds = old_hack->orig_cds;
        new_hack->orig_lastdrv = old_hack->orig_lastdrv;
    }
    else
    {
        // this is first time we've run XLASTDRV
        new_hack->orig_cds = old_cds;
        new_hack->orig_lastdrv = old_lastdrv;
    }

    // plug new values into SysVars!!
    *((BYTE far * far *) &sysvars[cdsptr_ofs]) = new_cds;
    sysvars[lastdrv_ofs] = new_lastdrv;
    
    _enable();
    
    // hope that no one has a dangling invalid pointer to old_cds!
    // Windows does, MSCDEX does!

free_old:       
    // if we allocated old one, we can free it
    if (_fstrcmp(old_hack->signature, XLASTDRV_SIGN) == 0)
    {
        if (_dos_freemem(FP_SEG(old_cds)) != 0)
            fail("Couldn't free up old CDS");
        else
            printf("Freed old XLASTDRV CDS at %Fp\n", old_cds);
    }
    
    return 0;
}

void set_genuine_dos_vers(void)
{
    union REGS r;
    r.x.ax = 0x3306;
    r.x.bx = 0;
    intdos(&r, &r);
    if (r.x.bx != 0)
    {
        _osmajor = r.h.bl;
        _osminor = r.h.bh;
    }
}

BYTE far *get_sysvars(void)
{
    union REGS r;
    struct SREGS s;
    r.h.ah = 0x52;
    segread(&s);
    s.es = r.x.bx = 0;
    intdosx(&r, &r, &s);
    return (BYTE far *) MK_FP(s.es, r.x.bx);
}

void set_mcb_owner(WORD para, WORD owner)
{
    *((WORD far *) MK_FP(para-1, 1)) = owner;
}

void set_mcb_name(WORD para, char far *name)
{
    _fmemcpy(MK_FP(para-1, 8), name, 8);
}

int is_win3e(void)
{
    union REGS r;
    r.x.ax = 0x1600;
    int86(0x2f, &r, &r);
    if ((r.h.al == 1) || (r.h.al == 0xFF))
        return 1;
    else 
        return ((r.h.al != 0) && (r.h.al != 0x80));
}

int get_lastdrive_doc(void)
{
#ifdef __TURBOC__
    return setdisk(0xFF);
#else
    unsigned lastdrv_doc;
    _dos_setdrive(0xFF, &lastdrv_doc);
    return lastdrv_doc;
#endif
}

