/*
copybios.c -- copy video bios to conv mem, TSRify, patch pointer
Andrew Schulman, April 1996
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef void far *FP;
typedef struct { BYTE type; WORD owner, size; } MCB;

void fail(const char *s) { puts(s); exit(1); }

typedef struct { 
    int intno; 
    WORD far *pprev_seg;
    WORD sig, far *sig_loc, prev_seg;
    } HOOK_INFO;

#define CALL_FAR    0x1EFF
#define JMP_FAR     0x2EFF

HOOK_INFO hook_info[] = {
    0x10, 0x0cc40024L, JMP_FAR, 0x0cc40297L, 0xc000,
    } ;

int rom_to_ram(WORD romseg)
{
    HOOK_INFO *info;
    WORD myseg;
    MCB far *fpmcb;
    FP fp;

    // check that configuration hasn't changed
    info = &hook_info[0];
    if (info->prev_seg != romseg) return 0;
    if (*(info->sig_loc) != info->sig) return 0;
    if (*(info->pprev_seg) != romseg) return 0;
    
    // allocate 64k block of RAM
    if (_dos_allocmem(4096, &myseg) != 0)
        return 0;
    fp = MK_FP(myseg, 0);
    
    // copy ROM to RAM
    _fmemcpy(fp, MK_FP(romseg, 0), 0xffff);

    // patch INT vector
    _disable();
    *(info->pprev_seg) = myseg;
    _enable();

    // TSRify it
    fpmcb = (MCB far *) MK_FP(myseg-1, 0);
    fpmcb->owner = 8;  // fake TSR

    return 1;
}

main(int argc, char *argv[])
{
    puts( rom_to_ram(0xC000) ? "Success" : "Failure" );
}
