
; FAKEDOS7  V1.04  (C) 1997 Caldera Inc.  All rights reserved
;
; Written by Matthias Paul
; eMail: <MPaul@calderauk.com>, <Matthias.Paul@post.rwth-aachen.de>
;
; ************************************************************************
; *   Copyright Unpublished Work of Caldera, Inc. All Rights Reserved.   *
; *                                                                      *
; *   THIS WORK IS A COPYRIGHT WORK AND CONTAINS CONFIDENTIAL,           *
; *   PROPRIETARY AND TRADE SECRET INFORMATION OF CALDERA, INC.          *
; *   ACCESS TO THIS WORK IS RESTRICTED TO (I) CALDERA, INC. EMPLOYEES   *
; *   WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF       *
; *   THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN CALDERA, INC. WHO   *
; *   HAVE ACCEPTED THE CALDERA OPENDOS SOURCE LICENSE OR OTHER CALDERA  *
; *   LICENSE AGREEMENTS. EXCEPT UNDER THE EXPRESS TERMS OF THE CALDERA  *
; *   LICENSE AGREEMENT NO PART OF THIS WORK MAY BE USED, PRACTICED,     *
; *   PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED, TRANSLATED,     *
; *   ABRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,*
; *   TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF        *
; *   CALDERA, INC. ANY USE OR EXPLOITATION OF THIS WORK WITHOUT         *
; *   AUTHORIZATION COULD SUBJECT THE PERPETRATOR TO CRIMINAL AND        *
; *   CIVIL LIABILITY.                                                   *
; ************************************************************************
;
; Preliminary experimental TSR for Caldera OpenDOS attempting to fake
; MS-DOS 7 so far, that Windows95 will start on top of it. (Still crude...)
;
; This version boots the whole way up into the GUI, but than the GUI freezes
; after a few seconds (Ctrl+Alt+Del displays Blue Screen "System busy").
; At this stage the mouse is still moving, and the lower layers still take
; keyboard input (e.g. to get the Blue Screen), but keyclicks and mouse
; clicks are not recognized by the GUI any more.
;
; Requirements:
;
; Will only install on Caldera OpenDOS, Novell DOS, PC-DOS 5.0+,
; or MS-DOS 5.0 - 6.22. (Only tested on OpenDOS 7.02 BETA2, so far.)
; (Would most probably also run with DR PalmDOS and DR DOS 6.0
; "business update 1993".)
;
; Excerpt of OpenDOS CONFIG.SYS (example):
;
;  DEVICE=c:\windows\himem.sys
;  REM Only required with USE_CURRENT_STATE eq 2:
;  REM DEVICE=c:\opendos\setver.exe (fake MS-DOS 7.xx version for FAKEDOS7.COM)
;  ?INSTALL=c:\prj\fakedos\fakedos7.com
;  ?DEVICE=c:\windows\ifshlp.sys
;  SHELL=c:\opendos\command.com c:\opendos\ /p
;
; Last edit: 1997-12-17
;
; History:
;
; 97-11-20  MPAUL  n/a   Initial version of FAKEDOS7.COM.
; 97-11-3?  MPAUL  1.00  Developed INT21h/30xxh, INT21h/3001h, INT21h/3306h,
;                        INT2Fh/1611h, INT2Fh/1613, INT2Fh/1614h, and
;                        INT2Fh/4A33h interceptors.
;                        (INT2Fh/1614h and INT2Fh/4A33h are not actually
;                        needed here...)
; 97-12-11  MPAUL  1.01  Tidy-up's, some version checks.
;                        We now take (some) valuable info of the
;                        version calls from the running system, rather than
;                        using hard-wired data. (Nice idea taken from Andrew
;                        Schulman's CALDERA.EXE.)
;                        INT2Fh/1612h faked for VTD.VXD which otherwise fails
;                        fatally. This could be dangerous, because we don't
;                        actually support it. However, we come a big leap
;                        further right into Protected Mode...
; 97-12-12  MPAUL  1.02  Corrected NEW_FLAGS_DEFAULT for USE_CURRENT_STATE eq 0.
;                        INT2Fh/1612h does not point to a function but to 15
;                        bytes of data (Thanks, Andrew!), which we now also
;                        fake (might be obsolete). Added pseudo code for
;                        i2F1612_internal_func which looks somewhat similar
;                        to our Int13Trap from IBMBIO.COM's INIT.ASM.
; 97-12-13  MPAUL        Cosmetics. Moved i2F1612_data_struc below
;                        i2F1612_return_struc in memory, and changed DOS_seg
;                        default value from 0 to C9h.
; 97-12-15  MPAUL        Added INT2Fh/1690h implementation, but the current
;                        faked data does not work. A full implementation needs
;                        changes in our IBMBIO.COM to log certain device driver
;                        specific data in an internal structure. But up to
;                        now, I deciffered the purpose for only one of the
;                        pointer types as used for the sub-entries...
; 97-12-16  MPAUL  1.03  To come further added stub implementation for
;                        INT2Fh/1607h for VKD and changed INT2Fh/1690h
;                        structure to just fake a zero structure at the
;                        moment (taken from Andrew's CALDERA.EXE), but now
;                        setting the data offset to i2F1612_050B. Removed
;                        REROUTE1612 conditional define, now always re-routing
;                        it. Rearranged the different structures to get a
;                        better impression of the actual memory image under
;                        MS-DOS. This version now boots up the whole way into
;                        GUI (including login to NetWare), but the keyboard/
;                        system freezes after a few seconds, mouse is still
;                        working. A message box complains about IFSHLP, which
;                        would decrease system performance. (Ctrl+Alt+Del
;                        produces Blue Screen "System busy").
; 97-12-17  MPAUL  1.04  Added USER_IDs to more easily switch between different
;                        configurations (still to lazy to implement parameter
;                        evaluation). The TSR now frees its environment. Shrunk
;                        registry filename buffer. Added USE_CURRENT_STATE eq 2.
;                        (The current minimal implementation as TSR takes about
;                        600 bytes, which probably could be further reduced if
;                        put into the kernel.)
; 97-12-18  MPAUL        Located the corresponding function in IO.SYS and added
;                        its pseudo code, but enabling this doesn't work so
;                        far... (Could be, that the returned offset is actually
;                        used in segment 70h, overwriting something???)
; 97-12-19  MPAUL        Andrew found out, that the code excerpt is part of
;                        an INT16h handler (very interesting, since comparing
;                        the fixed offsets in INIT.ASM, I thought I saw a call
;                        to the strategy entry point of disk IO in there, which
;                        didn't make sense...). However, FAKEDOS7 became less
;                        predictable after adding this stuff.

; Macros:
Jmps macro address
jmp short address
endm

; Equates:
LF                    equ 10
CR                    equ 13
JMPF_Instr            equ 0EAh
NOP_Instr             equ 90h

DOS_VER_MAX           equ 0616h                     ; max. supported DOS is 6.22
DOS_VER_MIN           equ 0500h                     ; min. supported DOS is 5.0
                                                    ; 031Fh = 3.31 (for DR PalmDOS & DR DOS 6.0)
CPM_VER_MIN           equ 72h                       ; min. supported BDOS is 7.2
                                                    ; 70h = DR PalmDOS
                                                    ; 71h = DR DOS 6.0 "business update 1993"
                                                    ; 72h = Novell DOS and OpenDOS 7.01
                                                    ; 73h = Caldera OpenDOS 7.02
VKD_ID                equ 000Dh                     ; VxD ID for virtual keyboard (VKD)
OEM_ID                equ 0FFh                      ; Microsoft OEM ID for version check

; User configurable settings (hard-wired at the moment): ----------------------
USER_ID               equ 10                        ; 0    = default
                                                    ; 1..9 = reserved
                                                    ; 10   = my Pentium II with 4DOS 6.00
                                                    ; 11   = my Pentium II with COMMAND.COM
                                                    ; > 100 = free
if USER_ID eq 10      ; my Pentium II (OpenDOS 7.02 & 4DOS 6.00 & Windows95 (7.10))
USE_CURRENT_STATE     equ 1                         ; 0 = use internal data
                                                    ; 1 = take current status
                                                    ; 2 = also use SETVER
REAL_VER              equ 0A07h                     ; this is for MS-DOS 7.10 (4.00.1111 REV A)
REAL_VER_STR          equ <"7.10">                  ;              ""
REGISTRY_FILENAME     equ <"C:\WINDOWS\SYSTEM.DAT">
SHELL_FILENAME        equ <"C:\SYS\SHELL\4DOS.600\4DOS.COM">
SHELL_PARAMETERS      equ <"C:\SYS\SHELL\4DOS.600\ @C:\SYS\SHELL\4DOS.600\4DOS.INI /E:3000 /P">

elseif USER_ID eq 11  ; my Pentium II (OpenDOS 7.02 & Windows95 (7.10))
USE_CURRENT_STATE     equ 1                         ; 0 = use internal data
                                                    ; 1 = take current status
                                                    ; 2 = also use SETVER
REAL_VER              equ 0A07h                     ; this is for MS-DOS 7.10 (4.00.1111 REV A)
REAL_VER_STR          equ <"7.10">                  ;              ""
REGISTRY_FILENAME     equ <"C:\WINDOWS\SYSTEM.DAT">
SHELL_FILENAME        equ <"C:\DOS\OPENDOS.702\COMMAND.COM">
SHELL_PARAMETERS      equ <"C:\DOS\OPENDOS.702\ /E:4000 /P">

else                  ; Default (OpenDOS 7.02 & Windows95 (7.0))
USE_CURRENT_STATE     equ 2                         ; 0 = use internal data
                                                    ; 1 = take current status
                                                    ; 2 = also use SETVER

;REAL_VER             equ 0007h                     ; this is for MS-DOS 7.0
;REAL_VER_STR         equ <"7.0">                   ;              ""
REGISTRY_FILENAME     equ <"C:\WINDOWS\SYSTEM.DAT">
SHELL_FILENAME        equ <"C:\OPENDOS\COMMAND.COM">
SHELL_PARAMETERS      equ <"C:\OPENDOS\ /P">

endif
; End of user configurable settings -------------------------------------------

if USE_CURRENT_STATE  eq 0
NEW_FLAGS_DEFAULT     equ 1000h                     ; 0000h = DOS low
else                                                ; 1000h = DOS in HMA
NEW_FLAGS_DEFAULT     equ 0
endif

OBSOLETE              equ 1                         ; 0 = minimal implementation
                                                    ;     (without optimization)
                                                    ; 1 = include additional
                                                    ;     obsolete code

if OBSOLETE eq 1
REG_PATH_LEN          equ 4Eh
endif

Code SEGMENT
                      ASSUME cs:code, ds:code
org 002Ch
env_seg:              dw    ?                       ; pointer to env. segment

;org 0080h            ; command line parameter

org 0100h                                           ; start of .COM program

fakedos7:             jmp   install                 ; goto installer routine


                      ; #######################################################
                      ; ###   Data area   #####################################
                      ; #######################################################


                      ; =======================================================
                      ; Location unknown (indirectly used via API):


registry_path         db    REGISTRY_FILENAME,0
if OBSOLETE eq 1
registry_def_path_end db    (REG_PATH_LEN + registry_path - registry_def_path_end) dup (0)
endif

                      ; =======================================================
                      ; Located in low BIOS/DOS code (global memory):

INT16_entry:
if OBSOLETE eq 1
                      ; <code entry point> CS=70h
                      ; Andrew says, this must be some kind of INT16h handler...
                      ;
                      ; However, this is the code:
                      ; entry_point:
                      ; sti
                      ; pushf
                      ; test byte ptr cs:[0009h],20h
                      ;  jnz skip_code
                      ; cli
                      ; call dword ptr cs:[0174h] ; (F000:E82E)
                      ;                           ; strat-entry of IO.SYS
                      ;                           ; disk driver???
                      ;                           ; (IntDisk in our INIT.ASM)
endif
i2F1607_return_struc: ; Located at 0070h:043Dh in MS-DOS 7:
                      ;---- start code signature A "dd 0EFF2E9Ch" + "050Bh"
                      pushf       ; db 9Ch
                      dec   word ptr cs:i2F1612_050B ; dec cs:["050Bh"]
                                  ; db 2Eh, 0FFh, 0Eh
                                  ; dw offset i2F1612_050B ; (050Bh in MS-DOS 7)
                      ;---- end code signature A + "050Bh"
                      ;---- start code signature B "dd 9D900175h"
                      jnz   skip_nop; db 75h, 01h
                      nop         ; db 90h (ES:DI+8) patched with V86 breakpoint:
                                  ; arpl [di+02CAh],bx
skip_nop:                         ; skip breakpoint if "050Bh" is zero
                      popf        ; db 9Dh
                      ;---- end code signature B
if OBSOLETE eq 1
                      ; retf 2
                      ; skip_code:
                      ; call unknown_function     ; db 0E8h, 39h, 00h
                      ; call dword ptr cs:[0154h] ; appears to be the entry
                      ;                           ; point to another device
                      ;                           ; driver in IO.SYS
                      ; jmps entry_point
endif

                      db    JMPF_Instr              ; chain to previous handler
int16_ofs             dw    0000                    ; JMPF seg:ofs
int16_seg             dw    0000

i2F1612_050B:         dw    0 ; moving this to i2F1612_data_struc below
                              ; doesn't work any longer...

                      ; -------------------------------------------------------
                      ; Located in low BIOS/DOS data:


i2F1612_data_struc:   ; Located at 0070h:0501h in MS-DOS 7:
                      ; N.B. These are the values found here on my machine.
                      ;      However, they differ from machine to machine.
                      ;      The startup values are stored in the IO.SYS file
                      ;      at file pos +0D01h (in my copy).
                      dw    199Ah ; days since 1980-01-01
                      dw    0     ; ???
                      dw    0     ; ticks [sec/100] (I'm not sure)??? or
                                  ; event counter (upwards) for idle???
                      db    00h,  00h, 0C8h,  00h
;i2F1612_050B: ; this doesn't work any longer, moved to above
                      dw    0     ; Once changed every boot, could be start
                                  ; time? Will be counted down in some way.
                                  ; But since some time it does not
                                  ; change any more.
                                  ; However, the transition from 1 to 0 is
                                  ; a special case, since it bypasses the
                                  ; V86 breakpoint.
                      db    0     ; unknown, always zero???
DOS_seg:              dw    0C9h  ; segment value of 1st DOS segment in IO.SYS
                                  ; (containing NUL, and 5 handles)
                                  ; Usually it's C9h, but sometimes it's 70h.

if OBSOLETE eq 1
                      ; The following code might be obsolete here...

i2F1612_internal_func:int 3       ; give control to debugger
                      db  1Dh dup (NOP_Instr); patch area...
                      retf        ; return

                      ; This excerpt is the original code from MS-DOS 7.10' IO.SYS:
                      ; Partially it looks quite close to our BIOS' Int13Trap code
                      ; in INIT.ASM. Maybe it has something to do with direct
                      ; disk access, 32bit disk access, or drive locking?
                      ;
                      ; mov     ds,word ptr cs:DOS_seg ; get DS for IO.SYS
                      ; lea     bp,[bp+00]
                      ; pushf
                      ; cls                             ; simulate INT13h through
                      ; call    far dword ptr ds:[0B4h] ; BIOS' I13Pointer
                      ; mov     word ptr cs:DOS_seg
                      ; push    cs
                      ; pop     ds                      ; align DS
                      ; pushf
                      ; call    A20Enable
                      ; popf
                      ; lea     bp,[bp+00]
                      ; retf                            ; return
endif


                      ; =======================================================
                      ; Apparently located in DOS data (can be loaded high)


                      ; Located at FFFFh:xxxx (about 1xxxh or 2xxxh)
i2F1612_return_struc  dw    1     ; ID or size indicator???
                      dw    offset i2F1612_data_struc; (0501h in MS-DOS 7)
bios_func_seg         dw    0     ; runtime fixup (0070h in MS-DOS 7)

                      ; about 70h bytes following to reach i2F1690_return_struc...


                      ; -------------------------------------------------------


                      ; Located at FFFFh:xxxxh (about 1xxxh or 2xxxh)
i2F1690_return_struc  dw    offset i2F1690_data_struc
bios_high_seg         dw    0     ; runtime fixup (FFFFh if MS-DOS 7 in HMA)

if OBSOLETE eq 1
                      dw    3 dup (0) ; unknown data
                      ; In MS-DOS IO.SYS this is followed by the path to
                      ; the shell (SHELL=), followed by it's parameters...
endif


                      ; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


shell_exec            db    SHELL_FILENAME,0
shell_param           db    (shell_param_end - shell_param_start)
shell_param_start     db    SHELL_PARAMETERS,0
shell_param_end:


                      ; -------------------------------------------------------


i2F1690_data_struc:   ; Located at FFFFh:xxxxh (might be at end of the
                      ; used HMA area, so that it can grow during boot...)
                      ; Maybe this info is stored here during IO.SYS'
                      ; CONFIG.SYS processing, also using the INT2Fh/1690h
                      ; pointer. Then we could easily code this by
                      ; implementing a callout API in IBMBIO.COM, received
                      ; by an external device driver...

                      ; Header:
                      db    7     ; ID or compatibility level???
                      dd    0     ; unknown (always zero)
                      db    1     ; count of following entries

                      ; 1st entry (only one in this case):
                      db    "12345678123" ; SPACE padded filename which
                                  ; contains loaded device drivers
                      db    2 dup (0); unknown???
                      db    0     ; count of block device drivers in this file
                      db    1     ; count of following sub-entries

                      ; 1st sub-entry (only one in this case):
                      db    0     ; type of entry (must not be 2 or 3 here)
                                  ; Maybe for instancing parts of the driver,
                                  ; saving special driver attributes, or to
                                  ; store info to dynamically deinstall/bypass
                                  ; drivers...
                                  ;
                                  ; These sub-entries must not necessarily
                                  ; be in ascending order...
                                  ; type 00h: (no occurence)
                                  ; type 01h: pointer to device driver header
                                  ;           (first sub-entry for all entries)
                                  ; type 02h: pointer to code entry in driver???
                                  ;           (some dispatcher) (IO.SYS, EMM386.EXE)
                                  ; type 03h: ??? (IO.SYS)
                                  ; type 04h: ??? (IO.SYS, EMM386.EXE)
                                  ; type 05h: pointer to code entry in driver???
                                  ;           (some dispatcher) (IO.SYS)
                                  ; type 06h: (no occurence)
                                  ; type 07h: (no occurence)
                                  ; type 08h: 0000:0000 (for several drivers) or
                                  ;           some kind of counter, maybe forced
                                  ;           loading/unloading order???
                                  ;           (last sub-entry for all entries)
                                  ; type 09h: ??? (HIMEM.SYS, S-ICE.EXE)
                                  ; type 0Ah: ??? (Interesting... That's my
                                  ;                FreeKEYB.COM/K3PLUS.COM...)
                                  ;
                      dd    0     ; pointer to item (0 = unused?)

                      ; 2nd, 3rd sub-entries were following here...

                      ; 2nd, 3rd entries, each containing a various number
                      ;      of sub-entries, were following here...


                      ; #######################################################
                      ; ###  INT21h interceptor code  #########################
                      ; #######################################################


INT21_entry:          pushf                         ; INT21h ISR routine:
                      cmp   ax,3001h                
                       je   INT21_3001              ; Get DOS version and flags
                      cmp   ah,30h                  
                       je   INT21_3000              ; Get DOS version and OEM
                      cmp   ax,3306h                
                       je   INT21_3306              ; Get real DOS version
                      popf

                      db    JMPF_Instr              ; chain to previous handler
int21_ofs             dw    0000                    ; JMPF seg:ofs
int21_seg             dw    0000


INT21_3000:           ; Get DOS version and OEM number:
                      ; ===============================
                      ; N.B. Since MS-DOS 7+ this is not only entered
                      ;      with AL=00h, but with AL<>01h.
                      ;      VMM32.VXD calls it with 308Dh, and one of
                      ;      it's VxDs calls it with 30CCh.
                      popf
dos_ver1:             mov   ax,REAL_VER
dos_oem_id:           mov   bh,OEM_ID               ; default: OEM "Microsoft"
                      jmps  INT21_30_exit


INT21_3001:           ; Get DOS version number and flags:
                      ; =================================
                      popf
dos_ver2:             mov   ax,REAL_VER
old_flags:            mov   bh,0                    ; traditional version flags
                                                    ; bit3 = DOS in ROM


INT21_30_exit:
serno_high:           mov   bl,0                    ; 24bit serial number (high)
serno_low:            mov   cx,0                    ; (low word)
                      jmps  INT21_exit


INT21_3306:           ; Get real DOS version number:
                      ; ============================
                      popf
dos_ver3:             mov   bx,REAL_VER
new_flags:            mov   dx,NEW_FLAGS_DEFAULT    ; 'new' revision and
                                                    ;  version flags
                                                    ; DL = revision
                                                    ; DH bit3 = DOS in ROM
                                                    ; DH bit4 = DOS in HMA
INT21_exit:           retf 2


                      ; #######################################################
                      ; ###  INT2Fh interceptor code  #########################
                      ; #######################################################


INT2F_entry:          pushf                         ; INT 2Fh ISR routine:
if OBSOLETE eq 1
                      cmp   ax,4A33h                ; MS-DOS 7 version check
                       je   INT2F_4A33
endif
                      cmp   ah,16h                  ; strip off anything but
                       jne  int2F_chain             ; Windows/DOS interface
                      cmp   al,07h                  ; Virtual device callout
                       je   INT2F_1607              ; (for VKD here)
                      cmp   al,11h                  
                       je   INT2F_1611              ; Get SHELL parameters
                      cmp   al,12h                  
                       je   INT2F_1612              ; Get BIOS data
                      cmp   al,13h                  
                       je   INT2F_1613              ; Get Registry path
if OBSOLETE eq 1
                      cmp   al,14h                  
                       je   INT2F_1614              ; Set Registry path
endif
                      cmp   al,90h                  
                       jne  int2F_chain
                      jmp   INT2F_1690              ; Get/Set MS-DOS 7 kernel data
int2F_chain:          popf

                      db    JMPF_Instr              ; chain to previous handler
int2F_ofs             dw    0000                    ;  JMPF seg:ofs
int2F_seg             dw    0000


INT2F_1607:           ; Virtual Device callout API (for VKD here):
                      ; ==========================================
                      cmp   bx,VKD_ID              ; is it for us?
                       jne  int2F_chain            ; no, so continue to chain
                      ; What about CX=function number???
                      push  cs
                      pop   es                     ; return ES:DI -> struc
                      mov   di, offset i2F1607_return_struc
                      jmps  INT2F_exit

INT2F_1611:           ; Get SHELL parameters:
                      ; =====================
if OBSOLETE eq 1
INT2F_4A33:           ; MS-DOS 7 version check:
                      ; =======================
endif
                      ; N.B. Both functions are identical

                      mov   ax,0                   ; believe what you want
                      push  cs                     ;  and fake MS-DOS 7+
                      pop   ds                     ; DS:DX -> shell_exec ASCIZ
                      mov   dx,offset shell_exec
                      mov   si,offset shell_param  ; DS:SI -> shell_param PAS-ASCII
                      mov   bx,0                   ; zero


INT2F_exit:           popf
                      retf 2


INT2F_1612:           ; Get BIOS data:
                      ; ==============
                      push  cs
                      pop   es                     ; return ES:BX -> struc
                      mov   bx,offset i2F1612_return_struc
                      mov   ax,0                   ; return "function implemented"
if OBSOLETE eq 1
                      ; A potential call to i2F1612_internal_func would crash
                      ; the system (never a good idea to play around with
                      ; INT 13h!). That's why I added code to reroute it
                      ; to a debugger, just for the case...
endif
                      jmps  INT2F_exit


INT2F_1613:           ; Get Registry ASCIZ path:
                      ; ========================
                      ; On entry: ES:DI, CX=buffer size
                      ; On exit:  CX=length (incl. Z)
                      ;           AX=0
                      push  ds
                      push  si
                      push  di
                      push  bx

                      cmp   cx,0
                       je   INT2F_1613_skip

                      push  cs
                      pop   ds
                      cld                          ; count up
                      mov   si,offset registry_path
                      mov   bx,0                   ; preset counter

INT2F_1613_loop:      inc   bx
                      lodsb                        ; AL <- DS:SI, INC SI
                      stosb                        ; AL -> ES:DI, INC DI
                      cmp   al,0                   ; EOL reached?
                      loopnz INT2F_1613_loop       ; DEC CX, AL=0 | end of buffer?

                      mov   cx,bx                  ; copied bytes from BX -> CX
INT2F_1613_skip:
                      mov   ax,0                   ; return "function implemented"

                      pop   bx
                      pop   di
                      pop   si
                      pop   ds

                      jmps  INT2F_exit

if OBSOLETE eq 1
INT2F_1614:           ; Set Registry ASCIZ path:
                      ; ========================
                      ; On exit: ES:DI -> ASCIZ path
                      ;          AX=0     ok
                      ;          AX=1614h unsupported
                      ;          AX=004Eh failed, max buf size
                      push  ds
                      push  es
                      push  si
                      push  di
                      push  bx
                      push  cx

                      cld                          ; count up
                      mov   al,0                   ; search for EOL
                      mov   cx,REG_PATH_LEN
                      mov   si,di

INT2F_1614_loop:      inc   bx
                      scasb                        ; AL=ES:DI?, INC DI
                      loopnz INT2F_1614_loop       ; buffer full
                      jnz   INT2F_1614_full

                      ;end found, BX=copy this many bytes (at least 1)
                      push  es                     ; DS:SI are source now
                      pop   ds
                      push  cs
                      pop   es
                      mov   di, offset registry_path; ES:DI is target now
                      mov   cx,bx
                      rep   movsb                  ; copy DS:SI -> ES:DI (CX times)

                      mov   ax,0                   ; return "success"
                      jmps  INT2F_1614_end

INT2F_1614_full:      mov   ax,REG_PATH_LEN

INT2F_1614_end:       pop   cx
                      pop   bx
                      pop   di
                      pop   si
                      pop   es
                      pop   ds

                      jmps  INT2F_exit
endif

INT2F_1690:           ; Get/Set MS-DOS 7 kernel data
                      ; ============================
                      ; On exit: ES:BX -> structure containing a pointer
                      ;                   to device driver list structure
                      push  cs
                      pop   es                     ; return ES:BX -> struc
                      mov   bx,offset i2F1690_return_struc
                      popf                         ; get back the flags
                      clc                          ; clear carry flag (necessary???)
                      retf 2


                      ; #######################################################
                      ; ###  Init & TSR installer code  #######################
                      ; #######################################################
install:              ; Init:
                      ; =====
                      mov   dx, offset Startup_Msg ; output startup message
                      mov   ah,9
                      int   21h

                      push  cs
                      pop   ds                     ; set DS

                      stc
                      mov   ax,4452h
                      int   21h                    ; determine DR DOS version
                       jc   check_dos
                      cmp   ax,(10h*256+CPM_VER_MIN)
                       jae  check_dos
                      jmp   wrong_CPM

check_dos:            mov   ax,3000h               ; get old DOS version and OEM
                      int   21h
                      xchg  ah,al
                      cmp   ax,DOS_VER_MIN
                       jb   wrong_DOS_10           ; filter out DOS below 5.0
                      cmp   bh,0                   ; is it PC-DOS?
                       je   dos_ok
                      cmp   ax,DOS_VER_MAX
                       jbe  dos_ok
wrong_DOS_10:         jmp   wrong_DOS              ; don't run on MS-DOS above 6.22
dos_ok:

if USE_CURRENT_STATE ge 1
                      ;we always leave OEM ID Microsoft
                      ;mov  byte ptr dos_oem_id+1,bh; store OEM ID

                      mov   byte ptr serno_high+1,bl; store high byte of serno
                      mov   word ptr serno_low+1,cx; store low word of serno

                      mov   ax,3001h               ; get old DOS version and flags
                      int   21h
                      mov   byte ptr old_flags+1,bh; store current state

                      ;mov  byte ptr serno_high+1,bl; we assume this was
                      ;mov  word ptr serno_low+1,cx ; always equal...

if USE_CURRENT_STATE eq 2
                      mov   word ptr dos_ver1+1,ax ; patch to current DOS
                      mov   word ptr dos_ver2+1,ax ; version (needs SETVER to
                      mov   word ptr dos_ver3+1,ax ; be installed)
endif
                      mov   ax,3306h               ; get old real DOS version
                      int   21h                    ; but only to get DX
                      mov   word ptr new_flags+1,dx; store current state
endif

                      mov   word ptr bios_func_seg,cs ; runtime fixup
                      mov   word ptr bios_high_seg,cs ; runtime fixup

                      ; Install:
                      ; ========
                      cld                          ; (obsolete)
                      cli                          ; enter mutex

                      mov   ax,3516h               ; install own INT16h handler
                      int   21h                    ; and store original handler
                      mov   ds:int16_ofs,bx
                      mov   ds:int16_seg,es
                      mov   dx,offset INT16_entry
                      mov   ax,2516h
                      int   21h

                      mov   ax,3521h               ; install own INT21h handler
                      int   21h                    ; and store original handler
                      mov   ds:int21_ofs,bx
                      mov   ds:int21_seg,es
                      mov   dx,offset INT21_entry
                      mov   ax,2521h
                      int   21h

                      push  cs
                      pop   ds
                      mov   ax,352Fh               ; install own INT2Fh handler
                      int   21h                    ; and store original handler
                      mov   ds:int2F_ofs,bx
                      mov   ds:int2F_seg,es
                      mov   dx,offset INT2F_entry
                      mov   ax,252Fh
                      int   21h

                      sti                          ; leave mutex

                      ; Make resident:
                      ; ==============
                      mov   dx,offset Install_Msg  ; display message
                      mov   ah,9
                      int   21h

                      mov   ax,word ptr ds:env_seg ; free environment block
                      push  ax
                      pop   es
                      mov   ah,49h
                      int   21h                    ; make zero since it's
                      mov   word ptr ds:env_seg,0  ; invalid now

                      mov   dx,offset install      ; terminate & stay resident
                      int   27h                    ; (keep 0..start resident)
                                                   ; never comes here...

                      ; Error exit:
                      ; ===========
wrong_DOS:            mov   dx,offset Wrong_DOS_Msg
                      jmps  error_exit
wrong_CPM:            mov   dx,offset Wrong_CPM_Msg

error_exit:           mov   ah,9                   ; display error message
                      int   21h

                      mov   ah,0
                      int   21h

                      ; #######################################################

Startup_Msg           db CR, LF
                      db "FAKEDOS7 V1.04  Glue to run the Windows95 GUI on top of OpenDOS [ALPHA]",CR,LF
                      db "Copyright (c) 1997 Caldera, Inc.  All rights reserved.",CR,LF,'$'
Install_Msg           db CR,LF
                      db "INT 16h, INT 21h, and INT 2Fh interceptors installed to fake "
if USE_CURRENT_STATE eq 0
                      db "MS-DOS ", REAL_VER_STR
elseif USE_CURRENT_STATE eq 1
                      db "DOS ", REAL_VER_STR
else
                      db "Windows95"
endif
                      db ".",CR,LF
if USE_CURRENT_STATE eq 2
                      db "(SETVER required to fake DOS version.)",CR,LF
endif
                      db '$'
Wrong_DOS_Msg         db CR,LF
                      db "Sorry, but at present this DOS configuration is unsupported.",CR,LF,'$'
Wrong_CPM_Msg         db CR,LF
                      db "Sorry, but at present only Caldera OpenDOS and Novell DOS are supported!",CR,LF,'$'
Code ends
end FAKEDOS7
