;
; WINGLUE  V1.05 [ALPHA] (C) 1997 Caldera Inc.  All rights reserved
;
; Written by Matthias Paul
; eMail: <MPaul@calderauk.com>, <Matthias.Paul@post.rwth-aachen.de>
;
; ************************************************************************
; *   Copyright Unpublished Work of Caldera, Inc. All Rights Reserved.   *
; *                                                                      *
; *   THIS WORK IS A COPYRIGHT WORK AND CONTAINS CONFIDENTIAL,           *
; *   PROPRIETARY AND TRADE SECRET INFORMATION OF CALDERA, INC.          *
; *   ACCESS TO THIS WORK IS RESTRICTED TO (I) CALDERA, INC. EMPLOYEES   *
; *   WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF       *
; *   THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN CALDERA, INC. WHO   *
; *   HAVE ACCEPTED THE CALDERA OPENDOS SOURCE LICENSE OR OTHER CALDERA  *
; *   LICENSE AGREEMENTS. EXCEPT UNDER THE EXPRESS TERMS OF THE CALDERA  *
; *   LICENSE AGREEMENT NO PART OF THIS WORK MAY BE USED, PRACTICED,     *
; *   PERFORMED, COPIED, DISTRIBUTED, REVISED, MODIFIED, TRANSLATED,     *
; *   ABRIDGED, CONDENSED, EXPANDED, COLLECTED, COMPILED, LINKED, RECAST,*
; *   TRANSFORMED OR ADAPTED WITHOUT THE PRIOR WRITTEN CONSENT OF        *
; *   CALDERA, INC. ANY USE OR EXPLOITATION OF THIS WORK WITHOUT         *
; *   AUTHORIZATION COULD SUBJECT THE PERPETRATOR TO CRIMINAL AND        *
; *   CIVIL LIABILITY.                                                   *
; ************************************************************************
;
; Preliminary experimental TSR for Caldera OpenDOS attempting to fake
; MS-DOS 7 so far, that Windows95 will start on top of it. (Still crude...)
;
; This version boots the whole way up into the GUI, but than the GUI freezes
; after a few seconds (Ctrl+Alt+Del displays Blue Screen "System busy").
; At this stage the mouse is still moving, and the lower layers still take
; keyboard input (e.g. to get the Blue Screen), but keyclicks and mouse
; clicks are not recognized by the GUI any more.
;
; Requirements:
;
; By default it will only install on Caldera OpenDOS, Novell DOS, PC-DOS 5.0+,
; or MS-DOS 5.0 - 6.22. (Only tested on OpenDOS 7.02 BETA2, so far. Would most
; probably also run with DR PalmDOS and DR DOS 6.0 "business update 1993",
; which both have a genuine CDS.)
;
; Excerpt of OpenDOS CONFIG.SYS (example):
;
;  DEVICE=c:\windows\himem.sys
;  ?DEVICE=c:\opendos\setver.exe (fake MS-DOS 7.xx version for WINGLUE.COM)
;  ?INSTALL=c:\prj\fakedos\winglue.com /G:7.xx /R:C:\WINDOWS\SYSTEM.DAT
;  ?DEVICE=c:\windows\ifshlp.sys
;  SHELL=c:\opendos\command.com c:\opendos\ /p
;
; Last edit: 1997-12-30
;
; History:
;
; 97-11-20  MPAUL  n/a   Initial version of "FAKEDOS7" (.ASM/.COM).
; 97-11-3?  MPAUL  1.00  Developed INT21h/30xxh, INT21h/3001h, INT21h/3306h,
;                        INT2Fh/1611h, INT2Fh/1613h, INT2Fh/1614h, and
;                        INT2Fh/4A33h interceptors.
;                        (INT2Fh/1614h and INT2Fh/4A33h are not actually
;                        needed here...)
; 97-12-11  MPAUL  1.01  Tidy-up's, some version checks.
;                        We now take (some) valuable info of the
;                        version calls from the running system, rather than
;                        using hard-wired data. (Nice idea taken from Andrew
;                        Schulman's CALDERA.EXE.)
;                        INT2Fh/1612h faked for VTD.VXD which otherwise fails
;                        fatally. This could be dangerous, because we don't
;                        actually support it. However, we come a big leap
;                        further right into Protected Mode...
; 97-12-12  MPAUL  1.02  Corrected NEW_FLAGS_DEFAULT for USE_CURRENT_STATE eq FALSE.
;                        INT2Fh/1612h does not point to a function but to 15
;                        bytes of data (Thanks, Andrew!), which we now also
;                        fake (might be obsolete). Added pseudo code for
;                        i2F1612_internal_func which looks somewhat similar
;                        to our Int13Trap from IBMBIO.COM's INIT.ASM.
; 97-12-13  MPAUL        Cosmetics. Moved i2F1612_data_struc below
;                        i2F1612_return_struc in memory, and changed DOS_seg
;                        default value from 0 to C9h.
; 97-12-15  MPAUL        Added INT2Fh/1690h implementation, but the current
;                        faked data does not work. A full implementation needs
;                        changes in our IBMBIO.COM to log certain device driver
;                        specific data in an internal structure. But up to
;                        now, I deciffered the purpose for only one of the
;                        pointer types as used for the sub-entries...
; 97-12-16  MPAUL  1.03  To come further added stub implementation for
;                        INT2Fh/1607h for VKD and changed INT2Fh/1690h
;                        structure to just fake a zero structure at the
;                        moment (taken from Andrew's CALDERA.EXE), but now
;                        setting the data offset to i2F1612_050B. Removed
;                        REROUTE1612 conditional define, now always re-routing
;                        it. Rearranged the different structures to get a
;                        better impression of the actual memory image under
;                        MS-DOS. This version now boots up the whole way into
;                        GUI (including login to NetWare), but the keyboard/
;                        system freezes after a few seconds, mouse is still
;                        working. A message box complains about IFSHLP, which
;                        would decrease system performance. (Ctrl+Alt+Del
;                        produces Blue Screen "System busy").
; 97-12-17  MPAUL  1.04  Added USER_IDs to more easily switch between different
;                        configurations (still to lazy to implement parameter
;                        evaluation). The TSR now frees its environment. Shrunk
;                        registry filename buffer. Added USE_CURRENT_STATE eq 2.
;                        (The current minimal implementation as TSR takes about
;                        600 bytes, which probably could be further reduced if
;                        put into the kernel.)
; 97-12-18  MPAUL        Located the corresponding function in IO.SYS and added
;                        its pseudo code, but enabling this doesn't work so
;                        far... (Could be, that the returned offset is actually
;                        used in segment 70h, overwriting something???)
; 97-12-19  MPAUL        Andrew found out, that the code excerpt is part of
;                        an INT16h handler (very interesting, since comparing
;                        the fixed offsets in INIT.ASM, I thought I saw a call
;                        to the strategy entry point of disk IO in there, which
;                        didn't make sense...). However, FAKEDOS7 became less
;                        predictable after adding this stuff.
; 97-12-21  MPAUL  1.05  Added IFSHLP.SYS to 1690h structure... (apparently
;                        no functional change).
;                        Fixed DR DOS multiuser version check.
; 97-12-22  MPAUL        Reformatted the source. Fixed INT2Fh/1614h.
;                        Added command line parameters (/F, /G, /R, /H, /?),
;                        SwitChar, and status messages. Removed USER_ID,
;                        USE_CURRENT_STATE, and NEW_FLAGS_DEFAULT.
; 97-12-29  MPAUL        Andrew found partial documentation for INT2Fh/1690h
;                        in the Win95 DDK. Great! Hence improved comments now
;                        using DDK names where applicable. This proved we were
;                        already on the right track and helped alot to much
;                        further decode the RMD API structures. Relocated TSR
;                        code before data for future enhancements. Changes in
;                        the faked INT2Fh/1690h data. Renamed TSR to "WINGLUE"
;                        (.ASM/.COM) which IMHO better describes its function.

FALSE           equ 0
TRUE            equ not FALSE

; User configurable settings (still hard-wired at the moment): ----------------

SHELL_FILENAME    equ <"C:\OPENDOS\COMMAND.COM">
SHELL_PARAMETERS  equ <"C:\OPENDOS\ /P">
;SHELL_FILENAME   equ <"C:\SYS\SHELL\4DOS.600\4DOS.COM">
;SHELL_PARAMETERS equ <"C:\SYS\SHELL\4DOS.600\ @C:\SYS\SHELL\4DOS.600\4DOS.INI /E:3000 /P">
PARSE_S         equ FALSE               ; include /S: parameter (NYI)

; Internal default settings: --------------------------------------------------

REG_FILENAME    equ <"C:\WINDOWS\SYSTEM.DAT">
REAL_VER        equ 0007h               ; 0007h = MS-DOS 7 (Windows95)
REAL_VER_STR    equ <"7.00">            ; 0A07h = MS-DOS 7.10 (4.00.1111 REV A)

OBSOLETE        equ TRUE                ; FALSE = minimal implementation
                                        ;         (with some optimization)
                                        ; TRUE  = include additional and
                                        ;         apparently obsolete code

TSR_NAME        equ <"WINGLUE">         ; alternative name: "OPENWIN "
TSR_FILENAME    equ <"WINGLUE ">        ; old name:         "FAKEDOS7"
TSR_VER         equ <"V1.05">           ; version

; Equates and macros: ---------------------------------------------------------

Jmps    macro   address                 ; Shortcut for short jumps
jmp     short   address
endm

DWO             equ dword ptr           ;
WO              equ word ptr            ;
BY              equ byte ptr            ;

                                        ; Special chars:
NUL             equ 0                   ; Null character
TAB             equ 9                   ; Tabulator
LF              equ 10                  ; Line feed
CR              equ 13                  ; Carriage return
SPACE           equ ' '                 ; Space
SWC             equ '/'                 ; Default SwitChar
SWC_REP         equ '%'                 ; SwitChar placeholder in help screens

                                        ; Opcodes:
JMPF_Instr      equ 0EAh                ; Jump far
NOP_Instr       equ 90h                 ; No operation

                                        ; Interrupts:
KBD_INT         equ 16h                 ; Keyboard ROS
DOS_INT         equ 21h                 ; DOS API
TERM_INT        equ 27h                 ; DOS Terminate
MPX_INT         equ 2Fh                 ; Multiplexer

                                        ; DOS_INT API functions:
CPM_TERM        equ 0                   ; Old CP/M like terminate
WRITE_CHAR      equ 2                   ; Write char to CON:
WRITE_STR       equ 9                   ; Write '$' terminated string to CON:
SET_VECT        equ 25h                 ; Set interrupt vector
GET_VER         equ 30h                 ; Get DOS version
GET_REAL_VER    equ 3306h               ; Get real DOS version
GET_VECT        equ 35h                 ; Get interrupt vector
GET_SWITCHAR    equ 3700h               ; Get SwitChar
GET_CPM_VER     equ 4452h               ; Get DR DOS BDOS version
FREE_MEM        equ 49h                 ; Free memory

                                        ; These are the Win95 DDK RMD.H names
                                        ; for the INT2Fh/1690h chain entries
                                        ; for RMDs (Real Mode Drivers)
                                        ; (all entries starting with
                                        ; "rmd_header_t")
;RMD_TYPE_PORT          equ 0           ; for "rmd_port_t"??? chain entry
;RMD_TYPE_ASPI          equ 1           ; for "rmd_aspi_t" chain entry
;RMD_TYPE_INT13         equ 2           ; for "rmd_int13_t" chain entry
;RMD_TYPE_CAM           equ 3           ; for "rmd_cam_t" chain entry
;RMD_TYPE_INT4B         equ 4           ; for "rmd_int4B_t" chain entry
;RMD_TYPE_BLOCK         equ 5           ; for "rmd_block_t" chain entry
;RMD_TYPE_SCSIPORT      equ 6           ; for "rmd_port_t"??? chain entry
RMD_TYPE_IOSYS          equ 7           ; for "rmd_bios_t" chain entry
                                        ; [A "rmd_bios_t" chain entry is the
                                        ; one and only entry encountered so
                                        ; far...]

                                        ; These are the Win95 DDK names for
                                        ; RMD_TYPE_IOSYS pointer types. However,
;MAX_RMDI_TYPES equ 10                  ; their actual purpose/usage/internals
                                        ; are not documented.
                                        ;v-- * indicates functionality, that
                                        ;      is now almost proved to be very
                                        ;      easy to implement.
                                        ;
RMDI_TYPE_DRVR_ADDR     equ 1           ;*Seg:Ofs pointer to device driver header
                                        ; (is always the 1st packet for all packets)
RMDI_TYPE_INT13         equ 2           ; Seg:Ofs pointer to code entry point
                                        ; (here, at 70h:3EEh for IO.SYS)
                                        ; [IO.SYS, EMM386.EXE]
RMDI_TYPE_BIOS_INT13    equ 3           ; pointer to list of ??? [IO.SYS]
RMDI_TYPE_INT4B         equ 4           ;*for VDS (Virtual DMA Specification):
                                        ; Seg:Ofs pointer to code entry point
                                        ; for VDS handler/interceptor
                                        ; (On my system IO.SYS just reports
                                        ; the same ROM entry point which
                                        ; is entered by RMDI_TYPE_INT4F. It
                                        ; appears to be just the INT4Bh vector
                                        ; from the IVT recorded at boot time.
                                        ; EMM386.EXE actually intercepts this
                                        ; vector, does some work on its own,
                                        ; and either passes it down to the
                                        ; current INT4Bh handler, or to the
                                        ; vector which was recorded at its own
                                        ; startup time, that is in our case
                                        ; the same ROM entry point as reported
                                        ; by IO.SYS in because IO.SYS did not
                                        ; intercept this vector.)
                                        ; [IO.SYS, EMM386.EXE]
RMDI_TYPE_INT4F         equ 5           ;*for CAM (Common Access Method) SCSI:
                                        ; Seg:Ofs pointer to code entry point
                                        ; (on my system at 70h:4FCh in IO.SYS,
                                        ; which far-jumps into ROM at F000h:F692h
                                        ; and re-initialized the PICs after
                                        ; writing FFh in 50h:6Bh. The code is
                                        ; exited using an IRET. Could it be,
                                        ; that this is just the default INT4Fh
                                        ; pointer from the IVT as recorded at
                                        ; IO.SYS boot-up time???)
                                        ; [IO.SYS]
RMDI_TYPE_ASPI_ENTRY    equ 6           ; [no occurence so far]
RMDI_TYPE_DRVR_HOOK     equ 7           ; [no occurence so far]
RMDI_TYPE_LINENUM       equ 8           ;*Line number of occurence in CONFIG.SYS
                                        ; or 00000000h if not in CONFIG.SYS
                                        ; (usually this is the last packet of all packets)
RMDI_TYPE_CONFIGSTAMP   equ 9           ; Apparently not an address, but some
                                        ; kind of 32bit checksum/finger-print
                                        ; of ???
                                        ; [HIMEM.SYS, S-ICE.EXE]
RMDI_TYPE_HWINTHOOK     equ 10          ;*Not an address (!!!), but apparently
                                        ; the hooked IRQ number+1 (This entry
                                        ; holds 2 for FreeKEYB.COM/K3PLUS.COM
                                        ; (my keyboard driver) which provides
                                        ; an ISR for IRQ1=INT9, and also hooks
                                        ; INT10h, INT15h, INT16h, INT1Ch,
                                        ; INT21h, INT28h, and INT2Fh.)
                                        ; [FreeKEYB/K3PLUS]

                                        ; rmde_Flags as found in Win95 DDK:
;RM_DEF_UNSAFE_BIT       equ 0          ; bit0 <0>
;RM_DEF_UMB_PROVIDER_BIT equ 1          ; bit1 <0>
;                                       ; (bit15-2 currently reserved <0>)
;RM_DEF_UNSAFE           equ (1 << RM_DEF_UNSAFE_BIT)
;RM_DEF_UMB_PROVIDER     equ (1 << RM_DEF_UMB_PROVIDER_BIT)


VKD_ID          equ 000Dh               ; VxD ID for virtual keyboard (VKD)
OEM_ID          equ 0FFh                ; Microsoft OEM ID for version check

DOS_VER_MAX     equ 0616h               ; Max. supported DOS is 6.22
DOS_VER_MIN     equ 0500h               ; Min. supported DOS is 5.0
;DOS_VER_MIN    equ 031Fh               ; (3.31 for DR PalmDOS & DR DOS 6.0)

CPM_VER_MIN     equ 72h                 ; Min. supported BDOS is 7.2
                                        ; 70h = DR PalmDOS
                                        ; 71h = DR DOS 6.0 "business update 1993"
                                        ; 72h = Novell DOS (all updates) and
                                        ;       Caldera OpenDOS 7.01
                                        ; 73h = Caldera OpenDOS 7.02

REG_PATH_LEN    equ 4Eh                 ; Length of registry path buffer


; #############################################################################
; ###  TSR start  #############################################################
; #############################################################################

Code SEGMENT
                ASSUME cs:code, ds:code

org 002Ch
env_seg:        dw ?                    ; pointer to env. segment


org 0080h                               ; also used as scratch pad
cmd_len:        db ?                    ; length of command line (PASCAL style)
cmd_line:       db 7Fh dup (?)          ; command line buffer


org 0100h                               ; start of .COM program


winglue:        jmp install             ; goto installer routine


; #############################################################################
; ###  INT 21h interceptor code  ##############################################
; #############################################################################


INT21_entry:
        pushf                           ; INT21h ISR routine:
        cmp     ax,(GET_VER*256+1)
         je     INT21_3001              ; Get DOS version and flags
        cmp     ah,GET_VER
         je     INT21_3000              ; Get DOS version and OEM
        cmp     ax,GET_REAL_VER
         je     INT21_3306              ; Get real DOS version
        popf

                db JMPF_Instr           ; chain to previous handler
int21_ofs       dw 0000                 ; JMPF seg:ofs
int21_seg       dw 0000


INT21_3000:
        ; Get DOS version and OEM number:
        ; ===============================
        ; N.B. Since MS-DOS 7+ this is not only entered
        ;      with AL=00h, but with AL<>01h.
        ;      VMM32.VXD calls it with 308Dh, and one of
        ;      it's VxDs calls it with 30CCh.
        popf
dos_ver1:
        mov     ax,REAL_VER
dos_oem_id:
        mov     bh,OEM_ID               ; default: OEM "Microsoft"
        jmps    INT21_30_exit


INT21_3001:
        ; Get DOS version number and flags:
        ; =================================
        popf
dos_ver2:
        mov     ax,REAL_VER
old_flags:
        mov     bh,0                    ; traditional version flags
                                        ; bit3 = DOS in ROM


INT21_30_exit:
serno_high:
        mov     bl,0                    ; 24bit serial number (high)
serno_low:
        mov     cx,0                    ; (low word)
        jmps    INT21_exit


INT21_3306:
        ; Get real DOS version number:
        ; ============================
        popf
dos_ver3:
        mov     bx,REAL_VER
new_flags:
        mov     dx,0                    ; 'new' revision and version flags
                                        ; DL = revision
                                        ; DH bit3 = DOS in ROM
                                        ; DH bit4 = DOS in HMA
if OBSOLETE eq TRUE
INT21_exit:
        retf    2
else
        jmps    INT21_exit
endif


; #############################################################################
; ###  INT 2Fh interceptor code  ##############################################
; #############################################################################
        

INT2F_entry:
        pushf                           ; INT 2Fh ISR routine:
if OBSOLETE eq TRUE
        cmp     ax,4A33h                ; MS-DOS 7 version check
         je     INT2F_4A33
endif
        cmp     ah,16h                  ; strip off anything but
         jne    int2F_chain             ; Windows/DOS interface
        cmp     al,07h                  ; Virtual device callout
         je     INT2F_1607              ; (for VKD here)
        ;cmp    al,0Fh
        ; je    INT2F_160F              ; INT 10h hook replacement
        cmp     al,11h
         je     INT2F_1611              ; Get SHELL info
        cmp     al,12h
         je     INT2F_1612              ; Get BIOS (IO.SYS) data
        cmp     al,13h
         je     INT2F_1613              ; Get Registry path
if OBSOLETE eq TRUE
        cmp     al,14h
         je     INT2F_1614              ; Set Registry path
endif
        cmp     al,90h
         jne    int2F_chain
        jmp     INT2F_1690              ; Get/Set MS-DOS 7 kernel data
int2F_chain:
        popf

                db JMPF_Instr           ; chain to previous handler
int2F_ofs       dw 0000                 ;  JMPF seg:ofs
int2F_seg       dw 0000


INT2F_1607:
        ; Virtual Device callout API (for VKD here):
        ; ==========================================
        ; Mentioned in Win95 DDK as W386_Device_Broadcast
        ; "Broadcast by virtual device"
        cmp     bx,VKD_ID              ; is it for us?
         jne    int2F_chain            ; no, so continue to chain
        push    cs                     ; (What about CX=function number???)
        pop     es                     ; return ES:DI -> struc
        mov     di, offset i2F1607_return_struc
        jmps    INT2F_exit

INT2F_1611:
        ; Get SHELL parameters:
        ; =====================
        ; Mentioned in Win95 DDK as W386_Get_Shell
        ; "IO.SYS service to return shell info"
if OBSOLETE eq TRUE
INT2F_4A33:
        ; MS-DOS 7 version check:
        ; =======================
endif
        ; N.B. Both functions are identical

        mov     ax,0                    ; believe what you want
        push    cs                      ;  and fake MS-DOS 7+
        pop     ds                      ; DS:DX -> shell_exec ASCIZ
        mov     dx,offset shell_exec
        mov     si,offset shell_param   ; DS:SI -> shell_param PAS-ASCII
        mov     bx,0                    ; zero


INT2F_exit:
        popf
if OBSOLETE eq FALSE
INT21_exit:
endif
        retf    2


INT2F_1612:
        ; Get BIOS data:
        ; ==============
        ; Mentioned in Win95 DDK as W386_Get_BIOS_Data
        ; "IO.SYS service to return IO.SYS data"
        push    cs
        pop     es                      ; return ES:BX -> struc
        mov     bx,offset i2F1612_return_struc
        mov     ax,0                    ; return "function implemented"
        jmps    INT2F_exit


INT2F_1613:
        ; Get Registry ASCIZ path:
        ; ========================
        ; Mentioned in Win96 DDK as W386_Get_SYSDATA_Path
        ; "IO.SYS service to return path to SYSTEM.DAT"
        ; On entry: ES:DI, CX=buffer size
        ; On exit:  CX=length (incl. Z)
        ;           ES:DI buffer filled (not necessarily uppercased)
        ;           AX=0
        push    ds
        push    si
        push    di
        push    bx

        cmp     cx,0
         je     INT2F_1613_skip

        push    cs
        pop     ds
        cld                             ; count up
        mov     si,offset registry_path
        mov     bx,0                    ; preset counter

INT2F_1613_loop:
        inc     bx
        lodsb                           ; AL <- DS:SI, INC SI
        stosb                           ; AL -> ES:DI, INC DI
        cmp     al,0                    ; EOL reached?
        loopnz  INT2F_1613_loop         ; DEC CX, AL=0 | end of buffer?

        mov     cx,bx                   ; copied bytes from BX -> CX
INT2F_1613_skip:
        mov     ax,0                    ; return "function implemented"

        pop     bx
        pop     di
        pop     si
        pop     ds

        jmps    INT2F_exit

if OBSOLETE eq TRUE
INT2F_1614:
        ; Set Registry ASCIZ path:
        ; ========================
        ; Mentioned in Win95 DDK as W386_Set_SYSDAT_Path
        ; "IO.SYS service to set path to SYSTEM.DAT"
        ;
        ; On entry: ES:DI -> ASCIZ path (case preserving)
        ; On exit : AX=0     ok
        ;           AX=1614h unsupported
        ;           AX=004Eh failed, max buf size
        push    ds
        push    es
        push    si
        push    di
        push    bx
        push    cx

        cld                             ; count up
        mov     al,0                    ; search for EOL
        mov     cx,REG_PATH_LEN
        mov     si,di
        xor     bx,bx                   ; preset BX=0

INT2F_1614_loop:
        inc     bx
        scasb                           ; AL=ES:DI?, INC DI
        loopnz  INT2F_1614_loop         ; buffer full
         jnz    INT2F_1614_full

        ;end found, BX=copy this many bytes (at least 1)
        push    es                      ; DS:SI are source now
        pop     ds
        push    cs
        pop     es
        mov     di,offset registry_path ; ES:DI is target now
        mov     cx,bx
        rep     movsb                   ; copy DS:SI -> ES:DI (CX times)

        mov     ax,0                    ; return "success"
        jmps    INT2F_1614_end

INT2F_1614_full:
        mov     ax,REG_PATH_LEN

INT2F_1614_end:
        pop     cx
        pop     bx
        pop     di
        pop     si
        pop     es
        pop     ds

        jmps    INT2F_exit
endif

INT2F_1690:
        ; Get/Set MS-DOS 7 kernel data
        ; ============================
        ; On exit: ES:BX -> structure containing a pointer
        ;                   to device driver list structure
        ;
        ; Partially documented in Win95 DDK as W386_Return_RMD "return RMD list"
        ; However, its purpose/usage/internals are not documented.
        ;
        push    cs
        pop     es                      ; return ES:BX -> struc
        mov     bx,offset i2F1690_Return_RMD_Struc
        popf                            ; get back the flags
        clc                             ; clear carry flag (necessary???)
        retf    2


; #############################################################################
; ###   TSR data area   #######################################################
; #############################################################################


        ; =====================================================================
        ; Location unknown (indirectly used via API):


registry_path:
        db      REG_FILENAME,0
if OBSOLETE eq TRUE
registry_def_path_end:
        db      (REG_PATH_LEN + registry_path - registry_def_path_end) dup (0)
endif

        ; =====================================================================
        ; Located in low BIOS/DOS code (global memory):

INT16_entry:
if OBSOLETE eq TRUE
        ; <code entry point> CS=70h
        ; Andrew says, this must be some kind of INT16h handler...
        ;
        ; However, this is the code:
        ; entry_point:
        ; sti
        ; pushf
        ; test BY cs:[0009h],20h
        ;  jnz skip_code
        ; cli
        ; call DWO cs:[0174h]       ; (F000:E82E)
        ;                           ; strat-entry of IO.SYS
        ;                           ; disk driver???
        ;                           ; (IntDisk in our INIT.ASM)
endif
i2F1607_return_struc:                   ; Located at 0070h:043Dh in MS-DOS 7:
        ;---- start code signature A "dd 0EFF2E9Ch" + "050Bh"
        pushf                           ; db 9Ch
        dec   WO cs:i2F1612_050B        ; dec cs:["050Bh"]
                                        ; db 2Eh, 0FFh, 0Eh
                                        ; dw offset i2F1612_050B ; (050Bh in MS-DOS 7)
        ;---- end code signature A + "050Bh"
        ;---- start code signature B "dd 9D900175h"
        jnz   skip_nop                  ; db 75h, 01h
        nop                             ; db 90h (ES:DI+8) patched with V86 breakpoint:
                                        ; arpl [di+02CAh],bx
skip_nop:                               ; skip breakpoint if "050Bh" is zero
        popf                            ; db 9Dh
        ;---- end code signature B
if OBSOLETE eq TRUE
        ; retf 2
        ; skip_code:
        ; call unknown_function         ; db 0E8h, 39h, 00h
        ; call DWO cs:[0154h]           ; appears to be the entry
        ;                               ; point to another device
        ;                               ; driver in IO.SYS
        ; jmps entry_point
endif

        db      JMPF_Instr              ; chain to previous handler
int16_ofs:
        dw      0000                    ; JMPF seg:ofs
int16_seg:
        dw      0000

i2F1612_050B:
        dw      0                       ; moving this to i2F1612_data_struc
                                        ; below doesn't work any longer...

        ; ---------------------------------------------------------------------
        ; Located in low BIOS/DOS data:


i2F1612_data_struc:
        ; Located at 0070h:0501h in MS-DOS 7:
        ; N.B. These are the values found here on my machine.
        ;      However, they differ from machine to machine.
        ;      The startup values are stored in the IO.SYS file
        ;      at file pos +0D01h (in my copy).
        dw      0; 199Ah                ; days since 1980-01-01
        dw      0                       ; ???
        dw      0                       ; ticks [sec/100] (I'm not sure)??? or
                                        ; event counter (upwards) for idle???
        db      00h,  00h, 0C8h,  00h
;i2F1612_050B: ; this doesn't work any longer, moved to above
        dw      0                       ; Once changed every boot, could be start
                                        ; time? Will be counted down in some way.
                                        ; But since some time it does not
                                        ; change any more.
                                        ; However, the transition from 1 to 0 is
                                        ; a special case, since it bypasses the
                                        ; V86 breakpoint.
        db      0                       ; unknown, always zero???
DOS_seg:dw      0C9h                    ; segment value of 1st DOS segment in IO.SYS
                                        ; (containing NUL, and 5 handles)
                                        ; Usually it's C9h, but sometimes it's 70h.

if OBSOLETE eq TRUE
                      ; The following code might be obsolete here...

i2F1612_internal_func:
        int     3                       ; give control to debugger
        db      1Dh dup (NOP_Instr)     ; patch area...
        retf                            ; return

        ; This excerpt is the original code from MS-DOS 7.10' IO.SYS:
        ; Partially it looks quite close to our BIOS' Int13Trap code
        ; in INIT.ASM. Maybe it has something to do with direct
        ; disk access, 32bit disk access, or drive locking?
        ;
        ; mov     ds,WO cs:DOS_seg        ; get DS for IO.SYS
        ; lea     bp,[bp+00]
        ; pushf
        ; cls                             ; simulate INT13h through
        ; call    far DWO ds:[0B4h]       ; BIOS' I13Pointer
        ; mov     WO cs:DOS_seg
        ; push    cs
        ; pop     ds                      ; align DS
        ; pushf
        ; call    A20Enable
        ; popf
        ; lea     bp,[bp+00]
        ; retf                            ; return
endif


        ; =====================================================================
        ; Apparently located in DOS data (can be loaded high)


        ; Located at FFFFh:xxxx (at about 1xxxh or 2xxxh)
i2F1612_return_struc:
        dw      1                       ; ID or size indicator???
        dw      offset i2F1612_data_struc; (0501h in MS-DOS 7)
bios_func_seg:
        dw      0                       ; runtime fixup (0070h in MS-DOS 7)

        ; about 70h bytes following to reach i2F1690_Return_RMD_Struc...


        ; ---------------------------------------------------------------------
        ; Located at FFFFh:xxxxh (at about 1xxxh or 2xxxh)


i2F1690_Return_RMD_Struc:
        ; This structure is called "Return_RMD_Struc" in Win95 DDK. The
        ; convention is called "RMD API". It contains two DWORDs (Seg:Ofs),
        ; the 1st ("RRS_RMD") pointing to the RMD chain, and apparently the
        ; 2nd ("RRS_Next_Ptr") pointing to next RMD chain (no info on end
        ; marker, but appears to be a zero entry).

RRS_RMD:                                ; Seg:Ofs of RMD chain
        dw      offset i2F1690_data_struc,0; Runtime segment fixup (FFFFh if MS-DOS 7 in HMA)
RRS_Next_Ptr:                           ; Seg:Ofs of next RMD chain in list
        dw      0,0                     ; (zero indicates end of list)
                                        ; (the segment value was non-zero
                                        ; during my tests, but pointed into
                                        ; free memory...)


if OBSOLETE eq TRUE
        dw      0                       ; ??? unknown data
        ; In MS-DOS IO.SYS this is followed by the path to
        ; the shell (SHELL=), followed by it's parameters...
endif


        ; -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --


shell_exec:
        db      SHELL_FILENAME,0
if OBSOLETE eq TRUE
        ; (For Windows95 the maximum reserved length including pending zero
        ; is 4Eh bytes.)
endif
shell_param:
        db      (shell_param_end - shell_param_start)
shell_param_start:
        db      SHELL_PARAMETERS,0
if OBSOLETE eq TRUE
        ; (For Windows95 the maximum reserved length including pending zero,
        ; but without length byte is 7Fh bytes.)
endif
shell_param_end:


        ; ---------------------------------------------------------------------


i2F1690_data_struc:
        ; Located at FFFFh:xxxxh (might be at end of the
        ; used HMA area, so that it can grow during CONFIG.SYS parsing...)
        ;
        ; Maybe this info is stored here during IO.SYS' CONFIG.SYS processing,
        ; also using the INT2Fh/1690h pointer. Then we could easily code this
        ; by implementing a callout API in IBMBIO.COM, received by an external
        ; driver, loaded at the upper end of conventional memory and removing
        ; itself from memory after finishing CONFIG.SYS. The structure itself
        ; must still be located somewhere in memory (HMA, maybe EMS page) for
        ; later usage.
        ;
        ; According to the Win95 DDK the general format of chain entries
        ; appears to be:
        ;
        ; rmd_header_t: db ? ; "rh_rmd_type" (see RMD_TYPE_*[1..7] for types)
        ;               dd ? ; "rh_next_rmd" (pointer to next "rmd_header_t"
        ;                                    or zero, if end of chain)
        ;
        ; In our case this means for "rmd_bios_t":

        ; "rbios_rmdhdr":
        db      RMD_TYPE_IOSYS          ; type flag for "rmd_bios_t"
        dd      0                       ; zero for end of chain
        ; "rmd_bios_t" entry:
        db      1                       ; "rbios_NumDriverEntries" # of entries

        ; Array of "rmd_driver_entry_t" entries follows:
        ; 1st "rmd_driver_entry_t" entry (only one in this case):
        db      "SETVER  "              ; "rmde_DriverName" \
        db      "EXE"                   ; "rmde_Driver_Ext" / SPACE padded
                                        ; file name of loaded device driver
        dw      0                       ; "rmde_Flags"
                                        ; bit0    = RM_DEF_UNSAFE_BIT
                                        ; bit1    = RM_DEF_UMB_PROVIDER_BIT
        db      0                       ; "rmde_NumUnits"
                                        ; number of block device drivers,
                                        ; 0 for char devices
        db      1                       ; "rmde_NumInfo"
                                        ; number of address packets in
                                        ; following array of address packets
                                        ; (at least one; in its "rmd_max_
                                        ; _driver_entry_t" IO.SYS can cope
                                        ; with a maximum of 26 entries!)

        ; 1st "rmd_driver_info_t" packet (only one in this case):
        db      RMDI_TYPE_LINENUM       ; "rmdi_Type" Type of entry
                                        ; (must not be RMDI_TYPE_INT13
                                        ; or RMDI_TYPE_BIOS_INT132 here)
        dd      0                       ; "rmdi_Addr" Seg:Ofs pointer to item
                                        ; (in *this* case it's just a number
                                        ; rather than an address)


        ; 2nd, 3rd, ... packets would follow here...

        ;        ; 2nd entry (only one in this case):
        ;        db      "IFSHLP  SYS"
        ;        dw      0
        ;        db      0
        ;        db      1
        ;
        ;        ; 1st packet (only one in this case):
        ;        db      RMDI_TYPE_LINENUM
        ;        dd      0

        ; 2nd, 3rd, ... entries, each containing a various number
        ;      of packets, would follow here...


; #############################################################################
; ###  Init & Installer code  #################################################
; #############################################################################

install:; Init:
        ; =====
        mov     dx, offset Startup_Msg  ; output startup message
        mov     ah,WRITE_STR
        int     DOS_INT

        push    cs
        pop     ds                      ; set DS

        cld                             ; (obsolete)
        call    update_switchar         ; patch current SwitChar
        call    parse_cmd               ; evaluate command line parameters

        mov     BY cmd_len,0            ; since we use the command line buffer
        mov     BY cmd_line,CR          ; as scratch pad, we fake an empty
                                        ; command line for analysis tools
                                        ; investigating this...
        stc
        mov     ax,GET_CPM_VER          ; determine DR DOS version
        int     DOS_INT                 ; (client/single-user nature)
         jc     check_dos
        cmp     ah,10h                  ; x86 single-user issue?
         je     is_drdos
        cmp     ah,14h                  ; x86 multi-user issue?
         jne    wrong_CPM_10
is_drdos:
        cmp     al,CPM_VER_MIN
         jae    check_dos
wrong_CPM_10:
        test    BY install_flags,FORCE
         jnz    check_dos
        jmp     wrong_CPM

check_dos:
        mov     ax,(GET_VER*256+0)      ; get old DOS version and OEM
        int     DOS_INT
        xchg    ah,al
        cmp     ax,DOS_VER_MIN
         jb     wrong_DOS_10            ; filter out DOS below 5.0
        cmp     bh,0                    ; is it PC-DOS?
         je     dos_ok
        cmp     ax,DOS_VER_MAX
         jbe    dos_ok
wrong_DOS_10:
        test    BY install_flags,FORCE
         jnz    dos_ok
        jmp     wrong_DOS               ; don't run on MS-DOS above 6.22
dos_ok:

        ;we always leave OEM ID Microsoft
        ;mov    BY dos_oem_id+1,bh      ; store OEM ID

        mov     BY serno_high+1,bl      ; store high byte of serno
        mov     WO serno_low+1,cx       ; store low word of serno

        mov     ax,(GET_VER*256+1)      ; get old DOS version and flags
        int     DOS_INT
        mov     BY old_flags+1,bh       ; store current state

        ;mov    BY serno_high+1,bl      ; we assume this was
        ;mov    WO serno_low+1,cx       ; always equal...

        test    BY install_flags,NEWVER ; new version set?
         jnz    skip_setver             ; then skip
        xchg    ah,al                   ;
        cmp     ax,DOS_VER_MAX          ; is the version number above usual?
         jna    skip_setver             ; no, skip (use internal default)
        xchg    ah,al                   ; yes, use the current version number
        call    setver                  ; update version number
skip_setver:

        mov     ax,GET_REAL_VER         ; get old real DOS version
        int     DOS_INT                 ; but only to get DX
        mov     WO new_flags+1,dx       ; store current state

        mov     WO bios_func_seg,cs     ; runtime fixup
        mov     WO RRS_RMD+2,cs         ; runtime fixup


        ; Install:
        ; ========
        cld                             ; (obsolete)
        cli                             ; enter mutex

        mov     ax,(GET_VECT*256+KBD_INT); install own INT16h handler
        int     DOS_INT                 ; and store original handler
        mov     ds:WO int16_ofs,bx
        mov     ds:WO int16_seg,es
        mov     dx,offset INT16_entry
        mov     ax,(SET_VECT*256+KBD_INT)
        int     DOS_INT

        mov     ax,(GET_VECT*256+DOS_INT); install own INT21h handler
        int     DOS_INT                 ; and store original handler
        mov     ds:WO int21_ofs,bx
        mov     ds:WO int21_seg,es
        mov     dx,offset INT21_entry
        mov     ax,(SET_VECT*256+DOS_INT)
        int     DOS_INT

        mov     ax,(GET_VECT*256+MPX_INT); install own INT2Fh handler
        int     DOS_INT                 ; and store original handler
        mov     ds:WO int2F_ofs,bx
        mov     ds:WO int2F_seg,es
        mov     dx,offset INT2F_entry
        mov     ax,(SET_VECT*256+MPX_INT)
        int     DOS_INT

        sti                             ; leave mutex

        ; Make resident:
        ; ==============
        mov     dx,offset Install_Msg   ; display message
        mov     ah,WRITE_STR
        int     DOS_INT

        mov     dx,offset Version_Msg
        mov     ah,WRITE_STR
        int     DOS_INT

        mov     ax,(GET_VER*256+0)      ; actually use API to get
        int     DOS_INT                 ; version number in AX
        call    disp_ver

        mov     dx,offset Registry_Msg
        mov     ah,WRITE_STR
        int     DOS_INT

        push    cs
        pop     es

        mov     di,offset cmd_line-1    ; we use cmd line buffer as scratch pad
        mov     cx,80h                  ; max. buffer length
        mov     ax,1613h
        int     MPX_INT                 ; get registry path in ES:DI (CX)
        ;cmp    ax,0
        ; jnz   error...
        dec     cx                      ; make zero-based
        add     di,cx                   ; point to EOL
        mov     BY es:[di],'$'          ; place terminator here
        mov     dx,offset cmd_line-1
        mov     ah,WRITE_STR
        int     DOS_INT

if PARSE_S eq TRUE
        ;mov     dx,offset Shell_Msg
        ;mov     ah,WRITE_STR
        ;int     DOS_INT
        ;
        ;<display shell path & parameters here>...
endif

        mov     dx,offset Finish_Msg
        mov     ah,WRITE_STR
        int     DOS_INT

        mov     ax,WO ds:env_seg        ; free environment block
        push    ax
        pop     es
        mov     ah,FREE_MEM
        int     DOS_INT                 ; make zero since it's
        mov     WO ds:env_seg,0         ; invalid now

        mov     dx,offset install       ; terminate & stay resident
        int     TERM_INT                ; (keep 0..start resident)


        ; Error exit:
        ; ===========
wrong_DOS:
        mov     dx,offset Wrong_DOS_Msg
        jmps    display_error
wrong_CPM:
        mov     dx,offset Wrong_CPM_Msg
display_error:
        mov     ah,WRITE_STR            ; display message
        int     DOS_INT

        test    BY install_flags,FORCE+NEWVER
         jnz    exit                    ; skip extra hint in these cases...
        mov     dx,offset Help_Hint_Msg

error_exit:
        mov     ah,WRITE_STR            ; display message
        int     DOS_INT

exit:
        mov     ah,CPM_TERM
        int     DOS_INT


        ; #####################################################################


parse_cmd:
;---------
        ;cmp     BY cmd_len,7Dh         ; command line length ok?
        ; jbe    parse_cmd_10
        ;mov     BY cmd_len,7Dh         ; trunc length
;parse_cmd_10:

        mov     BY cmd_line+7Eh,CR      ; force string termination
        mov     si,offset cmd_line      ; (some shells don't do this correctly)

        push    cs
        pop     es

parse_loop:
        call    get_switch              ; get next switch

        cmp     al,'?'
         je     parse_help
        cmp     al,'h'
         jne    parse_loop_01

parse_help:
        mov     dx,offset Help_Msg
        pop     ax                     ; align stack
        jmps    error_exit


parse_loop_01:
        cmp     al,'f'                  ; is it /F ?
         jne    parse_loop_05
        or      install_flags,FORCE
        jmps    parse_loop

        ; ---------------------------------------------------------------------


parse_loop_05:
        cmp     al,'g'                  ; is it /G:<xx.xx> ?
         jne    parse_loop_10

        call    skip_equal              ; (EOL handled implicitly)
        call    atoi                    ; get numeric value for main version
         jc     parse_error             ; must be a number
        cmp     ax,2                    ; below 2.xx ?
         jb     parse_error
        cmp     ax,9                    ; above 9.xx ?
         ja     parse_error
        mov     bx,ax                   ; store main version number
                                        ; (bl = main ver, bh = 0)
        lodsb
        cmp     al,'.'                  ; do we have a sub version?
         je     parse_num_sub
        cmp     al,','
         jne    parse_num_ok
parse_num_sub:
        call    whitespace              ; deblank string
        call    atoi                    ; get sub version number
         jc     parse_error
        and     al,7Fh                  ; strip off 128..255 to 0..127
        cmp     ax,99                   ; (for future enhancements)
         ja     parse_error             ; above x.99 ?
        cmp     ax,9
         ja     parse_num_ok
        inc     di
        cmp     si,di                   ; has there been only one digit?
         jne    parse_num_ok            ; no, skip (7.01 -> 7.01)
        mov     cl,10                   ; yes, so multiply by 10 (7.1 -> 7.10)
        mul     cl                      ; ax:=al*10
parse_num_ok:
        mov     bh,al                   ; build DOS version number in BX
        mov     ax,bx
        call    setver                  ; update DOS version
        or      BY install_flags,NEWVER
        jmps    parse_loop

parse_error:
        mov     dx,offset Error_Msg
        pop     ax
        jmp     error_exit


        ; ---------------------------------------------------------------------


parse_loop_10:
        cmp     al,'r'                  ; is it /R:<registry filename> ?
         jne    parse_loop_20

        call    skip_equal              ; (EOL handled implicitly)
        push    si                      ; store current position
parse_loop_11:
        lodsb                           ; search for end of parameter
        cmp     al,SPACE
         je     parse_loop_15
        cmp     al,TAB
         je     parse_loop_15
swc1:   cmp     al,SWC
         je     parse_loop_15
        cmp     al,CR                   ; is it EOL already?
         jne    parse_loop_11

parse_loop_15:
        dec     si                      ; unget last character
        mov     bl,al                   ; save current AL value

        mov     BY es:[si],NUL          ; force a NUL terminator
        pop     di                      ; restore old position in DI

if OBSOLETE eq TRUE
        mov     ax,1614h                ; simulate an interrupt
        pushf
        push    cs
        call    INT2F_entry
else
        xor     ax,ax                   ; preset AX with error code 0

        ;push    ds                     ; Set Registry ASCIZ path
        ;push    es                     ; (similar to INT2F_1614)
        push    si
        push    di
        push    bx
        push    cx

        mov     al,0                    ; search for EOL
        mov     cx,REG_PATH_LEN
        mov     si,di
        xor     bx,bx                   ; preset BX=0

INT2F_1614_loop:
        inc     bx
        scasb                           ; AL=ES:DI?, INC DI
        loopnz  INT2F_1614_loop         ; buffer full
         jnz    INT2F_1614_full

        ;end found, BX=copy this many bytes (at least 1)
        ;push    es                      ; DS:SI are source now
        ;pop     ds
        ;push    cs
        ;pop     es

        mov     di,offset registry_path ; ES:DI is target now
        mov     cx,bx
        rep     movsb                   ; copy DS:SI -> ES:DI (CX times)

        mov     ax,0                    ; return "success"
        jmps    INT2F_1614_end

INT2F_1614_full:
        mov     ax,REG_PATH_LEN

INT2F_1614_end:
        pop     cx
        pop     bx
        pop     di
        pop     si
        ;pop     es
        ;pop     ds
endif
        mov     BY es:[si],bl           ; restore old character
        cmp     ax,0                    ; error occured?
         jne    parse_error
parse_loop_19:
        jmp     parse_loop


        ; ---------------------------------------------------------------------


parse_loop_20:
if PARSE_S eq TRUE
        ;cmp    al,'s'                  ; is it /S:<shell filename & params> ?
        ; jne   parse_loop
        ;call   skip_equal              ; (EOL shall be handled implicitly)
        ;
        ;ret                            ; exit parsing
endif
        jmps parse_loop_19


        ; ---------------------------------------------------------------------


get_switch:                             ; get next command line switch
;----------
        call    whitespace              ; skip all spaces & tabs
                                        ; (EOL handled implicitly)
        lodsb                           ; get next character

swc2:   cmp     al,SWC                  ; did we get a switch?
         je     get_switch_10           ; yes, return the character
get_switch_exit:
        pop     ax                      ; align stack
        ret

get_switch_10:
        lodsb                           ; get next character
        or      al,('a' xor 'A')        ; return lower-cased character
        ret


whitespace:     ; skip leading spaces
;----------
;       entry:  ES:SI -> command line
;       exit:   ZF set if EOL (SI still points at EOL)

        lodsb                           ; get next character

        cmp     al,CR                   ; EOL reached?
         je     whitespace_exit

        cmp     al,SPACE                ; is it a space?
         je     whitespace              ; yes, skip it
        cmp     al,TAB                  ; is it a tab?
         je     whitespace              ; yes, skip it
whitespace_exit:
        dec     si                      ; else unget the character
        ret


skip_equal:     ; skip leading equal signs
;----------
;       entry:  ES:SI -> command line
;       exit:   ZF set if EOL (SI still points at EOL)

        call    whitespace              ; deblank command line
                                        ; (EOL handled implicitly)
        lodsb                           ; get next character

        cmp     al,CR                   ; EOL reached?
         je     skip_equal_exit

        cmp     al,':'                  ; is it a colon?
         je     skip_equal              ; yes, skip it
        cmp     al,'='                  ; is it a equal sign?
         je     skip_equal              ; yes, skip it
skip_equal_exit:
        dec     si                      ; else unget the character
        call    whitespace              ; deblank command line
        ret


atoi:
;----
; To convert a decimal number in the form of an ASCII string to
; a 16 bit integer.
;
; Entry
;       ES:SI -> ASCII decimal number
;                (the end of the number is taken as the first non-digit)
; Exit
;       Carry clear:
;               AX = converted number
;               ES:SI -> first non-digit
;               ES:DI -> first digit (if found)
;
;       Carry set:
;               Either the first character was not a digit
;               or the number could not be represented in 16 bits
;               ES:SI -> point at which error occured
;               AX, DI undefined

        push    bx
        push    cx
        push    dx
        call    whitespace              ; Deblank line
         je     atoi_error              ; exit on EOL
        mov     di,si                   ; save string start offset
        mov     cx,10                   ; for multiply
        xor     ax,ax                   ; number is formed in AX

atoi_loop:
        mov     bl,es:[si]              ; BL = next char from string

        sub     bl,'0'
         jc     atoi_done
        cmp     bl,9
         ja     atoi_done
        mov     bh,0                    ; BX = next digit

        mul     cx                      ; AX = 10 * AX
         jc     exit_atoi               ; check for 16 bit overflow
        add     ax,bx                   ; AX = (10 * AX) + BX
         jc     exit_atoi

        inc     si                      ; ES:SI -> next char in string
        jmps    atoi_loop

atoi_done:
        cmp     si,di                   ; was there at least one digit?
         jne    exit_atoi               ;  yes - exit with carry clear
atoi_error:
        stc                             ;  no - set error flag

exit_atoi:
        pop     dx
        pop     cx
        pop     bx
        ret



btoa:   ; Display 8bit digit in AL on screen
        ; AX, DX destroyed
        aam                             ; AH:=AL DIV 10, AL:=AL MOD 10
        push    ax                      ; push digit
        and     ah,ah                   ; all digits finished
         jz     btoa_10
        mov     al,ah                   ; no, get next digit
        call    btoa                    ; recursion
btoa_10:
        pop     dx                      ; restore digit in DL
        add     dl,'0'                  ; convert digit to ASCII
        mov     ah,WRITE_CHAR
        int     DOS_INT                 ; display digit (AL destroyed)
        ret


disp_ver: ; display version in AX
        ; destroys AX, BL, DX
        mov     bl,ah                   ; save sub-version in BL
        call    btoa                    ; display main version
        mov     dl,'.'
        mov     ah,WRITE_CHAR
        int     DOS_INT                 ; display separator
        cmp     bl,9                    ; more than 9 as sub-version?
         ja     skip_leadin
        mov     dl,'0'
        mov     ah,WRITE_CHAR
        int     DOS_INT                 ; display leading '0'
skip_leadin:
        mov     al,bl
        call    btoa                    ; display sub-version
        ret

update_switchar:
;---------------
        mov     dl,SWC
        mov     ax,GET_SWITCHAR
        int     DOS_INT                 ; Get current SwitChar setting

        mov     BY swc1+1,dl            ; update SwitChar in parameter
        mov     BY swc2+1,dl            ; scanner

        push    cs
        pop     es

        mov     di,offset Help_Hint_Msg
        mov     cx,(offset Wrong_DOS_Msg - offset Help_Hint_Msg)
        mov     al,SWC_REP
switchar_loop:
        scasb                           ; AL=ES:DI?, INC DI
         jne    switchar_10
        mov     BY es:[di-1],dl         ; update switchar in help screens
switchar_10:
        loop    switchar_loop           ; until patching is finished
        ret

setver:
;------
        mov     WO dos_ver1+1,ax        ; patch to current DOS version
        mov     WO dos_ver2+1,ax
        mov     WO dos_ver3+1,ax
        ret


; #############################################################################
; ###  Installer data area  ###################################################
; #############################################################################


; The installer also used the command line buffer as scratch pad.

FORCE           equ 1
NEWVER          equ 2

install_flags   db 0
 

; #############################################################################
; ###  Messages  ##############################################################
; #############################################################################


Startup_Msg:
        db      CR, LF
        db      TSR_FILENAME," ",TSR_VER,"  Glue to run the Windows95 GUI on top of OpenDOS [ALPHA]",CR,LF
        db      "Copyright (c) 1997 Caldera, Inc.  All rights reserved.",CR,LF,'$'

Install_Msg:
        db      CR,LF
        db      "INT 16h, INT 21h, and INT 2Fh interceptors installed.$"

Version_Msg:
        db      CR,LF
        db      "DOS version     : $"

Registry_Msg:
        db      CR,LF
        db      "Windows registry: $"

if PARSE_S eq TRUE
;Shell_Msg:
;       db      CR,LF
;       db      "Command shell   : $"
endif

Finish_Msg:
        db      CR,LF,'$'

Help_Hint_Msg:
        db      "Please try ",SWC_REP,"F to override this warning, or get more help with ",SWC_REP,"H.",CR,LF,'$'

Error_Msg:
        db      CR,LF
        db      "Syntax error or parameter value out of range!",CR,LF
Help_Msg:
        db      CR,LF
        db      TSR_NAME," [",SWC_REP,"F] [",SWC_REP,"G:n.nn] [",SWC_REP,"R:registry]"
if PARSE_S eq TRUE
        db      " [",SWC_REP,"S:shell]"
endif
        db      CR,LF,CR,LF
        db      "  ",SWC_REP,"F            Forced installation of the driver",CR,LF
        db      "  ",SWC_REP,"G:n.nn       DOS version number to be faked in the range of 2.00..9.99",CR,LF
        db      "                Default: As SETVERed for ",TSR_NAME,", global version, or ",REAL_VER_STR,CR,LF
        db      "  ",SWC_REP,"R:registry   Full path and filename to Windows95 registry",CR,LF
        db      "                Default: ",REG_FILENAME,CR,LF
if PARSE_S eq TRUE
        db      "  ",SWC_REP,"S:shell      Full path, filename, and parameters to command processor shell",CR,LF
        db      "                Default: ",SHELL_FILENAME," ",SHELL_PARAMETERS," (last parameter)",CR,LF
endif
        db      '$'

Wrong_DOS_Msg:
        db      CR,LF
        db      "Sorry, but this DOS configuration is unsupported by default.",CR,LF,'$'

Wrong_CPM_Msg:
        db      CR,LF
        db      "Sorry, but only Caldera OpenDOS and Novell DOS are supported by default!",CR,LF,'$'

Code ends
end WINGLUE

